var fs = require('fs'),
    jsdom = require('jsdom'),
    PDFDocument = require('pdfkit');

function createMappingWindow(staticBasepath, callback) {
    jsdom.env({
        html: "<html><head></head><body></body></html>",
        src: [
            fs.readFileSync(staticBasepath + "js/contrib/jquery-2.1.0.min.js").toString(),
            fs.readFileSync(staticBasepath + "js/contrib/jg_global.js").toString(),
            fs.readFileSync(staticBasepath + "js/contrib/jg_library.js").toString(),
            fs.readFileSync(staticBasepath + "js/splunk_mapping2.js").toString()
        ],
        done: function(err, window) {
            if (err) {
                callback(err, null);
                return;
            }

            // Set up mock createElementNS
            window.document.createElementNS = function(ns, tagName) {
                var elem = window.document.createElement(tagName);

                elem.getBBox = function() {
                    if (this.textContent) {

                        var doc = new PDFDocument(),
                            fontSize = parseInt(elem.getAttribute("font-size"), 10),
                            font = doc.font('Helvetica', fontSize);

                        return ({
                            x: elem.offsetLeft,
                            y: elem.offsetTop,
                            // need to compute max width per line
                            width: font.widthOfString(this.textContent),
                            // need to compute currentLineHeight()+(numLines-1)*currentLineHeight(true)
                            height: font.currentLineHeight() + 4  // offset by 4 to produce results consistent with browser rendering
                        });
                    }

                    return ({
                        x: elem.offsetLeft,
                        y: elem.offsetTop,
                        width: elem.offsetWidth,
                        height: elem.offsetHeight
                    });
                };

                return elem;
            };

            // Set up mock console logging
            var mockConsole = {
                log: function() {
                    mockConsole.addMessage.apply(mockConsole, arguments);
                },
                warn: function() {
                    mockConsole.addMessage.apply(mockConsole, arguments);
                },
                debug: function() {
                    mockConsole.addMessage.apply(mockConsole, arguments);
                },
                error: function() {
                    mockConsole.addMessage.apply(mockConsole, arguments);
                },

                messages: [],
                addMessage: function(args) {
                    var i, strSegments = [];
                    for (i = 0; i < arguments.length; i++) {
                        strSegments.push(arguments[i]);
                    }
                    mockConsole.messages.push('JSDOM CONSOLE: ' + strSegments.join(' '));
                }
            };
            window.console = mockConsole;

            callback(null, window);
        }
    });
}

function prepareRowData(data) {
    var mapData = {};
    if (data) {
        var fields = data.fields;
        var rows = data.rows;
        if (fields && rows) {
            var numFields = fields.length;
            var numRows = rows.length;
            var numEntries;
            var row;
            var obj;
            var i;
            var j;

            mapData.fields = filterFields(fields);
            mapData.data = [];
            for (i = 0; i < numRows; i++) {
                row = rows[i];
                numEntries = Math.min(row.length, numFields);
                obj = {};
                for (j = 0; j < numEntries; j++)
                    obj[fields[j]] = row[j];
                mapData.data.push(obj);
            }
        }
    }
    return mapData;
}

function prepareColumnData(data) {
    var mapData = {};
    if (data) {
        var fields = data.fields;
        var columns = data.columns;
        if (fields && columns) {
            var numColumns = Math.min(fields.length, columns.length);
            var numRows = (numColumns > 0) ? columns[0].length : 0;
            var obj;
            var i;
            var j;

            for (i = 1; i < numColumns; i++)
                numRows = Math.min(numRows, columns[i].length);

            mapData.fields = filterFields(fields);
            mapData.data = [];
            for (i = 0; i < numRows; i++) {
                obj = {};
                for (j = 0; j < numColumns; j++)
                    obj[fields[j]] = columns[j][i];
                mapData.data.push(obj);
            }
        }
    }
    return mapData;
}

function filterFields(fields) {
    if (!fields)
        return null;

    var filteredFields = [];
    var field;
    for (var i = 0, l = fields.length; i < l; i++) {
        field = fields[i];
        if (field && (field.charAt(0) !== "_"))
            filteredFields.push(field);
    }
    return filteredFields;
}

function resolveURL(propertyValue, staticBasepath, splunkdBasepath) {
    var propertyValue2 = propertyValue ? propertyValue.replace(/^\s\s*/, "").replace(/\s\s*$/, "") : propertyValue;
    if (propertyValue2) {
        var rStatic = /^\/static\//;
        var rSplunkd = /^\/splunkd\/__raw\//;
        if (rStatic.test(propertyValue2))
            propertyValue = propertyValue2.replace(rStatic, staticBasepath.replace(/\/$/, "") + "/");
        else if (rSplunkd.test(propertyValue2))
            propertyValue = propertyValue2.replace(rSplunkd, splunkdBasepath.replace(/\/$/, "") + "/");
    }
    return propertyValue;
}

function toDOMArray(a, window) {
    return a ? window.Array.prototype.slice.call(a) : a;
}

function getSVG(data, staticBasepath, callback) {

    createMappingWindow(staticBasepath, function(err, window) {
        if (err) {
            callback(err, null);
            return;
        }

        var getConsoleMessages = function() {
            if (window.console.messages && window.console.messages.length > 0) {
                return window.console.messages;
            }
            return null;
        };

        process.on("uncaughtException", function (error) {
            var fullError = {
                consoleMessages: getConsoleMessages(),
                message: error.stack
            };
            callback(fullError, null);
        });

        try {
            var splunk = window.splunk;

            var splunkdBasepath = (data.system && data.system.splunkdUri) ? data.system.splunkdUri : "https://localhost:8089/";
            var width = parseInt(data.width, 10);
            var height = parseInt(data.height, 10);
            var props = data.props;
            var series = data.series;

            var mapData = prepareColumnData(series);
            mapData.data = toDOMArray(mapData.data, window);
            mapData.fields = toDOMArray(mapData.fields, window);

            var booleanParser = splunk.parsers.BooleanParser.getInstance();
            var numberParser = splunk.parsers.NumberParser.getInstance();
            var stringParser = splunk.parsers.StringParser.getInstance();
            var numberArrayParser = splunk.parsers.ArrayParser.getInstance(numberParser);
            var stringArrayParser = splunk.parsers.ArrayParser.getInstance(stringParser);
            var numberObjectParser = splunk.parsers.ObjectParser.getInstance(numberParser);
            var latLonParser = splunk.mapping2.parsers.LatLonParser.getInstance();
            var latLonBoundsParser = splunk.mapping2.parsers.LatLonBoundsParser.getInstance();

            var defaultSeriesColors = [ 0x6CB8CA, 0xFAC61D, 0xD85E3D, 0x956E96, 0xF7912C, 0x9AC23C, 0x5479AF, 0x999755, 0xDD87B0, 0x65AA82,
                                        0xA7D4DF, 0xFCDD77, 0xE89E8B, 0xBFA8C0, 0xFABD80, 0xC2DA8A, 0x98AFCF, 0xC2C199, 0xEBB7D0, 0xA3CCB4,
                                        0x416E79, 0x967711, 0x823825, 0x59425A, 0x94571A, 0x5C7424, 0x324969, 0x5C5B33, 0x85516A, 0x3D664E ];
            defaultSeriesColors = toDOMArray(defaultSeriesColors, window);

            var seriesColorPalette = new splunk.palettes.ListColorPalette(defaultSeriesColors);
            var fieldColorPalette = new splunk.palettes.FieldColorPalette(null, seriesColorPalette);

            var markerLayer = new splunk.mapping2.layers.PieMarkerLayer();
            markerLayer.set("markerColorPalette", fieldColorPalette);
            markerLayer.set("data", mapData.data);
            markerLayer.set("fields", mapData.fields);

            var tileLayer = new splunk.mapping2.layers.SVGTileLayer();
            tileLayer.set("url", resolveURL("/splunkd/__raw/services/mbtiles/splunk-tiles/{z}/{x}/{y}", staticBasepath, splunkdBasepath));
            tileLayer.set("maxZoom", 7);

            var attributionControl = new splunk.mapping2.controls.SVGAttributionControl();

            var map = new splunk.mapping2.Map();
            map.set("width", width);
            map.set("height", height);
            map.addLayer(tileLayer);
            map.addLayer(markerLayer);
            map.addControl(attributionControl);
            map.fitWorld(true);

            // global properties
            if (typeof props['fieldColors'] === "string")
                fieldColorPalette.set("fieldColors", numberObjectParser.stringToValue(props['fieldColors']));
            if (typeof props['seriesColors'] === "string")
                seriesColorPalette.set("colors", numberArrayParser.stringToValue(props['seriesColors']));

            // map properties
            if (typeof props['map.center'] === "string") {
                var center = latLonParser.stringToValue(props['map.center']);
                if (center)
                    map.set("center", center);
            }
            if (typeof props['map.zoom'] === "string") {
                var zoom = numberParser.stringToValue(props['map.zoom']);
                if (!isNaN(zoom))
                    map.set("zoom", zoom);
            }
            if (typeof props['map.fitBounds'] === "string") {
                var fitBounds = latLonBoundsParser.stringToValue(props['map.fitBounds']);
                if (fitBounds)
                    map.fitBounds(fitBounds);
            }

            // tileLayer properties
            if (typeof props['tileLayer.url'] === "string")
                tileLayer.set("url", resolveURL(props['tileLayer.url'], staticBasepath, splunkdBasepath));
            if (typeof props['tileLayer.subdomains'] === "string")
                tileLayer.set("subdomains", stringArrayParser.stringToValue(props['tileLayer.subdomains']));
            if (typeof props['tileLayer.minZoom'] === "string")
                tileLayer.set("minZoom", numberParser.stringToValue(props['tileLayer.minZoom']));
            if (typeof props['tileLayer.maxZoom'] === "string")
                tileLayer.set("maxZoom", numberParser.stringToValue(props['tileLayer.maxZoom']));
            if (typeof props['tileLayer.invertY'] === "string")
                tileLayer.set("invertY", booleanParser.stringToValue(props['tileLayer.invertY']));
            if (typeof props['tileLayer.attribution'] === "string")
                tileLayer.set("attribution", stringParser.stringToValue(props['tileLayer.attribution']));

            // markerLayer properties
            if (typeof props['markerLayer.markerOpacity'] === "string")
                markerLayer.set("markerOpacity", numberParser.stringToValue(props['markerLayer.markerOpacity']));
            if (typeof props['markerLayer.markerMinSize'] === "string")
                markerLayer.set("markerMinSize", numberParser.stringToValue(props['markerLayer.markerMinSize']));
            if (typeof props['markerLayer.markerMaxSize'] === "string")
                markerLayer.set("markerMaxSize", numberParser.stringToValue(props['markerLayer.markerMaxSize']));

            // fit data
            var dataBounds = null;
            if (typeof props['data.bounds'] === "string")
                dataBounds = latLonBoundsParser.stringToValue(props['data.bounds']);
            dataBounds = markerLayer.getLatLonBounds(dataBounds ? dataBounds.getCenter() : null);
            if (dataBounds)
                map.fitBounds(dataBounds);

            var svg1 = tileLayer.toSVGString();
            var svg2 = markerLayer.toSVGString();
            var svg3 = attributionControl.toSVGString();
            var svg = splunk.vectors.VectorUtils.concatSVGStrings(svg1, svg2, svg3);

            callback({ consoleMessages: getConsoleMessages() }, svg);
        }
        catch(err) {
            window.console.log('exception in splunk_mapping::getSVG' + JSON.stringify(err));
            var fullError = {
                consoleMessages: getConsoleMessages(),
                message: err
            };
            callback(fullError, null);
        }
    });
}

exports.getSVG = getSVG;
