/**
 * Singleton to manage a shared schema for all available visualization types in pivot.
 *
 * Each visualization type has an entry in the schema with the following structure:
 *
 * type {String} the id of the visualization type
 * icon {String} the icon CSS class for the button in the menu
 * description {String} the human-readable description of the visualization type
 * showConfigMenu {Boolean} whether to show the config menu side bar when this visualization type is active
 * requiresIndexTime {Boolean} whether the visualization type depends on the current object having _time as a field, defaults to false
 * masterView {View constructor} the view to instantiate to render the content
 * dataOutputMode {String} the output_mode argument to pass to the results endpoint
 * reportLevelAttributes {Object} custom attributes to be set on the report when this visualization type is selected
 *
 * configMenuPanels {Array<Object>} list of available panels for the visualization type
 *                  each object has the following fields:
 *
 *     title {String} the title display for the panel
 *     shortTitle {String} the title to use for the panel's "add new" button
 *     elementType {String} the pivot report element type (if any) the panel maps to
 *     setAttributes {Object} a dictionary containing values to be set on the element when the panel is initialized
 *     dataTypes {String or Array<String>} a whitelist of allowed data types for fields in the panel
 *     outputType {String} an optional output type that an element must match for this panel, "metric" or "dimension"
 *     required {Boolean} whether the panel is required to have at least one field in it
 *     maxLength {Number} the maximum number of fields allowed in the panel, if omitted there is no maximum
 *
 *     elementsSelector {Function} a callback to use for advanced configurations, accepts one argument which is a list of
 *         matching element models (after other configuration options like dataTypes and elementType have been applied),
 *         and returns a list of elements that should be mapped to the panel
 *     newElementHandler {Function} a callback to use for advanced configurations, accepts two arguments which are the new
 *         element being added and a boolean indicating if the element was just created or is being transferred, can make
 *         modifications to the model as necessary and/or return options which will be passed to the collection add method
 *
 *     controlGroups {Array<Object>} a list of control groups, each with the following fields:
 *
 *         type {String} either 'reportControl' or 'vizControl', controls which models will be used to create the control group
 *         view {View constructor} the view to instantiate for the control group
 *         options {Object} options to pass to the view constructor
 *         visibility {Object} key/value pairs that control when the group will be visible
 *                    the key is a model property name
 *                    the value is a string or list of strings that the property value must match for the control to be visible
 */

define([
            'jquery',
            'underscore',

            // content views
            'views/pivot/content/TableContent',
            'views/pivot/content/ChartContent',
            'views/pivot/content/SingleValueContent',

            // viz control groups
            'views/shared/vizcontrols/custom_controls/DrilldownRadioGroup',
            'views/shared/vizcontrols/custom_controls/AxisLabelRotationControlGroup',
            'views/shared/vizcontrols/custom_controls/AxisLabelElisionControlGroup',
            'views/shared/vizcontrols/custom_controls/AxisScaleControlGroup',
            'views/shared/vizcontrols/custom_controls/AxisIntervalControlGroup',
            'views/shared/vizcontrols/custom_controls/AxisMinValueControlGroup',
            'views/shared/vizcontrols/custom_controls/AxisMaxValueControlGroup',
            'views/shared/vizcontrols/custom_controls/LegendPlacementControlGroup',
            'views/shared/vizcontrols/custom_controls/LegendTruncationControlGroup',
            'views/shared/vizcontrols/custom_controls/StackModeControlGroup',
            'views/shared/vizcontrols/custom_controls/NullValueModeControlGroup',
            'views/shared/vizcontrols/custom_controls/SliceMinimumSizeControlGroup',
            'views/shared/vizcontrols/custom_controls/GaugeStyleControlGroup',
            'views/shared/vizcontrols/custom_controls/GaugeColorRangesControlGroups',
            'views/shared/vizcontrols/custom_controls/SingleValueBeforeLabelControlGroup',
            'views/shared/vizcontrols/custom_controls/SingleValueAfterLabelControlGroup',
            'views/shared/vizcontrols/custom_controls/SingleValueUnderLabelControlGroup',
            'views/shared/vizcontrols/custom_controls/BubbleMinSizeControlGroup',
            'views/shared/vizcontrols/custom_controls/BubbleMaxSizeControlGroup',

            // pivot control groups
            'views/pivot/custom_controls/TimeRangeControlGroup',
            'views/pivot/custom_controls/FilterTypeControlGroup',
            'views/pivot/custom_controls/FilterMatchControlGroup',
            'views/pivot/custom_controls/FilterLimitByControlGroup',
            'views/pivot/custom_controls/FilterLimitControlGroup',
            'views/pivot/custom_controls/LabelControlGroup',
            'views/pivot/custom_controls/AxisLabelControlGroup',
            'views/pivot/custom_controls/TimePeriodControlGroup',
            'views/pivot/custom_controls/CreateRangesControlGroup',
            'views/pivot/custom_controls/SplitSortControlGroup',
            'views/pivot/custom_controls/SplitLimitControlGroup',
            'views/pivot/custom_controls/MaxRangesControlGroup',
            'views/pivot/custom_controls/RangeSizeControlGroup',
            'views/pivot/custom_controls/RangeStartControlGroup',
            'views/pivot/custom_controls/RangeEndControlGroup',
            'views/pivot/custom_controls/TrueLabelControlGroup',
            'views/pivot/custom_controls/FalseLabelControlGroup',
            'views/pivot/custom_controls/CellValueControlGroup'
        ],
        function(
            $,
            _,

            // content views
            TableContent,
            ChartContent,
            SingleValueContent,

            // viz control groups
            DrilldownRadioGroup,
            AxisLabelRotationControlGroup,
            AxisLabelElisionControlGroup,
            AxisScaleControlGroup,
            AxisIntervalControlGroup,
            AxisMinValueControlGroup,
            AxisMaxValueControlGroup,
            LegendPlacementControlGroup,
            LegendTruncationControlGroup,
            StackModeControlGroup,
            NullValueModeControlGroup,
            SliceMinimumSizeControlGroup,
            GaugeStyleControlGroup,
            GaugeColorRangeControls,
            SingleValueBeforeLabelControlGroup,
            SingleValueAfterLabelControlGroup,
            SingleValueUnderLabelControlGroup,
            BubbleMinSizeControlGroup,
            BubbleMaxSizeControlGroup,

            // pivot control groups
            TimeRangeControlGroup,
            FilterTypeControlGroup,
            FilterMatchControlGroup,
            FilterLimitByControlGroup,
            FilterLimitControlGroup,
            LabelControlGroup,
            AxisLabelControlGroup,
            TimePeriodControlGroup,
            CreateRangesControlGroup,
            SplitSortControlGroup,
            SplitLimitControlGroup,
            MaxRangesControlGroup,
            RangeSizeControlGroup,
            RangeStartControlGroup,
            RangeEndControlGroup,
            TrueLabelControlGroup,
            FalseLabelControlGroup,
            CellValueControlGroup
        ) {

    // constants for available visualization types
    // where applicable these must match what will be set as 'display.visualizations.charting.chart' on the report
    var VIZ_TYPES = {
        STATISTICS_TABLE: 'statsTable',
        BAR_CHART: 'bar',
        COLUMN_CHART: 'column',
        AREA_CHART: 'area',
        LINE_CHART: 'line',
        PIE_CHART: 'pie',
        SCATTER_CHART: 'scatter',
        BUBBLE_CHART: 'bubble',
        SINGLE_VALUE: 'singlevalue',
        RADIAL_GAUGE: 'radialGauge',
        MARKER_GAUGE: 'markerGauge',
        FILLER_GAUGE: 'fillerGauge'
    };
    VIZ_TYPES.SIMPLE_CARTESIAN_TYPES = [VIZ_TYPES.BAR_CHART, VIZ_TYPES.COLUMN_CHART, VIZ_TYPES.LINE_CHART, VIZ_TYPES.AREA_CHART];
    VIZ_TYPES.BAR_COLUMN_TYPES = [VIZ_TYPES.BAR_CHART, VIZ_TYPES.COLUMN_CHART];
    VIZ_TYPES.LINE_AREA_TYPES = [VIZ_TYPES.LINE_CHART, VIZ_TYPES.AREA_CHART];
    VIZ_TYPES.GAUGE_TYPES = [VIZ_TYPES.RADIAL_GAUGE, VIZ_TYPES.MARKER_GAUGE, VIZ_TYPES.FILLER_GAUGE];

    // constants for data output modes
    // these must match what will be sent to the results/results_preview endpoint
    var DATA_MODES = {
        COLUMN_MAJOR: 'json_cols',
        ROW_MAJOR: 'json_rows'
    };

    var CONTROL_TYPES = {
        REPORT_CONTROL: 'reportControl',
        VISUALIZATION_CONTROL: 'vizControl'
    };

    // constants for field data types
    // these must match what the pivot report and its associated element collections are using
    var FIELD_DATA_TYPES = {
        STRING: 'string',
        NUMBER: 'number',
        BOOLEAN: 'boolean',
        TIMESTAMP: 'timestamp',
        IPV4: 'ipv4',
        OBJECT_COUNT: 'objectCount',
        CHILD_COUNT: 'childCount'
    };

    var AGGREGATION_OUTPUT_TYPES = {
        METRIC: 'metric',
        DIMENSION: 'dimension'
    };

    // constants for report element types
    // these must match what the pivot report and its associated element collections are using
    var REPORT_ELEMENT_TYPES = {
        FILTER: 'filter',
        COLUMN_SPLIT: 'column',
        ROW_SPLIT: 'row',
        CELL_VALUE: 'cell'
    };

    var extendPanelConfig = function(basePanelConfig, newPanelConfig, controlGroupsCallback, context) {
        var extendedConfig = $.extend({}, basePanelConfig, newPanelConfig);
        if(controlGroupsCallback) {
            context = context || null;
            extendedConfig.controlGroups = controlGroupsCallback.call(context, basePanelConfig.controlGroups);
        }
        else {
            extendedConfig.controlGroups = $.extend(true, [], basePanelConfig.controlGroups);
        }
        return extendedConfig;
    };

    var updateSplitLimitGroups = function(controlGroups, additionalOptions) {
        controlGroups = $.extend(true, [], controlGroups);
        _(controlGroups).chain().where({ view: SplitLimitControlGroup  }).each(function(group) {
            group.options = $.extend(group.options, additionalOptions);
        });
        return controlGroups;
    };

    var removeSplitLimitAndSortGroups = function(controlGroups) {
        return _(controlGroups).filter(function(group) {
            return (group.view !== SplitLimitControlGroup && group.view !== SplitSortControlGroup);
        });
    };

    var removeSplitSortGroups = function(controlGroups) {
        return _(controlGroups).filter(function(group) {
            return group.view !== SplitSortControlGroup;
        });
    };

    var removeLabelRotationGroups = function(controlGroups) {
        return _(controlGroups).filter(function(group) {
            return group.view !== AxisLabelElisionControlGroup && group.view !== AxisLabelRotationControlGroup;
        });
    };

    var GENERAL_PANEL = {
        title: _('General').t(),
        controlGroups: [
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: DrilldownRadioGroup
            }
        ]
    };

    var GAUGE_GENERAL_PANEL = {
        title: _('General').t(),
        controlGroups: [
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: GaugeStyleControlGroup
            }
        ]
    };

    var TIME_RANGE_PANEL = {
        title: _('Time Range').t(),
        elementType: REPORT_ELEMENT_TYPES.FILTER,
        dataTypes: FIELD_DATA_TYPES.TIMESTAMP,
        maxLength: 1,
        controlGroups: [
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: TimeRangeControlGroup
            }
        ]
    };

    var FILTER_PANEL = {
        title: _('Filter').t(),
        elementType: REPORT_ELEMENT_TYPES.FILTER,
        dataTypes: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER, FIELD_DATA_TYPES.BOOLEAN],
        controlGroups: [
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: FilterTypeControlGroup,
                visibility: { type: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER] }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: FilterMatchControlGroup,
                visibility: { filterType: 'match' }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: FilterLimitByControlGroup,
                visibility: { filterType: 'limit' }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: FilterLimitControlGroup,
                visibility: { filterType: 'limit' }
            }
        ]
    };

    var INDEPENDENT_AXIS_PANEL = {
        elementType: REPORT_ELEMENT_TYPES.ROW_SPLIT,
        dataTypes: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER, FIELD_DATA_TYPES.BOOLEAN,
                        FIELD_DATA_TYPES.TIMESTAMP],
        controlGroups: [
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: AxisLabelControlGroup,
                options: { inputClassName: 'input-medium', xAxis: true },
                visibility: { type: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER,
                                     FIELD_DATA_TYPES.BOOLEAN, FIELD_DATA_TYPES.TIMESTAMP] }
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: AxisLabelRotationControlGroup,
                options: { axisType: AxisLabelRotationControlGroup.X_AXIS }
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: AxisLabelElisionControlGroup,
                options: { axisType: AxisLabelElisionControlGroup.X_AXIS },
                visibility: { 'display.visualizations.charting.axisLabelsX.majorLabelStyle.rotation': ['-90', '-45', '45', '90'] }
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: SplitSortControlGroup,
                visibility: { type: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER,
                                        FIELD_DATA_TYPES.BOOLEAN, FIELD_DATA_TYPES.TIMESTAMP] }
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: SplitLimitControlGroup,
                options: { elementType: 'row' },
                visibility: { type: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4] }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: CreateRangesControlGroup,
                visibility: { type: FIELD_DATA_TYPES.NUMBER }
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: SplitLimitControlGroup,
                options: { elementType: 'row' },
                visibility: { type: FIELD_DATA_TYPES.NUMBER, display: 'all' }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: MaxRangesControlGroup,
                visibility: { type: FIELD_DATA_TYPES.NUMBER, display: 'ranges' }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: RangeSizeControlGroup,
                visibility: { type: FIELD_DATA_TYPES.NUMBER, display: 'ranges' }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: RangeStartControlGroup,
                visibility: { type: FIELD_DATA_TYPES.NUMBER, display: 'ranges' }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: RangeEndControlGroup,
                visibility: { type: FIELD_DATA_TYPES.NUMBER, display: 'ranges' }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: TrueLabelControlGroup,
                visibility: { type: FIELD_DATA_TYPES.BOOLEAN }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: FalseLabelControlGroup,
                visibility: { type: FIELD_DATA_TYPES.BOOLEAN }
            },
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: TimePeriodControlGroup,
                visibility: { type: FIELD_DATA_TYPES.TIMESTAMP },
                options: { menuWidth: 'narrow' }
            }
        ]
    };

    var BASE_CELL_VALUE_PANEL = {
        elementType: REPORT_ELEMENT_TYPES.CELL_VALUE,
        dataTypes: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER,
                        FIELD_DATA_TYPES.OBJECT_COUNT, FIELD_DATA_TYPES.CHILD_COUNT],
        outputType: AGGREGATION_OUTPUT_TYPES.METRIC,
        controlGroups: [
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: CellValueControlGroup,
                options: { outputType: AGGREGATION_OUTPUT_TYPES.METRIC },
                visibility: {
                    type: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER]
                }
            }
        ]
    };

    var DEPENDENT_AXIS_PANEL = extendPanelConfig(BASE_CELL_VALUE_PANEL,
        {
            elementType: REPORT_ELEMENT_TYPES.CELL_VALUE,
            dataTypes: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER,
                            FIELD_DATA_TYPES.OBJECT_COUNT, FIELD_DATA_TYPES.CHILD_COUNT],
            outputType: AGGREGATION_OUTPUT_TYPES.METRIC
        },
        function(controlGroups) {
            controlGroups = controlGroups.concat([
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: AxisScaleControlGroup,
                    options: { axisType: AxisScaleControlGroup.Y_AXIS }
                },
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: AxisIntervalControlGroup,
                    options: { axisType: AxisIntervalControlGroup.Y_AXIS }
                },
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: AxisMinValueControlGroup,
                    options: { axisType: AxisMinValueControlGroup.Y_AXIS }
                },
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: AxisMaxValueControlGroup,
                    options: { axisType: AxisMaxValueControlGroup.Y_AXIS }
                }
            ]);
            controlGroups.unshift({
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: AxisLabelControlGroup,
                options: { inputClassName: 'input-medium', xAxis: false }
            });
            return controlGroups;
        }
    );

    var SCATTER_Y_AXIS_PANEL = extendPanelConfig(DEPENDENT_AXIS_PANEL, {
        title: _('Y-Axis').t(),
        maxLength: 1,
        required: true,
        elementsSelector: function(matchingElements) {
            return matchingElements.length > 1 ? [matchingElements[1]] : [];
        },
        newElementHandler: function(newElement, isExisting) {
            if(!isExisting && newElement.get('type') === FIELD_DATA_TYPES.NUMBER) {
                newElement.set({ value: 'avg' });
            }
            return { at: 1 };
        },
        isEnabledForAdd: function(matchingElements) {
            if(matchingElements.length === 0) {
                return _('Select an X-Axis before adding Y-Axis.').t();
            }
        }
    });

    var SCATTER_X_AXIS_PANEL = extendPanelConfig(SCATTER_Y_AXIS_PANEL,
        {
            title: _('X-Axis').t(),
            elementsSelector: function(matchingElements) {
                return matchingElements.length > 0 ? [matchingElements[0]] : [];
            },
            newElementHandler: function(newElement, isExisting) {
                if(!isExisting && newElement.get('type') === FIELD_DATA_TYPES.NUMBER) {
                    newElement.set({ value: 'avg' });
                }
                return { at: 0 };
            },
            isEnabledForAdd: function() {
                // Must override isEnabledForAdd implementation in parent SCATTER_Y_AXIS_PANEL
                return false;
            }
        },
        function(controlGroups) {
            controlGroups = $.extend(true, [], controlGroups);
            var axisControlGroupViews = [
                AxisScaleControlGroup,
                AxisIntervalControlGroup,
                AxisMinValueControlGroup,
                AxisMaxValueControlGroup
            ];
            _(controlGroups).each(function(controlGroup) {
                if(_(axisControlGroupViews).contains(controlGroup.view)) {
                    controlGroup.options = $.extend({}, controlGroup.options, { axisType: controlGroup.view.X_AXIS });
                }
            });
            var labelControlGroup = _(controlGroups).find(function(controlGroup) {
                    return controlGroup.view === AxisLabelControlGroup;
                });
            labelControlGroup.options.xAxis = true;
            var rotationGroups = [
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: AxisLabelRotationControlGroup,
                    options: { axisType: AxisLabelRotationControlGroup.X_AXIS }
                },
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: AxisLabelElisionControlGroup,
                    options: { axisType: AxisLabelElisionControlGroup.X_AXIS },
                    visibility: { 'display.visualizations.charting.axisLabelsX.majorLabelStyle.rotation': ['-90', '-45', '45', '90'] }
                }
            ];
            controlGroups = controlGroups.slice(0, 1).concat(rotationGroups, controlGroups.slice(1));
            return controlGroups;
        }
    );

    var COLOR_PANEL = extendPanelConfig(INDEPENDENT_AXIS_PANEL, {
            elementType: REPORT_ELEMENT_TYPES.COLUMN_SPLIT,
            dataTypes: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER, FIELD_DATA_TYPES.BOOLEAN,
                            FIELD_DATA_TYPES.TIMESTAMP],
            title: _('Color').t(),
            shortTitle: _('Color').t()
        },
        function(controlGroups) {
            controlGroups = removeSplitSortGroups(controlGroups);
            controlGroups = removeLabelRotationGroups(controlGroups);
            controlGroups = updateSplitLimitGroups(controlGroups, {
                label: _('Max Colors').t(),
                elementType: 'column'
            });
            var labelControlGroup = _(controlGroups).find(function(controlGroup) {
                    return controlGroup.view === AxisLabelControlGroup;
                });

            var legendControlGroups = [
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: LegendPlacementControlGroup,
                    options: {
                        popdownOptions: { attachDialogTo: 'body' },
                        label: _('Legend Position').t()
                    }
                },
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: LegendTruncationControlGroup,
                    options: {
                        label: _('Legend Truncation').t()
                    }
                }
            ];
            return _.union(
                _(controlGroups).without(labelControlGroup),
                legendControlGroups
            );
        }
    );

    var SCATTER_MARK_PANEL = extendPanelConfig(INDEPENDENT_AXIS_PANEL, {
            maxLength: 1,
            title: _('Mark').t(),
            required: true,
            elementsSelector: function(matchingElements) {
                return matchingElements.length > 0 ? [matchingElements[0]] : [];
            },
            newElementHandler: function(newElement, isExisting) {
                return { at: 0 };
            }
        },
        function(controlGroups) {
            controlGroups = removeSplitSortGroups(controlGroups);
            controlGroups = removeLabelRotationGroups(controlGroups);
            return _(controlGroups).filter(function(controlGroup) {
                return controlGroup.view !== AxisLabelControlGroup;
            });
        }
    );

    var SCATTER_COLOR_PANEL = extendPanelConfig(INDEPENDENT_AXIS_PANEL, {
            title: _('Color').t(),
            maxLength: 1,
            elementsSelector: function(matchingElements) {
                return matchingElements.length > 1 ? [matchingElements[1]] : [];
            },
            isEnabledForAdd: function(matchingElements) {
                if(matchingElements.length === 0){
                    return _('Select a mark before adding color.').t();
                }
            }
        },
        function(controlGroups) {
            controlGroups = removeSplitLimitAndSortGroups(controlGroups);
            controlGroups = removeLabelRotationGroups(controlGroups);
            controlGroups = _(controlGroups).filter(function(group) {
                return group.view !== AxisLabelControlGroup;
            });
            var legendControlGroups = [
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: LegendPlacementControlGroup,
                    options: {
                        popdownOptions: { attachDialogTo: 'body' },
                        label: _('Legend Position').t()
                    }
                },
                {
                    type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                    view: LegendTruncationControlGroup,
                    options: {
                        label: _('Legend Truncation').t()
                    }
                }
            ];
            return _.union(controlGroups, legendControlGroups);
        }
    );

    var GAUGE_VALUE_PANEL = {
        title: _('Value').t(),
        elementType: REPORT_ELEMENT_TYPES.CELL_VALUE,
        dataTypes: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER,
                        FIELD_DATA_TYPES.OBJECT_COUNT, FIELD_DATA_TYPES.CHILD_COUNT],
        required: true,
        maxLength: 1,
        controlGroups: [
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: CellValueControlGroup,
                options: { outputType: AGGREGATION_OUTPUT_TYPES.METRIC },
                visibility: {
                    type: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER]
                }
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: GaugeColorRangeControls
            }
        ]
    };

    var SINGLE_VALUE_VALUE_PANEL = {
        title: _('Value').t(),
        elementType: REPORT_ELEMENT_TYPES.CELL_VALUE,
        dataTypes: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER, FIELD_DATA_TYPES.BOOLEAN,
                        FIELD_DATA_TYPES.OBJECT_COUNT, FIELD_DATA_TYPES.CHILD_COUNT],
        required: true,
        maxLength: 1,
        controlGroups: [
            {
                type: CONTROL_TYPES.REPORT_CONTROL,
                view: CellValueControlGroup,
                visibility: {
                    type: [FIELD_DATA_TYPES.STRING, FIELD_DATA_TYPES.IPV4, FIELD_DATA_TYPES.NUMBER, FIELD_DATA_TYPES.BOOLEAN]
                }
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: SingleValueBeforeLabelControlGroup
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: SingleValueAfterLabelControlGroup
            },
            {
                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                view: SingleValueUnderLabelControlGroup
            }
        ]
    };

    var SCHEMA = [
        {
            type: VIZ_TYPES.STATISTICS_TABLE,
            icon: 'icon-table',
            description: _('Statistics Table').t(),
            showConfigMenu: false,
            masterView: TableContent,
            dataOutputMode: DATA_MODES.ROW_MAJOR
        },
        {
            type: VIZ_TYPES.COLUMN_CHART,
            icon: 'icon-chart-column',
            description: _('Column Chart').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            reportLevelAttributes: { showRowSummary: 'false', showColSummary: 'false' },
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(INDEPENDENT_AXIS_PANEL, 
                    { title: _('X-Axis').t(), maxLength: 1, required: true },
                    function(controlGroups) { return updateSplitLimitGroups(controlGroups, { label: _('Max Columns').t() }); }
                ),
                extendPanelConfig(DEPENDENT_AXIS_PANEL, { title: _('Y-Axis').t(), maxLength: 1, required: true }),
                extendPanelConfig(COLOR_PANEL,
                    { maxLength: 1 },
                    function(controlGroups) {
                        var stackModeGroup = {
                            type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                            view: StackModeControlGroup
                        };
                        return _.union(controlGroups, [stackModeGroup]);
                    }
                ),
                extendPanelConfig(GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.BAR_CHART,
            icon: 'icon-chart-bar',
            description: _('Bar Chart').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            reportLevelAttributes: { showRowSummary: 'false', showColSummary: 'false' },
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(INDEPENDENT_AXIS_PANEL, {
                        title: _('X-Axis (Bars)').t(),
                        shortTitle: _('X-Axis').t(),
                        maxLength: 1,
                        required: true
                    },
                    function(controlGroups) {
                        controlGroups = _(controlGroups).filter(function(group) {
                            return group.view !== AxisLabelRotationControlGroup && group.view !== AxisLabelElisionControlGroup;
                        });
                        return updateSplitLimitGroups(controlGroups, { label: _('Max Bars').t() });
                    }
                ),
                extendPanelConfig(DEPENDENT_AXIS_PANEL, {
                    title: _('Y-Axis (Bar Width)').t(),
                    shortTitle: _('Y-Axis').t(),
                    maxLength: 1,
                    required: true
                }),
                extendPanelConfig(COLOR_PANEL,
                    { maxLength: 1 },
                    function(controlGroups) {
                        var stackModeGroup = {
                            type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                            view: StackModeControlGroup
                        };
                        return _.union(controlGroups, [stackModeGroup]);
                    }
                ),
                extendPanelConfig(GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.SCATTER_CHART,
            icon: 'icon-chart-scatter',
            description: _('Scatter Chart').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            reportLevelAttributes: { showRowSummary: 'false', rowLimitType: 'descending', rowLimitAmount: 500 },
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(SCATTER_MARK_PANEL, {}),
                extendPanelConfig(SCATTER_X_AXIS_PANEL, { }),
                extendPanelConfig(SCATTER_Y_AXIS_PANEL, {}),
                extendPanelConfig(SCATTER_COLOR_PANEL, {}),
                extendPanelConfig(GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.BUBBLE_CHART,
            icon: 'icon-chart-bubble',
            description: _('Bubble Chart').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            reportLevelAttributes: { showRowSummary: 'false', rowLimitType: 'descending', rowLimitAmount: 500 },
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(SCATTER_MARK_PANEL, {}),
                extendPanelConfig(SCATTER_X_AXIS_PANEL, { }),
                extendPanelConfig(SCATTER_Y_AXIS_PANEL, {}),
                extendPanelConfig(BASE_CELL_VALUE_PANEL,
                    {
                        title: _('Size').t(),
                        maxLength: 1,
                        required: true,
                        elementsSelector: function(matchingElements) {
                            return matchingElements.length > 2 ? [matchingElements[2]] : [];
                        },
                        newElementHandler: function(newElement, isExisting) {
                            if(!isExisting && newElement.get('type') === FIELD_DATA_TYPES.NUMBER) {
                                newElement.set({ value: 'avg' });
                            }
                            return { };
                        },
                        isEnabledForAdd: function(matchingElements) {
                            if(matchingElements.length < 2) {
                                // Only X-Axis has been selected, if at all
                                return _('Select a Y-Axis before adding size.').t();
                            }
                        }
                    },
                    function(controlGroups) {
                        return controlGroups.concat([
                            {
                                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                                view: BubbleMinSizeControlGroup
                            },
                            {
                                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                                view: BubbleMaxSizeControlGroup
                            }
                        ]);
                    }
                ),
                extendPanelConfig(SCATTER_COLOR_PANEL, {}),
                extendPanelConfig(GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.AREA_CHART,
            icon: 'icon-chart-area',
            description: _('Area Chart').t(),
            showConfigMenu: true,
            requiresIndexTime: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            reportLevelAttributes: { showRowSummary: 'false', showColSummary: 'false' },
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(INDEPENDENT_AXIS_PANEL, {
                    title: _('X-Axis (Time)').t(),
                    shortTitle: _('X-Axis').t(),
                    dataTypes: [FIELD_DATA_TYPES.TIMESTAMP],
                    maxLength: 1,
                    required: true
                }),
                extendPanelConfig(DEPENDENT_AXIS_PANEL,
                    { title: _('Y-Axis').t(), maxLength: 1, required: true },
                    function(controlGroups) {
                        var nullValueModeGroup = {
                            type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                            view: NullValueModeControlGroup
                        };

                        return _.union(controlGroups, [nullValueModeGroup]);
                    }

                ),
                extendPanelConfig(COLOR_PANEL,
                    { maxLength: 1, title: _('Color (Areas)').t() },
                    function(controlGroups) {
                        controlGroups = updateSplitLimitGroups(controlGroups, { label: _('Max Areas').t() });
                        var stackModeGroup = {
                            type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                            view: StackModeControlGroup
                        };
                        return _.union(controlGroups, [stackModeGroup]);
                    }),
                extendPanelConfig(GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.LINE_CHART,
            icon: 'icon-chart-line',
            description: _('Line Chart').t(),
            showConfigMenu: true,
            requiresIndexTime: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            reportLevelAttributes: { showRowSummary: 'false', showColSummary: 'false' },
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(INDEPENDENT_AXIS_PANEL, {
                    title: _('X-Axis (Time)').t(),
                    shortTitle: _('X-Axis').t(),
                    dataTypes: [FIELD_DATA_TYPES.TIMESTAMP],
                    maxLength: 1,
                    required: true
                }),
                extendPanelConfig(DEPENDENT_AXIS_PANEL,
                    { title: _('Y-Axis').t(), maxLength: 1, required: true },
                    function(controlGroups) {
                        var nullValueModeGroup = {
                            type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                            view: NullValueModeControlGroup
                        };
                        return _.union(controlGroups, [nullValueModeGroup]);
                    }

                ),
                extendPanelConfig(COLOR_PANEL, 
                    { maxLength: 1, title: _('Color (Lines)').t() },
                    function(controlGroups) { return updateSplitLimitGroups(controlGroups, { label: _('Max Lines').t() }); }
                ),
                extendPanelConfig(GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.PIE_CHART,
            icon: 'icon-chart-pie',
            description: _('Pie Chart').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            reportLevelAttributes: { showRowSummary: 'false' },
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                // naming got a little weird here, pie is using the independent axis panel for color
                extendPanelConfig(INDEPENDENT_AXIS_PANEL,
                    { title: _('Color').t(), maxLength: 1, required: true },
                    function(controlGroups) {
                        controlGroups = removeLabelRotationGroups(controlGroups);
                        var axisLabelControlGroup = _(controlGroups).find(function(controlGroup) {
                                return controlGroup.view === AxisLabelControlGroup;
                            }),
                            labelControlGroup = {
                                type: CONTROL_TYPES.REPORT_CONTROL,
                                view: LabelControlGroup
                            };

                        return _.union([labelControlGroup], _(controlGroups).without(axisLabelControlGroup));
                    }
                ),
                extendPanelConfig(DEPENDENT_AXIS_PANEL,
                    { title: _('Size').t(), maxLength: 1, required: true },
                    function(controlGroups) {
                        var axisLabelControlGroup = _(controlGroups).find(function(controlGroup) {
                                return controlGroup.view === AxisLabelControlGroup;
                            }),
                            axisScaleControlGroup = _(controlGroups).find(function(controlGroup) {
                                return controlGroup.view === AxisScaleControlGroup;
                            }),
                            axisInternalControlGroup = _(controlGroups).find(function(controlGroup) {
                                return controlGroup.view === AxisIntervalControlGroup;
                            }),
                            axisMinValueControlGroup = _(controlGroups).find(function(controlGroup) {
                                return controlGroup.view === AxisMinValueControlGroup;
                            }),
                            axisMaxValueControlGroup = _(controlGroups).find(function(controlGroup) {
                                return controlGroup.view === AxisMaxValueControlGroup;
                            }),
                            labelControlGroup = {
                                type: CONTROL_TYPES.REPORT_CONTROL,
                                view: LabelControlGroup
                            },
                            sliceMinimumSizeControlGroup = {
                                type: CONTROL_TYPES.VISUALIZATION_CONTROL,
                                view: SliceMinimumSizeControlGroup
                            };

                        return _.union(
                            [labelControlGroup],
                            _(controlGroups).without(axisLabelControlGroup, axisScaleControlGroup, axisInternalControlGroup, axisMinValueControlGroup, axisMaxValueControlGroup),
                            [sliceMinimumSizeControlGroup]
                        );
                    }
                ),
                extendPanelConfig(GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.SINGLE_VALUE,
            icon: 'icon-single-value',
            description: _('Single Value Display').t(),
            showConfigMenu: true,
            masterView: SingleValueContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(SINGLE_VALUE_VALUE_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.RADIAL_GAUGE,
            icon: 'icon-gauge-radial',
            description: _('Radial Gauge').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(GAUGE_VALUE_PANEL, {}),
                extendPanelConfig(GAUGE_GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.MARKER_GAUGE,
            icon: 'icon-gauge-marker',
            description: _('Marker Gauge').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(GAUGE_VALUE_PANEL, {}),
                extendPanelConfig(GAUGE_GENERAL_PANEL, {})
            ]
        },
        {
            type: VIZ_TYPES.FILLER_GAUGE,
            icon: 'icon-gauge-filler',
            description: _('Filler Gauge').t(),
            showConfigMenu: true,
            masterView: ChartContent,
            dataOutputMode: DATA_MODES.COLUMN_MAJOR,
            configMenuPanels: [
                extendPanelConfig(TIME_RANGE_PANEL, {}),
                extendPanelConfig(FILTER_PANEL, {}),
                extendPanelConfig(GAUGE_VALUE_PANEL, {}),
                extendPanelConfig(GAUGE_GENERAL_PANEL, {})
            ]
        }
    ];

    var PivotVizManager = function() {
        this.schema = this.normalizeSchema(SCHEMA);
        // add all constants to the singleton instance so they are available to consumers
        $.extend(this, VIZ_TYPES, DATA_MODES, CONTROL_TYPES, FIELD_DATA_TYPES, REPORT_ELEMENT_TYPES, AGGREGATION_OUTPUT_TYPES);
    };

    $.extend(PivotVizManager.prototype, {

        normalizeSchema: function(schema) {
            var arrayify = function(stringOrArray) {
                return _(stringOrArray).isArray() ? stringOrArray : [stringOrArray];
            };

            _(schema).each(function(vizConfig) {
                if(!vizConfig.hasOwnProperty('requiresIndexTime')) {
                    vizConfig.requiresIndexTime = false;
                }
                if(vizConfig.hasOwnProperty('configMenuPanels')) {
                    _(vizConfig.configMenuPanels).each(function(panelConfig) {
                        // if there's no shortTitle, set it to the title
                        if(!panelConfig.hasOwnProperty('shortTitle')) {
                            panelConfig.shortTitle = panelConfig.title || '';
                        }
                        // give the panel an id based on its shortTitle, can be used to provide hooks for automated testing
                        panelConfig.id = panelConfig.shortTitle.replace(/\W/g, '').toLowerCase();
                        if(panelConfig.hasOwnProperty('dataTypes')) {
                            panelConfig.dataTypes = arrayify(panelConfig.dataTypes);
                        }
                        _(panelConfig.controlGroups).each(function(controlGroup) {
                            _(controlGroup.visibility || {}).each(function(allowedValues, attribute) {
                                controlGroup.visibility[attribute] = arrayify(allowedValues);
                            });
                            // since the viz config menu uses an "auto-update" interaction,
                            // we want all controls to validate and force updates when they set
                            controlGroup.options = $.extend({ validate: true, forceUpdate: true }, controlGroup.options);
                        });
                    });
                }
            });
            return schema;
        },

        getConfigByVizType: function(vizType) {
            return _(this.schema).find(function(configObject) { return configObject.type === vizType; });
        }

    });

    return new PivotVizManager();

});
