define(
	[
		'jquery',
		'underscore',
		'models/shared/managementconsole/DistsearchGroup'
	],
	function(
		$,
		_,
		DistsearchGroupModel
	) {
		var VALID_PRIMARY_ROLES = [
			'search_head',
			'cluster_master',
			'license_master',
			'indexer',
			'deployment_server',
			'kv_store'
		];

		var getAllPrimaryRoles = function() {
			return VALID_PRIMARY_ROLES.slice();
		};

		return {
			distributedSearchGroupsCollection: null,
			getServerName: null,

			VALID_PRIMARY_ROLES: VALID_PRIMARY_ROLES,

			containsRole: function(partialRole) {
				// joins two group of roles into single array
				var serverRolesArr = this.getActiveServerRolesToString();
				var customRolesArr = this.entry.content.get('tags');
				var allRolesArr = _.union([], serverRolesArr, customRolesArr);

				// if the filter input matches even a single roles, return true
				return _.reduce(allRolesArr, function(memo, role) {
					return memo || role.toLowerCase().indexOf(partialRole.toLowerCase()) !== -1;
				}, false);
			},

			getServerRoleToString : function(roleId) {
				if (roleId === 'kv_store') {
					return 'KV Store';
				}
				return roleId.replace(/_/g, ' ').replace(
					/\b([a-z])/g, 
					function(letter) {
						return letter.toUpperCase();
					}
				);
			},

			getRoleGroupName: function(roleId) {
				return DistsearchGroupModel.GROUP_NAME_PREFIX + roleId;
			},
			getRoleFromGroupName: function(groupName) {
				return groupName.substr(DistsearchGroupModel.GROUP_NAME_PREFIX.length);
			},
			getCustomGroupNameFromGroupName: function(groupName) {
				return groupName.substr( DistsearchGroupModel.CUSTOM_GROUP_NAME_PREFIX.length);
			},
			getCustomGroupName: function(tagName) {
				return  DistsearchGroupModel.CUSTOM_GROUP_NAME_PREFIX + tagName;
			},

			initializeDistributedSearchGroups: function(
				distributedSearchGroupsCollection,
				getServerName
			) {
				this.distributedSearchGroupsCollection = distributedSearchGroupsCollection || this.distributedSearchGroupsCollection;
				this.getServerName = getServerName || this.getServerName;

				if (!this.distributedSearchGroupsCollection) {
					throw Error('Distributed search groups mixin requires a collection argument');
				}
				if (!this.getServerName) {
					throw Error('Distributed search groups mixin requires a getServerName argument');
				}

				// Setup 'active_server_roles' so it has the following features:
				// 1. It is the same as whatever was read from the stored DistsearchGroups
				var storedRoles = this._serverRolesFromGroups();
				var updateGroups = false;
				if (storedRoles.length === 0 && this.entry.content.get('status-toggle') !== 'Disabled') {
					storedRoles = this._serverRolesFromInternalRoles();
					updateGroups = storedRoles.length > 0;
				}

				this.entry.content.set('active_server_roles', storedRoles);

				if (updateGroups) {
					this._updateDistributedSearchGroups();
				}
				// 2. Whenever active_server_roles is changed on this model, change it on the DistsearchGroups
				// This way, the View API has no idea DistsearchGroups are involved
				// Also, we can always call .save() immediately on the DistsearchGroups since
				// they will be in sync
				this.entry.content.on('change:active_server_roles', this._updateDistributedSearchGroups, this);


				var storedCustomGroups = this._customGroupsFromGroups();
				this.entry.content.set('tags', storedCustomGroups);
				this.entry.content.on('change:tags', this._updateCustomSearchGroups, this);

				this.entry.content.validate = function(attrs) {
					if (!_.isUndefined(attrs.active_server_roles) &&
						(!_.isArray(attrs.active_server_roles) ||
						 attrs.active_server_roles.length === 0)) {
					
						return _("You must provide at least one server role.").t();
					}

					if (!_.isUndefined(attrs.tags)) {
						if (_.find(attrs.tags, function(name) { return /[^a-zA-Z0-9_\-]/.test(name); })) {
							return _("Group names must only contain alphanumeric characters, dashes, or underscores.").t();
						}
					}

					if (!_.isUndefined(attrs.host) &&
						(!_.isString(attrs.host) ||
						  $.trim(attrs.host) === '')) {

						return _("Please enter a valid instance name.").t();
					}

					if (!_.isUndefined(attrs.host_fqdn) &&
						(!_.isString(attrs.host_fqdn) ||
						  $.trim(attrs.host_fqdn) === '')) {

						return _("Please enter a valid machine name.").t();
					}
				}.bind(this);

			},

			getActiveServerRolesToString: function() {
				return _.map(
					this.entry.content.get('active_server_roles'),
					this.getServerRoleToString
				);
			},

            // resets the instance's internal server roles
            // called when and instance goes from 'Disabled' -> 'Enabled'
            resetInitialInternalServerRoles: function() {
				var storedRoles = this._serverRolesFromInternalRoles();
				this.entry.content.set('active_server_roles', storedRoles);
            },

			_serverRolesFromGroups: function() {
				return this.distributedSearchGroupsCollection.filter(function(group) {
					return !group.isCustomGroup() && _.contains(group.getServers(), this.getServerName());
				}, this).map(function(group) {
					return this.getRoleFromGroupName(group.getGroupName());
				}, this);
			},

			_serverRolesFromInternalRoles: function() {
				// In the case of local instance, server_roles is an object
				// Otherwise, it's an array
				var internalRoles = this.entry.content.get('server_roles');

				if (!_.isArray(internalRoles)) {
					internalRoles = _.keys(internalRoles);
				}

				var internalPrimaryRoles = _.intersection(
					internalRoles,
					getAllPrimaryRoles()
				);

				return _.union(internalPrimaryRoles, this.requiredRoles || []);
			},


			_customGroupsFromGroups: function() {
				return this.distributedSearchGroupsCollection.filter(function(group) {
					return group.isCustomGroup() &&
						_.contains(group.getServers(), this.getServerName());
				}, this).map(function(group) {
					return this.getCustomGroupNameFromGroupName(group.getGroupName());
				}, this);

			},


			_updateDistributedSearchGroups: function() {
				_.each(VALID_PRIMARY_ROLES, function(primaryRole) {
					var roleGroup = this.distributedSearchGroupsCollection.find(function(group) {
						return this.getRoleFromGroupName(group.getGroupName()) === primaryRole;
					}, this);
					
					if (!roleGroup) {
						roleGroup = new DistsearchGroupModel();
						roleGroup.setName(this.getRoleGroupName(primaryRole));
						roleGroup.setServers([]);
						roleGroup.setSearchMode(primaryRole === 'indexer' ? 'default' : 'light');
						roleGroup.setDefault(primaryRole === 'indexer');
						this.distributedSearchGroupsCollection.add(roleGroup);
					}
				}, this);

				// Ensure all the groups are now in sync with the latest change to this model's role
				_(this.distributedSearchGroupsCollection.filter(function(group) {
					return !group.isCustomGroup();
				}, this)).each(function(distsearch) {
					var groupName = distsearch.getGroupName();
					// The role is encoded in the name. 
					var groupRole = groupName.substr(DistsearchGroupModel.GROUP_NAME_PREFIX.length);

					// If this peer is now in this group's role, add it
					if (_.contains(this.entry.content.get('active_server_roles'), groupRole)) {
						distsearch.addServer(this.getServerName());

					// Otherwise, it's not in this group. Remove it.
					} else {
						distsearch.removeServer(this.getServerName());
					}
				}, this);
			},

			_updateCustomSearchGroups: function() {
				// TODO you should really investigate whether you could modify the groups attributes only on the
				// actual peer endpoint
				// Additionally, if you can't, make a "flush" mechanism that tallies up these "dirty" groups
				// and only saves them off

				var customGroups = _(this.distributedSearchGroupsCollection.filter(function(group) { 
					return group.isCustomGroup(); 
				}));

				_.each(this.entry.content.get('tags'), function(tagName) {
					var customGroup = customGroups.find(function(group) {
						return this.getCustomGroupNameFromGroupName(group.getGroupName()) === tagName;
					}, this);

					if (!customGroup) {
						customGroup = new DistsearchGroupModel();
						customGroup.setName(this.getCustomGroupName(tagName));
						customGroup.setServers([]);
						customGroup.setDefault(false);
						this.distributedSearchGroupsCollection.add(customGroup);
					}
				}, this);
				customGroups = _(this.distributedSearchGroupsCollection.filter(function(group) { 
					return group.isCustomGroup(); 
				}));

				customGroups.each(function(group) {
					var groupName = group.getGroupName();
					var customGroupName = this.getCustomGroupNameFromGroupName(groupName);

					if (_.contains(this.entry.content.get('tags'), customGroupName)) {
						group.addServer(this.getServerName());
					} else {
						group.removeServer(this.getServerName());
					}
				}, this);
			},
			staticMembers: {
				// TODO actual implementation
				getAllPrimaryRoles: getAllPrimaryRoles
			}


		};

	}
);
