define(
    [
        'jquery',
        'underscore',
        'models/Base'
    ],
    function($, _, BaseModel) {
        return BaseModel.extend({
            initialize: function() {
                BaseModel.prototype.initialize.apply(this, arguments);
            },
            defaults: {
                'charting.axisY.scale': 'linear',
                'charting.axisX.scale': 'linear',
                'charting.axisX.minimumNumber': '',
                'charting.axisX.maximumNumber': '',
                'charting.axisY.minimumNumber': '',
                'charting.axisY.maximumNumber': '',
                'charting.axisTitleX.text': '',
                'charting.axisTitleY.text': '',
                'charting.axisLabelsX.majorUnit': '',
                'charting.axisLabelsY.majorUnit': '',
                'charting.legend.placement': 'right',
                'charting.legend.labelStyle.overflowMode': 'ellipsisMiddle',
                'charting.chart.stackMode':	'default',
                'charting.chart.nullValueMode': 'zero',
                'charting.chart.rangeValues': '["0","30","70","100"]',
                'charting.chart.style': 'shiny',
                'charting.gaugeColors': '["84E900", "FFE800", "BF3030"]'
            },
            validation: {
                'charting.axisX.minimumNumber': [
                    {
                        pattern: 'number',
                        msg: _('Please enter a number').t(),
                        required: false
                    },
                    {
                        fn: 'validateXAxisExtremes',
                        required: false
                    }
                ],
                'charting.axisX.maximumNumber': [
                    {
                        pattern: 'number',
                        msg: _('Please enter a number').t(),
                        required: false
                    },
                    {
                        fn: 'validateXAxisExtremes',
                        required: false
                    }
                ],
                'charting.axisY.minimumNumber': [
                    {
                        pattern: 'number',
                        msg: _('Please enter a number').t(),
                        required: false
                    },
                    {
                        fn: 'validateYAxisExtremes',
                        required: false
                    }
                ],
                'charting.axisY.maximumNumber': [
                    {
                        pattern: 'number',
                        msg: _('Please enter a number').t(),
                        required: false
                    },
                    {
                        fn: 'validateYAxisExtremes',
                        required: false
                    }
                ],
                'charting.axisLabelsX.majorUnit': {
                    pattern: 'number',
                    min: Number.MIN_VALUE,
                    msg: _('Please enter a positive number').t(),
                    required: false
                },
                'charting.axisLabelsY.majorUnit': {
                    pattern: 'number',
                    min: Number.MIN_VALUE,
                    msg: _('Please enter a positive number').t(),
                    required: false
                },
                'charting.axisY.scale': {
                    fn: 'validateYScaleAndStacking',
                    required: false
                },
                'charting.chart.stackMode': {
                    fn: 'validateYScaleAndStacking',
                    required: false
                },
                'charting.chart.rangeValues': {
                    fn: 'validateRangeValues',
                    required: false
                }
            },

            validateXAxisExtremes: function(value, attr, computedState) {
                var min = parseFloat(computedState['charting.axisX.minimumNumber']),
                    max = parseFloat(computedState['charting.axisX.maximumNumber']);

                if(!_.isNaN(min) && !_.isNaN(max) && max <= min) {
                    return _('The minimum value must be less than maximum value').t();
                }
            },

            validateYAxisExtremes: function(value, attr, computedState) {
                var min = parseFloat(computedState['charting.axisY.minimumNumber']),
                    max = parseFloat(computedState['charting.axisY.maximumNumber']);

                if(!_.isNaN(min) && !_.isNaN(max) && max <= min) {
                    return _('The minimum value must be less than maximum value').t();
                }
            },

            validateYScaleAndStacking: function(value, attr, computedState) {
                var yAxisScale = computedState['charting.axisY.scale'],
                    stackMode = computedState['charting.chart.stackMode'];

                if(yAxisScale === 'log' && stackMode !== 'default') {
                    return _('Log scale and stacking cannot be enabled at the same time').t();
                }
            },

            validateRangeValues: function(value) {
                var ranges = _(value ? JSON.parse(value) : []).map(parseFloat);
                if(_(ranges).any(_.isNaN)) {
                    return _('All color ranges must be valid numbers').t();
                }

                var dedupedRanges = _.uniq(ranges),
                    sortedRanges = $.extend([], ranges).sort(function(a, b) { return a - b; });

                if(!_.isEqual(ranges, dedupedRanges) || !_.isEqual(ranges, sortedRanges)) {
                    return _('Color ranges must be entered from lowest to highest').t();
                }
            },

            attrToArray: function(attr) {
                var value = this.get(attr);
                if(!value){
                    return [];
                }
                return _.values(JSON.parse(value));
            },

            rangesValuesToArray: function() {
                return this.attrToArray('charting.chart.rangeValues');
            },

            gaugeColorsToArray: function() {
                return this.attrToArray('charting.gaugeColors');
            }
        },
        {
            CHARTING_ATTRS_FILTER: ['^charting\..*'],
            SINGLE_VALUE_ATTRS_FILTER: ['^singlevalue\..*']
        });
    }
);
