define(
    [
        'jquery',
        'underscore',
        'models/search/Report',
        'models/shared/Cron',
        'models/shared/TimeRange',
        'util/time',
        'util/splunkd_utils',
        'splunk.util',
        'util/math_utils'
    ],
    function(
        $,
        _,
        ReportModel,
        CronModel,
        TimeRangeModel,
        time_utils,
        splunkd_utils,
        splunkUtil,
        mathUtils
    ){
        return ReportModel.extend({
            initialize: function() {
                ReportModel.prototype.initialize.apply(this, arguments);

                this.cron = new CronModel();
                this.associated.cron = this.cron;

                this.workingTimeRange = new TimeRangeModel({enableRealTime:false});
                this.associated.workingTimeRange = this.workingTimeRange;
            },
            defaults: {
                'ui.name': void(0),
                'ui.type': 'scheduled',
                'ui.realtime.triggercondition': 'per_result',
                'ui.realtime.resultscomparator': 'greater than',
                'ui.realtime.resultstimeunit': 'm',
                'ui.realtime.resultscomparatorinput': '0',
                'ui.realtime.resultstime': '1',
                'ui.realtime.customsearch': void(0),
                'ui.scheduled.triggercondition': 'events',
                'ui.scheduled.resultscomparator': 'greater than',
                'ui.scheduled.resultsinput': '0',
                'ui.permissions': splunkd_utils.USER,
                'ui.executeactions': true,
                'ui.severity': 2,
                'ui.supresstime': 60,
                'ui.supresstimeunit': 's',
                'ui.include_results_link': 1,
                'ui.include_view_link': 1,
                'ui.include_search': 0,
                'ui.include_trigger': 0,
                'ui.include_trigger_time': 0,
                'ui.include_results_pdf': 0,
                'ui.include_results_csv': 0,
                'ui.include_results_inline': 0,
                'ui.include_results_inline_format': 'table'
            },
            /*
             * This is working validaiton code which will be uncommented when i come back to it post SKO.
             * NOTE --> for this to function we need to flip the switch and make controls opt-out of
             *          validaiton by default.  In testing I had flipped that switch locally.
             */
            validation: {
            //step: Trigger page
                'ui.name': {
                    required: true,
                    msg: _('Title is required.').t()
                },
                'ui.type': {
                    required: true
                },
                'ui.realtime.resultscomparatorinput': {
                    fn: 'validateRealtimeResultsComparatorInput'
                 },
                'ui.realtime.resultstime': {
                    fn: 'validateRealtimeResultsTime'
                },
                'ui.realtime.customsearch': {
                    fn: 'validateRealtimeCustomSearch'
                },
                'ui.scheduled.resultsinput': {
                    fn: 'validateScheduledResultsInput'
                },
                'ui.scheduled.customsearch': {
                    fn: 'validateScheduledCustomSearch'
                },
                'ui.showtriggeredalerts': {
                    fn: 'validateActions'
                },
            //step: Actions page
                'ui.supresstime': {
                    fn: 'validateSuppressTime'
                },
                'ui.fieldvalue': {
                    fn: 'validateFieldValue'
                },
                'ui.scriptfilename': {
                    fn: 'validateScriptFilename'
                }

            },
            validateRealtimeResultsComparatorInput: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'triggers' &&
                    computedState['ui.type'] === 'realtime' &&
                    (computedState['ui.realtime.triggercondition'] === 'events' ||
                        computedState['ui.realtime.triggercondition'] === 'hosts' ||
                        computedState['ui.realtime.triggercondition'] === 'sources') &&
                    _.isNaN(mathUtils.strictParseFloat(value))) {

                    return _('Trigger threshold must be a number.').t();
                }
            },
            validateRealtimeResultsTime: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'triggers' &&
                    computedState['ui.type'] === 'realtime' &&
                    (computedState['ui.realtime.triggercondition'] === 'events' ||
                        computedState['ui.realtime.triggercondition'] === 'hosts' ||
                        computedState['ui.realtime.triggercondition'] === 'sources' ||
                        computedState['ui.realtime.triggercondition'] === 'custom') &&
                    (_.isNaN(parseFloat(value)) || parseFloat(value) <= 0)) {

                    return _('Trigger time range must be a number greater than 0.').t();
                }
            },
            validateRealtimeCustomSearch: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'triggers' &&
                    computedState['ui.type'] === 'realtime' &&
                    computedState['ui.realtime.triggercondition'] === 'custom' &&
                    (_.isUndefined(value) || $.trim(value).length === 0)) {

                    return _('Custom condition is required.').t();
                }
            },
            validateScheduledResultsInput: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'triggers' &&
                    computedState['ui.type'] === 'scheduled' &&
                    (computedState['ui.scheduled.triggercondition'] === 'events' ||
                        computedState['ui.scheduled.triggercondition'] === 'hosts' ||
                        computedState['ui.scheduled.triggercondition'] === 'sources') &&
                    _.isNaN(parseFloat(value))) {

                    return _('Trigger threshold must be a number.').t();
                }
            },
            validateScheduledCustomSearch: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'triggers' &&
                    computedState['ui.type'] === 'scheduled' &&
                    computedState['ui.scheduled.triggercondition'] === 'custom' &&
                    (_.isUndefined(value) || $.trim(value).length === 0)) {

                    return _('Custom condition is required.').t();
                }
            },
            validateActions: function(value, attr, computedState) {
                var email = computedState['ui.email'],
                    script = computedState['ui.script'];
                if (computedState['ui.step'] === 'actions' &&
                    (!value && !email && !script)) {

                    return _('Enable at least one action.').t();
                }
            },
            validateSuppressTime: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'actions' &&
                    computedState['ui.throttle'] &&
                    (!mathUtils.isInteger(value) || value <= 0)) {

                    return _('Throttle suppression time range must be a integer greater than 0.').t();
                }
            },
            validateFieldValue: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'actions' &&
                    computedState['ui.throttle'] &&
                    ((!computedState['ui.executeactions'] ||
                        (computedState['ui.type'] === 'realtime' &&
                            computedState['ui.realtime.triggercondition'] === 'per_result')) &&
                    (_.isUndefined(value) || $.trim(value).length === 0))) {

                    return _('Per result alert throttling requires at least one throttling field; use * to throttle on all fields.').t();
                }
            },
            validateScriptFilename: function(value, attr, computedState) {
                if (computedState['ui.step'] === 'actions' &&
                    computedState['ui.script'] &&
                    (_.isUndefined(value) || $.trim(value).length === 0)) {

                    return _('A filename is required if script action is enabled').t();
                }
            }, 
            transposeToSavedsearch: function() {
                /*.:. STEP1 ...............................................................................................
                 *
                 * Branch based on the alert type: realtime | scheduled
                 */
                this.entry.content.set({
                    'name': this.get('ui.name'),
                    'description': this.get('ui.description'),
                    'is_scheduled': 1
                });

                switch(this.get('ui.type')){
                    case 'realtime':
                        if (this.get('ui.realtime.triggercondition') != 'per_result'){
                            if(this.get('ui.realtime.resultstime')){
                                this.entry.content.set({
                                    'dispatch.earliest_time': 'rt-' + this.get('ui.realtime.resultstime') + '' + this.get('ui.realtime.resultstimeunit'),
                                    'dispatch.latest_time': 'rt-0'+this.get('ui.realtime.resultstimeunit'),
                                    'cron_schedule': '* * * * *'
                                });
                            }
                        } else {
                            this.entry.content.set({
                                'dispatch.earliest_time': 'rt',
                                'dispatch.latest_time': 'rt',
                                'cron_schedule': '* * * * *'
                            });
                        }

                        switch(this.get('ui.realtime.triggercondition')){
                            case 'per_result':
                                this.entry.content.set('alert_type', 'always');
                                break;
                            case 'events':
                                this.entry.content.set({
                                    'alert_type': 'number of events',
                                    'alert_comparator': this.get('ui.realtime.resultscomparator'),
                                    'alert_threshold': this.get('ui.realtime.resultscomparatorinput')
                                });
                                break;
                            case 'hosts':
                                this.entry.content.set({
                                    'alert_type': 'number of hosts',
                                    'alert_comparator': this.get('ui.realtime.resultscomparator'),
                                    'alert_threshold': this.get('ui.realtime.resultscomparatorinput')
                                });
                                break;
                            case 'sources':
                                this.entry.content.set({
                                    'alert_type': 'number of sources',
                                    'alert_comparator': this.get('ui.realtime.resultscomparator'),
                                    'alert_threshold': this.get('ui.realtime.resultscomparatorinput')
                                });
                                break;
                            case 'custom':
                                this.entry.content.set({
                                    'alert_type': 'custom',
                                    'alert_condition': this.get('ui.realtime.customsearch')
                                });
                                break;
                        }
                        break;
                    case 'scheduled':
                        switch(this.get('ui.scheduled.triggercondition')){
                            case 'events':
                                this.entry.content.set({
                                    'alert_type': 'number of events',
                                    'alert_comparator': this.get('ui.scheduled.resultscomparator'),
                                    'alert_threshold': this.get('ui.scheduled.resultsinput')
                                });
                                break;
                            case 'hosts':
                                this.entry.content.set({
                                    'alert_type': 'number of hosts',
                                    'alert_comparator': this.get('ui.scheduled.resultscomparator'),
                                    'alert_threshold': this.get('ui.scheduled.resultsinput')
                                });
                                break;
                            case 'sources':
                                this.entry.content.set({
                                    'alert_type': 'number of sources',
                                    'alert_comparator': this.get('ui.scheduled.resultscomparator'),
                                    'alert_threshold': this.get('ui.scheduled.resultsinput')
                                });
                                break;
                            case 'custom':
                                this.entry.content.set({
                                    'alert_type': 'custom',
                                    'alert_condition': this.get('ui.scheduled.customsearch')
                                });
                                break;
                        }
                        switch(this.cron.get('cronType')){
                            case 'hourly':
                                this.entry.content.set({
                                    'dispatch.earliest_time': '-1h',
                                    'dispatch.latest_time': 'now'
                                });
                                break;
                            case 'daily':
                                this.entry.content.set({
                                    'dispatch.earliest_time': '-1d',
                                    'dispatch.latest_time': 'now'
                                });
                                break;
                            case 'weekly':
                                this.entry.content.set({
                                    'dispatch.earliest_time': '-1w',
                                    'dispatch.latest_time': 'now'
                                });
                                break;
                            case 'monthly':
                                this.entry.content.set({
                                    'dispatch.earliest_time': '-1mon',
                                    'dispatch.latest_time': 'now'
                                });
                                break;
                            case 'custom':
                                this.entry.content.set({
                                    'dispatch.earliest_time': this.workingTimeRange.get('earliest'),
                                    'dispatch.latest_time': this.workingTimeRange.get('latest')
                                });
                                break;
                        }
                        this.entry.content.set({'cron_schedule': this.cron.getCronString()});
                        break;
                }
                //.:. end step1............................................................................................

                /*.:. STEP2 ...............................................................................................
                 *
                 * Step2 is, for the most part, independent of step1.  We do not explicitly define attributes that will
                 * be handled by SplunkD (see comments)
                 *
                 */

                 //holds list of actions
                 var actions = [];

                if(this.get('ui.showtriggeredalerts')){
                    this.entry.content.set({
                        'alert.track': 1,
                        'alert.severity': this.get('ui.severity')
                    });
                } else {
                    this.entry.content.set({
                        'alert.track': 0
                    });
                }

                if(this.has('ui.executeactions') && this.get('ui.executeactions') && !(this.get('ui.type') === 'realtime' && this.get('ui.realtime.triggercondition') === 'per_result')){
                    this.entry.content.set({
                        'alert.digest_mode' : 1
                    });
                } else {
                    this.entry.content.set({
                        'alert.digest_mode' : 0
                    });
                }

                if(this.get('ui.throttle') == 1){
                    this.entry.content.set({
                        'alert.suppress': 1,
                        'alert.suppress.period': this.get('ui.supresstime') + this.get('ui.supresstimeunit')
                    });
                    if(((this.get('ui.type') == 'realtime') && (this.get('ui.realtime.triggercondition') == 'per_result')) || !this.get('ui.executeactions')) {
                        this.entry.content.set('alert.suppress.fields', this.get('ui.fieldvalue'));
                    } else {
                        this.entry.content.set('alert.suppress.fields', '');
                    }
                } else {
                    this.entry.content.set('alert.suppress', 0);
                }

                if(this.get('ui.email')){
                    var inlineResults = this.get('ui.include_results_inline') ? 1 : 0,
                        csvResults = this.get('ui.include_results_csv') ? 1 : 0,
                        pdfResults = this.get('ui.include_results_pdf') ? 1 : 0;

                    // add to actions list
                    actions.push('email');

                    this.entry.content.set({
                        'action.email': 1,
                        'action.email.subject': this.get('action.email.subject'),
                        'action.email.to': this.get('action.email.to'),
                        'action.email.cc': this.get('action.email.cc'),
                        'action.email.bcc': this.get('action.email.bcc'),
                        'action.email.message.alert': this.get('action.email.message.alert'),
                        'action.email.subject.alert': this.get('action.email.subject.alert'),
                        'action.email.priority': this.get('action.email.priority'),
                        'action.email.include.results_link': this.get('ui.include_results_link') ? 1 : 0,
                        'action.email.include.view_link': this.get('ui.include_view_link') ? 1 : 0,
                        'action.email.include.search': this.get('ui.include_search') ? 1 : 0,
                        'action.email.include.trigger': this.get('ui.include_trigger') ? 1 : 0,
                        'action.email.include.trigger_time': this.get('ui.include_trigger_time') ? 1 : 0,
                        'action.email.sendresults': +(inlineResults || csvResults || pdfResults),
                        'action.email.sendpdf': pdfResults,
                        'action.email.sendcsv': csvResults,
                        'action.email.inline': inlineResults,
                        'action.email.format': this.get('ui.include_results_inline_format')
                    });
                } else {
                    this.entry.content.set({
                        'action.email' : 0
                    });
                }

                if(this.get('ui.script')){
                    // add to actions list
                    actions.push('script');
                    this.entry.content.set({
                        'action.script': 1,
                        'action.script.filename': this.get('ui.scriptfilename')
                    });
                } else {
                    this.entry.content.set({
                        'action.script': 0
                    });
                }

                // set actions
                this.entry.content.set('actions', actions.join(', '));

                //.:. end step2...............................................................................................
            },
            /**
             * @param model:  <models.Report> type Alert,
             */
            transposeFromSavedsearchAlert: function(model) {
            /*.:. STEP1 ...............................................................................................
             *
             * Branch based on the alert type: realtime | scheduled
             */
             var triggercondition = model.entry.content.get('alert_type'),
                earliestTime = model.entry.content.get('dispatch.earliest_time'),
                latestTime = model.entry.content.get('dispatch.latest_time'),
                type = model.isRealTime() ? 'realtime' : 'scheduled',
                attr = {
                    'ui.name': model.entry.get('name'),
                    'ui.description': model.entry.content.get('description'),
                    'ui.type': type
                };

                this.workingTimeRange.set({
                    'earliest': earliestTime,
                    'latest': latestTime
                });

                this.entry.content.set('cron_schedule', model.entry.content.get('cron_schedule'));
             switch(type){
                case 'realtime' :
                    switch(triggercondition) {
                        case 'always' :
                            attr['ui.realtime.triggercondition'] = 'per_result';
                            break;
                        case 'number of events' :
                            attr['ui.realtime.triggercondition'] = 'events';
                            attr['ui.realtime.resultscomparator'] = model.entry.content.get('alert_comparator');
                            attr['ui.realtime.resultscomparatorinput'] = model.entry.content.get('alert_threshold');
                            attr['ui.realtime.resultstime'] = earliestTime.match(/\d+/)[0];
                            attr['ui.realtime.resultstimeunit'] = _.indexOf(['d', 'h', 'm'], earliestTime.match(/[a-z]$/)[0]) === -1 ? 'm' : earliestTime.match(/[a-z]$/)[0];
                            break;
                        case 'number of hosts' :
                            attr['ui.realtime.triggercondition'] = 'hosts';
                            attr['ui.realtime.resultscomparator'] = model.entry.content.get('alert_comparator');
                            attr['ui.realtime.resultscomparatorinput'] = model.entry.content.get('alert_threshold');
                            attr['ui.realtime.resultstime'] = earliestTime.match(/\d+/)[0];
                            attr['ui.realtime.resultstimeunit'] = _.indexOf(['d', 'h', 'm'], earliestTime.match(/[a-z]$/)[0]) === -1 ? 'm' : earliestTime.match(/[a-z]$/)[0];
                            break;
                        case 'number of sources' :
                            attr['ui.realtime.triggercondition'] = 'sources';
                            attr['ui.realtime.resultscomparator'] = model.entry.content.get('alert_comparator');
                            attr['ui.realtime.resultscomparatorinput'] = model.entry.content.get('alert_threshold');
                            attr['ui.realtime.resultstime'] = earliestTime.match(/\d+/)[0];
                            attr['ui.realtime.resultstimeunit'] = _.indexOf(['d', 'h', 'm'], earliestTime.match(/[a-z]$/)[0]) === -1 ? 'm' : earliestTime.match(/[a-z]$/)[0];
                            break;
                        case 'custom' :
                            attr['ui.realtime.triggercondition'] = 'custom';
                            attr['ui.realtime.customsearch'] = model.entry.content.get('alert_condition');
                            attr['ui.realtime.resultstime'] = earliestTime.match(/\d+/)[0];
                            attr['ui.realtime.resultstimeunit'] = _.indexOf(['d', 'h', 'm'], earliestTime.match(/[a-z]$/)[0]) === -1 ? 'm' : earliestTime.match(/[a-z]$/)[0];
                            break;
                    }
                    break;
                case 'scheduled' :
                    this.cron = CronModel.createFromCronString(model.entry.content.get('cron_schedule'));
                    //check if earliest latest are defaults if not set cronType to custom
                    switch(this.cron.get('cronType')){
                        case 'hourly':
                            if (earliestTime !== '-1h' || latestTime !== 'now') {
                                this.cron.set('cronType', 'custom');
                            }
                            break;
                        case 'daily':
                            if (earliestTime !== '-1d' || latestTime !== 'now') {
                                this.cron.set('cronType', 'custom');
                            }
                            break;
                        case 'weekly':
                            if (earliestTime !== '-1w' || latestTime !== 'now') {
                                this.cron.set('cronType', 'custom');
                            }
                            break;
                        case 'monthly':
                            if (earliestTime !== '-1mon' || latestTime !== 'now') {
                                this.cron.set('cronType', 'custom');
                            }
                            break;
                    }
                    switch(triggercondition) {
                        case 'always' :
                                attr['ui.scheduled.triggercondition'] = 'events';
                                attr['ui.scheduled.resultscomparator'] = 'greater than';
                                attr['ui.scheduled.resultsinput'] = '0';
                            break;
                        case 'number of events' :
                                attr['ui.scheduled.triggercondition'] = 'events';
                                attr['ui.scheduled.resultscomparator'] = model.entry.content.get('alert_comparator');
                                attr['ui.scheduled.resultsinput'] = model.entry.content.get('alert_threshold');
                            break;
                        case 'number of hosts' :
                                attr['ui.scheduled.triggercondition'] = 'hosts';
                                attr['ui.scheduled.resultscomparator'] = model.entry.content.get('alert_comparator');
                                attr['ui.scheduled.resultsinput'] = model.entry.content.get('alert_threshold');
                            break;
                        case 'number of sources' :
                                attr['ui.scheduled.triggercondition'] = 'sources';
                                attr['ui.scheduled.resultscomparator'] = model.entry.content.get('alert_comparator');
                                attr['ui.scheduled.resultsinput'] = model.entry.content.get('alert_threshold');
                            break;
                        case 'custom' :
                            attr['ui.scheduled.triggercondition'] = 'custom';
                            attr['ui.scheduled.customsearch'] = model.entry.content.get('alert_condition');
                            break;
                    }
                    break;
             }
            //.:. end step1............................................................................................

            /*.:. STEP2 ...............................................................................................
             *
             * Step2 is, for the most part, independent of step1.  We do not explicitly define attributes that will
             * be handled by SplunkD (see comments)
             *
             */

            // Enable Actions

            // Show triggered in alert manger
            attr['ui.showtriggeredalerts'] = model.entry.content.get('alert.track');
            
            // ui.severity default is 2
            if (model.entry.content.get('alert.severity')) {
                attr['ui.severity'] = model.entry.content.get('alert.severity');
            }

            // send email
            attr['ui.email'] = model.entry.content.get('action.email');
            attr['action.email.subject'] = model.entry.content.get('action.email.subject');
            attr['action.email.subject.alert'] = model.entry.content.get('action.email.subject.alert');
            attr['action.email.useNSSubject'] = model.entry.content.get('action.email.useNSSubject');
            attr['action.email.to'] = model.entry.content.get('action.email.to');
            attr['action.email.cc'] = model.entry.content.get('action.email.cc');
            attr['action.email.bcc'] = model.entry.content.get('action.email.bcc');
            attr['action.email.message.alert'] = model.entry.content.get('action.email.message.alert');
            attr['action.email.priority'] = model.entry.content.get('action.email.priority');
            attr['ui.include_results_link'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.include.results_link'));
            attr['ui.include_view_link'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.include.view_link'));
            attr['ui.include_search'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.include.search'));
            attr['ui.include_trigger'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.include.trigger'));
            attr['ui.include_trigger_time'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.include.trigger_time'));

            attr['ui.include_results_pdf'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.sendpdf'));
            attr['ui.include_results_csv'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.sendcsv'));
            attr['ui.include_results_inline'] = splunkUtil.normalizeBoolean(model.entry.content.get('action.email.inline'));
            attr['ui.include_results_inline_format'] = model.entry.content.get('action.email.format');

            // run script
            attr['ui.script'] = model.entry.content.get('action.script');
            attr['ui.scriptfilename'] = model.entry.content.get('action.script.filename');

            // Action Option

            // Execute ations
            attr['ui.executeactions'] = splunkUtil.normalizeBoolean(model.entry.content.get('alert.digest_mode'));

            // Throttle
            attr['ui.throttle'] = model.entry.content.get('alert.suppress');

            var supresstime = model.entry.content.get('alert.suppress.period');
            if (supresstime) {
                attr['ui.supresstime'] = supresstime.match(/^\d*/)[0];
                attr['ui.supresstimeunit'] = supresstime.match(/s|m|h|d/)[0];
            }

            attr['ui.fieldvalue'] = model.entry.content.get('alert.suppress.fields');

            // Sharing
            attr['ui.permissions'] = splunkd_utils.USER;

            this.set(attr);

            }
        });
    }
);
