define(
    [
        'jquery',
        'underscore',
        'util/math_utils',
        'models/Base',
        'util/general_utils',
        'splunk.util'
    ],
    function($, _, math_utils, BaseModel, general_utils, splunk_util) {
        var parseFloat = math_utils.strictParseFloat;

        var minMaxValidatorFactory = function(minAttr, maxAttr, errorMessage) {
            var validator = function(value, attr, computedState) {
                var min = attr === minAttr ? parseFloat(value) : parseFloat(computedState[minAttr]),
                    max = attr === maxAttr ? parseFloat(value) : parseFloat(computedState[maxAttr]);

                if(!_.isNaN(min) && !_.isNaN(max) && max <= min) {
                    return errorMessage;
                }
            };
            return validator;
        };

        var Visualization = BaseModel.extend({
            initialize: function() {
                BaseModel.prototype.initialize.apply(this, arguments);

                // SPL-77030, if a line chart is set to log scale on the y-axis, clear any stack mode
                this.on('change:display.visualizations.charting.axisY.scale', function() {
                    if(this.get('display.visualizations.charting.chart') === 'line'
                                && this.get('display.visualizations.charting.axisY.scale') === 'log') {
                        this.set({ 'display.visualizations.charting.chart.stackMode': 'default' });
                    }
                }, this);

                // SPL-77499, if all overlay fields are removed, disable the second y-axis
                this.on('change:display.visualizations.charting.chart.overlayFields', function() {
                    if(!this.get('display.visualizations.charting.chart.overlayFields')) {
                        this.set({ 'display.visualizations.charting.axisY2.enabled': '0' });
                    }
                });
            },
            components: {
                'single': ['gen'],
                'line': ['gen', 'x', 'y', 'overlay', 'leg'],
                'area': ['gen', 'x', 'y','overlay', 'leg'],
                'column': ['gen', 'x', 'y','overlay', 'leg'],
                'bar': ['gen', 'x', 'y','overlay', 'leg'],
                'pie': ['gen'],
                'scatter':  ['gen', 'x', 'y', 'leg'],
                'bubble': ['gen', 'x', 'y', 'leg', 'mark'],
                'radialGauge': ['gen', 'ranges'],
                'fillerGauge': ['gen', 'ranges'],
                'markerGauge': ['gen', 'ranges'],
                'statistics': ['statistics'],
                'events': ['events'],
                'mapping': ['gen', 'map', 'markers', 'tiles']
            },
            defaults: {
                'display.visualizations.charting.axisY.scale': 'linear',
                'display.visualizations.charting.axisY2.scale': 'inherit',
                'display.visualizations.charting.axisX.scale': 'linear',
                'display.visualizations.charting.axisX.minimumNumber': '',
                'display.visualizations.charting.axisX.maximumNumber': '',
                'display.visualizations.charting.axisY.minimumNumber': '',
                'display.visualizations.charting.axisY.maximumNumber': '',
                'display.visualizations.charting.axisTitleX.text': '',
                'display.visualizations.charting.axisTitleY.text': '',
                'display.visualizations.charting.axisLabelsX.majorUnit': '',
                'display.visualizations.charting.axisLabelsY.majorUnit': '',
                'display.visualizations.charting.axisLabelsX.majorLabelStyle.overflowMode': 'ellipsisNone',
                'display.visualizations.charting.axisLabelsX.majorLabelStyle.rotation': '0',
                'display.visualizations.charting.legend.placement': 'right',
                'display.visualizations.charting.legend.labelStyle.overflowMode': 'ellipsisMiddle',
                'display.visualizations.charting.chart.stackMode': 'default',
                'display.visualizations.charting.layout.splitSeries': '0',
                'display.visualizations.charting.drilldown': 'all',
                'display.visualizations.charting.chart.bubbleMinimumSize': 10,
                'display.visualizations.charting.chart.bubbleMaximumSize': 50,
                'display.visualizations.charting.chart.bubbleSizeBy': 'area',
                'display.visualizations.charting.chart.nullValueMode': 'gaps',
                'display.visualizations.charting.chart.sliceCollapsingThreshold': 0.01,
                'display.visualizations.charting.chart.rangeValues': '["0","30","70","100"]',
                'display.visualizations.charting.chart.style': 'shiny',
                'display.visualizations.charting.gaugeColors': [0x84E900,0xFFE800,0xBF3030],
                'display.visualizations.mapping.drilldown': 'all',
                'display.visualizations.mapping.map.center': '(0,0)',
                'display.visualizations.mapping.map.zoom': '2',
                'display.visualizations.mapping.markerLayer.markerOpacity': '0.8',
                'display.visualizations.mapping.markerLayer.markerMinSize': '10',
                'display.visualizations.mapping.markerLayer.markerMaxSize': '50',
                'display.visualizations.mapping.data.maxClusters': '100',
                'display.visualizations.mapping.tileLayer.minZoom': '0',
                'display.visualizations.mapping.tileLayer.maxZoom': '7',
                'display.prefs.events.count': '10',
                'display.prefs.statistics.count': '10',
                'display.events.type': 'list',
                'display.events.rowNumbers': '0',
                'display.events.maxLines': '5',
                'display.events.raw.drilldown': 'full',
                'display.events.list.drilldown': 'full',
                'display.events.list.wrap': '1',
                'display.events.table.drilldown': '1',
                'display.events.table.wrap': '1',
                'display.statistics.drilldown': 'cell',
                'display.statistics.rowNumbers': '0',
                'display.statistics.wrap': '1',
                'display.statistics.overlay': 'none'
            },
            validation: {

                // General chart attribute validation rules

                'display.visualizations.charting.chart.stackMode': [
                    {
                        oneOf: ['default', 'stacked', 'stacked100'],
                        required: true,
                        msg: _('Stack Mode must be default, stacked, or stacked100').t()
                    },
                    {
                        fn: 'validateYScaleAndStacking'
                    }
                ],

                'display.visualizations.charting.layout.splitSeries': {
                    fn: 'validateBoolean',
                    required: true
                },

                'display.visualizations.charting.drilldown': {
                    oneOf: ['all', 'none'],
                    required: true,
                    msg: _('Chart Drilldown must be all or none').t()
                },

                'display.visualizations.charting.chart.nullValueMode': {
                    oneOf: ['zero', 'connect', 'gaps'],
                    required: true,
                    msg: _('Null Value Mode must be zero, connect, or gaps').t()
                },

                // Legend validation rules

                'display.visualizations.charting.legend.placement': {
                    oneOf: ['right', 'left', 'top', 'bottom', 'none'],
                    required: true,
                    msg: _('Legend Placement must be right, left, top, bottom, or none').t()
                },

                'display.visualizations.charting.legend.labelStyle.overflowMode': {
                    oneOf: ['ellipsisMiddle', 'ellipsisStart', 'ellipsisEnd'],
                    required: true,
                    msg: _('Legend Truncation must be ellipsisMiddle, ellipsisStart, or ellipsisEnd').t()
                },

                // Axis validation rules

                'display.visualizations.charting.axisX.minimumNumber': [
                    {
                        fn: 'validateNumberOrAuto',
                        required: false
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.axisX.minimumNumber',
                            'display.visualizations.charting.axisX.maximumNumber',
                            _('The X-Axis Min Value must be less than the Max Value.').t()
                        ),
                        required: false
                    }
                ],
                'display.visualizations.charting.axisX.maximumNumber': [
                    {
                        fn: 'validateNumberOrAuto',
                        required: false
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.axisX.minimumNumber',
                            'display.visualizations.charting.axisX.maximumNumber',
                            _('The X-Axis Min Value must be less than the Max Value.').t()
                        ),
                        required: false
                    }
                ],
                'display.visualizations.charting.axisY.minimumNumber': [
                    {
                        fn: 'validateNumberOrAuto',
                        required: false
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.axisY.minimumNumber',
                            'display.visualizations.charting.axisY.maximumNumber',
                            _('The Y-Axis Min Value must be less than the Max Value.').t()
                        ),
                        required: false
                    }
                ],
                'display.visualizations.charting.axisY.maximumNumber': [
                    {
                        fn: 'validateNumberOrAuto',
                        required: false
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.axisY.minimumNumber',
                            'display.visualizations.charting.axisY.maximumNumber',
                            _('The Y-Axis Min Value must be less than the Max Value.').t()
                        ),
                        required: false
                    }
                ],
                'display.visualizations.charting.axisY2.minimumNumber': [
                    {
                        fn: 'validateNumberOrAuto',
                        required: false
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.axisY2.minimumNumber',
                            'display.visualizations.charting.axisY2.maximumNumber',
                            _('The Y-Axis Min Value must be less than the Max Value.').t()
                        ),
                        required: false
                    }
                ],
                'display.visualizations.charting.axisY2.maximumNumber': [
                    {
                        fn: 'validateNumberOrAuto',
                        required: false
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.axisY2.minimumNumber',
                            'display.visualizations.charting.axisY2.maximumNumber',
                            _('The Y-Axis Min Value must be less than the Max Value.').t()
                        ),
                        required: false
                    }
                ],
                'display.visualizations.charting.axisLabelsX.majorUnit': [
                    {
                        fn: 'validatePositiveNumberOrAuto',
                        required: false
                    }
                ],
                'display.visualizations.charting.axisLabelsY.majorUnit': [
                    {
                        fn: 'validatePositiveNumberOrAuto',
                        required: false
                    }
                ],
                'display.visualizations.charting.axisLabelsY2.majorUnit': [
                    {
                        fn: 'validatePositiveNumberOrAuto',
                        required: false
                    }
                ],
                'display.visualizations.charting.axisLabelsX.majorLabelStyle.rotation': {
                    oneOf: ['-90', '-45', '0', '45', '90'],
                    required: true,
                    msg: _('Axis Label Rotation must be -90, -45, 0, 45, or 90').t()
                },
                'display.visualizations.charting.axisLabelsX.majorLabelStyle.overflowMode': {
                    oneOf: ['ellipsisNone', 'ellipsisMiddle'],
                    required: true,
                    msg: _('Axis Label overflow mode must be ellipsisNone or ellipsisMiddle').t()
                },
                'display.visualizations.charting.axisY.scale': [
                    {
                        oneOf: ['linear', 'log'],
                        required: true,
                        msg: _('Y-Axis Scale must be linear or log').t()
                    },
                    {
                        fn: 'validateYScaleAndStacking'
                    }
                ],
                'display.visualizations.charting.axisX.scale': {
                    oneOf: ['linear', 'log'],
                    required: true,
                    msg: _('X-Axis Scale must be linear or log').t()
                },
                'display.visualizations.charting.axisY2.scale': {
                    oneOf: ['linear', 'log', 'inherit'],
                    required: true,
                    msg: _('X-Axis Scale must be linear, log, or inherit').t()
                },

                // Pie chart validation rules

                'display.visualizations.charting.chart.sliceCollapsingThreshold': {
                    pattern: 'number',
                    min: 0,
                    msg: _('Minimum Size must be a non-negative number.').t(),
                    required: true
                },

                // Gauge validation rules

                'display.visualizations.charting.chart.rangeValues': {
                    fn: 'validateRangeValues',
                    required: false
                },

                'display.visualizations.charting.chart.style': {
                    oneOf: ['shiny', 'minimal'],
                    required: true,
                    msg: _('Gauge Style must be shiny or minimal').t()
                },

                // Map renderer validation rules

                'display.visualizations.mapping.drilldown': {
                    oneOf: ['all', 'none'],
                    required: true,
                    msg: _('Drilldown must be all or none').t()
                },

                'display.visualizations.mapping.map.center': {
                    pattern: /^\s*\(([-+.0-9]+)\s*,\s*([-+.0-9]+)\)\s*$/,
                    required: true,
                    msg: _('Latitude and Longitude must be valid numbers').t()
                },

                'display.visualizations.mapping.map.zoom': {
                    pattern: 'digits',
                    required: true,
                    min: 0,
                    msg: _('Zoom must be a non-negative integer').t()
                },

                'display.visualizations.mapping.markerLayer.markerOpacity': {
                    pattern: 'number',
                    required: true,
                    min: 0,
                    max: 1,
                    msg: _('Marker Opacity must be between 0 and 100%').t()
                },

                'display.visualizations.mapping.markerLayer.markerMinSize': [
                    {
                        pattern: 'digits',
                        required: true,
                        min: 1,
                        msg: _('Minimum Size must be a positive integer').t()
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.mapping.markerLayer.markerMinSize',
                            'display.visualizations.mapping.markerLayer.markerMaxSize',
                            _('The Min Size must be less than the Max Size.').t()
                        )
                    }
                ],

                'display.visualizations.mapping.markerLayer.markerMaxSize': [
                    {
                        pattern: 'digits',
                        required: true,
                        min: 1,
                        msg: _('Maximum Size must be a positive integer').t()
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.mapping.markerLayer.markerMinSize',
                            'display.visualizations.mapping.markerLayer.markerMaxSize',
                            _('The Min Size must be less than the Max Size.').t()
                        )
                    }
                ],

                'display.visualizations.mapping.data.maxClusters': {
                    pattern: 'digits',
                    required: true,
                    min: 0,
                    msg: _('Max Clusters must be a non-negative integer').t()
                },

                'display.visualizations.mapping.tileLayer.minZoom': [
                    {
                        pattern: 'digits',
                        required: true,
                        min: 0,
                        msg: _('Minimum Zoom must be a non-negative integer').t()
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.mapping.tileLayer.minZoom',
                            'display.visualizations.mapping.tileLayer.maxZoom',
                            _('The Min Zoom must be less than the Max Zoom.').t()
                        )
                    }
                ],

                'display.visualizations.mapping.tileLayer.maxZoom': [
                    {
                        pattern: 'digits',
                        required: true,
                        min: 0,
                        msg: _('Maximum Zoom must be a non-negative integer').t()
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.mapping.tileLayer.minZoom',
                            'display.visualizations.mapping.tileLayer.maxZoom',
                            _('The Min Zoom must be less than the Max Zoom.').t()
                        )
                    }
                ],

                // Event renderer validation rules

                'display.prefs.events.count': {
                    pattern: 'digits',
                    min: 1,
                    max: 100,
                    msg: _('Rows Per Page must be a positive number no larger than 100.').t(),
                    required: true
                },

                'display.events.type': {
                    oneOf: ['list', 'raw', 'table'],
                    required: true,
                    msg: _('Event Display Type must be list, raw, or table').t()
                },

                'display.events.raw.drilldown': {
                    oneOf: ['full', 'none', 'inner', 'outer'],
                    required: true,
                    msg: _('Event Drilldown must be full, none, inner, or outer').t()
                },

                'display.events.list.drilldown': {
                    oneOf: ['full', 'none', 'inner', 'outer'],
                    required: true,
                    msg: _('Event Drilldown must be full, none, inner, or outer').t()
                },

                'display.events.table.drilldown': {
                    fn: 'validateBoolean',
                    required: true
                },

                'display.events.rowNumbers': {
                    fn: 'validateBoolean',
                    required: true
                },

                'display.events.list.wrap': {
                    fn: 'validateBoolean',
                    required: true
                },

                'display.events.table.wrap': {
                    fn: 'validateBoolean',
                    required: true
                },

                'display.events.maxLines': {
                    oneOf: ['0', '5', '10', '20', '50', '100', '200'],
                    required: true,
                    msg: _('Max Lines must 0, 5, 10, 20, 50, 100, or 200').t()
                },

                // Statistics table validation rules

                'display.prefs.statistics.count': {
                    pattern: 'digits',
                    min: 1,
                    max: 100,
                    msg: _('Rows Per Page must be a positive number no larger than 100.').t(),
                    required: true
                },

                'display.statistics.drilldown': {
                    oneOf: ['row', 'cell', 'none'],
                    required: true,
                    msg: _('Statistics Drilldown must be row, cell, or none').t()
                },

                'display.statistics.overlay': {
                    oneOf: ['none', 'heatmap', 'highlow'],
                    required: true,
                    msg: _('Data Overlay must be none, heatmap, or highlow').t()
                },

                'display.statistics.rowNumbers': {
                    fn: 'validateBoolean',
                    required: true
                },

                'display.statistics.wrap': {
                    fn: 'validateBoolean',
                    required: true
                },

                // Bubble chart validation rules

                'display.visualizations.charting.chart.bubbleSizeBy' : {
                    oneOf: ['area', 'diameter'],
                    required: true
                },

                'display.visualizations.charting.chart.bubbleMaximumSize': [
                    {
                        pattern: 'digits',
                        min: 1,
                        msg: _('Maximum Size must be a positive integer.').t(),
                        required: true
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.chart.bubbleMinimumSize',
                            'display.visualizations.charting.chart.bubbleMaximumSize',
                            _('The Min Size must be less than the Max Size.').t()
                        )
                    }
                ],

                'display.visualizations.charting.chart.bubbleMinimumSize': [
                    {
                        pattern: 'digits',
                        min: 1,
                        msg: _('Minimum Size must be a positive integer.').t(),
                        required: true
                    },
                    {
                        fn: minMaxValidatorFactory(
                            'display.visualizations.charting.chart.bubbleMinimumSize',
                            'display.visualizations.charting.chart.bubbleMaximumSize',
                            _('The Min Size must be less than the Max Size.').t()
                        )
                    }
                ]
            },

            validateYScaleAndStacking: function(value, attr, computedState) {
                // SPL-77030, since line charts ignore stack mode, this validation rule does not apply
                if(computedState['display.visualizations.charting.chart'] === 'line') {
                    return false;
                }
                var yAxisScale = attr === 'display.visualizations.charting.axisY.scale' ? value :
                                                computedState['display.visualizations.charting.axisY.scale'],
                    stackMode = attr === 'display.visualizations.charting.chart.stackMode' ? value :
                                                computedState['display.visualizations.charting.chart.stackMode'];

                if(yAxisScale === 'log' && stackMode !== 'default') {
                    return _('Log scale and stacking cannot be enabled at the same time.').t();
                }
            },

            validateRangeValues: function(value) {
                var ranges;
                try {
                    ranges = _(value ? JSON.parse(value) : []).map(parseFloat);
                }
                catch (e) {
                    return _('Ranges must be of the form: ["0","30","70","100"]').t();
                }
                if(_(ranges).any(_.isNaN)) {
                    return _('All color ranges must be valid numbers.').t();
                }
                var filteredRanges = Visualization.filterToValidRangeValues(ranges);
                if(!_.isEqual(ranges, filteredRanges)) {
                    return _('Color ranges must be entered from lowest to highest.').t();
                }
            },

            validateNumberOrAuto: function(value, attr, computedState) {
                var num = parseFloat(value);
                if(value && _.isNaN(num) && !value.match(/^(auto)$/)){
                    return splunk_util.sprintf(
                        _('%s must be a number or "auto".').t(),
                        this.humanizeAttr(attr)
                    );
                }
            },

            validatePositiveNumberOrAuto: function(value, attr, computedState) {
                var numberOrAutoValidation = this.validateNumberOrAuto(value, attr, computedState);
                if(numberOrAutoValidation) {
                    return numberOrAutoValidation;
                }
                var num = parseFloat(value);
                if(!_.isNaN(num) && num <= 0) {
                    return splunk_util.sprintf(
                        _('%s must be a positive number or "auto".').t(),
                        this.humanizeAttr(attr)
                    );
                }
            },

            validateBoolean: function(value, attr) {
                if(!general_utils.isBooleanEquivalent(value)) {
                    return splunk_util.sprintf(
                        _('%s must be a valid boolean value').t(),
                        this.humanizeAttr(attr)
                    );
                }
            },

            humanizeAttr: function(attr) {
                switch(attr){
                    case 'display.visualizations.charting.axisLabelsX.majorUnit':
                        return _('X-Axis Interval').t();
                    case 'display.visualizations.charting.axisLabelsY.majorUnit':
                    case 'display.visualizations.charting.axisLabelsY2.majorUnit':
                        return _('Y-Axis Interval').t();
                    case 'display.visualizations.charting.axisX.minimumNumber':
                        return _('X-Axis Min Value').t();
                    case 'display.visualizations.charting.axisY.minimumNumber':
                    case 'display.visualizations.charting.axisY2.minimumNumber':
                        return _('Y-Axis Min Value').t();
                    case 'display.visualizations.charting.axisX.maximumNumber':
                        return _('X-Axis Max Value').t();
                    case 'display.visualizations.charting.axisY.maximumNumber':
                    case 'display.visualizations.charting.axisY2.maximumNumber':
                        return _('Y-Axis Max Value').t();
                    case 'display.visualizations.charting.layout.splitSeries':
                        return _('Split Series Mode').t();
                    case 'display.events.table.drilldown':
                        return _('Event Drilldown').t();
                    case 'display.statistics.wrap':
                    case 'display.events.list.wrap':
                    case 'display.events.table.wrap':
                        return _('Wrap Results').t();
                    case 'display.statistics.rowNumber':
                    case 'display.events.rowNumbers':
                        return _('Row Numbers').t();
                    default:
                        return attr;
                }
            },

            attrToArray: function(attr) {
                return Visualization.parseStringifiedArray(this.get(attr));
            },

            rangesValuesToArray: function() {
                return this.attrToArray('display.visualizations.charting.chart.rangeValues');
            },

            deserializeGaugeColorArray: function() {
                return Visualization.parseStringifiedColorArray(this.get('display.visualizations.charting.gaugeColors'));
            },

            // use auto mode only if ranges and colors are both not defined
            gaugeIsInAutoMode: function() {
                return !this.get('display.visualizations.charting.chart.rangeValues') && !this.get('display.visualizations.charting.gaugeColors');
            }
        },
        {
            CHARTING_ATTRS_FILTER: ['^display.visualizations.charting...*'],
            SINGLE_VALUE_ATTRS_FILTER: ['^singlevalue\..*'],

            filterToValidRangeValues: function(rangeValues) {
                if(_(rangeValues).isString()) {
                    rangeValues = Visualization.parseStringifiedArray(rangeValues);
                }
                rangeValues = _(rangeValues).chain().map(parseFloat).filter(function(val) { return !_(val).isNaN(); }).uniq().value();
                var runningMax = -Infinity;
                rangeValues = _(rangeValues).filter(function(val) {
                    if(val < runningMax) {
                        return false;
                    }
                    runningMax = val;
                    return true;
                });
                return rangeValues;
            },

            parseStringifiedArray: function(value) {
                if(!value){
                    return [];
                }
                var parsedValues = [];
                // SPL-80318 wrap this call to JSON.parse in a try-catch because the input could be invalid JSON
                try {
                    parsedValues = _.values(JSON.parse(value));
                }
                catch(e) {}
                return parsedValues;
            },

            parseStringifiedColorArray: function(value) {
                if(_(value).isArray()) {
                    return value;
                }
                if(!value || value.charAt(0) !== '[' || value.charAt(value.length - 1) !== ']') {
                    return false; //need to find a different way to bail
                }
                return splunk_util.stringToFieldList(value.substring(1, value.length - 1));
            }

        });

        return Visualization;
    }
);
