define(
    [
        'backbone',
        'underscore',
        'uri/route',
        'splunk.config',
        'splunk.util',
        'models/shared/Application',
        'models/classicurl',
        'models/services/server/ServerInfo',
        'models/services/configs/Web',
        'models/shared/User',
        'models/account/Session',
        'models/shared/UpdateChecker',
        'models/account/Login',
        'views/account/Master'
    ],
    function(
        Backbone,
        _,
        route,
        splunkConfig,
        util,
        ApplicationModel,
        classicurlModel,
        ServerInfoModel,
        WebModel,
        UserModel,
        SessionModel,
        UpdateCheckerModel,
        LoginModel,
        MasterView
    ) {
        // These translations are used in the /templates/pages/static.html page but need to be declared here so that
        // they will be extracted and added to the javascript translation scope.
        _('This browser is not supported by Splunk.').t();
        _('Please refer to the list of %s.').t();
        _('Supported Browsers').t();

        return Backbone.Router.extend({
            routes: {
                ':locale/account/:page': 'page',
                ':locale/account/:page?*params': 'page',
                ':locale/account/:page/': 'page',
                ':locale/account/:page/?*params': 'page',
                '*root/:locale/account/:page': 'pageRooted',
                '*root/:locale/account/:page?*params': 'pageRooted',
                '*root/:locale/account/:page/': 'pageRooted',
                '*root/:locale/account/:page/?*params': 'pageRooted'
            },
            initialize: function() {
                this.model = {};
                this.model.application = new ApplicationModel(),
                this.model.classicurl = classicurlModel,
                this.model.serverInfo = new ServerInfoModel({}, {splunkDPayload: __splunkd_partials__['/services/server/info']}),
                this.model.session = new SessionModel({}, {splunkDPayload: __splunkd_partials__['/services/session']}),
                this.model.web = new WebModel({}, {splunkDPayload: __splunkd_partials__['/configs/conf-web']}),
                this.model.updateChecker = new UpdateCheckerModel(),
                this.model.login = new LoginModel(),
                this.model.user = new UserModel({}, {serverInfoModel: this.model.serverInfo});
                this.model.user.urlRoot = undefined;//urlRoot is relied on by other consumers; required to be deleted in order to set a fully qualified link as id 
                this.model.login.on('error', function() {
                    if (this.model.login.isPasswordChangeRequired()) {
                        this.model.application.set('page', 'passwordchange');
                    }
                }, this);
                this.model.login.on('sync', function() {
                    //optional first time run password change for authenticated user
                    if (!this.model.session.entry.content.get('hasLoggedIn') && !this.model.serverInfo.isFreeLicense() && this.model.login.isAdmin()) {
                        var id = '/services/' + this.model.user.url + '/' + encodeURIComponent(this.model.login.get('username'));
                        this.model.user.set('id', id);
                        this.model.application.set('page', 'passwordchange');
                    } else {
                        var delay = (this.model.serverInfo.isFreeLicense() && this.model.updateChecker.get('url')) ? 5000 : 0;
                        this.successRedirect({delay: delay});
                    }
                }, this);
                this.model.user.on('sync', function() {
                    this.successRedirect();
                }, this);
                this.model.updateChecker.on('sync', function() {
                    if (this.model.serverInfo.isFreeLicense() && !this.model.updateChecker.get('hasNewVersion')) {
                        this.model.login.saveAsFree(this.model.session.entry.content.get('cval'), this.model.application.get('return_to'));
                    }
                }, this);
                //ensure globals ie., window.$C sync'd with splunkd session partial values
                splunkConfig.LOCALE = this.model.session.entry.content.get('lang');
            },
            successRedirect: function(options) {
                options || (options = {});
                var url = route.returnTo(this.model.application.get('root'), this.model.application.get('locale'), this.model.application.get('return_to') || '');
                if (options.delay) {
                    setTimeout(function() {
                        window.location = url;
                    }, options.delay);
                } else {
                    window.location = url;
                }
            },
            bootstrapUpdateChecker: function() {
                this.model.updateChecker.fetchHelper(this.model.serverInfo, this.model.web, this.model.application, 'login', this.model.session.entry.content.get('uid'), {timeout: 10000});
            },
            page: function(locale, page) {
                document.title = _('Login | Splunk').t();
                this.model.classicurl.fetch(); //is syncronous
                this.model.application.set({
                    locale: locale,
                    app: '-',
                    page: page,
                    return_to: this.model.classicurl.get('return_to')
                });
                var masterView = new MasterView({
                    model: {
                        application: this.model.application,
                        serverInfo: this.model.serverInfo,
                        session: this.model.session,
                        web: this.model.web,
                        updateChecker: this.model.updateChecker,
                        login: this.model.login,
                        user: this.model.user
                    }
                });
                masterView.render().attachToDocument(document.body, 'appendTo');
                if (this.model.application.get('page')==='login') {
                    if (this.model.serverInfo.isFreeLicense()) {
                        setTimeout(function() {
                            if (util.normalizeBoolean(this.model.web.entry.content.get('updateCheckerBaseURL'))) {
                                this.bootstrapUpdateChecker();
                            } else {
                                this.model.login.saveAsFree(this.model.session.entry.content.get('cval'), this.model.application.get('return_to'));
                            }
                        }.bind(this), 1500);

                    } else if (util.normalizeBoolean(this.model.web.entry.content.get('updateCheckerBaseURL'))) {
                        this.bootstrapUpdateChecker();
                    }
                }
            },
            pageRooted: function(root, locale, page) {
                this.model.application.set({
                    root: root
                }, {silent: true});
                this.page(locale, page);
            }
        });
    }
);
