define(function(require)
{

	var jg_extend = require("jg_global").jg_extend;
	var Brush = require("jgatt").graphics.brushes.Brush;
	var ObservableArrayProperty = require("jgatt").properties.ObservableArrayProperty;
	var ObservableProperty = require("jgatt").properties.ObservableProperty;
	var NumberUtils = require("jgatt").utils.NumberUtils;

	return jg_extend(Brush, function(BorderStrokeBrush, base)
	{

		// Public Properties

		this.colors = new ObservableArrayProperty("colors", Number, [ 0x000000, 0x000000, 0x000000, 0x000000 ])
			.readFilter(function(value)
			{
				return value.concat();
			})
			.writeFilter(function(value)
			{
				var length = value ? value.length : 0;
				var top = ((length > 0) && !isNaN(value[0])) ? NumberUtils.minMax(Math.floor(value[0]), 0x000000, 0xFFFFFF) : 0x000000;
				var right = ((length > 1) && !isNaN(value[1])) ? NumberUtils.minMax(Math.floor(value[1]), 0x000000, 0xFFFFFF) : top;
				var bottom = ((length > 2) && !isNaN(value[2])) ? NumberUtils.minMax(Math.floor(value[2]), 0x000000, 0xFFFFFF) : top;
				var left = ((length > 3) && !isNaN(value[3])) ? NumberUtils.minMax(Math.floor(value[3]), 0x000000, 0xFFFFFF) : right;
				return [ top, right, bottom, left ];
			});

		this.alphas = new ObservableArrayProperty("alphas", Number, [ 1, 1, 1, 1 ])
			.readFilter(function(value)
			{
				return value.concat();
			})
			.writeFilter(function(value)
			{
				var length = value ? value.length : 0;
				var top = ((length > 0) && !isNaN(value[0])) ? NumberUtils.minMax(value[0], 0, 1) : 1;
				var right = ((length > 1) && !isNaN(value[1])) ? NumberUtils.minMax(value[1], 0, 1) : top;
				var bottom = ((length > 2) && !isNaN(value[2])) ? NumberUtils.minMax(value[2], 0, 1) : top;
				var left = ((length > 3) && !isNaN(value[3])) ? NumberUtils.minMax(value[3], 0, 1) : right;
				return [ top, right, bottom, left ];
			});

		this.thicknesses = new ObservableArrayProperty("thicknesses", Number, [ 1, 1, 1, 1 ])
			.readFilter(function(value)
			{
				return value.concat();
			})
			.writeFilter(function(value)
			{
				var length = value ? value.length : 0;
				var top = ((length > 0) && (value[0] < Infinity)) ? Math.max(value[0], 0) : 1;
				var right = ((length > 1) && (value[1] < Infinity)) ? Math.max(value[1], 0) : top;
				var bottom = ((length > 2) && (value[2] < Infinity)) ? Math.max(value[2], 0) : top;
				var left = ((length > 3) && (value[3] < Infinity)) ? Math.max(value[3], 0) : right;
				return [ top, right, bottom, left ];
			});

		this.caps = new ObservableProperty("caps", String, "none")
			.writeFilter(function(value)
			{
				switch (value)
				{
					case "none":
					case "round":
					case "square":
						return value;
					default:
						return "none";
				}
			});

		this.joints = new ObservableProperty("joints", String, "miter")
			.writeFilter(function(value)
			{
				switch (value)
				{
					case "miter":
					case "round":
					case "bevel":
						return value;
					default:
						return "miter";
				}
			});

		this.miterLimit = new ObservableProperty("miterLimit", Number, 10)
			.writeFilter(function(value)
			{
				return ((value > 0) && (value < Infinity)) ? value : 10;
			});

		this.pixelHinting = new ObservableProperty("pixelHinting", Boolean, true);

		// Constructor

		this.constructor = function(colors, alphas, thicknesses, caps, joints, miterLimit, pixelHinting)
		{
			base.constructor.call(this);

			if (colors != null)
				this.set(this.colors, colors);
			if (alphas != null)
				this.set(this.alphas, alphas);
			if (thicknesses != null)
				this.set(this.thicknesses, thicknesses);
			if (caps != null)
				this.set(this.caps, caps);
			if (joints != null)
				this.set(this.joints, joints);
			if (miterLimit != null)
				this.set(this.miterLimit, miterLimit);
			if (pixelHinting != null)
				this.set(this.pixelHinting, pixelHinting);
		};

		// Protected Methods

		this.draw = function(properties, commands, graphics, matrix, bounds)
		{
			var numCommands = commands.length;
			if (numCommands === 0)
				return;

			var x1 = Infinity;
			var x2 = -Infinity;
			var y1 = Infinity;
			var y2 = -Infinity;
			var command;
			var i;

			for (i = 0; i < numCommands; i++)
			{
				command = commands[i];
				if ((command.name === "moveTo") || (command.name === "lineTo"))
				{
					x1 = Math.min(x1, command.x);
					x2 = Math.max(x2, command.x);
					y1 = Math.min(y1, command.y);
					y2 = Math.max(y2, command.y);
				}
				else if (command.name === "curveTo")
				{
					x1 = Math.min(x1, command.controlX, command.anchorX);
					x2 = Math.max(x2, command.controlX, command.anchorX);
					y1 = Math.min(y1, command.controlY, command.anchorY);
					y2 = Math.max(y2, command.controlY, command.anchorY);
				}
			}

			var borderPoints = [ { x: x1, y: y1 }, { x: x2, y: y1 }, { x: x2, y: y2 }, { x: x1, y: y2 }, { x: x1, y: y1 } ];
			var thickness = NaN;
			var color = NaN;
			var alpha = NaN;
			var newStroke = true;

			for (i = 0; i < 4; i++)
			{
				if (properties.thicknesses[i] !== thickness)
				{
					thickness = properties.thicknesses[i];
					newStroke = true;
				}

				if (thickness === 0)
					continue;

				if (properties.colors[i] !== color)
				{
					color = properties.colors[i];
					newStroke = true;
				}

				if (properties.alphas[i] !== alpha)
				{
					alpha = properties.alphas[i];
					newStroke = true;
				}

				if (newStroke)
				{
					graphics.endStroke();
					graphics.setStrokeStyle(thickness, properties.caps, properties.joints, properties.miterLimit, properties.pixelHinting);
					graphics.beginSolidStroke(color, alpha);
					graphics.moveTo(borderPoints[i].x, borderPoints[i].y);
					newStroke = false;
				}

				graphics.lineTo(borderPoints[i + 1].x, borderPoints[i + 1].y);
			}

			graphics.endStroke();
		};

	});

});
