define(function(require, exports, module) {
    var _ = require('underscore');
    var $ = require('jquery');
    var mvc = require("./mvc");
    var BaseSplunkView = require("./basesplunkview");
    var Messages = require("./messages");
    var Drilldown = require('./drilldown');
    var utils = require('./utils');
    var console = require('util/console');
    var SplunkUtil = require('splunk.util');
    var SplunkLegend = require('splunk.legend');
    var jschartingUtils = require('util/jscharting_utils');
    var splunkConfig = require('splunk.config');
    var resizable = require('jquery.ui.resizable');
    var JSCharting;
    var GeneralUtils = require('util/general_utils');
    var TokenAwareModel = require('./tokenawaremodel');
    var Route = require('uri/route');
    var Printer = require('helpers/Printer');
    
    require('css!../css/jschart');

    var ChartView = BaseSplunkView.extend({
        moduleId: module.id,

        className: "splunk-chart",
        chartOptionPrefix: 'charting.',

        options: {
            'height': '250px',
            'data': 'preview',
            'type': 'column',
            'drilldownRedirect': true,
            'resizable': false
        },

        omitFromSettings: ['el', 'reportModel'],

        normalizeOptions: function(settings) {

            this._normalizeDrilldownType(settings, settings.get(settings.has('charting.drilldown') ? 'charting.drilldown' : 'drilldown'));

            if (settings.has("charting.layout.splitSeries")) {
                if(GeneralUtils.isBooleanEquivalent(settings.get("charting.layout.splitSeries"))) {
                    settings.set("charting.layout.splitSeries", GeneralUtils.normalizeBoolean(settings.get("charting.layout.splitSeries")) ? "1" : "0");
                }
            }

            if (settings.has("show")) {
                settings.set("show", SplunkUtil.normalizeBoolean(settings.get('show')) ? "1" : "0");
            }

            if (settings.has("charting.axisY2.enabled")) {
                if(GeneralUtils.isBooleanEquivalent(settings.get("charting.axisY2.enabled"))) {
                    settings.set("charting.axisY2.enabled", GeneralUtils.normalizeBoolean(settings.get("charting.axisY2.enabled")) ? "1" : "0");
                }
            }

            if (settings.has("charting.legend.labelStyle.overflowMode")) {
                if (settings.get("charting.legend.labelStyle.overflowMode") === "default") {
                    settings.set("charting.legend.labelStyle.overflowMode", "ellipsisMiddle");
                }
            }
        },

        _normalizeDrilldownType: function(settings, value) {
            settings.set('charting.drilldown', Drilldown.getNormalizedDrilldownType(value, { allowBoolean: true }));
        },

        initialize: function(options) {
            this.configure();
            this.model = this.options.reportModel || TokenAwareModel._createReportModel();
            this.settings._sync = utils.syncModels(this.settings, this.model, {
                auto: true,
                prefix: 'display.visualizations.',
                exclude: ['managerid','id','name','data', 'type', 'drilldownRedirect', 'height']
            });

            this.normalizeOptions(this.settings);
            this.settings.on('change:drilldown change:charting.drilldown', this._normalizeDrilldownType, this);

            // set our maxResultCount to the current value 'charting.chart.data', or default to 1000
            this.maxResultCount = 1000;
            if(this.settings.has('charting.data.count') && !_.isNaN(parseInt(this.settings.get('charting.data.count'), 10))) {
                this.maxResultCount = parseInt(this.settings.get('charting.data.count'), 10);
            }

            this._currentHeight = parseInt(this.settings.get('height'), 10);

            // initialize containers as detached DOM
            this.$chart = $('<div></div>');
            this.$msg = $('<div></div>');
            this.$inlineMsg = $('<div></div>');

            this.settings.on("change", this.render, this);
            var self = this;
            require(['js_charting/js_charting'], function(JSChartingLib){
                JSCharting = JSChartingLib;
                // Create the chart only if there is a pending create.
                if (self._chartCreationPending && self.chartData) {
                    self._chartCreationPending = false;
                    self._createChart();
                }
                self.createChart = self._createChart;
            });

            this.bindToComponentSetting('managerid', this._onManagerChange, this);
            SplunkLegend.register(this.cid);

            // Setup resizing
            this._onResizeMouseup = _.bind(this._onResizeMouseup, this);
            this.settings.on('change:resizable', function(model, value, options) {
                value ? this._enableResize() : this._disableResize();
            }, this);
            if (this.settings.get('resizable')) {
                this._enableResize();
            }

            // Height Handler
            this.settings.on('change:height', this._onChartHeightChange, this);

            this.listenTo(Printer, Printer.PRINT_START, this.resizeChart);
            this.listenTo(Printer, Printer.PRINT_END, this.resizeChart);

            // If we don't have a manager by this point, then we're going to
            // kick the manager change machinery so that it does whatever is
            // necessary when no manager is present.
            if (!this.manager) {
                this._onManagerChange(mvc.Components, null);
            }
        },

        _onManagerChange: function(managers, manager) {
            if (this.manager) {
                this.manager.off(null, null, this);
                this.manager = null;
            }
            if (this.resultsModel) {
                this.resultsModel.off(null, null, this);
                this.resultsModel = null;
            }

            if (!manager) {
                this.message("no-search");
                return;
            }

            // Clear any messages, since we have a new manager.
            this.message("empty");

            this._err = false;
            this.manager = manager;
            this.resultsModel = this.manager.data(this.settings.get("data"), {
                autofetch: true,
                output_mode: "json_cols",
                show_metadata: true,
                count: this.maxResultCount
            });
            this.resultsModel.on("data", this._onDataChanged, this);
            this.resultsModel.on("error", this._onSearchError, this);
            manager.on("search:start", this._onSearchStart, this);
            manager.on("search:progress", this._onSearchProgress, this);
            manager.on("search:done", this._onSearchProgress, this);
            manager.on("search:cancelled", this._onSearchCancelled, this);
            manager.on("search:refresh", this._onSearchRefresh, this);
            manager.on("search:fail", this._onSearchFailed, this);
            manager.on("search:error", this._onSearchError, this);
            manager.replayLastSearchEvent(this);
        },

        _onDataChanged: function() {
            if (!this.resultsModel.hasData()) {
                if (this._isJobDone) {
                    this.message('no-results');
                }
                return;
            }

            var chartData = this.resultsModel.data();
            console.log('chart data changed:', chartData);

            if (chartData.fields.length) {
                this.chartData = chartData;
                this.updateChart();
            }
        },

        _onSearchProgress: function(properties) {
            this._err = false;
            properties = properties || {};
            var content = properties.content || {};
            var previewCount = content.resultPreviewCount || 0;
            var isJobDone = this._isJobDone = content.isDone || false;

            if (previewCount === 0 && isJobDone) {
                this.message('no-results');
                return;
            }

            if (!this.chartData) {
                this.chartData = {
                    fields: [],
                    columns: []
                };
                this.updateChart();
            }

        },

        _onSearchCancelled: function() {
            this._isJobDone = false;
            this.message('cancelled');
        },

        _onSearchRefresh: function() {
            this._isJobDone = false;
            this.message('refresh');
        },

        _onSearchError: function(message, err) {
            this._isJobDone = false;
            this._err = true;
            var msg = Messages.getSearchErrorMessage(err) || message;
            this.message({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },

        _onSearchFailed: function(state) {
            this._isJobDone = false;
            this._err = true;
            var msg = Messages.getSearchFailureMessage(state);
            this.message({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },

        _onSearchStart: function() {
            this._isJobDone = false;
            this._err = false;
            this.destroyChart();
            this.message('waiting');
        },

        message: function(info) {
            this.$msg.detach();
            Messages.render(info, this.$msg);
            this.$msg.prependTo(this.$el);
        },

        inlineMessage: function(info) {
            info.compact = true;
            Messages.render(info, this.$inlineMsg);
        },

        render: function() {
            this.$el.height(this._currentHeight).css('overflow', 'hidden');
            this.$msg.height(this._currentHeight).css('overflow', 'hidden');
            this.$chart.appendTo(this.$el);
            this.$inlineMsg.appendTo(this.$el);
            if (this.chart) {
                this.destroyChart();
            }

            if(!this._boundInvalidateChart) {
                this._boundInvalidateChart = _.bind(this.invalidateChart, this);
            }
            SplunkLegend.removeEventListener('labelIndexMapChanged', this._boundInvalidateChart);
            SplunkLegend.addEventListener('labelIndexMapChanged', this._boundInvalidateChart);

            if(!this._debouncedResize) {
                this._debouncedResize = _.debounce(_.bind(this.resizeChart, this), 100);
            }
            $(window).off('resize', this._debouncedResize);
            $(window).on('resize', this._debouncedResize);

            this.createChart();

            return this;
        },

        show: function() {
            this.$el.css('display', '');
        },

        hide: function() {
            this.$el.css('display', 'none');
        },

        createChart: function() {
            this._chartCreationPending = true;
        },

        _enableResize: function() {
            if (this._canEnableResize()) {
                this.$el.resizable({ handles: "s", stop: this._onResizeStop.bind(this)});
                // workaround until jquery ui is updated
                this.$el.on('mouseup', this._onResizeMouseup);
            }
        },

        _disableResize: function() {
            if (this._canEnableResize()) {
                this.$el.resizable('destroy');
                this.$el.off('mouseup', this._onResizeMouseup);
            }
        },

        // NOTE: Bound to 'this' by constructor
        _onResizeMouseup: function(event) {
            $(this).width("100%");
        },

        _onResizeStop: function(event, ui) {
            $(event.target).width("100%");
            this._currentHeight = this.$el.height();
            this.settings.set('height', this._currentHeight, {silent:true});
            this.resizeChart();
        },

        _canEnableResize: function() {
            return true;
        },

        _createChart: function() {
            // Initialize the chart with the type if it is there. If somebody
            // actually specified charting.chart, that option will win.
            var displayProperties = {'chart': this.settings.get("type")};

            // Copy over the settings for everything that starts with the
            // prefix (charting.) by default.
            var prefix = this.chartOptionPrefix;
            _.each(this.settings.toJSON(), function(value, key){
                if(key.substring(0, prefix.length) == prefix) {
                    displayProperties[key.substring(prefix.length)] = value;
                }
            }, this);

            if(this._err) { return; }

            this.$msg.detach();
            console.log('Creating chart with data: ', displayProperties);
            var chart = this.chart = JSCharting.createChart(this.$chart, displayProperties);
            chart.on('pointClick', this.emitDrilldownEvent.bind(this));
            chart.on('legendClick', this.emitDrilldownEvent.bind(this));
            chart.on('chartRangeSelect', this.emitSelectionEvent.bind(this));
            this.updateChart();
        },

        _onChartHeightChange: function() {
            this._currentHeight = parseInt(this.settings.get('height'), 10);
            this.resizeChart();
        },

        resizeChart: function() {
            if(this.chart) {
                this.updateChartContainerHeight();
                this.chart.resize();
            }
        },

        updateChart: function() {
            if(this._err) { return; }
            console.log('updateChart data=%o this.chart=%o', this.chartData, this.chart);
            if (this.chartData) {
                if(this.chart) {
                    this.$msg.detach();
                    this.$inlineMsg.empty();
                    var processedChartData = jschartingUtils.preprocessChartData(this.chartData, this.chart.getCurrentDisplayProperties()),
                        chartReadyData = JSCharting.extractChartReadyData(processedChartData);

                    this.chart.prepare(
                        chartReadyData,
                        jschartingUtils.getCustomDisplayProperties(chartReadyData, splunkConfig)
                    );
                    // if the preprocessChartData method truncated the data, show a message to that effect
                    if(processedChartData.columns.length > 0 &&
                            (processedChartData.columns.length < this.chartData.columns.length ||
                                processedChartData.columns[0].length < this.chartData.columns[0].length)) {

                        var rawMessage = 'These results may be truncated. Your search generated too much data for' +
                                            ' the current visualization configuration.',
                            message = _(rawMessage).t(),
                            pageInfo = utils.getPageInfo(),
                            docsHref = Route.docHelp(pageInfo.root, pageInfo.locale, 'learnmore.charting.datatruncation'),
                            helpLink = ' <a href="<%- href %>" target="_blank">' +
                                            '<span><%- text %></span>' +
                                            '<i class="icon-external icon-no-underline"></i>' +
                                        '</a>';

                        message += _(helpLink).template({ href: docsHref, text: _('Learn More').t() });
                        this.inlineMessage({
                            level: 'warn',
                            message: message
                        });
                    }
                    // otherwise if the number of results matches the max result count that was used to fetch,
                    // show a message that we might not be displaying the full data set
                    else if(this.chartData.columns.length > 0 && this.maxResultCount > 0 &&
                            this.chartData.columns[0].length >= this.maxResultCount) {
                        this.inlineMessage({
                            level: 'warn',
                            message: SplunkUtil.sprintf(
                                _('These results may be truncated. This visualization is configured to display a maximum of %s results per series, and that limit has been reached.').t(),
                                this.maxResultCount
                            )
                        });
                    }
                    this.updateChartContainerHeight();
                    if(this.chart.requiresExternalColorPalette()) {
                        var fieldList = this.chart.getFieldList();
                        SplunkLegend.setLabels(this.cid, fieldList);
                    }
                    this.invalidateChart();
                } else {
                    this.createChart();
                }
            }
        },

        destroyChart: function() {
            if (this.chart) {
                this.chart.off();
                this.chart.destroy();
                this.chart = null;
                clearTimeout(this._redrawChartTimeout);
            }
        },

        updateChartContainerHeight: function() {
            this.$chart.height(this._currentHeight - this.$inlineMsg.outerHeight());
        },

        invalidateChart: function() {
            clearTimeout(this._redrawChartTimeout);
            if(!this.chart || !this.chartData) {
                return;
            }
            var self = this;
            this._redrawChartTimeout = setTimeout(function() {
                if(self.chart.requiresExternalColorPalette()) {
                    self.setChartColorPalette();
                }
                var startTime = new Date().getTime();
                self.chart.draw().done(function() {
                    if (console.DEBUG_ENABLED) {
                        console.log('Chart=%o drawn in duration=%o ms',
                                self.model.get('display.visualizations.charting.chart'), new Date().getTime() - startTime);
                    }
                    self.trigger('rendered', self);
                    self.model.set({
                        currentChartFields: self.chart.getFieldList()
                    });
                });
            }, 5);
        },

        setChartColorPalette: function() {
            var fieldIndexMap = {};
            _(this.chart.getFieldList()).each(function(field) {
                fieldIndexMap[field] = SplunkLegend.getLabelIndex(field);
            });
            this.chart.setExternalColorPalette(fieldIndexMap, SplunkLegend.numLabels());
        },

        emitDrilldownEvent: function(e) {
            var manager = this.manager;
            var drilldownType = Drilldown.getNormalizedDrilldownType(
                this.settings.get('charting.drilldown'), {allowBoolean: true});

            var payload = Drilldown.createEventPayload({
                field: e.name2 || e.name,
                data: Drilldown.normalizeDrilldownEventData(e, {
                    manager: manager,
                    contextProperty: 'rowContext'
                }),
                event: e
            }, function(){
                Drilldown.autoDrilldown(e, manager, {
                    drilldownType: (drilldownType === 'all') ? 'all' : 'none'
                });
            });
            
            this.trigger('drilldown click', payload, this);
            this.trigger(
                (e.type === "legendClick")
                    ? "clicked:legend click:legend"
                    : "clicked:chart click:chart",
                _.extend({}, e, {
                    preventDefault: payload.preventDefault,
                    drilldown: payload.drilldown
                }),
                this);

            if (drilldownType === 'all' && this.settings.get("drilldownRedirect") && !payload.defaultPrevented()) {
                payload.drilldown();
            }
        },

        emitSelectionEvent: function(e) {
            var data = {
                start: e.startXValue,
                end: e.endXValue
            };
            if (!this.resultsModel) {
                return;
            }
            var results = this.resultsModel.data();
            if (results) {
                _(results.fields).each(function (field, idx) {
                    data['start.' + field.name] = results.columns[idx][e.startXIndex];
                    data['end.' + field.name] = results.columns[idx][e.endXIndex];
                });
            }
            var eventObj = {
                data: data,
                preventDefault: function() {
                    e.preventDefault();
                },
                results: results,
                selection: function() {
                    var data = [];
                    for (var i = e.startXIndex; i <= e.endXIndex; i++) {
                        data.push(_(results.columns).pluck(i));
                    }
                    var fields = _(results.fields).pluck('name');
                    return _(data).map(function(d) { return _.object(fields, d); });
                },
                startIndex: e.startXIndex,
                endIndex: e.endXIndex,
                startValue: e.startXValue,
                endValue: e.endXValue,
                isReset: e.isReset,
                event: e
            };
            this.trigger('selection', eventObj, this);
        },

        remove: function() {
            SplunkLegend.unregister(this.cid);
            if(this._boundInvalidateChart) {
                SplunkLegend.removeEventListener('labelIndexMapChanged', this._boundInvalidateChart);
            }
            if(this._debouncedResize) {
                $(window).off('resize', this._debouncedResize);
            }
            if(this.chart) {
                this.destroyChart();
            }
            if(this.settings) {
                this.settings.off();
                if(this.settings._sync) {
                    this.settings._sync.destroy();
                }
            }
            BaseSplunkView.prototype.remove.call(this);
        }
    });

    return ChartView;
});
