define(function(require) {
    var _ = require('underscore');
    var $ = require('jquery');
    var utils = require('./utils');
    var route = require('uri/route');
    var console = require('util/console');
    var util = require('splunk.util');
    var sharedModels = require('./sharedmodels');
    var splunkConfig = require('splunk.config');
    var IntentionsParser = require('models/services/search/IntentionsParser');
    var GeneralUtils = require('util/general_utils');

    var Drilldown = {
        /**
         * Calls the intentions parser which then generates a new search string and time range based on the original
         * search string of the manager and the click information.
         *
         * @param clickInfo {Object} - {
         *      name/value, name2/value2 {String}
         *      _span {Number}
         *      type {String}
         *  }
         * @param manager - the search manager use to derive the query and time range of the search
         * @param options {Object} - {
         *      drilldownType {String} - type of drilldown "row", "cell"
         *      sync {Boolean} - true to make the XHR call synchronous to allow for popups to open in the click
         *                     event bubble - which prevents popup blockers from interfering
         *      params {Object} - Additional intention params
         *      negate {Boolean} - Invert the drilldown intention
         *  }
         *
         * @returns A promise for the result of the intentions parser - which is an object containing the search
         *              string (q) and the time range (earliest and latest)
         */
        applyDrilldownIntention: function(clickInfo, manager, options) {
            options || (options = {});
            var applicationModel = sharedModels.get('app');
            var query = manager.settings.resolve({ qualified: true });
            var earliest = manager.job.properties().searchEarliestTime;
            var latest = manager.job.properties().searchLatestTime;

            if (!earliest && manager.job.properties().earliestTime) {
                earliest = util.getEpochTimeFromISO(manager.job.properties().earliestTime);
            }
            if (!latest && manager.job.properties().latestTime) {
                latest = util.getEpochTimeFromISO(manager.job.properties().latestTime);
            }
            var intentionsParser = new IntentionsParser();

            var intentionParams;
            if (options.drilldownType === 'geoviz') {
                intentionParams = _.extend({
                    action: 'geoviz'
                }, options.params);
                if (options.negate) {
                    intentionParams.negate = 'true';
                }
            } else if (clickInfo.type === 'legendClick') {
                if (clickInfo.hasOwnProperty('name') && clickInfo.hasOwnProperty('value')) {
                    // if the legend click has a name and value (this happens for scatter charts), use them
                    intentionParams = {
                        action: 'fieldvalue',
                        field: clickInfo.name,
                        value: clickInfo.value
                    };
                } else {
                    // otherwise do a keyword search based on name2
                    intentionParams = {
                        action: 'keyword',
                        field: clickInfo.name2
                    };
                }
            } else {
                var field, value;
                if (clickInfo._span) {
                    intentionsParser.set({
                        'dispatch.earliest_time': clickInfo.value,
                        'dispatch.latest_time': JSON.stringify(parseFloat(clickInfo.value) + clickInfo._span)
                    }, {silent: true});
                }
                if (options.drilldownType === 'row') {
                    field = _(clickInfo.rowContext).keys().map(function(f) { return f.replace('row.', ''); });
                    value = _(clickInfo.rowContext).values();
                } else if (clickInfo.name === '_time') {
                    field = clickInfo.name2;
                    value = clickInfo.value2;
                } else {
                    if (clickInfo.name && clickInfo.name2) {
                        field = [clickInfo.name, clickInfo.name2];
                        value = [clickInfo.value, clickInfo.value2];
                    } else if (clickInfo.name) {
                        field = clickInfo.name;
                        value = clickInfo.value;
                    } else {
                        field = clickInfo.name2;
                        value = clickInfo.value2;
                    }
                }
                intentionParams = {
                    action: 'fieldvalue',
                    field: field || '',
                    value: value || ''
                };
            }

            console.info('Applying intention', intentionParams);
            var dfd = $.Deferred();
            intentionsParser.fetch({
                data: _.extend({
                    q: query,
                    fieldMetaData: JSON.stringify(manager.job.properties().fieldMetadataResults),
                    app: applicationModel.get('app'),
                    owner: applicationModel.get('owner')
                }, intentionParams),
                sync: options.sync
            }).done(function() {
                    dfd.resolve({
                        'q': util.stripLeadingSearchCommand(intentionsParser.get('fullSearch')),
                        'earliest': intentionsParser.get('dispatch.earliest_time') || earliest || 0,
                        'latest': intentionsParser.get('dispatch.latest_time') || latest || ''
                    }, clickInfo.modifierKey);
                }).fail(function() {
                    dfd.reject.apply(dfd, arguments);
                });

            return dfd.promise();
        },
        // DEPRECATED - use handleAutoDrilldown instead
        handleDrilldown: function(click, drilldownType, manager) {
            console.warn('DEPRECATED Drilldown.handleDrilldown() - use Drilldown.autoDrilldown() instead');
            return Drilldown.autoDrilldown(click, manager, { drilldownType: drilldownType });
        },
        /**
         * Applies the drilldown intention and redirects to the search page when the result is ready.
         * If there is a user-defined global drilldown, that gets called instead of redirecting.
         * @see Drilldown#applyDrilldownIntention()
         */
        autoDrilldown: function(click, manager, options) {
            Drilldown.applyDrilldownIntention(click, manager, options)
                .done(splunkConfig.ON_DRILLDOWN || Drilldown.redirectToSearchPage);
        },
        /**
         * Redirect to the search page
         * @param params {Object} query string parameters to send to search view
         * @param newWindow {Boolean} true to open the search page in a new window
         */
        redirectToSearchPage: function(params, newWindow) {
            var app = sharedModels.get('app').toJSON();
            var url = route.search(app.root, app.locale, app.app, {
                data: params
            });
            utils.redirect(url, newWindow);
        },

        getNormalizedTimerange: function(manager, e) {
            if (e && e.name === "_time" && e._span) {
                var span = parseFloat(e._span);
                var earliest = parseInt(e.value, 10);
                return {
                    earliest: earliest || '',
                    latest: earliest + span || ''
                };
            } else {
                return Drilldown.getSearchJobTimerange(manager);
            }
        },

        getSearchJobTimerange: function(manager){
            var earliest, latest;
            if (manager.job) {
                earliest = manager.job.properties().searchEarliestTime;
                latest = manager.job.properties().searchLatestTime;
                if (!earliest && manager.job.properties().earliestTime) {
                    earliest = util.getEpochTimeFromISO(manager.job.properties().earliestTime);
                }
                if (!latest && manager.job.properties().latestTime) {
                    latest = util.getEpochTimeFromISO(manager.job.properties().latestTime);
                }
            }
            return {
                earliest: earliest || '',
                latest: latest || ''
            };
        },

        /**
         * Normalizes information of a core view drilldown event into a data object that is passed to drilldown
         * listeners
         *
         * @param e - event payload from a drilldown event of core view
         * @param options - {
         *      contextProperty {String}
         *      jobTimeOnly {Boolean} - force the use of the search job time range for drilldown
         * }
         * @returns {Object} - the normalized event data
         */
        normalizeDrilldownEventData: function(e, options) {
            options || (options = {});
            var data = {}, manager = options.manager;
            _.extend(data, {
                "click.name": e.name,
                "click.value": e.value
            });
            if (e.hasOwnProperty('name2')) {
                _.extend(data, {
                    "click.name2": e.name2,
                    "click.value2": e.value2
                });
            } else {
                _.extend(data, {
                    "click.name2": e.name,
                    "click.value2": e.value
                });
            }
            if (options.contextProperty) {
                _.extend(data, e[options.contextProperty]);
                if (e._span) {
                    data['row._span'] = e._span;
                }
            }

            _.extend(data, options.jobTimeOnly ?
                Drilldown.getSearchJobTimerange(manager) :
                Drilldown.getNormalizedTimerange(manager, e, options)
            );
            return data;
        },

        /**
         * Create a drilldown event payload from the given data object and additionally injects 3 methods:
         *
         * 1) preventDefault() - can be called by an event listener to prevent the default drilldown action
         * 2) defaultPrevented() - to detected whether preventDefault() has been called at least once
         * 3) drilldown() - to explicitly invoke the default drilldown action. This makes sure that the default action
         *                  is only invoked once no matter how of often drilldown() is being called.
         *
         * @param data {Object} - additional data for the event payload
         * @param defaultHandler {Function} - the default drilldown handler function
         * @return {Object} the enriched event payload
         */
        createEventPayload: function(data, defaultHandler) {
            var defaultPrevented = false;
            return _.extend(data, {
                preventDefault: function() {
                    defaultPrevented = true;
                },
                defaultPrevented: function() {
                    return defaultPrevented;
                },
                drilldown: _.once(defaultHandler)
            });
        },

        /**
         * Normalizes the given drilldown type option value
         * @param value the user-specified value
         * @param options {Object} - {
         *      validValues {Array} - all accepted string values
         *      default {String} - the default value if the specified value is not defined or invalid
         *      allowBoolean {Boolean} - if user-specified boolean values are to be accepted
         *      aliasMap {Object} - a name mapping
         * }
         */
        getNormalizedDrilldownType: function(value, options) {
            options = _.extend({ allowBoolean: false, validValues: ['all', 'none'], 'default': 'all' }, options);
            options.aliasMap = _.extend(_.object(options.validValues, options.validValues), options.aliasMap);
            if (options.aliasMap.hasOwnProperty(value)) {
                return options.aliasMap[value];
            } else if (options.allowBoolean && GeneralUtils.isBooleanEquivalent(value)) {
                return GeneralUtils.normalizeBoolean(value) ? 'all' : 'none';
            }
            return options['default'];
        }
    };

    return Drilldown;

});
