define(function(require, exports, module) {
    var _ = require('underscore');
    var mvc = require('./mvc');
    var BaseSplunkView = require("./basesplunkview");
    var Messages = require("./messages");
    
    require("async!http://maps.googleapis.com/maps/api/js?sensor=false");
    require("css!../css/googlemap.css");

    var GoogleMapView = BaseSplunkView.extend({
        moduleId: module.id,
        
        className: "splunk-googlemap",

        options: {
            managerid: null,
            data: "preview"
        },
        
        initialize: function() {
            this.configure();
            
            this.bindToComponentSetting('managerid', this._onManagerChange, this);
            
            this.map = null;
            this.markers = [];
            
            // If we don't have a manager by this point, then we're going to
            // kick the manager change machinery so that it does whatever is
            // necessary when no manager is present.
            if (!this.manager) {
                this._onManagerChange(mvc.Components, null);
            }
        },
        
        _onManagerChange: function(managers, manager) {
            if (this.manager) {
                this.manager.off(null, null, this);
                this.manager = null;
            }
            if (this.resultsModel) {
                this.resultsModel.off(null, null, this);
                this.resultsModel.destroy();
                this.resultsModel = null;
            }

            if (!manager) {
                this.message('no-search');
                return;
            }
            
            // Clear any messages, since we have a new manager.
            this.message("empty");

            this.manager = manager;            
            this.resultsModel = manager.data(this.settings.get("data"), {
                output_mode: "json_rows"
            });
            manager.on("search:start", this._onSearchStart, this);
            manager.on("search:progress", this._onSearchProgress, this);
            manager.on("search:cancelled", this._onSearchCancelled, this);
            manager.on("search:error", this._onSearchError, this);
            this.resultsModel.on("data", this.render, this);
            this.resultsModel.on("error", this._onSearchError, this);
            
            manager.replayLastSearchEvent(this);
        },

        _onSearchCancelled: function() { 
            this._isJobDone = false;
            this.message('cancelled');
        },

        _onSearchError: function(message, err) {
            this._isJobDone = false;
            var msg = message;
            if (err && err.data && err.data.messages && err.data.messages.length) {
                msg = _(err.data.messages).pluck('text').join('; ');
            }
            this.message({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },

        _onSearchProgress: function(properties) {
            properties = properties || {};
            var content = properties.content || {};
            var previewCount = content.resultPreviewCount || 0;
            var isJobDone = this._isJobDone = content.isDone || false;
            
            if (previewCount === 0 && isJobDone) {
                this.message('no-results');
                return;
            }
            
            if (previewCount === 0) {
                this.message('waiting');
                return;
            }
        },

        _onSearchStart: function() { 
            this._isJobDone = false;
            this.message('waiting');
        },

        clearMarkers: function() {
            var count = this.markers.length;
            for (var i = 0; i < count; ++i)
                this.markers[i].setMap(null);
            this.markers.length = 0;
        },

        createMap: function() {
            this.map = new google.maps.Map(this.el, {
                center: new google.maps.LatLng(47.60, -122.32),
                zoom: 2,
                mapTypeId: google.maps.MapTypeId.ROADMAP
            });
            this.map.setOptions(this.options);
        },

        message: function(info) {
            this.map = null;
            Messages.render(info, this.$el);
        },

        render: function() {
            if (!this.manager) {
                return;
            }
            
            if (!this.resultsModel || !this.resultsModel.hasData()) {
                if (this.resultsModel && !this.resultsModel.hasData() && this._isJobDone) {
                    this.message("no-results");
                }
                return this;
            }
            
            if (!this.map) {
                this.createMap();
            }

            var that = this;
            this.clearMarkers();
            this.resultsModel.collection().each(function(row) {
                var lat = parseFloat(row.get("lat"));
                var lng = parseFloat(row.get("lng"));
                var latlng = new google.maps.LatLng(lat, lng);
                var marker = new google.maps.Marker({
                    position: latlng,
                    map: that.map
                });
                that.markers.push(marker);
            });

            return this;
        }
    });
    
    return GoogleMapView;
});
