define(function(require, exports, module) {
    var _ = require("underscore");
    var mvc = require('./mvc');
    var utils = require('./utils');
    var BaseManager = require('./basemanager');
    var SearchModels = require('./searchmodel');
    var SplunkResultsModel = require('./splunkresultsmodel');
    var Messages = require('./messages');

    function mergeSearch(base, sub) {
        if (!sub) {
            return base;
        }
        return [ base.replace(/[\|\s]$/g,''), sub.replace(/^[\|\s]/g,'') ].join(' | ');
    }

    var PostProcessResultsModel = SplunkResultsModel.extend({
        _requestOptions: function() {
            var options = SplunkResultsModel.prototype._requestOptions.call(this);
            var manager = this.get('manager');
            
            var postProcessSearch = manager.settings.postProcessResolve();
            if (postProcessSearch) {
                options.search = mergeSearch(postProcessSearch, options.search);
            }
            return options;
        }
    });
    
    var PostProcessSearchSettingsModel = SearchModels.SearchQuery.extend({
        resolve: function(options) {
            options || (options = {});
            if (!this.isResolved() && options.tokens !== true) {
                return undefined;
            }
            
            var parentSearch;
            if (this._manager.parent) {
                parentSearch = this._manager.parent.settings.resolve(options);
            }
            if (parentSearch === undefined) {
                return undefined;
            }
            // Fetch the post-process search but force it to be unqualified - ie. we don't want a leading search command
            // since we're about to merge it with the base search
            var thisSearch = SearchModels.SearchQuery.prototype.resolve.call(this, _.defaults({ qualified: false }, options));
            
            return mergeSearch(parentSearch, thisSearch);
        },
        
        postProcessResolve: function() {
            return SearchModels.SearchQuery.prototype.resolve.apply(this, arguments);
        },
        
        isResolved: function() {
            var resolvedSearch = SearchModels.SearchQuery.prototype.resolve.apply(this, arguments);
            var tokens = this.get('search', {tokens: true});
            return resolvedSearch !== undefined || tokens === undefined;
        }
    });

    var PostProcessSearchManager = BaseManager.extend({
        moduleId: module.id,

        defaults: {
            managerid: null,
            search: ''
        },        
        
        constructor: function(attributes, options) {
            attributes = attributes || {};
            options = options || {};
            
            // This has to be in the constructor, otherwise
            // we will call Model.set before we have created this sub-model.
            // Note that the Backbone.Model constructor will call 'set', 
            // which in turn will call our 'set' override, which will forward
            // the search query to the query model.
            // We also make sure to deal with the legacy 'postProcess' attribute.
            attributes.search = attributes.postProcess || attributes.search;
            this.settings = new PostProcessSearchSettingsModel({}, options);
            this.settings._manager = this;
            
            // Permit deprecated access via the 'query' field
            this.query = this.settings;

            return BaseManager.prototype.constructor.apply(this, arguments);
        },
        
        initialize: function(attrs, options) {
            attrs = attrs || {};
            options = options || {};
            
            // Initialize the data attribute to the empty object, so that we
            // never look at our parent's data attribute.
            this.set("data", {});
            
            this.settings.on('change', this._handleSettingsChange, this);
            
            // Whenever the query changes, we can start our search.
            this.settings.on('change:search', this.startSearch, this);
            
            this.on('all', function(evt) {
                if (evt.indexOf("search:") === 0) {
                    this._lastSearchEvent = [ evt, Array.prototype.slice.call(arguments, 1) ];
                }
            }, this);
            
            // Listen to our parent manager being created
            var managerid = attrs.managerid || attrs.manager;
            mvc.Components.bind('change:' + managerid, this.onManagerChange, this);
            if (mvc.Components.has(managerid)) {
                this.onManagerChange(mvc.Components, mvc.Components.get(managerid));
            }
        },
        
        set: function(key, value, options) {
            // This is reduced logic from what exists in SearchManager, as we 
            // only allow for a query model, we don't allow for a custom one,
            // etc. We do need to be able to forward the 'set' calls though,
            // so it is necessary.
            var attrs;
            
            // Normalize the key-value into an object
            if ((_.isObject(key) && !_.isArray(key)) || key === null) {
                attrs = key;
                options = value;
            } else {
                attrs = {};
                attrs[key] = value;
            }
            
            // Forward all attributes to settings model
            this.settings.set(attrs, options);
            
            // Continue propagating attributes to self for backward compatibility
            BaseManager.prototype.set.call(this, attrs, options);
        },
        
        _handleSettingsChange: function() {
            var managerPropertiesChanged = _.omit(
                this.settings.changed,
                SearchModels.SearchSettingsModel.ALLOWED_ATTRIBUTES);
            
            // Continue propagating non-job attributes to self for backward compatibility
            BaseManager.prototype.set.call(this, managerPropertiesChanged);
        },
        
        startSearch: function() {
            if (this.parent && this.parent.job) {
                var job = this.parent.job;
                var state = job.state();
                
                // We have to trigger a full cycle of start/progress/done if
                // appropriate.
                this._onSearchStart(job);
                this._onSearchTick("search:progress", state, job);
                
                if (state && state.content && state.content.isDone) {
                    this._onSearchTick("search:done", state, job);
                }
            }
        },
        
        onManagerChange: function(x, ctx) {
            if (this.parent) {
                this.parent.off(null, null, this);
                delete this.job;
            }
            
            if (!ctx) {
                return;
            }
            
            this.parent = ctx;
            
            // We need to be able to calculate the count for the post process
            this._countData = this.data("preview", {
                autofetch: false,
                output_mode: "json",
                search: 'stats count'
            });
            this._countData.on("error", _.partial(this.trigger, "search:error"), this);
            
            // We use the parent's search, so that any search paramater changes
            // get propagated to it.
            this.search = this.parent.search;
            
            // We set up handlers for the various search events
            ctx.on('search:start',      this._onSearchStart,                                    this);
            ctx.on('search:progress',   _.partial(this._onSearchTick,   "search:progress"),     this);
            ctx.on('search:done',       _.partial(this._onSearchTick,   "search:done"),         this);
            ctx.on('search:cancelled',  _.partial(this.trigger,         "search:cancelled"),    this);
            ctx.on('search:fail',       _.partial(this.trigger,         "search:fail"),         this);
            ctx.on('search:error',      _.partial(this.trigger,         "search:error"),        this);
            
            ctx.replayLastSearchEvent(this);
        },
        
        get: function(/* args */) {
            // Try to get property from own settings, then from parent.
            var setting = this.settings.get.apply(this.settings, arguments);
            if (setting === undefined && this.parent) {
                setting = this.parent.get.apply(this.parent, arguments);
            }
            return setting;
        },
        
        _onSearchStart: function(job) {
            this.job = job;
            
            if (this.settings.isResolved()) {
                this.trigger("search:start", job);
            }
            else {
                this.trigger("search:error", Messages.resolve('unresolved-search').message);
            }
        },
        
        _onSearchTick: function(eventName, properties, job) {
            // If we don't have a query (e.g. it is not set, or it is not resolved),
            // then do nothing.
            if (!this.settings.isResolved()) {
                return;
            }
            
            // We have to create a clone of properties and content, as we are
            // going to be modifying them.
            properties = _.clone(properties || {});
            var content = properties.content = _.clone(properties.content || {});
            var previewCount = content.resultPreviewCount || 0;

            var that = this;
            var trigger = function() {
                that.trigger(eventName, properties, job);
                
                if (eventName === "search:progress") {
                    // Backbone does a deep equality check on sets, and since
                    // properties is an object, and the object contents may have
                    // not changed (e.g. due to a cached job), Backbone might not
                    // fire a "change" event. This ensures that a change
                    // event is fired every time by doing a silent unset followed
                    // by a set.
                    that.unset("data", {silent: true});
                    that.set("data", properties.content);
                }
            };

            // Regardless if the parent search has any results, we always look
            // at the post process result count. This is because the post process
            // can both remove results and/or create results, and as such, the
            // parent search counts are not reliable.
            this._countData.fetch({
                success: function() {
                    var count = 0;
                    if (that._countData.hasData()) {
                        var data = that._countData.data();
                        count = parseInt(data.results[0].count, 10);
                    }
                    
                    content.resultPreviewCount = count;
                    content.resultCount = count;
                    trigger();
                }
            });
        },
        
        data: function(source, args) {
            if (!source) {
                throw new Error("Cannot get a results model without specifying the source.");
            }
            
            args = _.defaults({ manager: this, source: source }, args);
            return new PostProcessResultsModel(args);
        },
        
        replayLastSearchEvent: function(listener) {
            // We want to look at our own search events (as they have our modified
            // counts), and not our parent's search events.
            var replayed = false;
            if (this._lastSearchEvent) {
                var lastSearchEvent = this._lastSearchEvent;
                var evtName = lastSearchEvent[0];
                var listenerObj = _(this._events[evtName]).find(function(attachedListener) { 
                    return attachedListener.ctx === listener; 
                });
                if (listenerObj) {
                    var args = lastSearchEvent[1];
                    replayed = true;
                    listenerObj.callback.apply(listener, args);
                }
            }
            return replayed;
        }
    });

    return PostProcessSearchManager;
});