define(function(require, exports, module) {
    var mvc = require('./mvc');
    var SearchManager = require('./searchmanager');
    var _ = require("underscore");    
    var console = require('util/console');
    var sdk = require('splunkjs/splunk');
    var Messages = require('./messages');
    var Async = require('../splunk').Async;
                   
    // When dispatching saved searches, we use the 'dispatch.<property>=X' form,
    // as opposed to when creating search jobs, where we use just '<property>=X'
    // form. Furthermore, some 'dispatch.' property names differ from their
    // search job counterparts: 
    //  * dispatch.buckets ~= status_buckets
    //  * dispatch.lookups ~= enable_lookups
    // Given that, the following two lists have both what properties can be
    // dispatched, as well as their variants and mappings.
    var DISPATCHABLE_PROPERTIES = {
        "status_buckets": true,
        "buckets": true, 
        "earliest_time": true,
        "indexedRealtime": true,
        "latest_time": true,
        "enable_lookups": true,
        "lookups": true,
        "max_count": true,
        "max_time": true,
        "reduce_freq": true,
        "rt_backfill": true,
        "spawn_process": true,
        "time_format": true,
        "ttl": true,
        "auto_cancel": true,
        "auto_pause": true
    };

    // Dispatchable properties specific to the savedsearch endpoint
    var ENDPOINT_PROPERTIES = {
        'trigger_actions': true
    };

    // Some of the dispatch properties have shortened names
    var DISPATCH_MAPPINGS = {
        'status_buckets': 'buckets',
        'enable_lookups': 'lookups'
    };

    var SavedSearchManager = SearchManager.extend({
        moduleId: module.id,

        defaults: {
            autostart: true,
            cache: false,
            cancelOnUnload: true,
            /**
             * When false, this manager will not run its search if its 'earliest_time' or 'latest_time'
             * settings are bound to unresolved tokens.
             *
             * Currently defaults to true, but is likely to default to false in future versions of
             * the framework.
             *
             * initialization only
             *
             * @deprecated
             */
            runWhenTimeIsUndefined: true
        },

        handleAutostart: function() {
            if (this.get("autostart")) {
                this.on("change:searchname", this._debouncedStartSearch, this);
            }
            else {
                this.off("change:searchname", this._debouncedStartSearch, this);
            }
            
            return SearchManager.prototype.handleAutostart.apply(this, arguments);
        },
        
        startSearch: function() {
            // We do not cancel any searches, as this is a saved search
            // and others might be using the same search

            if (!this.get("searchname")) {
                this.trigger("search:error", _("No saved search provided.").t());
                return;
            }

            // Get the cache and convert it to a number:
            // false -> 0, never use cache
            // true -> -1, always use cache
            // positive integer -> positive integer, use cache if newer than
            var cache = this.get("cache") || false;
            if (_.isBoolean(cache)) {
                cache = cache ? -1 : 0;
            }
            
            // Get the threshold for cache (which could be infinity)
            var threshold = cache < 0 ? 
                (new Date(0)).valueOf() :
                ((new Date()).valueOf() - cache * 1000);
            
            var search = new sdk.Service.SavedSearch(this.service, this.get("searchname"), {
                owner: this.service.owner,
                app: this.service.app,
                sharing: this.service.sharing
            });
            
            var that = this;
            search.fetch(function(err, search) {
                if (err) {
                    if (err.status === 404) {
                        var message = _("Warning: saved search not found: ").t() + '"' + that.get("searchname") + '"';
                        console.log(message);
                        that.trigger("search:error", message);
                    } else {
                        console.log("Error fetching saved searches");
                        that.trigger("search:error", _("Error fetching saved searches").t());
                    }
                    return;
                }
                
                // Set the query so that someone can fetch it
                that.settings.set("search", search.properties().search, {silent: true});
                
                // Skip fetching the saved search history if cache is disabled 
                // and directly dispatch it
                var cacheDisabled = cache === 0 || (cache === 'scheduled' && !search.properties().is_scheduled);
                if (cacheDisabled) {
                    that._startSearchWithNewJob(search);
                    return;
                }
                
                search.history(function(err, list) {
                    // Find all jobs that match
                    var matches = [];
                    for (var i = 0; i < list.length; ++i) {
                        var potentialJob = list[i];
                        var properties = potentialJob.properties() || {};
                        var published = (new Date(potentialJob.published())).valueOf();
                        var isScheduled = (cache === "scheduled") && properties.isScheduled;

                        // We should reuse running realtime searches 
                        // This is behavior is consistent with 5.0.x
                        // be aware of realtime alerts search isRealtime but isDone
                        var isRunningRealTimeSearch = properties.isRealTimeSearch && !properties.isDone;

                        // If we are reusing scheduled jobs and this one was scheduled
                        // or this job is fresh enough, then we can reuse it
                        // of this one is realtime and still running.
                        if (isScheduled || published > threshold || isRunningRealTimeSearch) {
                            matches.push(potentialJob);
                        }
                    }
                    
                    // Sort to easily find the latest one
                    matches.sort(function(a,b) { 
                        var aValue = (new Date(a.published())).valueOf();
                        var bValue = (new Date(b.published())).valueOf();
                        return bValue - aValue;
                    });
                    
                    // If a job exists, then we will use it, otherwise,
                    // create a new job.
                    var job = matches[0];
                    if (job) {
                        that._startSearchWithExistingJob(job);
                    }                
                    else {
                        that._startSearchWithNewJob(search);
                    }
                });
            });

            return this;
        },
        
        // Mocked by unit tests.
        _startSearchWithExistingJob: function(job) {
            this.createManager(job);
        },
        
        // Mocked by unit tests.
        _startSearchWithNewJob: function(search) {
            var that = this;

            // Don't start search if waiting for time range
            var timeRangeIsUnresolved =
                (this.settings.get("earliest_time", {tokens: false}) === undefined &&
                 this.settings.get("earliest_time", {tokens: true})) ||
                (this.settings.get("latest_time", {tokens: false}) === undefined &&
                 this.settings.get("latest_time", {tokens: true}));
            if (timeRangeIsUnresolved && !this.settings.get("runWhenTimeIsUndefined")) {
                this.trigger("search:error", Messages.resolve("unresolved-search").message);
                return;
            }

            search.dispatch(this._createOptionsForSearchJob(), function(err, job) {
                if (err) {
                    that.trigger("search:error", _("Error dispatching saved search.").t(), err, job);
                    return;
                }
                
                // SPL-88606 - If enablepreview is invoked to quickly we sometimes get a 204 response and preview
                // is not enabled after that. In this case we just retry it until we get a different response 
                // code (such as 200). 
                var waitForPreviewEnabled = true;
                Async.whilst(
                    function(){ return waitForPreviewEnabled; },
                    function(done){
                        job.post('control', { action: 'enablepreview'}, function(err, resp){
                            waitForPreviewEnabled = resp.status === 204;
                            done();
                        });
                    },
                    function(err){
                        console.log('Error enabling preview for saved search job', err);
                    }
                );

                that.createManager(job);
            });
        },
        
        _createOptionsForSearchJob: function() {
            var options = {};
            _.each(this.settings.toJSON(), function(val, key) {
                var dispatchableProperty = _.has(DISPATCHABLE_PROPERTIES, key);
                var endpointProperty = _.has(ENDPOINT_PROPERTIES, key);
                // If there is no value or this is not a valid dispatch.* key,
                // then we ignore it
                if (val === undefined) {
                    return;
                } 
                
                // Prefix the key and map it if necessary (e.g. status_buckets -> buckets)
                if (dispatchableProperty) {
                    var prefixedKey = "dispatch." + (DISPATCH_MAPPINGS[key] || key);
                    options[prefixedKey] = val;
                } else if (endpointProperty) {
                    options[key] = val;
                }

            });
            
            return options;
        }
    });
    
    return SavedSearchManager;
});
