define(function(require, exports, module) {
    var $ = require("jquery");
    var _ = require("underscore");
    var mvc = require('./mvc');
    var Backbone = require("backbone");
    var BaseSplunkView = require("./basesplunkview");
    var console = require('util/console');
    var TimeRangeView = require("./timerangeview");
    var InternalSearchBar = require("views/shared/searchbar/Master");
    var TimeRangeModel = require('models/shared/TimeRange');
    var utils = require('./utils');
    var sharedModels = require('./sharedmodels');
    var splunkConfig = require('splunk.config');

    require("css!../css/search-bar");
    
    var SearchBarView = BaseSplunkView.extend({
        moduleId: module.id,
        
        className: "splunk-searchbar",
        
        options: {
            "default": undefined,
            managerid: null,
            /**
             * Whether or not to display an embedded time range view
             * in this search bar.
             */
            timerange: true,
            /**
             * Whether to display the search assistant.
             * 
             * When the framework is in independent-mode the search assistant
             * does not function and this option will be ignored.
             * 
             * Initialization-only.
             */
            useAssistant: true,
            /**
             * Whether to automatically open the search assistant
             * while the search bar is being typed in.
             * 
             * Only applies if useAssistant is true.
             * 
             * Initialization-only.
             */
            autoOpenAssistant: true,
            value: undefined
        },
        
        initialize: function() {
            var that = this;
            
            this.configure();
            this.settings.enablePush("value");
            
            if (this.settings.has('timepicker')) {
                console.warn(
                    'The "%s" setting of class "%s" is deprecated. Use "%s" instead.',
                    'timepicker', 'SearchBarView', 'timerange');
                this.settings.set('timerange', this.settings.get('timepicker'));
                this.settings.unset('timepicker');
            }
            
            // Initialize value with default, if provided
            this._onDefaultChange();
            
            this._state = new Backbone.Model({
                'dispatch.earliest_time': this.settings.get("earliest_time"),
                'dispatch.latest_time': this.settings.get("latest_time"),
                'search': this.settings.get('value') || ""
            });
            
            // Get the shared models
            var appModel = sharedModels.get("app");
            var userModel = sharedModels.get("user");
            var appLocalModel = sharedModels.get("appLocal");
            var timesCollection = sharedModels.get("times");
            var serverInfo = sharedModels.get("serverInfo");

            var timeRangeModel = new TimeRangeModel();
            timeRangeModel.save({
                'earliest': this.settings.get("earliest_time"),
                'latest': this.settings.get("latest_time")
            });
            
            // Create embedded time range view even if we don't
            // plan to actually show it.
            this.timerange = this._createTimeRange(this.settings);
            
            // Permit deprecated access to the 'timepicker' field
            this.timepicker = this.timerange;
            
            // We cannot create the searchbar until these internal models
            // have been fetched, and so we wait on them being done.
            this._dfd = $.when(timesCollection.dfd, userModel.dfd, appLocalModel.dfd, serverInfo.dfd).done(function() {
                var useAssistant = 
                    that.settings.get('useAssistant') &&
                    // JIRA: Remove search assistant's dependency on splunkweb,
                    //       making it depend on splunkd directly.
                    //       Then enable the search assistant in independent
                    //       mode. (SPL-80734)
                    !splunkConfig.INDEPENDENT_MODE;
                
                var autoOpenAssistant = 
                    that.settings.get('autoOpenAssistant') &&
                    useAssistant;
                
                that.searchbar = new InternalSearchBar({
                    showTimeRangePicker: true,
                    useAssistant: useAssistant,
                    autoOpenAssistant: autoOpenAssistant,
                    disableOnSubmit: false,
                    collection: {
                        times: timesCollection
                    },
                    model: {
                        state: that._state,
                        timeRange: timeRangeModel,
                        user: userModel,
                        appLocal: appLocalModel,
                        application: appModel,
                        serverInfo: serverInfo
                    }
                });
            });
            
            // Update view if model changes
            this.settings.on("change:value", function(model, value, options) {
                options = options || {};
                var suppressValSet = options._self;
                if (!suppressValSet) {
                    that.val(value || "");
                }
            });
            
            this.bindToComponentSetting('managerid', this._onManagerChange, this);
            
            this._state.on("change:search", this._onSearchChange, this);
            this.settings.on("change:timerange", this._onDisplayTimeRangeChange, this);
            this.settings.on("change:default", this._onDefaultChange, this);
        },
        
        _createTimeRange: function(settings) {
            var timeRangeOptions = settings.extractWithPrefix('timerange_');
            var timePickerOptions = settings.extractWithPrefix('timepicker_');
            if (!_.isEmpty(timePickerOptions)) {
                console.warn(
                    'The "%s" settings of class "%s" are deprecated. Use "%s" instead.',
                    'timepicker_*', 'SearchBarView', 'timerange_*');
            }
            
            var options = _.extend(
                { managerid: settings.get("managerid") },
                timeRangeOptions,
                timePickerOptions);
            
            return new TimeRangeView(options);
        },
        
        _onDefaultChange: function(model, value, options) {
            // Initialize value with default, if provided
            var oldDefaultValue = this.settings.previous("default");
            var defaultValue = this.settings.get("default");
            var currentValue = this.settings.get('value');
            
            if (defaultValue !== undefined &&
                (currentValue === oldDefaultValue || currentValue === undefined))
            {
                this.settings.set('value', defaultValue);
            }
        },
        
        _onDisplayTimeRangeChange: function() {
            // We cannot work with the searchbar/timerange until they
            // are done being created.
            var that = this;
            $.when(this._dfd).done(function() {
                if (that.settings.get("timerange")) {
                    that.searchbar.children.timeRangePicker.$el.show();
                }
                else {
                    that.searchbar.children.timeRangePicker.$el.hide();
                }
            });
        },
        
        _onManagerChange: function(managers, manager) {
            if (this.manager) {
                this.manager.off(null, null, this);
                this.manager = null;
            }

            this.manager = manager;
            
            // We defer setting the query to let the underlying search bar
            // have enough to finish setting up. Since we might get a 
            // a manager synchronously, it may have not finished setting up
            // (e.g. it is setting its own deferred actions).
            var that = this;
            _.defer(function() {
                that._updateQuery();
            });
        },
        
        render: function() {
            // We cannot work with the searchbar/timerange until they
            // are done being created.
            var that = this;
            $.when(this._dfd).done(function() {
                if (!that._rendered) {
                    that.searchbar.render();
                    that._patchTimePicker();
                    
                    that.$el.append(that.searchbar.el);
                    
                    // Ensure we properly show/hide the timerange
                    that._onDisplayTimeRangeChange();
                    
                    // Prevent multiple renderings
                    that._rendered = true;
                }
            });
                            
            return this;
        },
        
        _patchTimePicker: function() {
            // Patch the internal search bar to display our time range view
            // (which may have special customizations) instead of the search
            // bar's original internal time range view.
            // 
            // The MVC time range view is patched in instead of its
            // underlying internal time range for backward compatibility
            // with existing CSS. We may break this compatibility later.
            this.searchbar.children.timeRangePicker.setElement(
                this.timerange.el);
            
            // Manually rerender the time range view in the DOM
            this.timerange.render();
            this.searchbar.$('.search-timerange')
                .empty()
                .append(this.timerange.el);
            
            // Add back the 'btn-group' CSS class, which is necessary
            // for the vertical divider between the time range view
            // and the search button to display.
            $(this.timerange.el).addClass('btn-group');
            
            // Patch our time range view to synchronize its state with
            // the internal search bar's internal time range view.
            utils.syncModels(
                // embedded internal time range's model
                this.timerange.timepicker.model.timeRange,
                // internal search bar's internal time range's model
                this.searchbar.children.timeRangePicker.model.timeRange,
                // bidirectional sync; initialize with first model
                { auto: true });
        },
        
        val: function(value) {
            if (value !== undefined) {
                this._setSearch(value);
                
                /* 
                 * Force firing of a new change event, even if the new
                 * value is the same as the old value. This provides
                 * the expected behavior if the user presses enter in
                 * a search box to refresh the search.
                 */
                this.settings.unset("value", {_self: true});
                this.settings.set("value", value, {_self: true});
                
                this.trigger("change", value, this);
            }
            else {
                return this._getSearch();
            }
        },
        
        _onSearchChange: function(model, value, options) {
            options = options || {};
            var suppressValSet = options._self;
            if (!suppressValSet) {
                this.val(value);
            }
        },
        
        _getSearch: function() {
            return this._state.get("search");
        },
        
        _setSearch: function(newSearch) {
            this._state.set("search", newSearch, {_self: true});
        },
        
        _updateQuery: function() {
            // If we have a previous search query set, display it
            if (this.manager) {
                var currentSearch = this._state.get("search") || "";
                var newSearch = (this.manager.settings.resolve() || "").trim();
                
                if (!currentSearch && newSearch) {
                    this._setSearch(newSearch);
                }
            }
        }
    });
    
    return SearchBarView;
});
