define(function(require, exports, module) {
    var _ = require("underscore");
    var $ = require("jquery");
    var mvc = require('./mvc');
    var Async = require('../splunk').Async;
    var BaseManager = require('./basemanager');
    var SplunkResultsModel = require('./splunkresultsmodel');
    var SearchModels = require('./searchmodel');
    var sdk = require('splunkjs/splunk');
    var splunkConfig = require('splunk.config');
    var TokenSafeString = require("./tokensafestring");
    var console = require('util/console');
    var JobTracker = require('./jobtracker');
    var Messages = require('./messages');
    
    var PROPERTIES_IGNORED_BY_CACHE = [
        // Doesn't affect results. Could be randomly generated which would
        // unintentionally bust the cache.
        'label',
        // Doesn't affect results. Ignoring this keeps old unit tests passing.
        'webframework.cache.hash'
    ];
    
    // DOC: The public 'settings' field contains this component's settings model.
    // DOC: The public 'search' field is deprecated. Use the 'settings' field instead.
    // DOC: The public 'query' field is deprecated. Use the 'settings' field instead.
    var SearchManager = BaseManager.extend({ 
        moduleId: module.id,

        defaults: {
            autostart: true,
            cache: false,
            cancelOnUnload: true,
            runWhenTimeIsUndefined: true
        },

        /**
         * Creates a SearchManager.
         * 
         * @param attributes {
         *      id: The ID of this component.
         *      *: See SearchSettingsModel.ALLOWED_ATTRIBUTES.
         * }
         */
        constructor: function(attributes, options) {
            attributes = attributes || {};
            options = options || {};
            
            // Create settings model
            if (attributes.queryModel && attributes.searchModel) {
                // There is no sensible way to reinterpret a simultaneous
                // usage of both the deprecated 'queryModel' and 'searchModel'
                // options
                throw new Error(
                    'Cannot specify two custom settings models via ' +
                    'deprecated options "queryModel" and "searchModel".');
            }
            this.settings =
                // Internal. Allows custom settings model to be provided.
                attributes.settings ||
                // Support deprecated internal option 'queryModel'
                attributes.queryModel ||
                // Support deprecated internal option 'searchModel'
                attributes.searchModel ||
                // Create standard settings model if no custom one provided
                new SearchModels.SearchSettingsModel({}, options);
            
            // Alias deprecated public field 'query' to the combined settings model
            this.query = this.settings;
            // Alias deprecated public field 'search' to the combined settings model
            this.search = this.settings;
            
            // No need to set it on our model
            delete attributes.settings;
            delete attributes.queryModel;
            delete attributes.searchModel;

            var returnValue = BaseManager.prototype.constructor.apply(this, arguments);
            
            return returnValue;
        },

        initialize: function(options) {
            options = options || {};
            
            this.service = mvc.createService({app: options.app, owner: options.owner});
            if (!this.settings.has('label')) {
                // Preserve backward compatibility with Framework 1.0 which
                // always created search jobs labeled the same as the
                // associated search manager's ID by default.
                this.settings.set('label', this.id);
            }
            
            this.listenTo(this.settings, 'change', this._handleSettingsChange);
            
            this._debouncedStartSearch = _.debounce(this.startSearch);

            // We want to change behavior depending on whether autostart is set
            // or not. So we first bind a listener to it changing, and then
            // we invoke that listener manually
            this.listenTo(this, "change:autostart", this.handleAutostart);
            this.listenTo(this, "search:error", function(msg) { this.lastError = msg; return true; });
            this.listenTo(this, 'all', function(evt) {
                if (evt.substring(0, 'search:'.length) === 'search:') {
                    this._lastSearchEvent = [ evt, Array.prototype.slice.call(arguments, 1) ];
                }
            });
            this.handleAutostart();

            // Kick off the search. This will only do anything if autostart is enabled
            this._start();

            this._cancelOnUnload = _.bind(this._cancelOnUnload, this);
            this.listenTo(this, 'change:cancelOnUnload change:cache', this.handleCancelOnUnload);
            this.handleCancelOnUnload();
        },
        
        _handleSettingsChange: function() {
            var managerPropertiesChanged = _.omit(
                this.settings.changed,
                SearchModels.SearchSettingsModel.ALLOWED_ATTRIBUTES);
            
            // Continue propagating non-job attributes to self for backward compatibility
            BaseManager.prototype.set.call(this, managerPropertiesChanged);
        },

        handleAutostart: function() {
            if (this.get("autostart")) {
                this.listenTo(this.settings, "change", this._handleSettingsChangeOnAutostart, this);
            }
            else {
                this.stopListening(this.settings, "change", this._handleSettingsChangeOnAutostart, this);
            }
        },
        
        _handleSettingsChangeOnAutostart: function() {
            var jobPropertiesChanged = _.pick(
                this.settings.changed,
                SearchModels.SearchSettingsModel.ALLOWED_ATTRIBUTES);
            
            if (!_.isEmpty(jobPropertiesChanged)) {
                this._debouncedStartSearch();
            }
        },

        handleCancelOnUnload: function() {
            this.stopListeningDOM($(window), 'unload', this._cancelOnUnload);
            if (this.get('cancelOnUnload') && !this.get('cache')) {
                this.listenToDOM($(window), 'unload', this._cancelOnUnload);
            }
        },

        _cancelOnUnload: function() {
            // Cancel job if it isn't already done
            if (this.job && !this.job.properties().isDone) {
                // JIRA: Should be delegating this to the SDK through mvc.createService()
                //       so that the assumption about proxy authentication doesn't
                //       need to be encoded here. (DVPL-4084)
                $.ajax({
                    url: splunkConfig.SPLUNKD_PATH + this.job.qualifiedPath + '/control',
                    data: { action: 'cancel' },
                    type: 'POST',
                    // Force a synchronous XHR call since an asynchronous request might have no
                    // effect in the unload phase
                    async: false
                });
            }
        },

        set: function(key, value, options) {         
            var attrs;
            
            // Normalize the key-value into an object
            if ((_.isObject(key) && !_.isArray(key)) || key === null) {
                attrs = key;
                options = value;
            } else {
                attrs = {};
                attrs[key] = value;
            }
            
            var settings = {};
            
            // If the deprecated 'search' or 'query' input attributes are
            // present then use them as the base set of output attributes
            // (and filter them out of the original input attributes).
            if (_.has(attrs, "search") && this._isDictionary(attrs.search)) {
                _.extend(settings, attrs.search);
                delete attrs.search;
            }
            if (_.has(attrs, "query") && this._isDictionary(attrs.query)) {
                _.extend(settings, attrs.query);
                delete attrs.query;
            }
            
            // Forward all attributes to settings model
            _.extend(settings, attrs);
            this.settings.set(settings, options);
            
            // Continue propagating attributes to self for backward compatibility
            BaseManager.prototype.set.call(this, attrs, options);
        },

        get: function(attr) {
            return this.settings.get.apply(this.settings, arguments);
        },
        
        _isDictionary: function(value) {
            // NOTE: This logic is fragile. Unfortunately there's no good
            //       way to detect a raw dictionary (via an object literal)
            //       as opposed to an object created through the 'new' keyword.
            return _.isObject(value) && !(value instanceof TokenSafeString);
        },
        
        _start: function() {         
            if (this.get("autostart")) {
                this.startSearch();
            }

            return this;
        },
        
        data: function(source, attrs) {
            if (!source) {
                throw new Error("Cannot get a results model without specifying the source.");
            }
            
            attrs = attrs || {};
            attrs.manager = this;
            attrs.source = source;
            return new SplunkResultsModel(attrs);
        },
        
        _startSearchFromSid: function(sid) {
            var that = this;
            var job = new sdk.Service.Job(this.service, sid);
            job.fetch(function(err) {
                if (err) {
                    that.trigger("search:error", _("Error in getting pre-existing search: ").t() + sid);
                    return;
                } 
                
                // Set our job and query model to the request parameters
                // for this search, in case we need to make any changes
                var request = job.properties().request;
                that.settings.set(
                    _.pick(request, SearchModels.SearchSettingsModel.ALLOWED_ATTRIBUTES), 
                    {silent: true, qualified: true}
                );
                
                that.createManager(job);
            });
        },
        
        startSearch: function(refresh) {
            this.lastError = '';
            this.refresh = refresh;
            this.cancel(); // Cancel any existing search

            // If we were provided with a SID, then we will
            // just try and use that
            if (this.has("sid")) {
                this._startSearchFromSid(this.get("sid"));
                return;
            }

            // Get the cache and convert it to a number:
            // false -> 0, never use cache
            // true -> -1, always use cache
            // positive integer -> positive integer, use cache if newer than
            var cache = this.get("cache") || false;
            if (_.isBoolean(cache)) {
                cache = cache ? -1 : 0;
            }
            
            var options = this._createOptionsForSearchJob();
            
            // Get the threshold for cache (which could be infinity)
            var threshold = cache < 0 ? 
                (new Date(0)).valueOf() :
                ((new Date()).valueOf() - cache * 1000);
            
            // Only start a search if we have a search string
            if (!options.search) {
                if (this.settings.get("search", {tokens: true})) {
                    this.trigger("search:error", Messages.resolve("unresolved-search").message);
                }
                else {
                    this.trigger("search:error", _("No search query provided.").t());   
                }
                return;
            }
            
            // Don't start search if waiting for time range
            var timeRangeIsUnresolved = 
                (this.settings.get("earliest_time", {tokens: false}) === undefined &&
                 this.settings.get("earliest_time", {tokens: true})) ||
                (this.settings.get("latest_time", {tokens: false}) === undefined &&
                 this.settings.get("latest_time", {tokens: true}));
            if (timeRangeIsUnresolved && !this.settings.get("runWhenTimeIsUndefined")) {
                this.trigger("search:error", Messages.resolve("unresolved-search").message);
                return;
            }

            // Skip fetching existing jobs if cache is disabled and directly create a new job
            if (cache === 0) {
                this.createSearch();
                return this;
            }
            
            // Filter on a hash derived from the job options as a performance
            // optimization to avoid enumerating every job on the server.
            // See _createOptionsForSearchJob() for this hash's computation.
            var filter = (options['webframework.cache.hash'])
                ? { search: 'request.webframework.cache.hash=' + options['webframework.cache.hash'] }
                : {};

            var that = this;

            // Check and see if a matching job already exists.
            this.service.jobs().fetch(filter, function(err, jobs) {
                if (err) {
                    console.log("Error fetching searches");
                    that.trigger("search:error", _("Error fetching searches").t());
                    return;
                }
                
                var list = jobs.list();
                
                // Splunk returns values as strings, so we have to make sure
                // our values are strings too, otherwise the comparison
                // will return false.
                _.each(options, function(val, key) {
                    options[key] = (_.isBoolean(val) || _.isNumber(val)) ? val.toString() : val;
                });
                
                // Find all jobs that match
                var matches = [];
                options = _.omit(options, PROPERTIES_IGNORED_BY_CACHE);
                for (var i = 0; i < list.length; ++i) {
                    var potentialJob = list[i];
                    var request = _.omit(
                        potentialJob.properties().request,
                        PROPERTIES_IGNORED_BY_CACHE);
                    var published = (new Date(potentialJob.published())).valueOf();
                    
                    if (_.isEqual(request, options) && (published > threshold)) {
                        matches.push(potentialJob);
                    }
                }
                
                // Sort to easily find the latest one
                matches.sort(function(a,b) { 
                    var aValue = (new Date(a.published())).valueOf();
                    var bValue = (new Date(b.published())).valueOf();
                    return bValue - aValue;
                });
                
                // If a job exists, then we will use it, otherwise,
                // create a new job.
                var job = matches[0];
                if (job) {
                    that.createManager(job);
                }
                else {
                    that.createSearch();
                }
            });

            return this;
        },
        
        _createOptionsForSearchJob: function() {
            var jobSettings = _.pick(
                this.settings.attributes,
                SearchModels.SearchSettingsModel.ALLOWED_ATTRIBUTES);
            
            // Let options = jobSettings, ignoring undefined values
            var options = {};
            _.each(jobSettings, function(v, k) {
                if (v !== undefined) {
                    options[k] = v;
                }
            });
            
            // Ensure the search is qualified
            options.search = this.settings.get('search', { qualified: true });
            
            // Can only create a hash if there's a search string
            if (options.search) {
                options['webframework.cache.hash'] = 'java5:' + 
                    SearchManager._createJava5HashForString(options.search).toString(16);
            }
            
            return options;
        },
        
        pause: function() {
            if (this.job) {
                this.job.pause();
            }
        },
        
        unpause: function() {
            if (this.job) {
                this.job.unpause();
            }
            return this;
        },
        
        finalize: function() {
            if (this.job) {
                this.job.finalize();
            }
            return this;
        },
        
        cancel: function() {
            // If this search is used with cache=true, then never cancel,
            // because somebody else might be using the same search.
            // Checks whether it was a refresh to load refresh messaging
            var msg = (this.refresh) ? 'refresh' : 'cancelled';

            if (this.get("cache")) {
                return;
            }
            
            if (this.job) {
                JobTracker.untrack(this.job.sid);
                this.job.cancel();
                this.job = null;
                this.trigger("search:" + msg);
            }
            return this;
        },
        
        createSearchPinger: function(job) {
            var that = this;

            // As long as the job on the manager is the same
            // as the one we were told to iterate over, keep going.
            Async.whilst(
                function() { return that.job === job; },
                function(done) {
                    // Just ping the job so that it doesn't die. We will wait
                    // a third of the time it takes for the job to expire or to 
                    // be auto-cancelled before pinging it
                    var props = job.properties();
                    var expireDelay, autoCancel;

                    if (props.isDone || props.request === undefined || props.request.auto_cancel === undefined) {
                        // Use TTL if the job is done or there is not auto_cancel
                        expireDelay = props.ttl;
                    } else if (props.request !== undefined) {
                        autoCancel = props.request && parseInt(props.request.auto_cancel, 10) || Infinity;
                        expireDelay = Math.min(autoCancel, props.ttl);
                    }

                    // Do the final calculation, and make sure we have a sensible
                    // default of 10 seconds.
                    var delay = parseInt(((expireDelay || 30) * 1000) / 3, 10);

                    Async.sleep(delay, function() {
                        // Make sure it's still our job
                        if (that.job === job) {
                            job.fetch(done);
                        }
                    });
                },
                function() {
                    // Do nothing, this will fail if the job is cancelled, 
                    // and that's fine.
                }
            );
        },
        
        // Create a search job manager for the given job.
        createManager: function(job) {
            var that = this;
            var check = _.bind(this._checkJob, this); 
            var cacheEnabled = this.get('cache');
            
            this.job = job;
            
            this.createSearchPinger(job);
            
            this.trigger("search:start", job);
            
            JobTracker.track(job, {
                progress: function(job) {
                    // Check if the event is from current active job
                    if (!check(job, cacheEnabled)) {
                        return;
                    }
                    var properties = job.properties();
                    var state = job.state();
                    
                    that.trigger("search:progress", state, job);
                    
                    // Backbone does a deep equality check on sets, and since
                    // properties is an object, and the object contents may have
                    // not changed (e.g. due to a cached job), Backbone might not
                    // fire a "change" event. This ensures that a change
                    // event is fired every time by doing a silent unset followed
                    // by a set.
                    that.unset("data", {silent: true});
                    that.set("data", properties);
                },
                done: function(job) {
                    // Check if the event is from current active job
                    if (!check(job, cacheEnabled)) {
                        return;
                    }
                    var state = job.state();
                    that.trigger("search:done", state, job);
                },
                failed: function(job) {
                    // Check if the event is from current active job
                    if (!check(job, cacheEnabled)) {
                        return;
                    }
                    var state = job.state();
                    that.trigger("search:fail", state, job);
                }
            });
        },

        _checkJob: function(job, cache){
            if (this.job !== job) {
                if (!cache) {
                    job.cancel();
                }
                return false;
            }
            return true;
        },

        // Create a new search job from current options.
        createSearch: function() {
            var options = this._createOptionsForSearchJob();
            var that = this;
            var cache = this.get('cache');
            // NOTE: If there was already an _inflightCreateXHR,
            //       it will be aborted immediately when it responds.
            var req = this._inflightCreateXHR = this.service.jobs().create(options, function(err, job) {
                // Check if response is from an obsolete job creation XHR
                if (req !== that._inflightCreateXHR) {
                    if (!(err || cache)) {
                        job.cancel();
                    }
                    return;
                } else {
                    that._inflightCreateXHR = null;
                }
                if (err) {
                    // We only want to trigger errors for non-abort cases. This is because
                    // we are the ones causing the abort (at the beginning of createSearch),
                    // and we should not cause an error for something we did. Furthermore,
                    // since we only abort in the case of creating a new search, we are guaranteed
                    // that the new search will cause some events to happen (even if those events
                    // are search:error ones).
                    if (err.status !== "abort") {
                        that.trigger("search:error", _("Could not create search.").t(), err, job);
                    }
                    return;
                }
                that.createManager(job);
            });
        },

        replayLastSearchEvent: function(listener) {
            var replayed = false;
            if (this._lastSearchEvent) {
                var evtName = this._lastSearchEvent[0];
                var listenerObj = _(this._events[evtName]).find(function(attachedListener){ return attachedListener.ctx === listener; });
                if (listenerObj) {
                    var args = this._lastSearchEvent[1];
                    listenerObj.callback.apply(listener, args);
                    replayed = true;
                }
            }
            return replayed;
        },
        
        getLastSearchEvent: function() {
            return this._lastSearchEvent;
        },

        dispose: function() {
            this.cancel();
            if (_.isFunction(this.settings.dispose)) {
                this.settings.dispose();
            }
            BaseManager.prototype.dispose.apply(this, arguments);
        }
    });

    /**
     * Hashes a string using the algorithm from the java.lang.String
     * hashCode function in Java SE 5.
     * 
     * Note that the returned hash may be negative.
     */
    SearchManager._createJava5HashForString = function(s) {
        var h = 0;
        for (var i = 0, n = s.length; i < n; i++) {
            // (x|0) converts to 32-bit int
            h = (((31*h)|0) + s.charCodeAt(i))|0;
        }
        return h;
    };
    
    return SearchManager;
});
