define([
    'underscore',
    'splunkjs/mvc',
    'splunkjs/mvc/basemanager',
    'splunkjs/mvc/basesplunkview',
    'splunkjs/mvc/settings',
    'splunkjs/mvc/tokenutils',
    'splunkjs/mvc/utils',
    'splunk.util'
], function (_, mvc, BaseManager, BaseSplunkView, Settings, TokenUtils, Utils, SplunkUtil) {

    var EventHandler = BaseManager.extend({
        initialize: function (options) {
            if (!options.event) {
                throw new Error('No event name specified for event handler instance');
            }
            if (options.actions && !options.conditions) {
                if (options.allowImplicitConditions) {
                    var implicit = options.allowImplicitConditions;
                    var conditions = [];
                    var defaultActions = [];
                    
                    _(options.actions).each(function(action) {
                        if(_(implicit.actions).contains(action.type)) {
                            var condAttr = _(implicit.attributes).find(function(attr){ return action[attr] != null; });
                            if (condAttr) {
                                conditions.push({ attr: implicit.attr, value: action[condAttr], actions: [action] });
                                return;
                            }
                        }
                        defaultActions.push(action);
                    });
                    
                    if (defaultActions.length > 0) {
                        conditions.push({ value: '*', actions: defaultActions });
                    }
                    options.conditions = conditions;
                } else {
                    options.conditions = [{ value: '*', actions: options.actions }];
                }
            }
            this.settings = new Settings(_.pick(options, 'viewid', 'event', 'conditions'));
            this.bindToViewSetting();
        },
        bindToViewSetting: function() {
            if (this.settings.has('viewid')) {
                this.bindToView(this.settings.get('viewid'));
            }
            this.listenTo(this.settings, 'change:viewid', this.bindToView);
        },
        bindToView: function(){
            var viewId = this.settings.get('viewid');
            this.listenTo(mvc.Components, 'change:' + viewId, this.onViewChange);
            var cur = mvc.Components.get(viewId);
            if(cur) {
                this.onViewChange(null, cur);
            }
        },
        onViewChange: function (components, view) {
            if (this._view) {
                this.stopListening(this._view);
            }
            if (view) {
                this._view = view;
                this.listenTo(view, this.settings.get('event'), this.handleEvent);
            }
        },
        handleEvent: function (e) {
            var conditions = this.settings.get('conditions');
            if (e && conditions && conditions.length) {
                var match = _(conditions).find(_.partial(_.bind(this.conditionMatches, this), _, e));
                if (match) {
                    if (_.isFunction(e.preventDefault)) {
                        e.preventDefault();
                    }
                    _(match.actions).each(_.partial(_.bind(this.executeAction, this), _, e.data));
                }
            }
        },
        conditionMatches: function (condition, e) {
            if (condition.value == '*') {
                return true;
            } else {
                return e[condition.attr] === condition.value;
            }
        },
        executeAction: function (action, data) {
            switch (action.type) {
                case "set":
                    return EventHandler.setToken(action.token, action.value, action, data);
                case "unset":
                    return EventHandler.unsetToken(action.token);
                case "link":
                    return EventHandler.linkTo(action.value, action.target, data);
                default:
                    throw new Error('Invalid drilldown action type ' + JSON.stringify(action.type));
            }
        }
    }, {
        setToken: function (name, format, options, data) {
            var defaultTokenModel = mvc.Components.get('default');
            var submittedTokenModel = mvc.Components.get('submitted');

            var value;
            if (options.delimiter != null && _.isArray(data)) {
                value = _(data).map(function(data){
                    var model = EventHandler.getTokenModel(data);
                    return TokenUtils.replaceTokenNames(format, model); 
                }).join(options.delimiter);
            } else {
                if (_.isArray(data)) {
                    data = _.first(data);
                }
                var model = EventHandler.getTokenModel(data);
                value = TokenUtils.replaceTokenNames(format, model);
            }
            
            if (options.prefix != null || options.suffix != null) {
                value = (options.prefix || '') + value + (options.suffix || '');
            }

            defaultTokenModel.set(name, value);
            if (submittedTokenModel) {
                submittedTokenModel.set(name, value);
            }
        },
        unsetToken: function (name) {
            var defaultTokenModel = mvc.Components.get('default');
            var submittedTokenModel = mvc.Components.get('submitted');
            defaultTokenModel.unset(name);
            if (submittedTokenModel) {
                submittedTokenModel.unset(name);
            }
        },
        linkTo: function (urlFormat, target, data) {
            if (urlFormat[0] === '/' && urlFormat[1] !== '/') {
                urlFormat = SplunkUtil.make_url(urlFormat);
            }
            var url = TokenUtils.replaceTokenNames(urlFormat, EventHandler.getTokenModel(data), TokenUtils.getEscaper('url'));
            if (target != null) {
                Utils.redirect(url, false, target);
            } else {
                Utils.redirect(url);
            }
        },
        getTokenModel: function (data) {
            var submittedTokenModel = mvc.Components.get('submitted');
            return _.extend(submittedTokenModel ? submittedTokenModel.toJSON() : {}, data);
        }
    });

    return EventHandler;
});