define([
    'underscore',
    'jquery',
    'splunkjs/mvc',
    'splunkjs/mvc/utils',
    'contrib/text!./factory_config.json',
    'util/console'
], function(_, $, mvc, utils, FACTORY_CONTENT_TEXT, console) {

    var SEARCH_MANAGER_CONSUMER_CLASSES = { viz: true, input: true };

    var DEFAULT_OPTIONS = {
        replaceDuplicateIDs: true
    };
    
    var DashboardFactory = function() {
        this.initialize.apply(this, arguments);
    };

    _.extend(DashboardFactory.prototype, {
        initialize: function(config, options) {
            options = _.defaults(options || {}, DEFAULT_OPTIONS);
            _.bindAll(this, "_calculateSettings", "_normalizeConfig", "_updateComponentSettings", "getComponentList",
                "getComponentTypeConfig", "getModuleForType", "getModulePaths", "groupComponents", "initialize",
                "instantiateComponent", "isComponentOfClass", "isDomComponent", "materialize", "materializeComponent");
            this.config = this._normalizeConfig(config);
            this.registry = options.registry || mvc.Components;
            this.idSeq = {};
            this.replaceDuplicateIDs = options.replaceDuplicateIDs;
            this._app = options.app;
        },
        /**
         * Instantiate the full set of components defined from the parsed dashboard output
         * @param {Object} normalizedParserOutput - the parsed and normalized object
         * @param {Node} targetDomNode - target DOM node to insert the the component into
         * @param options {Object}
         * {
         *      "waitForReady": wait for all child components being ready before resolving the promise (enabled by default)
         *      "loadPanels": trigger load for panelrefs after they are materialized (disabled by default)
         *      "idPrefix": a prefix for all identifiers (id and ref) for all materialized components 
         * }
         *
         * @returns {*} promise for instance of that component
         */
        materialize: function(normalizedParserOutput, targetDomNode, options) {
            options || (options = {});
            var parserOutput = $.extend(true, {}, normalizedParserOutput);
            var components = this.getComponentList(parserOutput, null, options);
            return this.materializeComponents(components, targetDomNode, options).promise();
        },
        materializeComponents: function(components, targetDomNode, options) {
            options || (options = {});
            var factory = this;
            var materializeResult = $.Deferred();
            require(this.getModulePaths(components), function() {
                var groupedComponents = factory.groupComponents(components);
                // Instantiate search managers first
                var managers = _(groupedComponents.managers || []).map(factory.instantiateComponent);
                // Materialize other components
                var materializedComponents = _(groupedComponents.others || [])
                    .map(_.partial(factory.materializeComponent, _, targetDomNode, options));

                if (options.waitForReady !== false) {
                    $.when.apply($, _.map(materializedComponents, DashboardFactory.whenComponentReady)).then(function(){
                        materializeResult.resolve(materializedComponents[0], managers, materializedComponents);
                    });
                } else {
                    materializeResult.resolve(materializedComponents[0], managers, materializedComponents);
                }
            });
            return materializeResult.promise();
        },
        materializeExisting: function(existingComponent, parserResult, options) {
            options || (options = {});
            parserResult = $.extend(true, {}, parserResult);
            parserResult.id = existingComponent.id;
            parserResult.omit = true;
            var components = this.getComponentList(parserResult, null, options);
            var root = components[0];

            var rootDomNode = _.result(existingComponent, 'getChildContainer');
            var materializeResult = $.Deferred();
            // Materialize all child components
            this.materializeComponents(components, rootDomNode, options).done(function(r, managers, components) {
                // Update existing components' settings once children are materialized
                if (_.isFunction(existingComponent.applySettings)) {
                    existingComponent.applySettings(root.settings);
                } else if (existingComponent.settings) {
                    existingComponent.settings.set(root.settings);
                }
                materializeResult.resolve(existingComponent, managers, components);
            }).fail(materializeResult.reject);
            return materializeResult.promise();
        },
        /**
         * Creates an instance of the given component
         * @param component the component definition
         * @returns {*} component instance
         */
        instantiateComponent: function(component) {
            var ComponentType = require(this.getModuleForType(component.type));
            var settings = _.defaults({ id: component.id }, component.settings);
            var settingsOptions = _.extend({}, component.settingsOptions);
            console.log('Instantiating component', component.type, 'with settings', settings, settingsOptions);
            return new ComponentType(settings, settingsOptions);
        },
        /**
         * Creates instance of the given component definition and renders it into the given DOM node
         * @param {Object} component - a definition of an individual component
         * @param {Object} rootDomNode
         * @param {Object} options
         * @returns {*} instance of that component
         */
        materializeComponent: function(component, rootDomNode, options) {
            options || (options = {});
            var parent, instance;
            if (!component.omit) {
                instance = this.instantiateComponent(component);

                if (this.isDomComponent(component)) {
                    if (component.parent) {
                        parent = this.registry.get(component.parent);
                        console.log('Attaching component to parent', component.parent, parent);
                        if (parent) {
                            DashboardFactory.addChildToContainer(instance, parent);
                        }
                    } else if (rootDomNode) {
                        instance.render().$el.appendTo(rootDomNode);
                    }
                    if (options.loadPanels === true && component.type  === "panelref") {
                        instance.load(component.settings.ref, component.settings.app);
                    }
                }
            }

            return instance;
        },
        isDomComponent: function(component) {
            return this.getComponentTypeConfig(component).dom === true;
        },
        isSearchManagerConsumer: function(component) {
            return !!SEARCH_MANAGER_CONSUMER_CLASSES[this.getComponentTypeConfig(component)['class']];
        },
        isComponentOfClass: function(component, cls) {
            return this.getComponentTypeConfig(component)['class'] == cls;
        },
        getComponentTypeConfig: function(component) {
            var typeSettings = this.config.typeMap[component.type];
            if (!typeSettings) {
                throw new Error('No factory settings for component type ' + JSON.stringify(component.type));
            }
            return typeSettings;
        },
        /**
         * Creates a flat list of components and replaces the children list with a list of IDs of the child components
         * @param parserOutput result of the dashboard parser/normalizer
         * @param parent - the parent component
         * @param options - options object
         * @returns {Array} of components
         */
        getComponentList: function(parserOutput, parent, options) {
            var result = [];
            if (parserOutput == null) {
                return result;
            }

            var component = _.omit(parserOutput, 'children', 'leaf');
            this._normalizeIdentifier(component, options);

            this._updateComponentSettings(component, parserOutput.children, parent, options);
            this._updateComponentSettingsOptions(component, parserOutput.children, parent);
            if (parserOutput.type) {
                result.push(component);
            }

            if (parserOutput.children && parserOutput.children.length) {
                if (parserOutput.leaf) {
                    component.children = [].concat(parserOutput.children);
                } else {
                    var subComponents = _(parserOutput.children).chain()
                        .each(function(child) {
                            child.parent = component.id;
                        })
                        .map(_.partial(this.getComponentList, _, component, options))
                        .flatten()
                        .value();
                    if (subComponents.length) {
                        result = result.concat.apply(result, subComponents);
                        component.children = _(parserOutput.children).pluck('id');
                    }
                }
            }

            return result;
        },
        /**
         * Group the given component list into categories of components (managers and others)
         * @param componentList
         * @returns {*} object containing the categories as keys and arrays of component as values
         */
        groupComponents: function(componentList) {
            var factory = this;
            return _(componentList).groupBy(function(component) {
                return factory.isComponentOfClass(component, "manager") ? "managers" : "others";
            });
        },
        _updateComponentSettings: function(component, children, parent, options) {
            var componentTypeConfig = this.getComponentTypeConfig(component);
            var configSettings = componentTypeConfig.settingsToCreate || {};
            var calculatedSettings = this._calculateSettings(component, children, parent, options);
            component.settings = _.extend(component.settings || {}, configSettings, calculatedSettings);

            if (componentTypeConfig.renameSettings) {
                _(componentTypeConfig.renameSettings).each(function(newName, oldName) {
                    if (component.settings.hasOwnProperty(oldName) && !component.settings.hasOwnProperty(newName)) {
                        component.settings[newName] = component.settings[oldName];
                        delete component.settings[oldName];
                    }
                });
            }

            component.settingsOptions = _.extend({}, componentTypeConfig.settingsOptions);

            return component;
        },
        _calculateSettings: function(component, children, parent, options) {
            var calculatedSettings = {};

            if (this.isSearchManagerConsumer(component)) {
                // if this is a viz grab the first manager child for the managerid
                var manager = _(children).chain().find(function(child) {
                    return this.isComponentOfClass(child, 'manager');
                }, this).value();
                if (manager) {
                    calculatedSettings['managerid'] = this._normalizeIdentifier(manager, options);
                }
            }

            if (this.isComponentOfClass(component, 'manager')) {
                calculatedSettings['app'] = this._app;
                if (options && options.idPrefix && component.settings.base) {
                    calculatedSettings.base = options.idPrefix + component.settings.base;
                }
            }

            if (this.isComponentOfClass(component, 'event') && parent) {
                calculatedSettings['viewid'] = parent.id;
            }
            
            if (this.isComponentOfClass(component, 'input') && component.settings.token) {
                if (component.type === 'time-input') {
                    calculatedSettings['earliest_time'] = '$form.' + component.settings.token + '.earliest$'; 
                    calculatedSettings['latest_time'] = '$form.' + component.settings.token + '.latest$'; 
                } else {
                    calculatedSettings['value'] = '$form.' + component.settings.token + '$';
                }
            }

            return calculatedSettings;
        },
        _updateComponentSettingsOptions: function(component, children, parent) {
            if (this.isComponentOfClass(component, "manager")) {
                if (this.isComponentOfClass(parent, "input")) {
                    // Search managers of inputs are bound to the default token namespace
                    component.settingsOptions.tokenNamespace = "default";
                } else {
                    // Search managers of elements around bound to the submitted token namespace
                    component.settingsOptions.tokenNamespace = "submitted";
                }
            }
            return component;
        },
        _normalizeIdentifier: function(component, options) {
            var existingIDs = this.registry.toJSON();
            var idPrefix = options && options.idPrefix || '';
            if (!component.id) {
                var config = this.getComponentTypeConfig(component);
                if (config.autoId) {
                    var id;
                    var prefix = config.autoId;
                    var seq = this.idSeq[prefix];
                    if (seq == null) {
                        seq = 1;
                    }
                    do {
                        id = idPrefix + prefix + (seq++);
                    } while (existingIDs[id]);
                    this.idSeq[prefix] = seq;
                    component.id = id;
                }
            } else if(this.replaceDuplicateIDs && existingIDs[component.id] && !component.omit) {
                console.warn('Replacing existing ID', component.id);
                delete component.id;
                return this._normalizeIdentifier(component, options);
            } else if (component.id.indexOf(idPrefix) !== 0 && !component.omit) {
                component.id = idPrefix + component.id;
            }
            return component.id;
        },
        /**
         * Find the require path for a component type
         * @param {String} type of component
         * @returns {String} path of required asset
         */
        getModuleForType: function(type) {
            var typeConfig = this.config.typeMap[type];
            if (!typeConfig) {
                throw new Error('Unknown component type ' + JSON.stringify(type));
            }
            return typeConfig.require;
        },
        /**
         * Find the unique require paths for a list of components
         * @param {Array} componentList - the list of components
         * @returns {Array} array of require module paths of the required assets
         */
        getModulePaths: function(componentList) {
            return _(componentList).chain()
                .pluck('type')
                .map(this.getModuleForType, this)
                .filter(_.identity) // filter null module paths
                .unique()
                .value();
        },
        _normalizeConfig: function(config) {
            var flatConfig = {typeMap: {}};

            _(config['typeMap']).each(function(value, key, obj) {
                var cls = {};
                if (value['class']) {
                    cls = config['classes'][value['class']];
                }
                flatConfig['typeMap'][key] = $.extend(true, {}, cls, value);
            });
            return flatConfig;
        }
    });

    var defaultInstance;
    // Class methods
    _.extend(DashboardFactory, {
        /**
         * Returns a shared instance of a dashboard factory using the default configuration
         * @returns {*} dashboardFactory
         */
        getDefault: function() {
            return defaultInstance || (defaultInstance = DashboardFactory.createDefault());
        },
        /**
         * Creates an instance of a dashboard factory using the default configuration
         * @returns {*} dashboardFactory
         */
        createDefault: function() {
            return new DashboardFactory(DashboardFactory.getDefaultConfig(), { app: utils.getCurrentApp() });
        },
        resetDefault: function() {
            defaultInstance = null;
        },
        /**
         * Returns the parsed json of the default configuration
         * @returns {Object} the default dashboard factory configuration
         */
        getDefaultConfig: _.once(function() {
            return JSON.parse(FACTORY_CONTENT_TEXT);
        }),

        getDashboardPanels: function(dashboardParserResult) {
            if (dashboardParserResult.type !== 'dashboard') {
                throw new Error('Expected root type to be a dashboard');
            }
            return _(dashboardParserResult.children).chain()
                .where({ type: 'row' })
                .pluck('children')
                .flatten()
                .filter( function(child) {
                    return child.type === "panel" || child.type === "panelref";
                })
                .value();
        },
        
        addChildToContainer: function(child, container) {
            if (typeof container.getChildContainer == 'function') {
                var el = container.getChildContainer();
                child.render().$el.appendTo(el);
            } else if (typeof container.addChild == 'function') {
                container.addChild(child);
            } else {
                console.warn("Can't add child to container - no addChild or getChildContainer method found", 
                    child, container);
            }
        },

        /**
         * Returns a promise for when the component is ready. This uses the components "componentReady" property, which
         * may be a promise itself or a method returning a promise.
         *
         * @param component
         * @returns {*} promise for when the component is ready
         */
        whenComponentReady: function(component) {
            return _.result(component, 'componentReady') || $.Deferred().resolve();
        }
    });

    return DashboardFactory;

});