define([
    'underscore',
    'jquery',
    'contrib/text!./parser_config.json',
    'util/console',
    'util/general_utils',
    'util/xml',
    'splunk.util',
    './htmlcleaner',
    '../tokenutils'
], function (_, $, DEFAULT_PARSER_CONFIG_TEXT, console, GeneralUtils, XML, SplunkUtil, HtmlCleaner, TokenUtils) {
    var sprintf = SplunkUtil.sprintf;

    /**
     * Dashboard XML parser
     */
    var DashboardParser = function () {
        this.initialize.apply(this, arguments);
    };

    /**
     * Component definition format

     {
         // Logical name of the component, also used as tagName if not explicitly specified 
         "name": "mycomponent",
 
         // Alternative tag name(s) for the component, can be a string or an array
         "alias": ["othername"],
 
         // Explicitly override the tag name to match for this component type
         "tagName": "foobar",
        
         TBD...
     }
     */

    _.extend(DashboardParser.prototype, {
        initialize: function (config) {
            this.config = config;
            _.bindAll(this, '_applyBaseDefinition', '_registerDefinition', '_registerComponentDefinition');
            var nodeTypes = this.nodeTypes = {};
            var definitions = this.definitions = {};
            _(config).chain()
                .flatten()
                .map(this._processDefinition)
                .each(function (nodeDef) {
                    definitions[nodeDef.name] = nodeDef;
                })
                .each(_.partial(this._applyBaseDefinition, _, definitions))
                .filter(function (nodeDef) {
                    return !nodeDef['abstract'];
                })
                .each(_.partial(this._registerDefinition, _, nodeTypes));
        },
        /**
         * Create an internal representation of the node definition from the parser config 
         * @private
         */
        _processDefinition: function (nodeDef) {
            if (nodeDef.settings) {
                _(nodeDef.settings).each(function (setting) {
                    setting = _.extend({}, setting);
                    switch (setting.type) {
                        case "attribute":
                            nodeDef.attributes || (nodeDef.attributes = {});
                            nodeDef.attributes[ setting.attributeName || setting.name ] = setting;
                            setting.type = setting.dataType || 'string';
                            break;
                        case "node":
                            nodeDef.valueNodes || (nodeDef.valueNodes = []);
                            nodeDef.valueNodes.push(setting);
                            setting.type = setting.dataType || 'string';
                            break;
                        case "option":
                            nodeDef.options || (nodeDef.options = {});
                            nodeDef.options[ setting.optionName || setting.name ] = setting;
                            if (setting.alias) {
                                nodeDef.options[setting.alias] = setting;
                            }
                            setting.type = setting.dataType || 'string';
                            break;
                    }
                });
            }
            return nodeDef;
        },
        /**
         * Apply configuration settings to a new definition its super-definition (if it extends one) 
         * @private
         */
        _applyBaseDefinition: function (nodeDef, definitions) {
            var extendsDef = nodeDef['extends'];
            if (extendsDef) {
                var superDef = definitions[extendsDef];
                if (!superDef) throw new Error('Config error: super type ' + extendsDef + ' not found!');
                if (superDef.tagName && !nodeDef.tagName) {
                    nodeDef.tagName = superDef.tagName;
                }
                if (superDef.options) {
                    nodeDef.options = _.defaults(nodeDef.options || {}, superDef.options);
                }
                if (superDef.attributes) {
                    nodeDef.attributes = _.defaults(nodeDef.attributes || {}, superDef.attributes);
                }
                if (superDef.valueNodes) {
                    nodeDef.valueNodes = _.union(superDef.valueNodes || [], nodeDef.valueNodes || []);
                }
                if (superDef.children) {
                    nodeDef.children = _.union(nodeDef.children || [], superDef.children);
                }
                if (superDef.normalization) {
                    nodeDef.normalization = _.extend({}, superDef.normalization, nodeDef.normalization);
                }
            }
            if (nodeDef.children) {
                nodeDef.children.sort();
            }
        },
        _registerDefinition: function (nodeDef, definitions) {
            var parser = this;
            var tagName = nodeDef.tagName || nodeDef.name;
            parser._registerComponentDefinition(tagName, nodeDef, definitions);
            if (nodeDef.alias) {
                _([nodeDef.alias]).chain().flatten().each(function (alias) {
                    parser._registerComponentDefinition(alias, nodeDef, definitions);
                });
            }
        },
        _registerComponentDefinition: function (tagName, def, definitions) {
            tagName = tagName.toLowerCase();
            if (definitions[tagName]) {
                var curDef = definitions[tagName];
                definitions[tagName] = _.isArray(curDef) ? curDef.concat(def) : [curDef, def];
            } else {
                definitions[tagName] = def;
            }
        },
        /**
         * Parse the given XML
         * @param xml {String|Node} - the XML as plain string or parsed document
         * @param options {Object} - {
         *      root {Array} - constrain valid root node (refers to definitions' names)
         *      validator {Validator} - pass in validator instance to collect parser warnings and errors 
         * }
         */
        parse: function (xml, options) {
            options || (options = {});
            var config = _.extend({}, this.config, options);
            var result;
            var doc = _.isString(xml) ? XML.parse(xml) : $(xml);
            var root = XML.root(doc);

            var rootNodeDefinition = this._findDefinitionForNode(root, null, options);

            if (rootNodeDefinition) {
                if (this._isRootNodeAllowed(rootNodeDefinition, config)) {
                    result = this._parseNode(root, rootNodeDefinition, options);
                } else {
                    this._error('Unable to parse dashboard XML: invalid root node', options.validator);
                }
            } else {
                this._error('Unable to parse dashboard XML: unknown root node', options.validator);
            }

            return options.normalize !== false ? this.normalize(result, options) : result;
        },
        parsePanel: function (xml, options) {
            options = _.extend({ root: ['panel'] }, options);
            return this.parse(xml, options);
        },
        parseDashboard: function (xml, options) {
            options = _.extend({ root: ['dashboard'] }, options);
            return this.parse(xml, options);
        },
        validate: function (xml, options) {
            var validationResult = { warnings: [], errors: [] };
            this.parse(xml, _.extend(options || {}, { validator: validationResult }));
            return validationResult;
        },
        /**
         * Normalize the result of a parser
         * @param parserResult
         * @param config
         * @param options
         * @returns {*}
         */
        normalize: function (parserResult, options) {
            options || (options = {});
            this._applyRowGrouping(parserResult, options);
            this._moveSettings(parserResult, options);
            this._inlineChildren(parserResult, options);
            this._convertResults(parserResult, options);
            this._normalizeIdentifiers(parserResult, options);
            this._omitFromResults(parserResult, options);
            this._decodeContent(parserResult, options);
            this._cleanupResults(parserResult, options);
            return parserResult;
        },
        /**
         * Walk the hierarchy of the parser result and call the given function for each of components in the result
         * @param parserResult the parser result
         * @param {Function} fn the function to call
         * @param {*} scope the scope to call the function with
         * @param {*} parent the parent 
         * @param {Boolean} depthFirst - true to talk the tree depth-first, otherwise breadth-first
         * @private
         */
        _visit: function (parserResult, fn, scope, parent, depthFirst) {
            if (!depthFirst) fn.call(scope, parserResult, parent);
            if (parserResult.children && parserResult.children.length) {
                _(parserResult.children.slice())
                    .each(_.partial(_.bind(this._visit, this), _, fn, scope, parserResult, depthFirst));
            }
            if (depthFirst) fn.call(scope, parserResult, parent);
        },
        /**
         * Detect duplicate IDs in the parser result 
         * @private
         */
        _normalizeIdentifiers: function (parserResult, options) {
            var parser = this;
            var seenIds = {};
            // capture all specified IDs - see if we have duplicates
            this._visit(parserResult, function (component) {
                var id = component.id = component.id || (component.settings || {}).id;
                if (id) {
                    if (seenIds[id]) {
                        parser._error('Duplicate ID: ' + JSON.stringify(id), options.validator);
                    }
                    seenIds[id] = true;
                }
            });
        },
        /**
         * Apply the "pullOutSettings" normalization rules on the parser result (moving settings from a component to
         * its parent)
         * @private
         */
        _moveSettings: function(parserResult, options){
            this._visit(parserResult, function(component, parent){
                var normalization = component.def.normalization;
                if (normalization && normalization.pullOutSettings) {
                    _(normalization.pullOutSettings).each(function(settingName){
                        var val = component.settings[settingName];
                        if(val !== undefined && parent.settings[settingName] === undefined) {
                            parent.settings[settingName] = val;
                            delete component.settings[settingName];
                        }
                    });
                }
            }, this);
        },
        /**
         * Apply the "convertTo" normalization rules on the parser result (ie. change the type of a component in the 
         * parser result and apply default settings of the target type)
         * @private
         */
        _convertResults: function (parserResult, options) {
            this._visit(parserResult, function (component, parent) {
                var normalization = component.def.normalization;
                if (normalization && normalization.convertTo) {
                    var convertTo = normalization.convertTo;
                    var convertToType = convertTo.type;
                    // Find existing target type instance if a fixed ID has been specified in the convertTo rule
                    var instance = _(parent.children).find(function (child) {
                        return child.type == convertToType &&
                            (convertTo.id === undefined || child.id == convertTo.id);
                    });

                    var idx = _.indexOf(parent.children, component);
                    if (instance) {
                        // remove item from children of parent
                        parent.children.splice(idx, 1);
                    } else {
                        var targetTypeDef = this.definitions[convertToType];
                        if (!targetTypeDef) throw new Error('Config error: Cannot convert ' + component.def.name +
                                                        ' - target type ' + convertToType + ' not found');
                        instance = this._createComponent(targetTypeDef);
                        instance.id = convertTo.id;
                        parent.children[idx] = instance;
                    }

                    // Apply settings to the converted node
                    _.extend(instance.settings, component.settings);
                }
            }, this);
        },
        /**
         * Apply legacy row grouping ("rowGrouping" normalization rule in the parser config) on the parser result. 
         * Note: row grouping is DEPRECATED 
         * @private
         */
        _applyRowGrouping: function(parserResult, options){
            options || (options = {});
            this._visit(parserResult, function(component, parent){
                var normalization = component.def.normalization;
                if (normalization && normalization.rowGrouping) {
                    var groupType = normalization.rowGrouping;
                    var groupTypeDef = this.definitions[groupType];
                    var grouping = component.settings.grouping || '';
                    
                    var curChildren = component.children;

                    var isPanel = function(child) { return child.def['extends'] == 'base-panel'; };
                    var isIgnored = function(child) { return child.def['extends'] == 'base-search'; };
                    var isElement = function(child) { return !isPanel(child) && !isIgnored(child); };
                    
                    if (_.some(curChildren, isPanel) && _.some(curChildren, isElement)) {
                        return this._error('Rows can only contain visualization elements or panels, not both.', options.validator);
                    }
                    
                    if (_.all(curChildren, function(child){ return isPanel(child) || isIgnored(child); })) {
                        if(grouping !== '') {
                            this._error('Row grouping can only applied to non-'+groupType+' children', options.validator);
                        }
                        return;
                    }
                    
                    var newChildren = _.filter(curChildren, isIgnored);
                    grouping = _(grouping.split(',')).map(function(g){ return parseInt($.trim(g), 10); });
                    
                    var childrenToGroup = _.filter(curChildren, isElement);
                    while(childrenToGroup.length) {
                        var count = grouping.shift() || 1;
                        var group = [];
                        while(count --) {
                            var child = childrenToGroup.shift();
                            if (child) {
                                group.push(child);
                            }
                        }
                        if (group.length) {
                            var groupComponent = this._createComponent(groupTypeDef);
                            groupComponent.children = group;
                            newChildren.push(groupComponent);
                        }
                    }
                    
                    component.children = newChildren;
                }       
            }, this);
        },
        _inlineChildren: function(parserResult, options){
            this._visit(parserResult, function(component, parent) {
                var normalization = component.def.normalization;
                if (normalization && normalization.inlineChildren) {
                    _(normalization.inlineChildren).each(function(inlineSettings){
                        var children = _(component.children).partition(function(child){ 
                            return child.type == inlineSettings.type ||
                                child.def['extends'] == inlineSettings.type; 
                        });
                        var inlineChildren = children[0];
                        if (inlineChildren.length) {
                            component.settings[inlineSettings.setting] = inlineSettings.single ? 
                                inlineChildren[0].settings : 
                                _(inlineChildren).pluck('settings');
                            component.children = children[1];
                        }
                    });
                }
            }, this, null, true);
        },
        _omitFromResults: function(parserResult){
            this._visit(parserResult, function(component, parent){
                var normalization = component.def.normalization;
                if (normalization && normalization.omitFromResults) {
                    parent.children.splice(_.indexOf(parent.children, component), 1);
                }
            }, this);
        },
        _decodeContent: function(parserResult, options) {
            this._visit(parserResult, function(component) {
                var normalization = component.def.normalization;
                if (normalization && normalization.decodeContent) {
                    if (component.settings.encoded) {
                        component.settings.content = _.unescape(component.settings.content);
                    }
                }                                    
            }, this);
        },
        _cleanupResults: function (results) {
            this._visit(results, function (component, parent) {
                delete component.def;
                component.settings || (component.settings = {});
                delete component.settings.id;
                if (component.id === undefined) {
                    delete component.id;
                }
            }, this);
        },
        getNodeDefinition: function (nodeName) {
            return this.nodeTypes[nodeName];
        },
        _isRootNodeAllowed: function (rootNodeDef, config) {
            return (!config.root) || _.contains(config.root, rootNodeDef.name);
        },
        _createComponent: function (def) {
            var cmp = {
                type: def.name,
                settings: {},
                children: [],
                def: def
            };

            var parser = this;
            _(def.settings).each(function (setting) {
                if (setting['default'] !== undefined) {
                    parser._applySetting(cmp, setting, setting['default']);
                }
            });

            return cmp;
        },
        _applySetting: function(result, settingDef, value){
            var target = result.settings;
            if(settingDef.parentObj) {
                target = target[settingDef.parentObj] || (target[settingDef.parentObj] = {});
            }
            target[settingDef.name] = value;
        },
        _parseNode: function (node, def, options) {
            var parser = this;
            var result = this._createComponent(def);

            if (def.attributes) {
                _(def.attributes).each(function (attributeDef, attributeName) {
                    var value = parser._parseValue(node.attr(attributeName), attributeDef, options);
                    if (value !== undefined) {
                        parser._applySetting(result, attributeDef, value);
                    }
                });
            }
            
            var contentSetting = _(def.settings || []).where({ type: 'content' })[0];
            if (contentSetting) {
                XML.replaceCdataNodes(node, true);
                var value = contentSetting.raw ? XML.serialize(node) : node.text();
                result.settings[contentSetting.name || 'content'] =
                    this._parseValue(value, contentSetting, options);
            } else {
                var excludeChildNodes = [];
                
                if (def.valueNodes) {
                    _(def.valueNodes).each(function (valNodeDef) {
                        var nodeName = valNodeDef.nodeName || valNodeDef.name;
                        excludeChildNodes.push(nodeName);
                        var valNode = node.children(nodeName);
                        if (valNode.length > 1) {
                            parser._warn(sprintf("Ignoring duplicate <%s> node in %s", nodeName, def.name));
                            valNode = $(valNode[0]);
                        }
                        if (valNode.length) {
                            var children = valNode.children();
                            if (!children.length && !valNodeDef.child) {
                                var value = parser._parseValue(valNode.text(), valNodeDef, options);
                                if (value !== undefined) {
                                    parser._applySetting(result, valNodeDef, value);
                                }
                            } else if (valNodeDef.child && children.length) {
                                var childNode = valNode.children(valNodeDef.child);
                                if(childNode.length) {
                                    var childVal = parser._parseValue(childNode.text(), valNodeDef, options);
                                    if (childVal !== undefined) {
                                        parser._applySetting(result, valNodeDef, childVal);
                                    }
                                }
                            } 
                        }
                    });
                }
                             
                if (def.options) {
                    excludeChildNodes.push('option');
                    _(node.children('option')).each(_.bind(function (node) {
                        var option = $(node);
                        var name = option.attr('name');
        
                        if (!name) {
                            this._warn('Invalid option (no name attribute)', options.validator);
                            return;
                        }
                        var optDef = def.options[name];
                        if (optDef) {
                            var value = this._parseValue(option.text(), optDef, options);
                            if (value !== undefined) {
                                parser._applySetting(result, optDef, value);
                            }
                        } else {
                            parser._warn('Unknown option name=' + JSON.stringify(name) 
                                + ' for node=' + JSON.stringify(def.name), options.validator);
                        }
                    }, this));
                    
                }
                
                var formatSetting = _(def.settings || []).where({ type: 'format' })[0];
                if (formatSetting) {
                    excludeChildNodes.push('format');
                    var format = this._parseFormatSetting(node, formatSetting, options);
                    if (format !== undefined) {
                        parser._applySetting(result, formatSetting, format);
                    }
                }
                
                var parseChildOptions = _.defaults({ excludeNodes: excludeChildNodes }, options);

                if (def.children !== undefined) {
                    result.children = this._parseChildren(def, def.children, node, parseChildOptions);
                    this._validateChildren(result.children, def.children, def, options);
                }
            }
            return result;
        },
        _parseChildren: function (parentDef, childList, node, options) {
            var allowedChildren = _(childList).pluck('type');
            var parser = this;
            var children = [];
            _(node.children()).each(_.bind(function (childNode) {
                var child = $(childNode);
                if (!_.contains(options.excludeNodes, this._nodeName(child))) {
                    var childDef = parser._findDefinitionForNode(child, allowedChildren, options);
                    if (childDef) {
                        if (_(allowedChildren).any(function(t){ return parser._isChildOfType(childDef, t); })) {
                            children.push(this._parseNode(child, childDef, options));
                        } else {
                            parser._warn(
                                    'Invalid child=' + JSON.stringify(childDef.name) +
                                    ' is not allowed in node=' + JSON.stringify(parentDef.name),
                                options.validator);
                        }
                    }
                }
            }, this));
            return children;
        },
        _validateChildren: function (childResults, allowedChildren, parentDef, options) {
            var warn = _.bind(this._warn, this);
            var parser = this;
            return _(allowedChildren).all(function (allowedChild) {
                var childCount = _.once(function () {
                    return _(childResults).filter(function(res){
                        return parser._isChildOfType(res.def, allowedChild.type); 
                    }).length;
                });
                
                if (allowedChild.min > 0 && childCount() < allowedChild.min) {
                    warn(sprintf('Expected at least %d children of %s in %s, instead saw %d',
                        allowedChild.min, allowedChild.type, parentDef.name, childCount()), options.validator);
                    return false;
                }
                if (allowedChild.max !== undefined && childCount() > allowedChild.max) {
                    warn(sprintf('Expected at most %d children of %s in %s, instead saw %d',
                        allowedChild.max, allowedChild.type, parentDef.name, childCount()), options.validator);
                    return false;
                }
                return true;
            });
        },
        _isChildOfType: function(def, type){
            return def.name === type || def['extends'] === type;
        },
        _findDefinitionForNode: function (node, allowedChildren, options) {
            var nodeName = this._nodeName(node);
            var nodeDef = this.getNodeDefinition(nodeName);
            if (!nodeDef) {
                this._warn('No node type definition found for node=' + nodeName, options && options.validator);
            }
            if (_.isArray(nodeDef)) {
                nodeDef = _(nodeDef).find(function (def) {
                    if (allowedChildren == null || _.contains(allowedChildren, def.name) ||
                        _.contains(allowedChildren, def['extends'])) {
                        if (def.match) {
                            return node.is(def.match) || node.find(def.match).length > 0;
                        } else {
                            return true;
                        }
                    }
                });
                if (!nodeDef) {
                    this._warn('No matching node type definition found for node=' + this._nodeToString(node), options && options.validator);
                }
            }
            return nodeDef;
        },
        _nodeToString: function(node){
            var n = XML.clone(node);
            n.empty();
            return XML.serialize(n);
        },
        _nodeName: function (node) {
            return node && node.prop('tagName').toLowerCase();
        },
        _parseOptionsHierarchy: function(node, options){
            var parser = this;
            var optionChildren = node.children('option');
            var listChildren = node.children('list');

            var hasOptions = optionChildren.length > 0;
            var hasList = listChildren.length > 0;
            
            if (hasOptions && hasList) {
                return this._error('Option cannot contain both option and list', options.validator);
            }
            
            var result;
            if (hasOptions) {
                result = {};
                _(optionChildren).each(function(opt){
                    var $opt = $(opt);
                    var name = $opt.attr('name');
                    if(name) {
                        result[name] = parser._parseOptionsHierarchy($opt, options);
                    } else {
                        parser._warn('Encountered option without name attribute', options.validator);
                    }
                });                    
            } else if(hasList) {
                result = [];
                _(listChildren).each(function(list){
                    var $list = $(list);
                    result.push(parser._parseOptionsHierarchy($list, options));
                });
            } else {
                result = parser._parseValue(node.text(), { dataType: 'auto' }, options);
            }
            
            return result;
        },
        _parseFormatSetting: function(node, setting, options) {
            var parser = this;
            var defaults = setting['default'] || {};
            var format = {};
            _(node.children('format')).each(function(formatNode){
                var $format = $(formatNode);
                var field = $format.attr('field') || defaults.field;
                var type = $format.attr('type') || defaults.type;
                
                if(!field || !type) {
                    parser._warn('Invalid format definition, both field and type need to be specified', 
                        options.validator);
                    return;
                }
                
                var fieldFormat = format[field] || (format[field] = []); 
                fieldFormat.push({ type: type, options: parser._parseOptionsHierarchy($format, options) });
            });
            return _.isEmpty(format) ? undefined : format;
        },
        _parseValue: function (value, definition, options) {
            var validator = options && options.validator;
            if (value === undefined && definition['default'] !== undefined) {
                return definition['default'];
            }
            if (definition.trim !== false && value) {
                value = $.trim(value);
            }

            var dataType = definition.dataType || definition.type || 'string';
            
            // Process non-token-aware data types
            switch (dataType) {
                case 'identifier':
                    if (!DashboardParser.isValidIdentifier(value)) {
                        return this._error('Invalid identifier: ' + JSON.stringify(value), validator);
                    }
                    return value;
                case 'html':
                    // always clean html
                    return this._cleanHtml(value);
                case 'token-name':
                    if (!DashboardParser.isValidTokenName(value)) {
                        return this._error('Invalid token name: ' + JSON.stringify(value), validator);
                    }
                    return value;
            }

            // Return raw string value if it contains a token
            if (definition.allowTokens !== false && TokenUtils.hasToken(value)) {
                return value;
            }
            
            switch (dataType) {
                case 'string':
                    return value;
                case 'boolean':
                    if (value !== undefined && !GeneralUtils.isBooleanEquivalent(value)) {
                        this._warn('Non-boolean value ' + JSON.stringify(value) + ' specified');
                    }
                    if (value === undefined && definition['default'] === undefined) {
                        return undefined;
                    }
                    return GeneralUtils.normalizeBoolean(value, { "default": definition['default'] });
                case 'integer':
                    var intVal = parseInt(value, 10);
                    if (_.isNaN(intVal)) {
                        this._warn('Invalid integer value: ' + JSON.stringify(value), validator);
                        return definition['default'];
                    }
                    return intVal;
                case 'float':
                    var floatVal = parseFloat(value);
                    if (_.isNaN(floatVal)) {
                        this._warn('Invalid float value: ' + JSON.stringify(value), validator);
                        return definition['default'];
                    }
                    return floatVal;
                case 'search-string':
                    return value;
                case 'app-resource':
                    return value;
                case 'app-resources':
                    return value ? _(value.split(',')).map($.trim) : [];
                case 'enum':
                    if (!_.contains(definition.values, value)) {
                        this._warn(sprintf('Value %s is not in the list of allowed values %s',
                            JSON.stringify(value), JSON.stringify(definition.values)), validator);
                        value = definition['default'];
                    }
                    return value;
                case 'fields-list':
                    try {
                        value = DashboardParser.parseFieldsList(value);
                    } catch (e) {
                        value = definition['default'];
                    }
                    return value;
                case 'csv-string':
                    try {
                        value = DashboardParser.parseCSVString(value);
                    } catch (e) {
                        value = definition['default'];
                    }
                    return value;
                case 'auto':
                    return this._parseValueAutoType(value);
                default:
                    this._error('Unknown value type=' + JSON.stringify(dataType));
            }
        },
        _parseValueAutoType: function (val) {
            if (val == 'true' || val == 'false') return this._parseValue(val, { type: 'boolean' });
            else if (/^\d+$/.test(val)) return this._parseValue(val, { type: 'integer'});
            else if (/^\d*\.\d+$/.test(val)) return this._parseValue(val, { type: 'float'});
            else return val;
        },
        _cleanHtml: function (htmlStr) {
            return HtmlCleaner.clean(htmlStr);
        },
        _error: function (msg, validator) {
            if (validator && _.isArray(validator.errors)) {
                validator.errors.push(msg);
            } else {
                console.error(msg);
                throw new Error(msg);
            }
        },
        _warn: function (msg, validator) {
            if (validator && _.isArray(validator.warnings)) {
                validator.warnings.push(msg);
            } else {
                console.warn(msg);
            }
        }
    });

    // This regex will take a space or comma separated list of fields, with quotes
    // for escaping strings with spaces in them, and match each individual
    // field.
    var FIELD_SPLITTER_REGEX = /(["'].*?["']|[^"',\s]+)(?=\s*|\s*,|\s*$)/g;

    // This regex will take a string that may or may not have leading quotes,
    // and strip them.
    var QUOTE_STRIPPER_REGEX = /^["']|["|']$/g;

    var defaultParserInstance;
    _.extend(DashboardParser, {
        getDefault: function () {
            return defaultParserInstance || (defaultParserInstance = DashboardParser.createDefault());
        },
        createDefault: function() {
            return new DashboardParser(DashboardParser.getDefaultConfig());
        },
        resetDefault: function(){
            defaultParserInstance = null;
        },
        getDefaultConfig: _.once(function () {
            return JSON.parse(DEFAULT_PARSER_CONFIG_TEXT);
        }),
        parseFieldsList: function (fields) {
            if (!_.isArray(fields)) {
                if (fields[0] === '[' && fields.slice(-1) === ']') {
                    fields = JSON.parse(fields);
                } else {
                    // Since this is a string, we're going to treat it as a
                    // space separated list of strings, with quoting. This is
                    // similar to what Splunk's 'fields' command takes.
                    fields = _.map(fields.match(FIELD_SPLITTER_REGEX), function (field) {
                        return field.replace(QUOTE_STRIPPER_REGEX, "");
                    });
                }
            }
            return fields;
        },
        parseCSVString: function(str) {
            if (str == null) {
                return str;
            }
            str = str.trim();
            if (str === '') {
                return [];
            }
            var result = [];
            var chars = str.split('');
            var idx = 0, len = chars.length;
            var start, end;
            var isLineSep = function(ch) {
                return ch === '\r' || ch === '\n';
            };
            var nextSep = function() {
                while (idx < len && !isLineSep(chars[idx]) && chars[idx] !== ',') {
                    idx++;
                }
                return idx;
            };
            while (idx < len) {
                while (idx < len && !isLineSep(chars[idx])) {
                    start = end = idx;
                    if (chars[idx] === '"') {
                        // quoted
                        start = end = idx = idx + 1;
                        while (idx < len) {
                            // find next non-escaped quote char
                            if (chars[idx] === '"') {
                                if (chars[idx + 1] !== '"') {
                                    break;
                                } else {
                                    idx++;
                                    chars[idx] = '';
                                }
                            }
                            idx++;
                            end = idx;
                        }
                        if (chars[idx] === '"') {
                            idx++;
                        }
                        nextSep();
                    } else {
                        end = nextSep();
                    }
                    result.push(chars.slice(start, end).join('').trim());
                    if (chars[idx] === ',') {
                        idx++;
                    }
                }
                if (isLineSep(chars[idx])) {
                    idx++;
                }
            }
            return result;
        },
        isValidIdentifier: function (id) {
            return /^[a-z]\w+$/gi.test(id);
        },
        isValidTokenName: function (name) {
            return /^[^\$]+$/.test(name);
        }
    });

    return DashboardParser;
});