define(function(require, exports, module) {
    var _ = require('underscore');
    var $ = require('jquery');
    var mvc = require("./mvc");
    var BaseSplunkView = require("./basesplunkview");
    var SingleValue = require("views/shared/SingleValue");
    var Backbone = require('backbone');
    var Utils = require('./utils');
    var Messages = require('./messages');
    var TokenAwareModel = require('./tokenawaremodel');
    var SharedModels = require('./sharedmodels');
    var Drilldown = require('./drilldown');
    require("css!../css/single-value");

    var SingleView = BaseSplunkView.extend({
        moduleId: module.id,

        className: "splunk-single",

        options: {
            data: "preview",
            field: "",
            classField: "",
            linkView: "search",
            drilldownRedirect: true
        },

        omitFromSettings: ['el', 'reportModel'],

        initialize: function() {
            this.configure();
            this.model = this.options.reportModel || TokenAwareModel._createReportModel();
            this.settings._sync = Utils.syncModels(this.settings, this.model, {
                auto: true,
                prefix: 'display.visualizations.singlevalue.',
                include: ["additionalClass", "linkView", "field", "linkFields", "classField",
                    "beforeLabel", "afterLabel", "underLabel", "linkSearch", "drilldown"]
            });
            this.settings.set('drilldown',
                Drilldown.getNormalizedDrilldownType(this.settings.get('drilldown'), { 'default': 'none' }));

            this.results = new Backbone.Model({
                columns: [],
                fields: []
            });

            this.reportModel = TokenAwareModel._createReportModel();
            this.reportModel._sync = Utils.syncModels(this.model, this.reportModel, {
                auto: 'push',
                tokens: true
            });

            this.singleValue = new SingleValue({
                model: {
                    searchResultsColumn: this.results,
                    state: this.reportModel,
                    application: SharedModels.get('app')
                }
            });

            this.listenTo(this.singleValue, 'click', this.emitDrilldownEvent);

            this.bindToComponentSetting('managerid', this._onManagerChange, this);
            this.settings.on("change", this.render, this);

            // If we don't have a manager by this point, then we're going to
            // kick the manager change machinery so that it does whatever is
            // necessary when no manager is present.
            if (!this.manager) {
                this._onManagerChange(mvc.Components, null);
            }
        },

        emitDrilldownEvent: function(e) {
            var manager = this.manager;
            var payload = Drilldown.createEventPayload({
                field: e.name2 || e.name,
                data: Drilldown.normalizeDrilldownEventData(e, {
                    contextProperty: 'rowContext',
                    manager: manager
                }),
                event: e
            }, function() {
                Drilldown.autoDrilldown(e, manager, { sync: e.modifierKey });
            });

            this.trigger('drilldown click', payload, this);
            if (this.settings.get("drilldownRedirect") && !payload.defaultPrevented()) {
                payload.drilldown();
            }
        },

        _onManagerChange: function(managers, manager) {
            if (this.manager) {
                this.manager.off(null, null, this);
                this.manager = null;
            }
            if (this.resultsModel) {
                this.resultsModel.off(null, null, this);
                this.resultsModel = null;
            }

            this._searchStatus = null;

            this._clearResults();

            if (!manager) {
                this._searchStatus = { state: "nomanager" };
                this.render();
                return;
            }

            // Clear any messages, since we have a new manager.
            this._searchStatus = { state: "start" };

            this.manager = manager;

            manager.on('search:start search:progress', this.onSearchProgress, this);
            manager.on('search:done', this.onSearchDone, this);
            manager.on('search:error', this.onSearchError, this);
            manager.on('search:fail', this.onSearchFailed, this);

            this.resultsModel = this.manager.data(this.settings.get("data"), {
                output_mode: "json_cols",
                count: 1,
                offset: 0,
                show_empty_fields: "True"
            });
            this.resultsModel.on("data", this._onDataUpdate, this);
            this.resultsModel.on("error", this.onSearchError, this);

            var replayed = manager.replayLastSearchEvent(this);
            if (!replayed) {
                this.render();
            }
        },

        onSearchProgress: function(properties) {
            var previewCount = ((properties || {}).content || {}).resultPreviewCount || 0;
            this._searchStatus = { state: "running", previewCount: previewCount };
            this.render();
        },

        onSearchDone: function(properties) {
            var previewCount = ((properties || {}).content || {}).resultPreviewCount || 0;
            this._searchStatus = { state: "done", previewCount: previewCount };
            this.render();
        },

        onSearchError: function(message, err) {
            var msg = Messages.getSearchErrorMessage(err) || message;
            this._searchStatus = { state: "error", message: msg };
            this.render();
        },

        onSearchFailed: function(state) {
            var msg = Messages.getSearchFailureMessage(state);
            this._searchStatus = { state: "error", message: msg };
            this.render();
        },

        _clearResults: function() {
            this.results.set({
                columns: null,
                fields: null
            }, { unset: true });
        },

        _onDataUpdate: function() {
            if (this.resultsModel.hasData()) {
                this.results.set({
                    columns: this.resultsModel.data().columns,
                    fields: this.resultsModel.data().fields
                });
            } else {
                this._clearResults();
                // If there are no results, then we need to update our preview
                // count. If the search is not done yet, this will simply get
                // updated on the next progress/done event.
                if (this._searchStatus) {
                    this._searchStatus.previewCount = 0;
                }
            }
            this.render();
        },

        render: function() {
            var searchStatus = this._searchStatus || null;
            var haveResults = searchStatus && searchStatus.previewCount;
            var resultsLoaded = haveResults && this.resultsModel && this.resultsModel.hasData();

            var message = null;
            if (searchStatus) {
                switch (searchStatus.state) {
                    case "nomanager":
                        message = "no-search";
                        break;
                    case "start":
                        message = "empty";
                        break;
                    case "running":
                        if (!haveResults) {
                            message = "waiting";
                        }
                        break;
                    case "cancelled":
                        message = "cancelled";
                        break;
                    case "done":
                        if (!haveResults) {
                            message = "no-results";
                        } else if (!resultsLoaded) {
                            message = "waiting";
                        }
                        break;
                    case "error":
                        message = {
                            level: "error",
                            icon: "warning-sign",
                            message: searchStatus.message
                        };
                        break;
                }
            }

            if (message) {
                if (!this.messageElement) {
                    this.messageElement = $('<div class="msg"></div>');
                }
                if (message === 'waiting') {
                    this.messageElement.addClass('waiting').html('<div class="single-value"><span class="single-result">&hellip;</span></div>');
                } else if (message === 'no-results') {
                    this.messageElement.addClass('no-results').html('<div class="single-value"><span class="single-result">N/A</span></div>');
                } else {
                    Messages.render(message, this.messageElement.removeClass('waiting'));
                }
                this.$el.append(this.messageElement);
            } else {
                if (this.messageElement) {
                    this.messageElement.remove();
                    this.messageElement = null;
                }
            }

            if (searchStatus && resultsLoaded && !message) {
                this.singleValue.activate().render().appendTo(this.el);
            } else if (this.singleValue) {
                this.singleValue.deactivate().$el.detach();
            }

            this.trigger('rendered', this);

            return this;
        },

        remove: function() {
            if (this.singleValue) {
                this.singleValue.remove();
            }
            if(this.reportModel) {
                if(this.reportModel._sync) {
                    this.reportModel._sync.destroy();
                }
                this.reportModel.off();
            }
            if (this.settings) {
                this.settings.off();
                if (this.settings._sync) {
                    this.settings._sync.destroy();
                }
            }
            BaseSplunkView.prototype.remove.call(this);
        }
    });

    return SingleView;
});
