define(function(require, exports, module) {
    var $ = require("jquery");
    var _ = require("underscore");
    var mvc = require("./mvc");
    var BaseSplunkView = require("./basesplunkview");
    var Messages = require("./messages");
    var Map = require("views/shared/Map");
    var Backbone = require('backbone');
    var utils = require('./utils');
    var route = require("uri/route");
    var console = require('util/console');
    var resizable = require('jquery.ui.resizable');
    var util = require('splunk.util');
    var TokenAwareModel = require('./tokenawaremodel');
    var Drilldown = require('./drilldown');
    var splunkConfig = require('splunk.config');

    var TileSources = {
        openStreetMap: {
            url: 'http://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png',
            maxZoom: 18
        }
        // JIRA: dynamically add Splunk tile source info (DVPL-3318)
    };

    var SplunkMapView = BaseSplunkView.extend({
        moduleId: module.id,

        className: "splunk-map",

        options: {
            data: "preview",
            height: '400px',
            tileSource: undefined,
            drilldownRedirect: true,
            resizable: false
        },

        omitFromSettings: ['el', 'reportModel'],

        normalizeOptions: function(settings, options) {
            settings.set('drilldown', Drilldown.getNormalizedDrilldownType(settings.get('drilldown'), { allowBoolean: true }));
        },

        initialize: function(options) {
            this.configure();
            this.model = this.options.reportModel || TokenAwareModel._createReportModel();

            this.settings._sync = utils.syncModels(this.settings, this.model, {
                auto: true,
                prefix: 'display.visualizations.',
                alias: {
                    tileUrl: 'display.visualizations.mapping.tileLayer.url',
                    maxZoom: 'display.visualizations.mapping.tileLayer.maxZoom',
                    drilldown: 'display.visualizations.mapping.drilldown'
                },
                exclude: ['managerid', 'data', 'tileSource', 'type', 'drilldownRedirect']
            });

            this.normalizeOptions(this.settings, options);

            // For these next three properties, there are two ways for them to appear in the settings model, only one
            // of which is aliased to the its counterpart in the report model.  These listeners make sure the two
            // properties stay in sync within the settings model (SPL-81910).
            this.listenTo(this.settings, 'change:maxZoom', function(model, value) {
                this.settings.set({ 'mapping.tileLayer.maxZoom': value }, { silent: true });
            });
            this.listenTo(this.settings, 'change:tileUrl', function(model, value) {
                this.settings.set({ 'mapping.tileLayer.url': value }, { silent: true });
            });
            this.listenTo(this.settings, 'change:drilldown', function(model, value) {
                this.settings.set({ 'mapping.drilldown': value }, { silent: true });
            });

            // In independent mode we default to using OSM tiles
            // because we don't know where to find the splunk tiles 
            // on the server. When in integrated mode, this is handled
            // by the web proxy server.
            // JIRA: hopefully in the future we can default to splunk
            // tiles if we make the map less reliant on splunkweb (DVPL-3670)
            if (splunkConfig.INDEPENDENT_MODE) {
                if (!this.settings.has('tileSource')) {
                    this.settings.set('tileSource', 'openStreetMap');                    
                }
            }

            this.results = new Backbone.Model({
                columns: [],
                fields: []
            });

            this._currentHeight = parseInt(this.settings.get('height'), 10);

            // initialize containers as detached DOM
            this.$map = $('<div></div>');
            this.$msg = $('<div></div>');

            this.createMap = _.debounce(_.bind(this.createMap, this), 0);
            this.createMap();
            this.settings.on('change:tileSource', this._updateTileSource, this);
            this._updateTileSource();
            this.bindToComponentSetting('managerid', this._onManagerChange, this);
        },

        _updateTileSource: function() {
            if (this.settings.has('tileSource')) {
                var tileSource = TileSources[this.settings.get('tileSource')];
                if (tileSource) {
                    this.settings.set({
                        tileUrl: tileSource.url,
                        maxZoom: tileSource.maxZoom
                    });
                } else {
                    console.warn('Invalid tileSource parameter value=%o', tileSource);
                }
            }
        },

        destroyMap: function() {
            if (this.map) {
                this.map.off();
                this.map.remove();
                this.map = null;
            }
        },

        createMap: function() {
            this.destroyMap();
            this.map = new Map({
                el: this.$map.appendTo(this.el),
                model: {
                    searchResultsColumn: this.results,
                    state: this.model
                },
                height: this.settings.get('height')
            });
            this._readMapCenterAndZoom();
            this.map.on('boundsChanged', this._onBoundsChanged, this);
            this.map.on('mapClicked', this.emitDrilldownEvent, this);

            // force data to update according to new map bounds
            this._onBoundsChanged();

            // Setup resizing
            this._onResizeMouseup = _.bind(this._onResizeMouseup, this);
            this.settings.on('change:resizable', function(model, value, options) {
                value ? this._enableResize() : this._disableResize();
            }, this);
            if (this.settings.get('resizable')) {
                this._enableResize();
            }
        },

        _enableResize: function() {
            if (this._canEnableResize()) {
                this.map.$el.resizable({autoHide: true, handles: "s", stop: this._onResizeStop.bind(this)});
                // workaround until jquery ui is updated
                this.map.$el.on('mouseup', this._onResizeMouseup);
            }
        },

        _disableResize: function() {
            if (this._canEnableResize()) {
                this.map.$el.resizable('destroy');
                this.map.$el.off('mouseup', this._onResizeMouseup);
            }
        },

        // NOTE: Bound to 'this' by createMap
        _onResizeMouseup: function(event) {
            $(this).width("100%");
        },

        _onResizeStop: function(event, ui) {
            $(event.target).width("100%");
            this._currentHeight = parseInt(this.settings.get('height'), 10);
            this.$msg.height(this._currentHeight).css('overflow', 'hidden');
        },

        _canEnableResize: function() {
            return true;
        },

        _onManagerChange: function(managers, manager) {
            if (this.manager) {
                this.manager.off(null, null, this);
                this.manager = null;
            }
            if (this.resultsModel) {
                this.resultsModel.off(null, null, this);
                this.resultsModel = null;
            }

            this._clearResults();

            if (!manager) {
                this.message("no-search");
                return;
            }

            // Clear any messages, since we have a new manager.
            this.$msg.detach();
            this.$map.css("visibility", "");

            this.manager = manager;
            this.manager.on("search:start", this._onSearchStart, this);
            this.manager.on("search:fail", this._onSearchFail, this);
            this.manager.on("search:error", this._onSearchError, this);

            this.resultsModel = this.manager.data(this.settings.get("data"), {
                serialfetch: true,
                output_mode: "json_cols",
                offset: 0,
                count: this.map ? this.map.getMaxClusters() : 100,
                show_empty_fields: "True",
                search: this.map ? this.map.getPostProcessSearch() : ''
            });
            this.resultsModel.on("data", this._onDataUpdate, this);
            this.resultsModel.on("error", this._onSearchError, this);

            this.manager.replayLastSearchEvent(this);
        },

        _onSearchStart: function() {
            this._clearResults();
        },

        _onSearchFail: function(state) {
            var msg = Messages.getSearchFailureMessage(state);
            this.message({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },

        _onSearchError: function(message, err) {
            var msg = Messages.getSearchErrorMessage(err) || message;
            this.message({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },

        _onDataUpdate: function() {
            this.$msg.detach();
            this.$map.css("visibility", "");
            if (this.resultsModel.hasData()) {
                this.results.set(this.resultsModel.data());
            } else {
                this._clearResults();
            }
            this.render();
        },

        _onBoundsChanged: function() {
            this._readMapCenterAndZoom();
            if (this.resultsModel) {
                this.resultsModel.set('search', this.map.getPostProcessSearch());
            }
        },

        _readMapCenterAndZoom: function() {
            this.model.set({
                currentMapZoom: this.map.getZoom(),
                currentMapCenter: this.map.getCenter().clone()
            });
        },

        emitDrilldownEvent: function(e) {
            if (this.settings.get('drilldown') === 'none') {
                return;
            }
            e = e || {};
            // Normalize data for drilldown event
            var event = _.extend({
                modifierKey: e.ctrlKey,
                ctrlKey: e.ctrlKey,
                fields: e.fields
            }, e);
            var data = event.data = {};
            _(e.fields).each(function(key) {
                var value = e.data[key];
                data['row.' + key] = value;
                if (key !== 'latitude' && key != 'longitude') {
                    if (!data.hasOwnProperty('click.name')) {
                        data['click.name'] = key;
                        data['click.value'] = value;
                        data['click.name2'] = key;
                        data['click.value2'] = value;
                    }
                }
            });
            _(['south', 'west', 'north', 'east']).each(function(orientation) {
                data['click.bounds.' + orientation] = e.data['_geo_bounds_' + orientation];
            });
            data['click.lat.name'] = e.data['_geo_lat_field'];
            data['click.lon.name'] = e.data['_geo_long_field'];
            data['click.lat.value'] = e.data['latitude'];
            data['click.lon.value'] = e.data['longitude'];
            event.field = data['click.name'] || null;

            _.extend(data, Drilldown.getNormalizedTimerange(this.manager));
            var payload = Drilldown.createEventPayload(event, _.bind(this._doDrilldown, this, e, this.manager));
            this.trigger('drilldown click click:marker', payload, this);
            var drilldownType = Drilldown.getNormalizedDrilldownType(this.settings.get('drilldown'), { allowBoolean: true });
            if (this.settings.get("drilldownRedirect") && drilldownType === 'all' && !payload.defaultPrevented()) {
                payload.drilldown();
            }
        },

        _doDrilldown: function(event, manager) {
            if (!(event && manager)) {
                return;
            }

            var DRILLDOWN_PROPERTIES = {
                _geo_lat_field: '_geo_lat_field',
                _geo_lon_field: '_geo_long_field',
                _geo_bounds_south: '_geo_bounds_south',
                _geo_bounds_west: '_geo_bounds_west',
                _geo_bounds_north: '_geo_bounds_north',
                _geo_bounds_east: '_geo_bounds_east'
            };

            if (event.data) {
                if (!_.all(DRILLDOWN_PROPERTIES, function(srcKey) { return !!event.data[srcKey]; })) {
                    // The search results did not contain bound information for the marker. This happens when the
                    // results have not been produced by the geostats command. In this case we use the first 2 fields
                    // (typically latitude and longitude) and send them to the intentions parser
                    var latLonFields = event.fields.slice(0,2);
                    var values = _(latLonFields).map(function(f){ return event.data[f]; });
                    Drilldown.autoDrilldown({
                        name: latLonFields,
                        value: values
                    }, manager, {
                        sync: event.modifierKey,
                        negate: event.altKey
                    });
                } else {
                    var params = _(DRILLDOWN_PROPERTIES).chain().map(function(srcKey, prop) {
                        return [prop, event.data[srcKey]];
                    }).object().value();
                    event.modifierKey = event.ctrlKey;
                    Drilldown.autoDrilldown(event, manager, {
                        drilldownType: 'geoviz',
                        negate: event.altKey,
                        sync: event.modifierKey,
                        params: params
                    });
                }
            }
        },

        _clearResults: function() {
            this.results.set({
                columns: [],
                fields: []
            });
        },

        message: function(info) {
            // Cannot use .hide() on map container because it screws up initial map rendering.
            // Map container dimensions must always evaluate to non-zero.
            // Using visibility hidden instead.
            this.$map.css("visibility", "hidden");
            this.$msg.detach();
            Messages.render(info, this.$msg);
            this.$msg.prependTo(this.$el);
        },

        render: function() {
            this.$el.css({ 'overflow': 'hidden', 'position': 'relative' });
            this.$msg.height(this._currentHeight).css({ 'overflow': 'hidden', 'position': 'absolute', 'width': '100%' });
            return this;
        },

        remove: function() {
            this.destroyMap();
            if (this.settings) {
                this.settings.off();
                if (this.settings._sync) {
                    this.settings._sync.destroy();
                }
            }
            BaseSplunkView.prototype.remove.call(this);
        }
    });

    return SplunkMapView;
});
