define(function (require, exports, module) {
    var $ = require("jquery");
    var _ = require("underscore");
    var mvc = require('./mvc');
    var utils = require('./utils');
    var Backbone = require('backbone');
    var BaseSplunkView = require("./basesplunkview");
    var ResultsTableView = require('views/shared/results_table/ResultsTableMaster');
    var Messages = require("./messages");
    var Drilldown = require('./drilldown');
    var PaginatorView = require("./paginatorview");
    var BaseCellRenderer = require('views/shared/results_table/renderers/BaseCellRenderer');
    var BaseRowExpansionRenderer = require('views/shared/results_table/renderers/BaseRowExpansionRenderer');
    var console = require('util/console');
    var TokenAwareModel = require('./tokenawaremodel');
    var GeneralUtils = require('util/general_utils');
    var DashboardParser = require('./simplexml/parser');
    
    require("css!../css/results-table");
    
    var DEFAULT_PAGE_SIZE = 10;
    var MAX_PAGE_SIZE = 100;

    var TableView = BaseSplunkView.extend({
        moduleId: module.id,

        className: 'splunk-table',
        options: {
            data: 'preview',
            drilldownRedirect: true,
            /**
             * Location where the pager is displayed relative to the table.
             * Can be 'bottom' (default) or 'top'.
             */
            pagerPosition: 'bottom',
            fields: [],
            pageSize: DEFAULT_PAGE_SIZE
        },
        normalizeOptions: function(options) {
            if(options.hasOwnProperty('count')) {
                var count = parseInt(this.options.count, 10) || DEFAULT_PAGE_SIZE;
                this.settings.set('pageSize', (count > MAX_PAGE_SIZE || count < 1) ? DEFAULT_PAGE_SIZE : count);
            }
            if(options.hasOwnProperty('wrap')) {
                if(GeneralUtils.isBooleanEquivalent(options.wrap)) {
                    this.settings.set('wrap', GeneralUtils.normalizeBoolean(options.wrap) ? '1' : '0');
                }
                else {
                    this.settings.set('wrap', options.wrap);
                }
            }
            if(options.hasOwnProperty('rowNumbers')) {
                if(GeneralUtils.isBooleanEquivalent(options.rowNumbers)) {
                    this.settings.set('displayRowNumbers', GeneralUtils.normalizeBoolean(options.rowNumbers) ? '1' : '0');
                }
                else {
                    this.settings.set('displayRowNumbers', options.rowNumbers);
                }
            } else if(options.hasOwnProperty('displayRowNumbers')) {
                this.settings.set('displayRowNumbers', GeneralUtils.normalizeBoolean(options.displayRowNumbers) ? '1' : '0');
            }
            if(options.hasOwnProperty('overlay')) {
                this.settings.set('dataOverlayMode', options['display.statistics.overlay']);
            }
            this.settings.set('drilldown', Drilldown.getNormalizedDrilldownType(options.drilldown, {
                validValues: ['row', 'cell', 'none'],
                aliasMap: { all: 'cell', off: 'none' },
                'default': 'cell'
            }));
            if(options.format) {
                var sparklineSettings = {};
                _.each(options.format, function(formats,field) {
                    _.each(formats, function(format){
                        if(format.type === 'sparkline') {
                            sparklineSettings[field] = format.options;
                        }
                    });
                });

                if(!_(sparklineSettings).isEmpty()) {
                    this.settings.set({
                        "sparkline.format": sparklineSettings
                    });
                }
            }
        },
        initialize: function (options) {
            this.configure();
            this.model = this.options.reportModel || TokenAwareModel._createReportModel();
            this.settings._sync = utils.syncModels(this.settings, this.model, {
                auto: true,
                alias: {
                    'displayRowNumbers': 'display.statistics.rowNumbers',
                    'dataOverlayMode': 'display.statistics.overlay',
                    'pageSize': 'display.prefs.statistics.count'
                },
                prefix: 'display.statistics.',
                include: ['displayRowNumbers','dataOverlayMode','pageSize','stripedRows','drilldown','sortKey',
                    'sortDirection','overlay','count','sparkline.format','wrap','fields'],
                exclude: ['drilldownRedirect']
            });
            this.normalizeOptions(this.settings.toJSON());

            var results = this.results = new Backbone.Model({
                rows: [],
                fields: []
            });
            
            var pagesize = this.settings.get('pageSize') || DEFAULT_PAGE_SIZE;

            var metadata = this.metadata = new Backbone.Model({
                pageSize: (pagesize > MAX_PAGE_SIZE || pagesize < 1) ? DEFAULT_PAGE_SIZE : pagesize,
                offset: 0
            });

            var table = this.table = new ResultsTableView({
                enableTableDock: false,
                el: $('<div></div>').appendTo(this.el),
                model: {
                    results: results,
                    metadata: metadata,
                    format: this.model
                }
            });

            this._rendered = false;
            this.listenTo(this.table, 'rendered', function(){
                this._rendered = true;
                this.trigger('rendered', this);
            }, this);

            this.settings._syncMeta = utils.syncModels(this.settings, this.metadata, {
                include: ['sortKey', 'sortDirection', 'pageSize'],
                auto: 'push'
            });

            this.listenTo(metadata, 'change', this.applyFetchOptions);
            
            // Setup pager
            this._onShowPagerChanged();
            this.settings.on('change:showPager', this._onShowPagerChanged, this);
            this.settings.on('change:pagerPosition', this._onPaginatorPositionChange, this);

            this.transformFieldsToArray();
            this.listenTo(this.settings, 'change:fields', this.onFieldsChanged);

            table.on('cellClick', this.emitDrilldownEvent, this);
            this.bindToComponentSetting('managerid', this.onManagerChange, this);
            
            // If we don't have a manager by this point, then we're going to
            // kick the manager change machinery so that it does whatever is
            // necessary when no manager is present.
            if (!this.manager) {
                this.onManagerChange(mvc.Components, null);
            }
        },

        emitDrilldownEvent: function(e) {
            var manager = this.manager;
            var drilldownType = this.settings.get('drilldown');
            var data = Drilldown.normalizeDrilldownEventData(e, {
                manager: manager,
                contextProperty: 'rowContext',
                jobTimeOnly: true
            });

            // Get all original row data, as e.rowContext only has data
            // for the visible fields in the table.
            var originalRows = this.resultsModel ? this.resultsModel.data().rows : null;
            if (originalRows && (e.rowIndex < originalRows.length)) {
                var originalRow = originalRows[e.rowIndex];
                _(this.resultsModel.data().fields).each(function(originalField, i) {
                    data["row." + originalField] = originalRow[i];
                });
            }

            var payload = Drilldown.createEventPayload({
                field: e.name2 || e.name,
                data: data,
                event: e
            }, function(){
                Drilldown.autoDrilldown(e, manager, {
                    drilldownType: drilldownType
                });
            });

            this.trigger('drilldown click', payload, this);

            var model = this.collection.at(e.rowIndex);
            if (model) {
                var key = e.hasOwnProperty("name2") ? e.name2 : e.name;
                var value = e.hasOwnProperty("value2") ? e.value2 : e.value;
                var originalEvent = e.originalEvent;

                this.trigger("clicked:row click:row", _.extend({
                    model: model,
                    index: e.rowIndex,
                    originalEvent: originalEvent,
                    component: this,
                    event: originalEvent
                }, payload), this);

                this.trigger("clicked:cell click:cell", _.extend({
                    model: model,
                    index: e.rowIndex,
                    column: e.cellIndex,
                    key: key,
                    value: value,
                    originalEvent: originalEvent,
                    event: originalEvent,
                    component: this
                }, payload), this);
            }
            
            // Now that the event is triggered, if there is a default action of
            // redirecting, we will execute it (depending on whether the user
            // executed preventDefault()).
            if (drilldownType !== 'none' && this.settings.get("drilldownRedirect") && !payload.defaultPrevented()) {
                payload.drilldown();
            }
        },

        onManagerChange: function (managers, manager) {
            if(this.manager) {
                this.stopListening(this.manager);
                this.manager = null;
            }
            if(this.resultsModel) {
                this.stopListening(this.resultsModel);
                this.resultsModel = null;
            }
            
            if (!manager) {
                this.renderMessage("no-search");
                return;
            }
            
            // Clear any messages, since we have a new manager.
            this.renderMessage("empty");
            
            this.manager = manager;
            this.resultsModel = this.manager.data(this.settings.get("data"), {
                autofetch: true,
                output_mode: "json_rows",
                show_empty_fields: "True",
                count: this.metadata.get('pageSize'),
                offset: this.metadata.get('offset')
            });
            this.listenTo(this.manager, "search:start", this.onSearchStart);
            this.listenTo(this.manager, "search:progress", this.onSearchProgress);
            this.listenTo(this.manager, "search:cancelled", this.onSearchCancelled);
            this.listenTo(this.manager, "search:refresh", this.onSearchRefresh);
            this.listenTo(this.manager, "search:fail", this.onSearchFailed);
            this.listenTo(this.manager, "search:error", this.onSearchError);
            this.listenTo(this.manager, "search:done", this.onSearchDone);
            this.listenTo(this.resultsModel, "data", this.renderResults);
            this.listenTo(this.resultsModel, "error", this.onSearchError);

            // Handle existing job
            var content = this.manager.get('data');
            if(content) {
                this.onSearchStart(content);
                this.onSearchProgress({ content: content });
                if(content.isDone) {
                    this.onSearchDone({ content: content });
                }
            }

            this.manager.replayLastSearchEvent(this);
        },
        _showMessages: function() {
            this.$('.msg').show();
            this.table.$el.hide();
            if (this.paginator) {
                this.paginator.$el.hide();
            }
        },
        _hideMessages: function() {
            this.$('.msg').hide();
            this.table.$el.show();
            if (this.paginator) {
                this.paginator.$el.show();
            }
        },
        renderMessage: function(msg) {
            var el = this.$('.msg');
            if(!el.length) {
                el = $('<div class="msg"></div>').appendTo(this.$el);
            }
            Messages.render(msg, el);
            this._showMessages();
        },
        onSearchStart: function() {
            this._searchRunning = true;
            this.clearResults();
            this.renderMessage('waiting');
            if (this.paginator) {
                this.paginator.settings.set('page', 0);
            }
        },
        onSearchDone: function(properties) {
            this._searchRunning = false;
            var previewCount = ((properties || {}).content || {}).resultPreviewCount || 0;
            this.updateCount(previewCount);
        },
        onSearchProgress: function(properties) {
            properties = properties || {};
            var job = properties.content || {};
            var previewCount = job.resultPreviewCount || 0;

            if(previewCount > 0) {
                this.updateCount(previewCount);
            }
        },
        onSearchCancelled: function() {
            this.renderMessage('cancelled');
        },
        onSearchRefresh: function() {
            this.renderMessage('refresh');
        },
        onSearchError: function(message, err) {
            var msg = Messages.getSearchErrorMessage(err) || message;
            this.renderMessage({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },
        onSearchFailed: function(state, job) {
            var msg = Messages.getSearchFailureMessage(state);
            this.renderMessage({
                level: "error",
                icon: "warning-sign",
                message: msg
            });
        },
        applyFetchOptions: function() {
            if(!this.resultsModel) {
                return;
            }
            var args = {
                count: this.metadata.get("pageSize"),
                offset: this.metadata.get("offset")
            };
            if(this.metadata.has('sortKey')) {
                args.search = [
                    '|','sort', '0', this.metadata.get('sortDirection') === 'desc' ? '-':'',
                    JSON.stringify(this.metadata.get('sortKey'))
                ].join(' ');
            }
            this.resultsModel.set(args);
        },
        updateCount: function(count) {
            if (count === 0) {
                this.clearResults();
                if(this._searchRunning) {
                    this.renderMessage('waiting');
                } else {
                    this.renderMessage('no-results');
                }
            }
            if(this.paginator) {
                this.paginator.settings.set('itemCount', count);
            }
        },
        onFieldsChanged: function(){
            this.transformFieldsToArray();
            this.renderResults();
        },
        transformFieldsToArray: function(){
            var fields = this.settings.get('fields');
            if (_.isString(fields)) {
                fields = DashboardParser.parseFieldsList($.trim(fields));
                this.settings.set('fields', fields, { silent: true });
            }
        },
        renderResults: function() {
            var rows, fields;
            if(!this.resultsModel || !this.resultsModel.hasData()){
                return;
            }
            this.collection = this.resultsModel.collection();
            if (this.settings.has('fields') && !_.isEmpty(this.settings.get('fields'))) {
                fields = this.settings.get('fields');
                var fieldsIdx = {}, resultsFields = this.resultsModel.data().fields;

                if (_(fields).contains('*')) {
                    fields = _(fields).chain().map(function(field) {
                        return field === '*' ? _(resultsFields).select(function(resField) {
                            return !_(fields).contains(resField);
                        }) : field;
                    }).flatten().unique().value();
                }

                _(resultsFields).each(function(f, i) { fieldsIdx[f] = i; });
                rows = _(this.resultsModel.data().rows).map(function(o){
                    return _(fields).map(function(f){ return fieldsIdx.hasOwnProperty(f) ? o[fieldsIdx[f]] : null; });
                });
            } else {
                rows = this.resultsModel.data().rows;
                fields = this.resultsModel.data().fields;
            }
            this.results.set({
                rows: rows,
                fields: fields
            });
            this._hideMessages();
        },
        clearResults: function() {
            if(this.paginator) {
                this.paginator.settings.set('itemCount', 0);
            }
            this.results.set({
                rows: [],
                fields: []
            });
        },
        render: function () {
            this.table.render();
            if(this.paginator) {
                this.paginator.render();
            }
            return this;
        },
        remove: function() {
            if(this.table) {
                this.table.remove();
            }
            if(this.paginator) {
                this.paginator.remove();
            }
            if(this.settings) {
                this.settings.off();
                if(this.settings._sync) {
                    this.settings._sync.destroy();
                }
                if(this.settings._syncMeta) {
                    this.settings._syncMeta.destroy();
                }
            }
            BaseSplunkView.prototype.remove.call(this);
        },

        addRowExpansionRenderer: function(rowExpansionRenderer) {
            // internal method validates rowExpansionRenderer 
            this.table.addRowExpansionRenderer(rowExpansionRenderer);
            if (this._rendered) {
                this.render();
            }
        },

        removeRowExpansionRenderer: function(rowExpansionRenderer) {
            // internal method validates rowExpansionRenderer
            this.table.removeRowExpansionRenderer(rowExpansionRenderer);
            if (this._rendered) {
                this.render();
            }
        },

        getRowExpansionRenderers: function() {
            return this.table.getRowExpansionRenderers();
        },

        expandRow: function(rowIndex) {
            // internal method validates rowIndex
            this.table.expandRow(rowIndex);
        },

        collapseRow: function() {
            this.table.collapseRow();
        },
        
        addCellRenderer: function(cellRenderer) {
            // internal method validates cellRenderer 
            this.table.addCellRenderer(cellRenderer);
            if (this._rendered) {
                this.render();
            }
        },

        removeCellRenderer: function(cellRenderer) {
            // internal method validates cellRenderer
            this.table.removeCellRenderer(cellRenderer);
            if (this._rendered) {
                this.render();
            }
        },

        getCellRenderers: function(rowIndex) {
            // internal method validates rowIndex
            return this.table.getCellRenderers(rowIndex);
        },
        
        _onShowPagerChanged: function() {
            if (this.settings.get('showPager') !== false) {
                // Create and show pager
                if (!this.paginator) {
                    this.paginator = new PaginatorView({
                        id: _.uniqueId(this.id+'-paginator'),
                        el: $('<div></div>'),
                        pageSize: this.metadata.get('pageSize')
                    });
                    
                    this.metadata.on('change:pageSize', this._onPageSizeChanged, this);
                    
                    this.paginator.settings.on('change:page', function() {
                        var count = this.metadata.get('pageSize');
                        var page = this.paginator.settings.get('page');
                        this.metadata.set('offset', count * page);
                    }, this);
                    
                    this._onPaginatorPositionChange();
                }
            } else {
                // Hide and destroy pager
                if (this.paginator) {
                    this.metadata.off('change:pageSize', this._onPageSizeChanged, this);
                    this.paginator.settings.off('change:page');
                    this.paginator.remove();
                    this.paginator = null;
                }
            }
        },
        
        _onPageSizeChanged: function() {
            this.paginator.settings.set('pageSize', this.metadata.get('pageSize'));
        },

        _onPaginatorPositionChange: function() {
            var action = (this.settings.get('pagerPosition') === 'bottom')
                ? 'appendTo'
                : 'prependTo';
            
            if (this.paginator) {
                this.paginator.$el.detach();
                this.paginator.$el[action](this.$el);
            }
        }
    });

    TableView.BaseCellRenderer = BaseCellRenderer;
    TableView.BaseRowExpansionRenderer = BaseRowExpansionRenderer;

    return TableView;
});
