/**
 * @author jszeto
 * @date 3/22/2013
 *
 * Dialog to edit the acceleration settings of a Data Model.
 *
 * Inputs:
 *
 *     model: {
 *         dataModel {models/services/datamodel/DataModel}
 *         application {models/Application}
 *
 * @fires AccelerationDialog#action:saveModel
 */
define(
    [
        'jquery',
        'underscore',
        'backbone',
        'views/shared/dialogs/DialogBase',
        'views/shared/controls/ControlGroup',
        'views/shared/controls/SyntheticCheckboxControl',
        'views/shared/controls/SyntheticSelectControl',
        'module'
    ],
    function(
        $,
        _,
        Backbone,
        DialogBase,
        ControlGroup,
        SyntheticCheckboxControl,
        SyntheticSelectControl,
        module
        )
    {
        return DialogBase.extend({
            moduleId: module.id,

            initialize: function(options) {
                DialogBase.prototype.initialize.call(this, options);

                this.bodyClassName = "form form-horizontal modal-body-scrolling";

                this.children.nameLabel = new ControlGroup({
                    controlType: 'Label',
                    controlOptions: {
                        defaultValue: this.model.dataModel.entry.content.get("displayName")
                    },
                    label: _("Data Model:").t()
                });

                var items = [
                    {label: _("1 Day").t(), value: "-1d"},
                    {label: _("7 Days").t(), value: "-1w"},
                    {label: _("1 Month").t(), value: "-1mon"},
                    {label: _("3 Months").t(), value: "-3mon"},
                    {label: _("1 Year").t(), value: "-1y"},
                    {label: _("All Time").t(), value: "0"}
                ];

                this.enabledCheckBox = new SyntheticCheckboxControl({modelAttribute: 'enabled',
                    model: this.model.dataModel.entry.content.acceleration, updateModel:false});
                this.children.enabledGroup = new ControlGroup({label: _("Accelerate:").t(),
                    controls: [this.enabledCheckBox],
                    help:_("Acceleration may increase storage and processing costs.").t()
                });
                this.enabledCheckBox.on("change", this.acceleratedChangeHandler, this);

                // TODO [JCS] Maybe Control.js should use the defaultValue if the model attribute is undefined.
                // But too risky of a change to make right now. Instead, will add a hack that we should revisit later
                var accelerationModel = this.model.dataModel.entry.content.acceleration;
                if (_.isUndefined(accelerationModel.get("earliest_time"))) {
                    accelerationModel.set("earliest_time", "-1d");
                }
                this.earliestTimeSelect = new SyntheticSelectControl({modelAttribute: 'earliest_time',
                                                                model: accelerationModel,
                                                                toggleClassName: 'btn',
                                                                menuWidth: 'narrow',
                                                                items: items,
                                                                updateModel:false,
                                                                popdownOptions: {
                                                                    attachDialogTo: '.modal:visible',
                                                                    scrollContainer: '.modal:visible .modal-body:visible'
                                                                }});
                this.children.earliestTimeGroup = new ControlGroup({label: _("Summary Range:").t(),
                                                                controls: [this.earliestTimeSelect],
                                tooltip: _("Sets the range of time (relative to now) for which data is accelerated. " +
                                    "Example: 1 Month accelerates the last 30 days of data in your pivots.").t()});

                this.settings.set("primaryButtonLabel", _("Save").t());
                this.settings.set("cancelButtonLabel", _("Cancel").t());
                this.settings.set("titleLabel", _("Edit Acceleration").t());
            },

            /**
             * When the enabled checkbox value is changed, toggle the visibility of the Summary Range control
             */
            acceleratedChangeHandler: function(model, value, options) {
                this.updateView();
            },

            updateView: function() {
                this.enabledCheckBox.getValue() ? this.children.earliestTimeGroup.$el.show() :
                                                  this.children.earliestTimeGroup.$el.hide();
            },

            primaryButtonClicked: function() {
                DialogBase.prototype.primaryButtonClicked.apply(this, arguments);

                this.enabledCheckBox.updateModel();
                this.earliestTimeSelect.updateModel();

                // Save the results into our object
                /**
                 * Save the Data Model
                 *
                 * @event AccelerationDialog#action:saveModel
                 * @param {string} data model name
                 */
                this.trigger("action:saveModel", this.model.dataModel.get("id"));
                this.hide();
            },

            renderBody : function($el) {
                var html = _(this.bodyTemplate).template({});
                $el.html(html);
                $el.find(".nameLabel-placeholder").replaceWith(this.children.nameLabel.render().el);
                $el.find(".accelerate-checkbox-placeholder").replaceWith(this.children.enabledGroup.render().el);
                $el.find(".summary-range-dropdown-placeholder").append(this.children.earliestTimeGroup.render().el);

                this.updateView();
            },

            bodyTemplate: '\
                <div class="nameLabel-placeholder"></div>\
                <div class="accelerate-checkbox-placeholder"></div>\
                <div class="summary-range-dropdown-placeholder"></div>\
        '

        });

    });