define([
            'underscore',
            'module',
            'views/Base',
            'uri/route',
            'util/datamodel/acceleration_utils',
            'models/classicurl'
        ],
        function(
            _,
            module,
            Base,
            route,
            accelerationUtils,
            classicUrl
        ) {

    var LOADING_MESSAGE = _('Loading').t();

    return Base.extend({

        moduleId: module.id,
        className: 'data-source-menu dropdown-menu',

        events: {
            'click .rebuild-button': function(e) {
                this.model.searchJob.trigger('rebuildAcceleration');
                e.preventDefault();
            },
            'click .inspect-acceleration-button': function(e) {
                this.model.searchJob.trigger('inspectAcceleration');
                e.preventDefault();
            }
        },

        /**
         * @constructor
         * @param options {Object} {
         *     model: {
         *         application: <models/shared/Application> the application state model
         *         report: <models/pivot/PivotReport> the pivot report
         *         dataModel: <models/services/datamodel/DataModel> the data model being reported on
         *         searchJob: <models/pivot/PivotJob> the current pivot job
         *         collectJob: <models/search/Job> only defined if the pivot job is in adhoc acceleration mode,
         *                                  the collect job that is populating the acceleration namespace
         *         tstatsSummarization <models/services/summarization/TStatsSummarization>
         *                                  only defined if the pivot job is in managed acceleration model,
         *                                  the summarization info for the job's namespace
         *     }
         * }
         */

        initialize: function() {
            Base.prototype.initialize.apply(this, arguments);
            if(this.model.collectJob) {
                this.model.collectJob.on('sync', this.updateAdhocAcceleration, this);
            }
            else if(this.model.tstatsSummarization) {
                this.model.tstatsSummarization.on('sync', this.updateManagedAcceleration, this);
            }
        },

        render: function() {
            var root = this.model.application.get('root'),
                app = this.model.application.get('app'),
                locale = this.model.application.get('locale'),
                modelName = this.model.dataModel.id,
                modelDisplayName = this.model.dataModel.entry.content.get('displayName'),
                objectName = this.model.report.getPivotObjectName(),
                objectDisplayName = this.model.dataModel.objectByName(objectName).get('displayName'),
                editObjectHref = route.data_model_editor(root, locale, app, { data: { model: modelName } }),
                selectNewDataModelHref = route.pivot(root, locale, app),
                selectNewObjectHref = route.pivot(root, locale, app, { data: { model: modelName } });

            this.$el.html(this.compiledTemplate({
                modelName: modelName,
                modelDisplayName: modelDisplayName,
                objectDisplayName: objectDisplayName,
                tempDataModel: this.model.dataModel.isTemporary(),
                editObjectHref: editObjectHref,
                canEdit: this.model.dataModel.entry.acl.canWrite(),
                selectNewDataModelHref: selectNewDataModelHref,
                selectNewObjectHref: selectNewObjectHref
            }));

            if(this.model.collectJob) {
                this.updateAdhocAcceleration();
            }
            else if(this.model.tstatsSummarization) {
                this.updateManagedAcceleration();
            }
            else {
                this.$('.acceleration-status-container').remove();
            }

            this.$("a.select-data-model-button").click(function(e) {
                e.preventDefault();
                classicUrl.clear().save({}, { trigger: true });
            });
            this.$("a.select-data-object-button").click(function(e) {
                e.preventDefault();

                classicUrl.clear().save({ model: modelName }, { trigger: true });
            });

            return this;
        },

        updateAdhocAcceleration: function() {
            var entry = this.model.collectJob.entry,
                content = entry.content,
                latestTime = this.model.collectJob.latestTimeSafe();

            this.$('.acceleration-status-container').html(_(this.adhocAccelerationTemplate).template({
                progressMessage: accelerationUtils.formatProgressMessage(content.get('doneProgress'), latestTime) || LOADING_MESSAGE,
                startedTimestamp: accelerationUtils.formatTimestamp(entry.get('published')) || LOADING_MESSAGE,
                earliestTimestamp: accelerationUtils.formatTimestamp(content.get('earliestTime')) || LOADING_MESSAGE
            }));
        },

        updateManagedAcceleration: function() {
            var content = this.model.tstatsSummarization.entry.content;
            this.$('.acceleration-status-container').html(_(this.managedAccelerationTemplate).template({
                statusMessage: accelerationUtils.formatProgressMessage(content.get('summary.complete')) || LOADING_MESSAGE,
                range: accelerationUtils.formatRange(content.get('summary.time_range')) || LOADING_MESSAGE,
                earliestTimestamp: accelerationUtils.formatTimestamp(content.get('summary.earliest_time')) || LOADING_MESSAGE
            }));
        },

        template: '\
            <div class="arrow"></div>\
            <% if(!tempDataModel) { %>\
                <ul>\
                    <li>\
                        <h4><%- modelDisplayName %></h4>\
                    </li>\
                    <li>\
                        <a class="select-data-model-button" href="<%- selectNewDataModelHref %>"><%- _("Select another Data Model").t() %></a>\
                    </li>\
                </ul>\
                <ul>\
                    <li>\
                        <h4><%- objectDisplayName %></h4>\
                    </li>\
                    <li>\
                        <a class="select-data-object-button" href="<%- selectNewObjectHref %>"><%- _("Select another Object").t() %></a>\
                    </li>\
                    <% if(canEdit) { %>\
                        <li>\
                            <a href="<%- editObjectHref %>"><%- _("Edit Object").t() %></a>\
                        </li>\
                    <% } %>\
                </ul>\
            <% } %>\
            <ul class="acceleration-status-container"></ul>\
        ',

        adhocAccelerationTemplate: '\
            <li>\
                <h4><%- _("Acceleration").t() %></h4>\
                <dl class="list-dotted">\
                    <dt><%- _("Progress").t() %></dt>\
                    <dd class="status-value"><%- progressMessage %></dd>\
                    <dt><%- _("Started").t() %></dt>\
                    <dd class="started-timestamp"><%- startedTimestamp %></dd>\
                    <dt><%- _("Earliest").t() %></dt>\
                    <dd class="latest-timestamp"><%- earliestTimestamp %></dd>\
                </dl>\
            </li>\
            <li><a href="#" class="rebuild-button"><%- _("Rebuild Acceleration").t() %></a></li>\
            <li><a href="#" class="inspect-acceleration-button"><%- _("Inspect Acceleration Job").t() %></a></li>\
        ',

        managedAccelerationTemplate: '\
            <li>\
                <h4><%- _("Acceleration").t() %></h4>\
                <dl class="list-dotted">\
                    <dt><%- _("Status").t() %></dt>\
                    <dd class="progress-value"><%- statusMessage %></dd>\
                    <dt><%- _("Range").t() %></dt>\
                    <dd class="range-value"><%- range %></dd>\
                    <dt><%- _("Earliest").t() %></dt>\
                    <dd class="latest-timestamp"><%- earliestTimestamp %></dd>\
                </dl>\
            </li>\
        '

    });

});
