/**
 * @author sfishel
 *
 * The master view for the pivot interface.
 */

define([
            'jquery',
            'underscore',
            'backbone',
            'module',
            'models/Base',
            'models/pivot/PivotReport',
            'models/charting/Charting',
            'collections/shared/FlashMessages',
            'views/Base',
            'views/pivot/VisualizationTypeMenu',
            'views/pivot/DocumentActionBar',
            'views/pivot/VisualizationConfigMenu',
            'views/pivot/JobActionBar',
            'views/shared/FlashMessages',
            'helpers/FlashMessagesHelper',
            'helpers/pivot/PivotVisualizationManager',
            'helpers/Printer',
            'helpers/user_agent'
        ],
        function(
            $,
            _,
            Backbone,
            module,
            BaseModel,
            PivotReport,
            Charting,
            FlashMessagesCollection,
            Base,
            VisualizationTypeMenu,
            DocumentActionBar,
            VisualizationConfigMenu,
            JobActionBar,
            FlashMessagesView,
            FlashMessagesHelper,
            pivotVizManager,
            Printer,
            userAgent
        ) {

    return Base.extend({

        moduleId: module.id,

        /**
         * @constructor
         * @param options {Object} {
         *     model: {
         *         application: <models/shared/Application> the application state model
         *         report: <models/pivot/PivotReport> the pivot report
         *         dataModel: <models/services/datamodel/DataModel> the data model being reported on
         *         tableResults: <models/services/search/job/ResultJsonRows> the table results
         *         chartResults: <models/services/search/job/ResultsColumn> the chart results
         *         searchJob: <models/pivot/PivotJob> the current pivot job
         *         appLocal <models.services.AppLocal> the local splunk app
         *         user <models.services/admin.User> the current user
         *         pivotSearch <models.pivot.PivotSearch> the current pivot query
         *     }
         *     collection: {
         *         dataModels <collections/services/datamodel/DataModels> the list of available data models
         *         timePresets <collections/services/data/ui/Times> the user-specific preset time preferences
         *     }
         *     flashMessages: <views.shared.FlashMessagesV2> a messages view with upstream models/collections already registered
         * }
         */

        initialize: function(options) {
            Base.prototype.initialize.call(this, options);

            this.vizConfigMenuOffset = 0;
            var objectName = this.model.report.getPivotObjectName(),
                object = this.model.dataModel.objectByName(objectName),
                objectHasIndexTime = !!object.getIndexTimeField(),
                vizMenuItems = _(pivotVizManager.schema).map(function(vizConfigObject) {
                    var item = _(vizConfigObject).pick('type', 'icon', 'description');
                    if(!objectHasIndexTime && vizConfigObject.requiresIndexTime) {
                        item.disabled = true;
                    }
                    return item;
                });
            this.children.vizTypeMenu = new VisualizationTypeMenu({
                model: this.model.report,
                items: vizMenuItems
            });
            this.children.docStatusBar = new DocumentActionBar({
                model: {
                    report: this.model.report,
                    dataModel: this.model.dataModel,
                    application: this.model.application,
                    searchJob: this.model.searchJob
                }
            });
            this.children.jobStatusBar = new JobActionBar({
                model: {
                    searchJob: this.model.searchJob,
                    report: this.model.report,
                    application: this.model.application,
                    appLocal: this.model.appLocal,
                    dataModel: this.model.dataModel,
                    pivotSearch: this.model.pivotSearch
                }
            });

            var debouncedUpdateInitializingMessgae = _.bind(_.debounce(this.updateInitializingMessage, 100), this);
            this.model.searchJob.on('change:id', function() {
                if(!this.model.searchJob.isNew()) {
                    this.hideInitializingMessage();
                }
                debouncedUpdateInitializingMessgae();
            }, this);

            this.windowResizeHandler = _(_.debounce(this.reflow, 50)).bind(this);
            $(window).on('resize', this.windowResizeHandler);

            Printer.on(Printer.PRINT_START, this.reflow, this);
            Printer.on(Printer.PRINT_END, this.reflow, this);
        },

        remove: function() {
            $(window).off('resize', this.windowResizeHandler);
            Printer.off(null, null, this);
            return Base.prototype.remove.call(this);
        },

        render: function() {
            this.$el.html(this.compiledTemplate({}));
            this.children.vizTypeMenu.render().replaceAll(this.$('.viz-type-menu-placeholder'));
            this.children.docStatusBar.render().replaceAll(this.$('.doc-status-bar-placeholder'));
            this.children.jobStatusBar.render().replaceAll(this.$('.job-status-bar-placeholder'));
            var vizConfig = this.getCurrentVizConfig();
            if(vizConfig.showConfigMenu) {
                this.initVizConfigMenu();
                this.children.vizConfigMenu.render().appendTo(this.$('.viz-builder-wrapper'));
            }
            this.initVizContent();
            this.children.vizContent.render().appendTo(this.$('.viz-builder-wrapper'));
            this.updateInitializingMessage();
            return this;
        },

        onAddedToDocument: function() {
            Base.prototype.onAddedToDocument.apply(this, arguments);
            this.reflow();
        },

        renderContainer: function(options) {
            options = options || {};
            var vizConfig = this.getCurrentVizConfig();
            if(vizConfig.showConfigMenu) {
                this.initVizConfigMenu();
                this.children.vizConfigMenu.render().appendTo(this.$('.viz-builder-wrapper'));
                if(options.disabled) {
                    this.children.vizConfigMenu.disable();
                }
            }
            else {
                this.removeVizConfigMenu();
            }
            this.initVizContent();
            this.children.vizContent.appendTo(this.$('.viz-builder-wrapper'));
            this.children.vizContent.renderContainer();
            if(options.disabled) {
                this.children.vizContent.disable();
            }
        },

        renderData: function() {
            if(this.children.vizContent) {
                this.children.vizContent.renderData();
                this.children.vizContent.updateData();
            }
            this.children.jobStatusBar.render();
        },

        clearData: function() {
            // SPL-74219, kind of a hacky fix but the safest one...
            // before we empty out the job status bar view, detach its children to avoid un-binding their listeners
            _(this.children.jobStatusBar.children).invoke('detach');
            this.children.jobStatusBar.$el.empty();
            if(this.children.vizContent) {
                this.children.vizContent.clearData();
            }
        },

        renderErrors: function(flashMessages) {
            this.clearData();
            if(flashMessages) {
                flashMessages.appendTo(this.children.jobStatusBar.el);
            }
        },

        updateInitializingMessage: function() {
            if(this.model.searchJob.isNew()) {
                this.showInitializingMessage();
            }
            else {
                this.hideInitializingMessage();
            }
        },

        showInitializingMessage: function() {
            this.$initializingMessage = $('<div class="initializing-message">' + _('Starting job...').t() + '</div>').prependTo(this.children.jobStatusBar.el);
            if(this.children.vizContent) {
                this.children.vizContent.renderMessage('info', _('Starting job...').t());
            }
        },

        hideInitializingMessage: function() {
            if(this.$initializingMessage) {
                this.$initializingMessage.remove();
            }
        },

        initVizConfigMenu: function() {
            this.removeVizConfigMenu();
            this.children.vizConfigMenu = new VisualizationConfigMenu({
                panels: this.getCurrentVizConfig().configMenuPanels,
                model: {
                    report: this.model.report,
                    dataModel: this.model.dataModel,
                    application: this.model.application,
                    appLocal: this.model.appLocal,
                    user: this.model.user
                },
                collection: {
                    timePresets: this.collection.timePresets
                },
                scrollOffset: this.vizConfigMenuOffset
            });
        },

        removeVizConfigMenu: function() {
            if(this.children.vizConfigMenu) {
                this.vizConfigMenuOffset = this.children.vizConfigMenu.getScrollOffset();
                this.children.vizConfigMenu.remove();
                delete this.children.vizConfigMenu;
            }
        },

        initVizContent: function() {
            if(this.children.vizContent) {
                this.children.vizContent.remove();
                this.children.vizContent.off(null, null, this);
                delete this.children.vizContent;
            }
            var vizConfig = this.getCurrentVizConfig(),
                // if the report is incomplete, pass in empty chart results
                chartResults = this.model.report.isComplete() ? this.model.chartResults
                                : new BaseModel({ fields: [], columns: [] });

            this.children.vizContent = new vizConfig.masterView({
                model: {
                    report: this.model.report,
                    results: (vizConfig.dataOutputMode === pivotVizManager.ROW_MAJOR) ? this.model.tableResults : chartResults,
                    job: this.model.searchJob,
                    dataModel: this.model.dataModel,
                    application: this.model.application,
                    appLocal: this.model.appLocal,
                    user: this.model.user
                },
                collection: {
                    timePresets: this.collection.timePresets
                },
                flashMessages: this.options.flashMessages
            });
            this.children.vizContent.$el.addClass(vizConfig.showConfigMenu ? 'viz-content-with-menu' : 'viz-content-full-width');


            // SPL-67726 - in IE 7, we need to refresh the layout of the config menu any time the data container is updated
            if(userAgent.isIE7()) {
                this.children.vizContent.on('updateData', function() {
                    if(this.children.vizConfigMenu) {
                        this.children.vizConfigMenu.refreshLayout();
                    }
                }, this);
            }
        },

        getCurrentVizConfig: function() {
            return pivotVizManager.getConfigByVizType(this.model.report.getVisualizationType());
        },

        template: '\
            <div class="doc-status-bar-placeholder"></div>\
            <div class="viz-type-menu-placeholder"></div>\
            <div class="job-status-bar-placeholder"></div>\
            <div class="viz-builder-wrapper"></div>\
        '

    });

});
