define([
        'underscore',
        'backbone',
        'module',
        'views/Base',
        'views/shared/FlashMessages',
        'views/shared/JobDispatchState',
        'views/shared/jobstatus/Master',
        'views/report/results/EventsPane',
        'views/report/results/ResultsPane',
        'util/splunkd_utils'
    ],
    function(
        _,
        Backbone,
        module,
        Base,
        FlashMessage,
        JobDispatchState,
        JobStatus,
        EventsPane,
        ResultsPane,
        splunkd_utils
    ){
        return Base.extend({
            moduleId: module.id,
             /**
             * @param {Object} options {
             *      model: {
             *          searchJob: <models.services.search.Job>,
             *          report: <models.search.Report>,
             *          eventsViewerReport: <models.search.Report>
             *          result: <models.services.search.job.ResultsV2>,
             *          summary: summary: <model.services.search.job.SummaryV2>,             
             *          resultJsonCols: <models.services.search.jobs.ResultJsonCols>,
             *          resultJsonRows: <models.services.search.jobs.ResultJsonRows>,
             *          application: <models.Application>,
             *          appLocal: <models.services.AppLocal>,
             *          user: <models.services.admin.User>
             *      },
             *      collection: {
             *          times: <collections.services.data.ui.TimesV2>,
             *          selectedFields: <collections.SelectedFields>,
             *          workflowActions: <collections.services.data.ui.WorkflowActions>
             *       }
             * }
             */
            initialize: function() {
                Base.prototype.initialize.apply(this, arguments);

                this.errorTypes = [splunkd_utils.FATAL, splunkd_utils.ERROR];
                
                //views
                this.children.flashMessage = new FlashMessage({
                    className: 'results-message',
                    model: {
                        searchJob: this.model.searchJob,
                        searchJobControl: this.model.searchJob.control,
                        report: this.model.report,
                        eventsViewerReport: this.model.eventsViewerReport,
                        result: this.model.result,
                        summary: this.model.summary,
                        resultJsonCols: this.model.resultJsonCols,
                        resultJsonRows: this.model.resultJsonRows
                    },
                    whitelist: this.errorTypes
                });
                
                this.children.jobStatus = new JobStatus({
                    model: {
                        searchJob: this.model.searchJob,
                        state: this.model.report.entry.content,
                        application: this.model.application,
                        appLocal: this.model.appLocal,
                        report: this.model.report,
                        serverInfo: this.model.serverInfo
                    },
                    enableReload: true,
                    enableSearchMode: false,
                    allowDelete: false
                });

                this.children.jobDispatchState = new JobDispatchState({
                    mode: 'auto',
                    model: this.model.searchJob
                });

                this.children.eventsPane = new EventsPane({
                    model: {
                        result: this.model.result,
                        summary: this.model.summary,
                        report: this.model.report,
                        eventsViewerReport: this.model.eventsViewerReport,
                        searchJob: this.model.searchJob,
                        application: this.model.application
                    },
                    collection: {
                        selectedFields: this.collection.selectedFields,
                        eventRenderers: this.collection.eventRenderers,
                        workflowActions: this.collection.workflowActions
                   }
                });

                this.children.resultsPane = new ResultsPane({
                    model: {
                        report: this.model.report,
                        resultJsonCols: this.model.resultJsonCols,
                        resultJsonRows: this.model.resultJsonRows,
                        searchJob: this.model.searchJob,
                        appLocal: this.model.appLocal,
                        application: this.model.application
                    }
                });
            },
            startListening: function() {
                this.listenTo(this.model.searchJob, 'prepared', function(){
                    if (this.active) {
                        this.manageStateOfChildren();
                    }
                });
            },
            activate: function(options) {
                var clonedOptions = _.extend({}, (options || {}));
                delete clonedOptions.deep;
                
                this.ensureDeactivated({deep: true});
                
                Base.prototype.activate.call(this, clonedOptions); 
                
                this.manageStateOfChildren();
                
                return this;
            }, 
            deactivate: function(options) {
                if (!this.active) {
                    return Base.prototype.deactivate.apply(this, arguments);
                }
                Base.prototype.deactivate.apply(this, arguments);
                
                return this;               
            },
            manageStateOfChildren: function() {
                var jobPreparing = this.model.searchJob.isPreparing(),
                    isReportSearch;
                
                this.children.flashMessage.activate({deep: true});
                
                if (this.model.report.isRealTime() && !this.model.user.canRTSearch()) {
                    if (this.$resultsWrapper) {
                        this.$resultsWrapper.hide();
                    }
                    //hide and deactivate everything else
                    _.chain(this.children).omit(['flashMessage']).each(function(child) {
                        child.deactivate({deep: true}).$el.hide();
                    });
                    
                    return this;
                }

                this.children.jobStatus.activate({deep: true}).$el.show();
                if (this.$resultsWrapper) {
                    this.$resultsWrapper.show();
                }

                if (jobPreparing) {
                    this.children.jobDispatchState.activate({deep: true}).$el.show();
                    
                    //hide and deactivate everything else
                    _.chain(this.children).omit(['jobDispatchState', 'jobStatus', 'flashMessage']).each(function(child) {
                        child.deactivate({deep: true}).$el.hide();
                    });
                    
                    return this;
                }

                isReportSearch = this.model.searchJob.isReportSearch();
                this.children.jobDispatchState.deactivate({deep: true}).$el.hide();

                if (isReportSearch) {
                    if (!this.children.resultsPane.active) {
                        this.children.resultsPane.activate().$el.show();
                    }
                    this.children.eventsPane.deactivate().$el.hide();
                } else {
                    if (!this.children.eventsPane.active) {
                        this.children.eventsPane.activate().$el.show();
                    }
                    this.children.resultsPane.deactivate().$el.hide();
                }
                
                return this;
            },
            render: function() {
                if (!this.el.innerHTML) {
                    this.$el.html(this.template);
                    this.$resultsWrapper = this.$('.results-wrapper');
                    
                    this.children.flashMessage.render().prependTo(this.$el);
                    this.$('.job-bar').replaceWith(this.children.jobStatus.render().el);
                    this.children.jobDispatchState.render().appendTo(this.$resultsWrapper);
                    this.children.eventsPane.render().appendTo(this.$resultsWrapper);
                    this.children.resultsPane.render().appendTo(this.$resultsWrapper);

                    if (this.model.report.isRealTime() && !this.model.user.canRTSearch()) {
                        this.$resultsWrapper.hide();
                    }
                }
                
                return this;
            },
            template: '\
                <div class="job-bar"></div>\
                <div class="results-wrapper"></div>\
            '
        });
    }
);
