define([
        'underscore',
        'backbone',
        'module',
        'views/Base',
        'views/shared/JobDispatchState',
        'views/shared/JSChartDrilldown',
        'views/shared/SingleValue',
        'views/shared/LazyMap',
        'views/shared/LazyResultsTableDrilldown',
        'views/report/tablecontrols/Master',
        'models/services/search/IntentionsParser',
        'models/Base',
        'util/drilldown_utils',
        'util/splunkd_utils',
        'splunk.util',
        'uri/route'
    ],
    function(
        _,
        Backbone,
        module,
        Base,
        JobDispatchState,
        JSChartDrilldown,
        SingleValue,
        LazyMap,
        LazyResultsTableDrilldown,
        TableControls,
        IntentionsParser,
        BaseModel,
        drilldown_utils,
        splunkd_utils,
        splunkUtil,
        route
    ){
        return Base.extend({
            moduleId: module.id,
            initialize: function() {
                Base.prototype.initialize.apply(this, arguments);

                this.model.drilldownIntentionsParser = new IntentionsParser();
                this.model.state = new BaseModel();

                this.children.jobDispatchState = new JobDispatchState({
                    model: this.model.searchJob,
                    mode: this.model.searchJob.entry.content.get('isPreviewEnabled') ? 'results_preview' : 'results'
                });

                this.children.resultsTableDrilldown = new LazyResultsTableDrilldown({
                    model: {
                        results: this.model.resultJsonRows,
                        metadata: this.model.report.entry.content,
                        format: this.model.report.entry.content,
                        intentionsParser: this.model.drilldownIntentionsParser,
                        searchJob: this.model.searchJob,
                        application: this.model.application,
                        state: this.model.state
                    },
                    sortKeyAttr: 'display.statistics.sortColumn',
                    sortDirAttr: 'display.statistics.sortDirection',
                    countAttr: 'display.prefs.statistics.count',
                    offsetAttr: 'display.prefs.statistics.offset',
                    tableDockOffset: 36
                });

            },
            initializeSingleValue: function() {
                this.children.singleValue = new SingleValue({
                    model: {
                        searchResultsColumn: this.model.resultJsonCols,
                        state: this.model.report.entry.content
                    },
                    width: 800,
                    height: 300
                });
            },
            initializeChart: function() {
                this.children.jsChartDrilldown = new JSChartDrilldown({
                    model: {
                        searchResultsColumn: this.model.resultJsonCols,
                        state: this.model.report.entry.content,
                        application: this.model.application,
                        intentionsParser: this.model.drilldownIntentionsParser,
                        searchJob: this.model.searchJob
                    },
                    width: '100%',
                    height: this.model.report.entry.content.get('display.visualizations.chartHeight')
                });
            },
            initializeControls: function() {
                this.children.controls = new TableControls({
                    model: {
                        report: this.model.report,
                        searchJob: this.model.searchJob,
                        state: this.model.state
                    },
                    mode: this.model.searchJob.entry.content.get('isPreviewEnabled') ? 'results_preview' : 'results',
                    offsetKey: "display.prefs.statistics.offset",
                    countKey: "display.prefs.statistics.count"
                });
            },
            initializeMap: function() {
                this.children.map = new LazyMap({
                    model: {
                        searchResultsColumn: this.model.resultJsonCols,
                        state: this.model.report.entry.content
                    },
                    width: '100%',
                    height: this.model.report.entry.content.get('display.visualizations.mapHeight')
                });
                this.listenTo(this.children.map, 'boundsChanged', this.updateMapPostProcess);
                this.listenTo(this.children.map, 'mapClicked', function(e) {
                    if(this.model.report.entry.content.get('display.visualizations.mapping.drilldown') === 'none') {
                        return;
                    }
                    this.model.drilldownIntentionsParser.clear({silent: true});
                    var intentionsData = $.extend(drilldown_utils.convertMapClickToIntentions(e), {
                        q: this.model.report.entry.content.get('search'),
                        app: this.model.application.get('app'),
                        owner: this.model.application.get('owner')
                    });
                    this.model.drilldownIntentionsParser.fetch({ data: intentionsData });
                });
            },
            updateMapPostProcess: function() {
                this.model.resultJsonCols.fetchData.set({
                    search: this.children.map.wrappedView.getPostProcessSearch()
                });
            },
            startListening: function() {
                this.listenTo(this.model.searchJob.entry.content, "change:dispatchState change:resultCount change:resultPreviewCount", _.debounce( function() {
                    if (this.active) {
                        this.visibility();
                    }
                }, 0));

                this.listenTo(this.model.drilldownIntentionsParser, 'change', function() {
                    var routeString = route.search(
                            this.model.application.get('root'),
                            this.model.application.get('locale'),
                            this.model.application.get('app'),
                            {data: {
                                earliest: this.model.drilldownIntentionsParser.get('dispatch.earliest_time') || this.model.report.entry.content.get('dispatch.earliest_time'),
                                latest: this.model.drilldownIntentionsParser.get('dispatch.latest_time') || this.model.report.entry.content.get('dispatch.latest_time'),
                                q: this.model.drilldownIntentionsParser.get('fullSearch')
                            }}
                        ),
                        newTab = this.model.drilldownIntentionsParser.get('drilldownNewTab');
                    if (newTab) {
                        newTab.location = routeString;
                        this.model.drilldownIntentionsParser.clear({silent: true});
                    } else {
                        window.location = routeString;
                    }
                });
            },
            activate: function(options) {
                var clonedOptions = _.extend({}, (options || {}));
                delete clonedOptions.deep;
                this.ensureDeactivated({deep:true});
                
                this.render();

                if (!this.children.controls) {
                    this.initializeControls();
                    this.children.controls.render().insertBefore(this.children.resultsTableDrilldown.el);
                }

                Base.prototype.activate.call(this, clonedOptions);
                this.visibility();
                return this;
            },
            deactivate: function(options) {
                if (!this.active) {
                    return Base.prototype.deactivate.apply(this, arguments);
                }

                Base.prototype.deactivate.apply(this, arguments);
                
                this.children.controls.remove();
                delete this.children.controls;
                
                if (this.children.jsChartDrilldown) {
                    this.children.jsChartDrilldown.remove();
                    delete this.children.jsChartDrilldown;
                }
                
                if (this.children.singleValue) {
                    this.children.singleValue.remove();
                    delete this.children.singleValue;
                }
                
                if (this.children.map) {
                    this.children.map.remove();
                    delete this.children.map;
                }

                this.model.drilldownIntentionsParser.fetchAbort();
                this.model.drilldownIntentionsParser.clear();
                
                return this;
            },
            showVizChild: function(child) {
                var needsRender = false,
                    $wrapper = $('<div class="chart-viewer"></div>'),
                    wrapperSelector = 'div.chart-viewer';

                if (!this.$el.children(wrapperSelector)[0]) {
                    $wrapper.insertBefore(this.children.controls.el);
                }

                if (!child.isAddedToDocument()) {
                    child.appendTo(this.$(wrapperSelector));
                    needsRender = true;
                }
                
                if (!child.active) {
                    child.activate({deep: true });
                }
                if (needsRender) {
                    child.render();
                }
                child.$el.show();
            },
            visibility: function() {
                var resultCount = this.model.searchJob.entry.content.get('isPreviewEnabled') ?
                        this.model.searchJob.entry.content.get('resultPreviewCount') :
                        this.model.searchJob.entry.content.get('resultCount'),
                    vizType = this.model.report.entry.content.get('display.visualizations.type');

                if (resultCount === 0) {
                    this.children.jobDispatchState.activate({deep: true}).$el.show();
                    if (this.model.searchJob.entry.content.get('dispatchState') === 'DONE') {
                        if (this.children.jsChartDrilldown) {
                            this.children.jsChartDrilldown.deactivate({deep: true }).$el.hide();
                        }
                        if (this.children.singleValue) {
                            this.children.singleValue.deactivate({deep: true }).$el.hide();
                        }
                    }
                    if (this.children.map) {
                        this.children.map.deactivate({deep: true }).$el.hide();
                    }

                    this.children.controls.deactivate({deep: true}).$el.hide();
                    this.children.resultsTableDrilldown.deactivate({deep: true }).$el.hide();
                } else {
                    this.children.jobDispatchState.deactivate({deep: true}).$el.hide();
                    this.children.controls.activate({deep: true}).$el.show();
                    this.children.resultsTableDrilldown.load().activate({deep: true}).$el.show();
                    if (splunkUtil.normalizeBoolean(this.model.report.entry.content.get('display.visualizations.show'))) {
                        switch (vizType) {
                            case 'charting':
                                if (!this.children.jsChartDrilldown) {
                                    this.initializeChart();
                                }
                                this.showVizChild(this.children.jsChartDrilldown);
                                break;
                            case 'singlevalue':
                                if (!this.children.singleValue) {
                                    this.initializeSingleValue();
                                }
                                this.showVizChild(this.children.singleValue);
                                break;
                            case 'mapping':
                                if (!this.children.map) {
                                    this.initializeMap();
                                }
                                this.children.map.load();
                                this.showVizChild(this.children.map);
                                $.when(this.children.map.wrappedViewLoaded()).done(_.bind(function() {
                                    this.model.resultJsonCols.fetchData.set({
                                        // TODO [sff] this should be a report attribute and handled upstream
                                        count: this.children.map.wrappedView.getMaxClusters()
                                    }, { silent: true });
                                    this.updateMapPostProcess();
                                }, this));
                                break;
                        }
                    }
                }
            },
            render: function() {
                if (!this.el.innerHTML) {
                    this.children.resultsTableDrilldown.render().appendTo(this.$el);
                    this.children.jobDispatchState.render().appendTo(this.$el);
                }
                return this;
            }
        });
    }
);