define([
            'jquery',
            'underscore',
            'module',
            'models/config',
            'views/Base',
            'util/jscharting_utils',
            'util/deferred_loader',
            'splunk.util',
            'helpers/Printer',
            'uri/route'
        ],
        function(
            $,
            _,
            module,
            configModel,
            Base,
            jschartingUtils,
            deferredLoader,
            splunkUtils,
            Printer,
            route
        ) {

    var JSChart = Base.extend({
        VIZ_PROPERTY_PREFIX_REGEX: /^display\.visualizations\.charting\./,
        className: 'chart',
        moduleId: module.id,
        /**
         * @constructor
         * @param options {
         * model: {
         *     searchResultsColumn <models.services.jobs.ResultJsonCols>: the data
         *     state: <content from models.search.report>: the charting properties
         *     application: <models.Application>: the application state
         * },
         * width: initial width,
         * height: initial height
         * }
         */
        initialize: function(options) {
            Base.prototype.initialize.apply(this, arguments);
            this.options = options || {};
                        
            if(this.options.width) {
                this.$el.width(this.options.width);
            }
            if(this.options.height) {
                this.$el.height(this.options.height);
            }

            this.$chart = $('<div></div>');
            this.$inlineMessage = $('<div></div>').css('text-align', 'center');

            this.activate();
        },
        startListening: function() {
            this.listenTo(this.model.state, 'change',function(){
                if(this.model.state.get('display.visualizations.type') === 'charting') {
                    var changedVizProps = this.model.state.filterChangedByWildcards(this.VIZ_PROPERTY_PREFIX_REGEX, {allowEmpty: true});
                    if(!_(changedVizProps).isEmpty()) {
                        this.debouncedDrawChart();
                    }
                }
            });
            
            this.listenTo(this.model.searchResultsColumn, 'change:fields change:columns', this.debouncedDrawChart);
            this.listenTo(this.model.searchResultsColumn, 'destroy', this.empty);
            
            this.listenTo(Printer, Printer.PRINT_START, this.reflow);
            this.listenTo(Printer, Printer.PRINT_END, this.reflow);
        },
        activate: function(options) {
            if (this.active) {
                return Base.prototype.activate.apply(this, arguments);
            }
            this.awake = true;
            this.$el.height(this.model.state.get('display.visualizations.chartHeight'));
            return Base.prototype.activate.apply(this, arguments);
        },
        deactivate: function(options) {
            if (!this.active) {
                return Base.prototype.deactivate.apply(this, arguments);
            }
            Base.prototype.deactivate.apply(this, arguments);            
            this.awake = false;
            return this;
        },
        empty:  function() {
            this.destroyChart();
            this.$chart.empty();
            this.$inlineMessage.empty();
            return this;
        },
        remove: function() {
            this.destroyChart();
            // Set awake to false here in case there is a pending call to drawChart (SPL-78135).
            this.awake = false;
            return Base.prototype.remove.apply(this, arguments);
        },
        // TODO [sff] this needs more thought, we want render to be snappy since it is often called in response to user action
        // but the actual draw can take a long time
        render: function() {
            this.$chart.appendTo(this.el);
            this.$inlineMessage.appendTo(this.el);
            this.debouncedDrawChart();
            return this;
        },
        setHeight: function(height) {
            this.$el.height(height);
        },
        debouncedDrawChart: function() {
            if(this._lazyLoadInProgress) {
                return;
            }
            if(this._debouncedDrawChart) {
                this._debouncedDrawChart();
                return;
            }
            this._lazyLoadInProgress = true;
            deferredLoader.load(['js_charting/js_charting']).done(_.bind(function(lazyJSCharting) {
                JSChart.JSCHARTING_LIBRARY = lazyJSCharting;
                this.drawChart();
                this._debouncedDrawChart = _.debounce(this.drawChart, 25);
                this._lazyLoadInProgress = false;
            }, this));
        },
        drawChart: function() {
            if (!this.awake || !this.active) {
                // it's possible we were put to sleep between when render was called and when drawChart was called
                this.touch = true;
                return;
            }
            this.$inlineMessage.empty();

            var rawResults = {
                fields: this.model.searchResultsColumn.get('fields') || [],
                columns: this.model.searchResultsColumn.get('columns') || []
            };
            var displayProperties = {};
            _.each(this.model.state.toJSON(), function(value, key){
                if(this.VIZ_PROPERTY_PREFIX_REGEX.test(key)) {
                    displayProperties[key.replace(this.VIZ_PROPERTY_PREFIX_REGEX, '')] = value;
                }
            }, this);

            var results = jschartingUtils.preprocessChartData(rawResults, displayProperties),
                maxResultCount = _.isFunction(this.model.searchResultsColumn.getFetchCount) ?
                                        this.model.searchResultsColumn.getFetchCount() : Infinity;

            // if the preprocessChartData method truncated the data, show a message to that effect
            if(results.columns.length > 0 &&
                    (results.columns.length < rawResults.columns.length ||
                        results.columns[0].length < rawResults.columns[0].length)) {
                this.renderResultsTruncatedMessage();
            }
            // otherwise if the number of results matches the max result count that was used to fetch,
            // show a message that we might not be displaying the full data set
            else if(results.columns.length > 0 && maxResultCount > 0 && results.columns[0].length >= maxResultCount) {
                this.renderMaxResultCountMessage(maxResultCount);
            }
            var chartReadyData = JSChart.JSCHARTING_LIBRARY.extractChartReadyData(results);
            $.extend(displayProperties, jschartingUtils.getCustomDisplayProperties(chartReadyData, configModel.toJSON()));

            if(!this.chart || !_.isEqual(displayProperties, this.chart.getCurrentDisplayProperties())) {
                this.destroyChart();
                this.chart = JSChart.JSCHARTING_LIBRARY.createChart(this.$chart, displayProperties);
            }
            this.updateChartContainerHeight();
            var that = this;
            this.chart.prepareAndDraw(chartReadyData, {}, function(chart) {
                that.trigger('chartUpdate', { fields: chart.getFieldList() });
            });
            this.chart.on('pointClick', function(eventInfo) {
                that.trigger('chartClick', eventInfo);
            });
            this.chart.on('legendClick', function(eventInfo) {
                that.trigger('chartClick', eventInfo);
            });
        },
        destroyChart: function() {
            if(this.chart) {
                this.chart.off();
                this.chart.destroy();
                delete this.chart; // GC handler
            }
        },
        reflow: function() {
            if(this.chart && this.$el.height() > 0) {
                this.updateChartContainerHeight();
                this.chart.resize();
            }
            Base.prototype.reflow.call(this);
        },
        updateChartContainerHeight: function() {
            this.$chart.height(this.$el.height() - this.$inlineMessage.outerHeight());
        },
        renderResultsTruncatedMessage: function() {
            var rawMessage = 'These results may be truncated. Your search generated too much data for' +
                                ' the current visualization configuration.',
                message = _(rawMessage).t(),
                docsHref = route.docHelp(
                    this.model.application.get('root'),
                    this.model.application.get('locale'),
                    'learnmore.charting.datatruncation'
                ),
                helpLink = ' <a href="<%- href %>" target="_blank">' +
                                '<span><%- text %></span>' +
                                '<i class="icon-external icon-no-underline"></i>' +
                            '</a>';

            message += _(helpLink).template({ href: docsHref, text: _('Learn More').t() });
            this.$inlineMessage.html(_(this.inlineMessageTemplate).template({ message: message, level: 'warning' }));
        },
        renderMaxResultCountMessage: function(resultCount) {
            var message = splunkUtils.sprintf(
                _('These results may be truncated. This visualization is configured to display a maximum of %s results per series, and that limit has been reached.').t(),
                resultCount
            );
            this.$inlineMessage.html(_(this.inlineMessageTemplate).template({ message: message, level: 'warning' }));
        },
        inlineMessageTemplate: '\
            <div class="alert alert-inline alert-<%= level %> alert-inline"> \
                <i class="icon-alert"></i> \
                <%= message %> \
            </div> \
        '
    },
    {
        // A placeholder for the charting library, which will be lazy-loaded.
        // This is visible for testing only.
        JSCHARTING_LIBRARY: null
    });

    return JSChart;

 });
