define(function(require, exports, module) {

    var $ = require('jquery');
    var SplunkI18N = require('splunk.i18n');
    var _ = SplunkI18N._;
    var SplunkUtil = require('splunk.util');
    var sprintf = SplunkUtil.sprintf;
    var FunctionUtils = require('jgatt').utils.FunctionUtils;
    var ValidateQueue = require('jgatt').validation.ValidateQueue;
    var ExternalLegend = require('splunk/charting/ExternalLegend');
    var Map = require('splunk/mapping/Map');
    var PieMarkerLayer = require('splunk/mapping/layers/PieMarkerLayer');
    var LatLonBoundsParser = require('splunk/mapping/parsers/LatLonBoundsParser');
    var LatLonParser = require('splunk/mapping/parsers/LatLonParser');
    var FieldColorPalette = require('splunk/palettes/FieldColorPalette');
    var ListColorPalette = require('splunk/palettes/ListColorPalette');
    var ArrayParser = require('splunk/parsers/ArrayParser');
    var BooleanParser = require('splunk/parsers/BooleanParser');
    var NumberParser = require('splunk/parsers/NumberParser');
    var ObjectParser = require('splunk/parsers/ObjectParser');
    var StringParser = require('splunk/parsers/StringParser');
    var Base = require('views/Base');
    var console = require('util/console');

    var _DEFAULT_PROPERTY_VALUES = {
        "fieldColors": "",
        "seriesColors": "[" +
            "0x6CB8CA,0xFAC61D,0xD85E3D,0x956E96,0xF7912C,0x9AC23C,0x5479AF,0x999755,0xDD87B0,0x65AA82," +
            "0xA7D4DF,0xFCDD77,0xE89E8B,0xBFA8C0,0xFABD80,0xC2DA8A,0x98AFCF,0xC2C199,0xEBB7D0,0xA3CCB4," +
            "0x416E79,0x967711,0x823825,0x59425A,0x94571A,0x5C7424,0x324969,0x5C5B33,0x85516A,0x3D664E" +
        "]",
        "data.maxClusters": "100",
        "tileLayer.url": "/splunkd/__raw/services/mbtiles/splunk-tiles/{z}/{x}/{y}",
        "tileLayer.subdomains": "[a,b,c]",
        "tileLayer.minZoom": "0",
        "tileLayer.maxZoom": "7",
        "tileLayer.invertY": "false",
        "tileLayer.attribution": "",
        "markerLayer.markerOpacity": "1",
        "markerLayer.markerMinSize": "10",
        "markerLayer.markerMaxSize": "50"
    };

    var _R_PROPERTY_PREFIX = /^display\.visualizations\.mapping\./;

    return Base.extend({

        // Public Properties

        moduleId: module.id,

        // Private Properties

        _map: null,
        _markerLayer: null,
        _externalLegend: null,
        _fieldColorPalette: null,
        _seriesColorPalette: null,
        _propertyValues: null,
        _booleanParser: null,
        _numberParser: null,
        _stringParser: null,
        _numberArrayParser: null,
        _stringArrayParser: null,
        _numberObjectParser: null,
        _latLonParser: null,
        _latLonBoundsParser: null,
        _maxClusters: 100,
        _isPrinting: false,
        _prePrintCenter: null,
        _prePrintZoom: 0,

        // Constructor

        initialize: function(options) {
            Base.prototype.initialize.apply(this, arguments);

            this.options = options || {};
            this.$el.width(this.options.width || "100%");
            this.$el.height(this.options.height || "400px");

            this._map_boundsChanged = FunctionUtils.bind(this._map_boundsChanged, this);
            this._map_mapClicked = FunctionUtils.bind(this._map_mapClicked, this);
            this._document_PrintStart = FunctionUtils.bind(this._document_PrintStart, this);
            this._document_PrintEnd = FunctionUtils.bind(this._document_PrintEnd, this);

            this._propertyValues = {};

            this._booleanParser = BooleanParser.getInstance();
            this._numberParser = NumberParser.getInstance();
            this._stringParser = StringParser.getInstance();
            this._numberArrayParser = ArrayParser.getInstance(this._numberParser);
            this._stringArrayParser = ArrayParser.getInstance(this._stringParser);
            this._numberObjectParser = ObjectParser.getInstance(this._numberParser);
            this._latLonParser = LatLonParser.getInstance();
            this._latLonBoundsParser = LatLonBoundsParser.getInstance();

            this._seriesColorPalette = new ListColorPalette();
            this._fieldColorPalette = new FieldColorPalette(null, this._seriesColorPalette);

            this._externalLegend = new ExternalLegend();
            this._externalLegend.connect();

            this._markerLayer = new PieMarkerLayer();
            this._markerLayer.set("legend", this._externalLegend);
            this._markerLayer.set("markerColorPalette", this._fieldColorPalette);

            this._map = new Map();
            this._map.formatNumber = this._formatNumber;
            this._map.formatDegrees = this._formatDegrees;
            this._map.addLayer(this._markerLayer);
            this._map.addEventListener("boundsChanged", this._map_boundsChanged);
            this._map.addEventListener("mapClicked", this._map_mapClicked);
            this._map.appendTo(this.$el);
            this._map.fitWorld(true);

            $(document).bind("PrintStart", this._document_PrintStart);
            $(document).bind("PrintEnd", this._document_PrintEnd);

            this.updateData();
            this.updateProperties();
            this._updateDataBounds();
            this.activate();
        },

        startListening: function() {
            if (this.model.searchResultsRow)
                this.listenTo(this.model.searchResultsRow, "change destroy error", this.updateData);
            else if (this.model.searchResultsColumn)
                this.listenTo(this.model.searchResultsColumn, "change destroy error", this.updateData);

            if (this.model.state)
                this.listenTo(this.model.state, "change", this._state_change);
        },

        // Public Methods

        getCenter: function() {
            return this._map.get("center").normalize();
        },

        getZoom: function() {
            return this._map.get("zoom");
        },

        getLatLonBounds: function() {
            return this._map.getLatLonBounds().normalize();
        },

        getPostProcessSearch: function() {
            var bounds = this._map.getLatLonBounds().normalize();
            return "geofilter south=" + bounds.s + " west=" + bounds.w + " north=" + bounds.n + " east=" + bounds.e + " maxclusters=" + this._maxClusters;
        },

        getMaxClusters: function() {
            return this._maxClusters;
        },

        updateData: function() {
            var extractedData = null;
            if (this.model.searchResultsRow)
                extractedData = this._extractRowData(this.model.searchResultsRow);
            else if (this.model.searchResultsColumn)
                extractedData = this._extractColumnData(this.model.searchResultsColumn);

            console.debug('Updating map data to', extractedData);

            this._markerLayer.set("data", extractedData ? extractedData.data : null);
            this._markerLayer.set("fields", extractedData ? this._filterFields(extractedData.fields) : null);
        },

        updateProperties: function() {
            var curValues = this._propertyValues;
            var newValues = {};
            var p;

            // set null values for all existing properties
            // if they are not overridden by either the default or state properties, they will be cleared
            for (p in curValues) {
                if (curValues.hasOwnProperty(p))
                    newValues[p] = null;
            }

            // copy default property values
            var defaultValues = _DEFAULT_PROPERTY_VALUES;
            for (p in defaultValues) {
                if (defaultValues.hasOwnProperty(p))
                    newValues[p] = defaultValues[p];
            }

            // copy non-empty state property values
            var stateValues = this.model.state ? this.model.state.toJSON() : {};
            var stateValue;
            for (p in stateValues) {
                if (stateValues.hasOwnProperty(p) && _R_PROPERTY_PREFIX.test(p)) {
                    stateValue = stateValues[p];
                    if ((stateValue != null) && (stateValue !== ""))
                        newValues[p.replace(_R_PROPERTY_PREFIX, "")] = stateValue;
                }
            }

            // apply map viewport properties in order
            // zoom must be first for Leaflet to do the right thing
            if (newValues.hasOwnProperty("map.zoom"))
                this._setMapProperty("map.zoom", newValues["map.zoom"]);
            if (newValues.hasOwnProperty("map.center"))
                this._setMapProperty("map.center", newValues["map.center"]);
            if (newValues.hasOwnProperty("map.fitBounds"))
                this._setMapProperty("map.fitBounds", newValues["map.fitBounds"]);

            // apply remaining properties
            // the viewport properties haven't changed, so they will be ignored by _setMapProperty
            for (p in newValues) {
                if (newValues.hasOwnProperty(p))
                    this._setMapProperty(p, newValues[p]);
            }
        },

        render: function() {
            return this;
        },

        remove: function() {
            $(document).unbind("PrintStart", this._document_PrintStart);
            $(document).unbind("PrintEnd", this._document_PrintEnd);

            this._map.removeEventListener("boundsChanged", this._map_boundsChanged);
            this._map.removeEventListener("mapClicked", this._map_mapClicked);
            this._map.dispose();

            this._externalLegend.close();

            return Base.prototype.remove.apply(this, arguments);
        },

        onShow: function() {
            this._map.updateSize();
            this._map.validate();
            Base.prototype.onShow.call(this);
        },

        // Private Methods

        _updateDataBounds: function() {
            var model = this.model.state;
            if (!model)
                return;

            var bounds = this._map.getLatLonBounds().normalize();
            model.set({ "display.visualizations.mapping.data.bounds": this._latLonBoundsParser.valueToString(bounds) });
        },

        _needsPropertyUpdate: function(changedProperties) {
            if (!changedProperties)
                return false;

            for (var p in changedProperties) {
                if (changedProperties.hasOwnProperty(p) && (p !== "display.visualizations.mapping.data.bounds"))
                    return true;
            }

            return false;
        },

        _extractRowData: function(model) {
            var extractedData = {};

            var fields = model.get("fields");
            var rows = model.get("rows");
            if (fields && rows) {
                var numFields = fields.length;
                var numRows = rows.length;
                var numEntries;
                var row;
                var obj;
                var i;
                var j;

                extractedData.fields = fields.concat();
                extractedData.data = [];
                for (i = 0; i < numRows; i++) {
                    row = rows[i];
                    numEntries = Math.min(row.length, numFields);
                    obj = {};
                    for (j = 0; j < numEntries; j++)
                        obj[fields[j]] = row[j];
                    extractedData.data.push(obj);
                }
            }

            return extractedData;
        },

        _extractColumnData: function(model) {
            var extractedData = {};

            var fields = model.get("fields");
            var columns = model.get("columns");
            if (fields && columns) {
                var numColumns = Math.min(fields.length, columns.length);
                var numRows = (numColumns > 0) ? columns[0].length : 0;
                var obj;
                var i;
                var j;

                for (i = 1; i < numColumns; i++)
                    numRows = Math.min(numRows, columns[i].length);

                extractedData.fields = fields.slice(0, numColumns);
                extractedData.data = [];
                for (i = 0; i < numRows; i++) {
                    obj = {};
                    for (j = 0; j < numColumns; j++)
                        obj[fields[j]] = columns[j][i];
                    extractedData.data.push(obj);
                }
            }

            return extractedData;
        },

        _filterFields: function(fields) {
            if (!fields)
                return null;

            var filteredFields = [];
            var field;
            for (var i = 0, l = fields.length; i < l; i++) {
                field = fields[i];
                if (field && (field.charAt(0) !== "_"))
                    filteredFields.push(field);
            }
            return filteredFields;
        },

        _setMapProperty: function(propertyName, propertyValue) {
            propertyValue = (propertyValue != null) ? String(propertyValue) : null;
            if (this._propertyValues[propertyName] == propertyValue)
                return;

            if (propertyValue != null)
                this._propertyValues[propertyName] = propertyValue;
            else
                delete this._propertyValues[propertyName];

            switch (propertyName) {
                // global properties
                case "fieldColors":
                    this._fieldColorPalette.set("fieldColors", this._numberObjectParser.stringToValue(propertyValue));
                    break;
                case "seriesColors":
                    this._seriesColorPalette.set("colors", this._numberArrayParser.stringToValue(propertyValue));
                    break;

                // data properties
                case "data.maxClusters":
                    var maxClusters = this._numberParser.stringToValue(propertyValue);
                    this._maxClusters = (maxClusters < Infinity) ? Math.max(Math.floor(maxClusters), 64) : 100;
                    break;

                // map properties
                case "map.center":
                    var center = this._latLonParser.stringToValue(propertyValue);
                    if (center)
                        this._map.set("center", center);
                    break;
                case "map.zoom":
                    var zoom = this._numberParser.stringToValue(propertyValue);
                    if (!isNaN(zoom))
                        this._map.set("zoom", zoom);
                    break;
                case "map.fitBounds":
                    var fitBounds = this._latLonBoundsParser.stringToValue(propertyValue);
                    if (fitBounds)
                        this._map.fitBounds(fitBounds);
                    break;

                // tileLayer properties
                case "tileLayer.url":
                    this._map.set("tileURL", this._resolveURL(propertyValue));
                    break;
                case "tileLayer.subdomains":
                    this._map.set("tileSubdomains", this._stringArrayParser.stringToValue(propertyValue));
                    break;
                case "tileLayer.minZoom":
                    this._map.set("tileMinZoom", this._numberParser.stringToValue(propertyValue));
                    break;
                case "tileLayer.maxZoom":
                    this._map.set("tileMaxZoom", this._numberParser.stringToValue(propertyValue));
                    break;
                case "tileLayer.invertY":
                    this._map.set("tileInvertY", this._booleanParser.stringToValue(propertyValue));
                    break;
                case "tileLayer.attribution":
                    this._map.set("tileAttribution", this._stringParser.stringToValue(propertyValue));
                    break;

                // markerLayer properties
                case "markerLayer.markerOpacity":
                    this._markerLayer.set("markerOpacity", this._numberParser.stringToValue(propertyValue));
                    break;
                case "markerLayer.markerMinSize":
                    this._markerLayer.set("markerMinSize", this._numberParser.stringToValue(propertyValue));
                    break;
                case "markerLayer.markerMaxSize":
                    this._markerLayer.set("markerMaxSize", this._numberParser.stringToValue(propertyValue));
                    break;
            }
        },

        _resolveURL: function(propertyValue) {
            var propertyValue2 = propertyValue ? SplunkUtil.trim(propertyValue) : propertyValue;
            if (propertyValue2 && (propertyValue2.charAt(0) === "/")) {
                var hadTrailingSlash = (propertyValue2.charAt(propertyValue2.length - 1) === "/");
                propertyValue2 = SplunkUtil.make_url(propertyValue2);
                var hasTrailingSlash = (propertyValue2.charAt(propertyValue2.length - 1) === "/");
                if (hasTrailingSlash != hadTrailingSlash)
                    propertyValue2 = hadTrailingSlash ? propertyValue2 + "/" : propertyValue2.substring(0, propertyValue2.length - 1);
                propertyValue = propertyValue2;
            }
            return propertyValue;
        },

        _formatNumber: function(num) {
            var pos = Math.abs(num);
            if ((pos > 0) && ((pos < 1e-3) || (pos >= 1e9)))
                return SplunkI18N.format_scientific(num, "##0E0");
            return SplunkI18N.format_decimal(num);
        },

        _formatDegrees: function(degrees, orientation) {
            var deg = Math.abs(degrees);
            var degInt = Math.floor(deg);
            var degStr = ("" + degInt);
            var min = (deg - degInt) * 60;
            var minInt = Math.floor(min);
            var minStr = (minInt < 10) ? ("0" + minInt) : ("" + minInt);
            var sec = (min - minInt) * 60;
            var secInt = Math.floor(sec);
            var secStr = (secInt < 10) ? ("0" + secInt) : ("" + secInt);

            var dirStr = "";
            if (degrees > 0)
                dirStr = (orientation === "ns") ? _("N") : _("E");
            else if (degrees < 0)
                dirStr = (orientation === "ns") ? _("S") : _("W");

            if (secInt > 0)
                return sprintf("%(degrees)s\u00B0%(minutes)s'%(seconds)s\"%(direction)s", { degrees: degStr, minutes: minStr, seconds: secStr, direction: dirStr });
            if (minInt > 0)
                return sprintf("%(degrees)s\u00B0%(minutes)s'%(direction)s", { degrees: degStr, minutes: minStr, direction: dirStr });
            return sprintf("%(degrees)s\u00B0%(direction)s", { degrees: degStr, direction: dirStr });
        },

        _state_change: function() {
            if (this.model.state.get("display.visualizations.type") === "mapping") {
                var changedProperties = this.model.state.filterChangedByWildcards(_R_PROPERTY_PREFIX, { allowEmpty: true });
                if (this._needsPropertyUpdate(changedProperties))
                    this.updateProperties();
            }
        },

        _map_boundsChanged: function(e) {
            if (this._isPrinting)
                return;

            this._updateDataBounds();
            this.trigger("boundsChanged", {});
        },

        _map_mapClicked: function(e) {
            if (this._isPrinting)
                return;

            this.trigger("mapClicked", { data: e.data, fields: e.fields, altKey: e.altKey, ctrlKey: e.ctrlKey, shiftKey: e.shiftKey });
        },

        _document_PrintStart: function(e) {
            if (this._isPrinting)
                return;

            this._isPrinting = true;

            this._prePrintCenter = this._map.get("center");
            this._prePrintZoom = this._map.get("zoom");

            this._map.updateSize();

            var dataBounds = this._markerLayer.getLatLonBounds(this._prePrintCenter);
            if (dataBounds)
                this._map.fitBounds(dataBounds);

            ValidateQueue.validateAll();
        },

        _document_PrintEnd: function(e) {
            if (!this._isPrinting)
                return;

            this._map.set("center", this._prePrintCenter);
            this._map.set("zoom", this._prePrintZoom);

            this._prePrintCenter = null;
            this._prePrintZoom = 0;
            this._isPrinting = false;
        }

    }, {
        DEFAULT_PROPERTY_VALUES: _DEFAULT_PROPERTY_VALUES
    });

});
