/**
 * A subclass of the Page view that dynamically collapses the header for small screen sizes
 * and provides a way to click to expand to see the full header.
 *
 * For the relevant CSS see exposed/less/base/navs-responsive.less under "Styles for the responsive header"
 */

define([
            'jquery',
            'underscore',
            'module',
            'views/shared/Page',
            'views/shared/appbar/Master',
            'helpers/user_agent',
            'util/color_utils'
        ],
        function(
            $,
            _,
            module,
            Page,
            AppbarMaster,
            userAgent,
            colorUtils
        ) {

    var HEADER_ANIMATION_HEIGHT = 59,
        HEIGHT_OF_COLLAPSED_NAV_BAR = 6,
        ANIMATION_DURATION = 200,
        INITIAL_HIDE_DELAY = 500;

    return Page.extend({

        moduleId: module.id,

        events: {
            'click .expand-collapse-control': function(e) {
                e.preventDefault();
                if(this.isAnimating) {
                    return;
                }
                if(this.isExpanded) {
                    this.hideFullHeader();
                }
                else {
                    this.showFullHeader();
                }
            }
        },

        initialize: function() {
            Page.prototype.initialize.apply(this, arguments);
            this.isExpanded = false;
            this.isAnimating = false;
            // boolean to keep track of whether the view has done its initial hide-show animation
            this.hasInitialized = false;
            this.$screen = $('<div class="expanded-nav-screen" tabindex=-1></div>').hide().appendTo('body');

            var that = this;
            this.$screen.on('click.' + this.cid, function() {
                if(that.isAnimating) {
                    return;
                }
                that.hideFullHeader();
            });
            $(window).on('resize.' + this.cid, function() {
                that.$screen.hide();
                that.$('header').stop();
                that.$('header').css('top', '');
                if(that.isInCollapsibleMode() && that.hasInitialized) {
                    that.detachNavAndAppBar();
                }
                else {
                    that.attachNavAndAppBar();
                }
                that.isExpanded = false;
                that.isAnimating = false;
            });

            this.activate();
        },

        startListening: function() {
            this.listenTo(this.model.appNav.entry.content, 'change:color', function() {
                this.parseBannerColor();
                this.updateBannerColor();
            });
        },

        remove: function() {
            this.$screen.remove();
            $(window).off('resize.' + this.cid);
            return Page.prototype.remove.apply(this, arguments);
        },

        render: function() {
            this.renderLoadingMessage();

            if (this.options.modelRef && this.options.modelRef.tour) {
                this.$el.addClass("non-responsive");
                $.when(this.options.deferreds.tour).then(function() {
                    this.model.tour = this.options.modelRef.tour;
                    this.children.tourBar.resetModel(this.model.tour);
                    this.children.tourBar.setFixedHeight(HEADER_ANIMATION_HEIGHT + HEIGHT_OF_COLLAPSED_NAV_BAR);
                    this.renderTour();
                }.bind(this));
            } else {
                this.renderHeader();
                this.renderFooter();
                this.$header = this.$('header');
                this.$navBar = this.$('.navbar');
                this.$appBar = this.$('.app-bar');
                this.$header.append(this.expandCollapseControlTemplate);
                this.parseBannerColor();
                this.updateBannerColor();
            } 

            return this;
        },
        renderTour: function() {
            this.$('.main-section-body').prepend(this.children.tourBar.render().el);
        },
        parseBannerColor: function() {
            var appNavColor = this.model.appNav.entry.content.get('color');
            if(appNavColor) {
                this.bannerColor = colorUtils.normalizeHexString(appNavColor);
            }
        },

        updateBannerColor: function() {
            if(!this.bannerColor) {
                return;
            }
            this.$('.expand-collapse-control').css({ 'background-color': this.bannerColor });
        },

        showFullHeader: function() {
            if (!this.$navBar || !this.$appBar || !this.$screen || !this.$header) {
                return;
            }
            var that = this;
            this.isAnimating = true;
            this.attachNavAndAppBar();
            this.$screen.fadeIn(ANIMATION_DURATION);
            this.$screen.css({'filter': 'alpha(opacity=30)'}); // FIX IE ALPHA FILTER
            this.$header.animate({ top: '+=' + HEADER_ANIMATION_HEIGHT }, {
                duration: ANIMATION_DURATION,
                done: function() {
                    that.isExpanded = true;
                    that.isAnimating = false;
                }
            });
        },

        hideFullHeader: function(onSuccess) {
            if (!this.$navBar || !this.$appBar) return;
            if (!this.$screen || !this.$header) {
                onSuccess();
                return;
            }
            var that = this;
            this.isAnimating = true;
            this.$screen.fadeOut(ANIMATION_DURATION);
            this.$header.animate({ top: '-=' + HEADER_ANIMATION_HEIGHT  }, {
                duration: ANIMATION_DURATION,
                done: function() {
                    that.detachNavAndAppBar();
                    that.isExpanded = false;
                    that.isAnimating = false;
                    if(onSuccess) {
                        onSuccess();
                    }
                }
            });
        },

        onAddedToDocument: function() {
            if (!this.$navBar || !this.$appBar) return;
            // on initial reflow in collapsible mode, make the full header visible anyway and hide it after a delay
            // NOTE: assumes that render will only be called once, at the beginning of the view's lifecycle
            if(this.isInCollapsibleMode()) {
                var showThenHide = _(function() {
                    this.$('header').css({ top: '+=' + HEADER_ANIMATION_HEIGHT });
                    this.isExpanded = true;
                    setTimeout(_(function() {
                        this.hideFullHeader(_(function() { this.hasInitialized = true; }).bind(this));
                    }).bind(this), INITIAL_HIDE_DELAY);
                }).bind(this);

                // SPL-67779 - IE 7 and 8 won't allow modifying the header synchronously here, so we have to defer the work,
                // this can sometimes cause a flicker, but oh well
                if(userAgent.isIELessThan(9)) {
                    _.defer(showThenHide);
                }
                else {
                    showThenHide();
                }
            }
            else {
                this.hasInitialized = true;
            }
            Page.prototype.onAddedToDocument.call(this);
        },

        isInCollapsibleMode: function() {
            // sniff out the CSS of the expand/collapse control element
            return this.$('.expand-collapse-control').css('display') !== 'none';
        },

        detachNavAndAppBar: function() {
            if (!this.$navBar || !this.$appBar) return;
            this.$navBar.detach();
            this.$appBar.detach();
        },

        attachNavAndAppBar: function() {
            if (this.$navBar && this.$appBar) {
                this.$navBar.prependTo(this.$header);
                this.$appBar.insertAfter(this.$navBar);
            }
        },

        expandCollapseControlTemplate: '\
            <div class="expand-collapse-control">\
                <a href="#" class="tab"><i class="icon-menu"></i></a>\
            </div>\
        '

    });

});
