define(
    [
        'jquery',
        'underscore',
        'module',
        'views/Base',
        'splunk.util',
        'uri/route'
    ],
    function($, _, module, Base, splunkUtil, route) {

        var getStringRepresentation = function(label, formatted) {
            if (_.isFunction(label)) {
                try {
                    label = label(formatted);
                } catch (e) {
                    return '';
                }
            }

            return label;
        };

        return Base.extend({
            moduleId: module.id,
            className: "single-value",
            initialize: function(options) {
                Base.prototype.initialize.apply(this, arguments);
                this.activate();
            },
            events: {
                'click a.single-drilldown': function(e) {
                    e.preventDefault();
                    var target = $(e.currentTarget);
                    var resModel = this.model.searchResultsColumn;

                    var rowContext = _.object(
                        _(resModel.get('fields')).map(function(f) { return 'row.' + f; }),
                        _(resModel.get('columns')).pluck(0)
                    );

                    var drilldownInfo = {
                        modifierKey: !!(e.ctrlKey || e.metaKey),
                        name: target.data('field'),
                        value: target.data('value'),
                        rowContext: rowContext
                    };

                    var specificEventNames = _(target.children('span').attr('class').split(' ')).map(function(cls) {
                        return 'click:' + cls;
                    }).join(' ');
                    this.trigger('click ' + specificEventNames, drilldownInfo);
                }
            },
            startListening: function() {
                this.listenTo(this.model.state, 'change:display.visualizations.singlevalue.beforeLabel change:display.visualizations.singlevalue.afterLabel change:display.visualizations.singlevalue.underLabel change:display.visualizations.singlevalue.additionalClass', this.render);
                this.listenTo(this.model.searchResultsColumn, 'change', this.render);
            },
            getResultField: function(field) {
                return this.getFieldValue(this.determineResultFieldName(field));
            },
            // fields can either be a list of strings or a list of dictionaries each with a 'name' entry
            // depending on whether 'show_metadata' is enabled
            getFieldNames: function() {
                var fields = this.model.searchResultsColumn.get('fields');
                if (!fields || fields.length === 0) {
                    return [];
                }
                if (_.isObject(fields[0])) {
                    return _(fields).pluck('name');
                }
                return $.extend([], fields);
            },
            setSeverity: function() {
                if (this.model.searchResultsColumn.get('fields')) {
                    var fields = _(this.model.searchResultsColumn.get('fields')).pluck('name'),
                        i = fields.indexOf('range');
                    if (i > -1) {
                        this.$('span.single-result').addClass(this.model.searchResultsColumn.get('columns')[i][0]);
                    }
                }
            },
            determineResultFieldName: function(configuredField) {
                var fields = this.getFieldNames();
                if (configuredField && _(fields).contains(configuredField)) {
                    return configuredField;
                } else {
                    return _(fields).find(function(f) {
                        return f === '_time' || f === '_raw' || f[0] !== '_';
                    });
                }
            },
            getFieldValue: function(field, defaultValue) {
                var fields = this.getFieldNames();
                var idx = _(fields).indexOf(field);
                var columns = this.model.searchResultsColumn.get('columns') || [];

                var column = columns[idx];
                if (!(column && column.length)) {
                    return defaultValue;
                } else {
                    return column[0];
                }
            },
            render: function() {
                if(!this._originalClass) {
                    this._originalClass = this.$el.attr('class');
                }
                var resultField = this.determineResultFieldName(this.model.state.get("display.visualizations.singlevalue.field"));
                var resultFieldValue = this.getFieldValue(resultField, _('N/A').t());
                var html = this.compiledTemplate({
                    result: resultFieldValue,
                    beforeLabel: getStringRepresentation(this.model.state.get("display.visualizations.singlevalue.beforeLabel") || "", resultFieldValue),
                    afterLabel: getStringRepresentation(this.model.state.get("display.visualizations.singlevalue.afterLabel") || "", resultFieldValue),
                    underLabel: getStringRepresentation(this.model.state.get("display.visualizations.singlevalue.underLabel") || "", resultFieldValue)
                });
                var cls = [
                    this._originalClass,
                    this.model.state.get('display.visualizations.singlevalue.additionalClass'),
                    this.getResultField(this.model.state.get('display.visualizations.singlevalue.classField'))
                ];
                this.$el.attr('class', cls.join(' '));
                this.$el.html(html);
                this.wrapLinks(resultField, resultFieldValue);
                this.setSeverity();
                return this;
            },
            wrapLinks: function(field, value) {
                if (!this.model.application) {
                    return;
                }
                var linkFields = this.model.state.get('display.visualizations.singlevalue.linkFields');
                linkFields = linkFields ? $.trim(linkFields).split(/\s*,\s*/) : [];
                var linkView = this.model.state.get('display.visualizations.singlevalue.linkView');
                var linkSearch = this.model.state.get('display.visualizations.singlevalue.linkSearch');
                var drilldown = this.model.state.get('display.visualizations.singlevalue.drilldown') || 'none';
                var app = this.model.application.toJSON();

                var LINK_FIELDS = {
                    beforelabel: '.before-label',
                    afterlabel: '.after-label',
                    underlabel: '.under-label',
                    result: '.single-result'
                };

                var link;
                if ((linkView !== 'search' && linkFields.length) || linkSearch) {
                    var url;
                    if (linkView) {
                        var params = linkSearch ? { q: linkSearch } : undefined;
                        if (linkView.charAt(0) === '/') {
                            url = splunkUtil.make_full_url(linkView, params);
                        } else {
                            url = route.page(app.root, app.locale, app.app, linkView, { data: params });
                        }
                    } else {
                        url = route.search(app.root, app.locale, app.app, { data: { q: linkSearch }});
                    }
                    link = $('<a />').attr('href', url);
                } else if (drilldown !== 'none') {
                    link = $('<a class="single-drilldown" href="#"></a>');
                    if (!_(linkFields).contains('result')) {
                        linkFields.push('result');
                    }
                }

                if (link) {
                    link.data({
                        field: field,
                        value: value
                    });
                    var el = this.$el;
                    _(linkFields).each(function(lf) {
                        var cls = LINK_FIELDS[lf];
                        if (cls) {
                            el.children(cls).wrap(link);
                        }
                    });
                }
            },
            template: '\
                <span class="before-label">\
                    <%- beforeLabel %>\
                </span>\
                <span class="single-result">\
                    <%- result %>\
                </span>\
                <span class="after-label">\
                    <%- afterLabel %>\
                </span>\
                <br>\
                <span class="under-label">\
                    <%- underLabel %>\
                </span>\
            '
        });
    }
);
