define(
    [
         'underscore',
         'backbone',
         'module',
         'views/Base',
         'views/shared/Modal',
         'views/shared/controls/ControlGroup',
         'views/shared/FlashMessages',
         'views/shared/alertcontrols/dialogs/edit_dialog/step1/RealTime',
         'views/shared/alertcontrols/dialogs/edit_dialog/step1/Scheduled',
         'util/splunkd_utils',
         'splunk.util',
         'uri/route'
     ],
     function(
        _,
        Backbone,
        module,
        BaseView,
        Modal,
        ControlGroup,
        FlashMessagesView,
        RealTime,
        Scheduled,
        splunkd_utils,
        splunkUtils,
        route
     ){
        return BaseView.extend({
            moduleId: module.id,
            className: 'alert-step-1',
            /**
             * @param {Object} options {
             *     model: {
             *         alert: <models.Alert>,
             *         user: <models.services.admin.User>,
             *         application: <models.Application>
             *     },
             *     {string} dialogType: edit|new  and indicates if it is a save or edit dialog.
             *     {string} dialogTitle: Text used for the dialogs title.
             * }
             */
            initialize: function() {
                BaseView.prototype.initialize.apply(this, arguments);

                var titleControlType = this.options.dialogType === 'edit' ? 'Label' : 'Text';

                this.children.flashMessages = new FlashMessagesView({
                    model: {
                        alert: this.model.alert,
                        cron: this.model.alert.cron,
                        workingTimeRange: this.model.alert.workingTimeRange
                    }
                });

                this.children.name = new ControlGroup({
                    className: 'alert-name control-group',
                    controlType: titleControlType,
                    controlClass: 'controls-block',
                    controlOptions: {
                        modelAttribute: 'ui.name',
                        model: this.model.alert,
                        save: false
                    },
                    label: _('Title').t()
                });

                this.children.descriptionField = new ControlGroup({
                    className: 'alert-description control-group',
                    controlType: 'Textarea',
                    controlClass: 'controls-block',
                    controlOptions: {
                        modelAttribute: 'ui.description',
                        model: this.model.alert,
                        placeholder: _('optional').t()
                    },
                    label: _('Description').t()
                });

                if (this.model.user.canRTSearch()) {
                    this.children.type = new ControlGroup({
                        className: 'alert-type control-group',
                        controlType: 'SyntheticRadio',
                        controlClass: 'controls-halfblock',
                        controlOptions: {
                            modelAttribute: 'ui.type',
                            model: this.model.alert,
                            items: [
                                {
                                    label: _('Scheduled').t(),
                                    value: 'scheduled'
                                },
                                {
                                    label: _('Real Time').t(),
                                    value: 'realtime'
                                }
                            ],
                            save: false
                        },
                        label: _('Alert type').t()
                    });
                } else {
                    this.children.type = new ControlGroup({
                        className: 'alert-type control-group',
                        controlType: 'Label',
                        controlOptions: {
                            defaultValue: _("Scheduled").t()
                        },
                        label: _('Alert type').t()
                    });
                }

                this.children.realtime  = new RealTime({
                    model: this.model.alert
                });

                this.children.scheduled = new Scheduled({
                    model: {
                        alert: this.model.alert,
                        application: this.model.application
                    }
                });

                this.model.alert.on('change:ui.type', this.toggleChildren, this);
            },
            toggleChildren: function() {
                switch(this.model.alert.get('ui.type')){
                    case 'realtime':
                        this.children.realtime.$el.show();
                        this.children.scheduled.$el.hide();
                        break;
                    case 'scheduled':
                        this.children.realtime.$el.hide();
                        this.children.scheduled.$el.show();
                        break;
                }
            },
            saveTimeRange: function() {
                var earliest = this.$('.earliest-timepicker input').val() || "" ,
                    latest = this.$('.latest-timepicker input').val() || "" ;

                this.model.alert.workingTimeRange.save(
                    {
                        earliest: earliest,
                        latest: latest
                    },
                    {
                        success: function(model) {
                            this.model.alert.entry.content.trigger('next');
                        }.bind(this)
                    }
                );
            },
            events: {
                'click .next' : function(e) {
                    if (this.model.alert.cron.get('cronType') === 'custom' &&
                        this.model.alert.get('ui.type') === 'scheduled') {
                        this.saveTimeRange();
                    } else {
                        this.model.alert.entry.content.trigger('next');
                    }
                    e.preventDefault();
                }
            },
            render: function() {
                this.$el.html(Modal.TEMPLATE);
                this.$(Modal.HEADER_TITLE_SELECTOR).html(this.options.dialogTitle);
                this.$(Modal.BODY_SELECTOR).append(Modal.FORM_HORIZONTAL_COMPLEX);
                this.$(Modal.BODY_SELECTOR).addClass('modal-body-scrolling');
                this.children.flashMessages.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.name.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.descriptionField.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.type.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.realtime.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.scheduled.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.$(Modal.BODY_FORM_SELECTOR).show();
                this.$(Modal.FOOTER_SELECTOR).append(Modal.BUTTON_CANCEL);
                this.$(Modal.FOOTER_SELECTOR).append('<a href="#" class="btn next btn-primary">' + _("Next").t() + '</a>');

                this.toggleChildren();
                
                if (this.model.inmem.isAlert() && (this.options.dialogType === 'edit') && splunkUtils.normalizeBoolean(this.model.inmem.entry.content.get('action.summary_index'))) {
                    var managerRouteToAlert = route.managerEdit(
                            this.model.application.get('root'),
                            this.model.application.get('locale'),
                            this.model.application.get('app'),
                            ['saved', 'searches', this.model.inmem.entry.get('name')],
                            this.model.inmem.id),
                        managerLinkToAlert = '<a href="' + managerRouteToAlert + '">' + _('Settings').t()  + '</a>',
                        summaryIndexMessage = splunkUtils.sprintf(_('This alert has summary indexing enabled. Summary indexing will be disabled if the alert is saved through this form. To retain summary indexing, edit this alert in %s.').t(), managerLinkToAlert);
                    
                    this.children.flashMessages.flashMsgHelper.addGeneralMessage(this.model.alert.id, {
                        type: splunkd_utils.WARNING,
                        html: summaryIndexMessage
                    });
                }

                return this;
            }
        }
        );
    });
