/**
 * A delegate view to handle docking table header, footer, and scroll bar.
 */

define(['jquery', 'underscore', 'views/shared/delegates/DetachedTableHeader'], function($, _, DetachedTableHeader) {

    return DetachedTableHeader.extend({
        awake: true,
        touch: false,
        initialize: function(options) {

            var defaults = {
                    table: "> table",
                    offset: 0,
                    dockScrollBar: true,
                    flexWidthColumn: -1,
                    defaultLayout:'auto'
                };
    
            _.defaults(this.options, defaults);
            DetachedTableHeader.prototype.initialize.call(this, options);
            
            this.disabled = false;
            this.dockedScrollHidden = true;
            this.left = 0;
            this.scrollLeft = false;
            
            this.eventNS = 'table-dock-' + this.cid;
            this.activate();
        },
        activate: function(options) {
            options || (options = {});
            if(this.active) {
                return DetachedTableHeader.prototype.activate.call(this, options);
            }
            
            var debouncedResizeHandler = _.debounce(this.handleWindowResize, 50);
            $(window).on('resize.' + this.eventNS, _(debouncedResizeHandler).bind(this));
            $(window).on('scroll.' + this.eventNS, _(this.handleWindowScroll).bind(this));
            
            _.defer(this.update.bind(this));
            
            this.options.dockScrollBar && this.$el.on('scroll.' + this.eventNS, _(this.handleContainerScroll).bind(this));
            
            return DetachedTableHeader.prototype.activate.call(this, options);
        },
        deactivate: function(options) {
            options || (options = {});
            if(!this.active) {
                return DetachedTableHeader.prototype.deactivate.call(this, options);
            }
            DetachedTableHeader.prototype.deactivate.call(this, options);
            
            if (this.$el.html()){
                this.$el.scrollLeft(0);
            }
            this.left = 0;
            this.scrollLeft = false;
            
            $(window).off('resize.' + this.eventNS);
            $(window).off('scroll.' + this.eventNS);
            this.$el.off('.' + this.eventNS);
            this.$headerTable && this.$headerTable.off('.' + this.eventNS);
            this.$dockedScroll && this.$dockedScroll.off('.' + this.eventNS);

            return this;
        },
        update: function() {
            if (!this._updateDebounced) {
                this._updateDebounced = _.debounce(this._update);
            }
            this._updateDebounced();
        },
        _update: function() {
            this.$table = this.$(this.options.table ).first();
            
            this.updateHeaders();
            this.options.dockScrollBar && this.updateScroll();

            if(this.$table.is(':visible')) {
                this.syncMeasurements();
            } else {
                _.defer(this.syncMeasurements.bind(this));
            }
        },
        syncMeasurements: function() {
            this.scrollLeft = false;
            this.syncColumnWidths();
            this.options.dockScrollBar && this.syncScrollWidths();
            this.handleWindowScroll();
            this.options.dockScrollBar && this.handleContainerScroll();
            this.trigger('updated', this.$headerTable);
        },
        disable: function() {
            this.disabled = true;
            this.$disable && this.$disable.show();
        },
        enable: function() {
            this.disabled = false;
            this.$disable && this.$disable.hide();
        },
        updateHeaders: function() {
            if(this.$header) {
                this.$header.remove();
            }
            
            this.$header = $('<div class="header-table-docked" style="display:none"><table></table><div class="disable"></div></div>').css({top: this.options.offset, left:this.left, right: 0});
            this.$disable = this.$header.find('> .disable')[this.disabled ? 'show' : 'hide']();
            this.$headerTable = this.$header.find('> table');
            this.$headerTable.attr('class', this.$table.attr('class'));
            this.$table.find('> thead').clone().appendTo(this.$headerTable);
            this.$header.prependTo(this.el);
            this.$headerTable.on('click.' + this.eventNS, 'th', function(e) {
                e.preventDefault();
                e.stopPropagation();
                var colIndex = $(e.currentTarget).prevAll().length + 1;
                this.$table.find('> thead > tr > th:nth-child(' + colIndex + ')').click();
            }.bind(this));
            this.$headerTable.on('click.' + this.eventNS, 'th a', function(e) {
                e.preventDefault();
            });
        },
                
        updateScroll: function() {
            if(this.$dockedScroll) {
                this.$dockedScroll.remove();
            }
            
            this.$dockedScroll = $('<div />').addClass('table-scroll-bar-docked').hide().appendTo(this.$el).css('left', this.$el.position().left);
            this.$dockedScroll.on('scroll.' + this.eventNS, _(this.handleDockedScrollChange).bind(this));
        },

        // ------ private methods ------ //

        handleWindowResize: function() {
            // no-op if update has not been called yet
            if(!this.$table || !this.$table.is(':visible')) {
                return;
            }
            this.syncColumnWidths();
            
            if(this.options.dockScrollBar) {
                this.updateDockedScrollBar();
                this.syncScrollWidths();
            }
        },

        handleWindowScroll: function(e) {
            // no-op if update has not been called yet
            if(!this.$table || !this.$table.is(':visible')) {
                return;
            }
            var scrollTop = $(window).scrollTop(),
                tableOffsetTop = this.$table.offset().top;

            if(scrollTop >= tableOffsetTop - this.options.offset) {
                if(!this.$header.is(':visible')) {
                    this.$header.css('top', this.options.offset).show();
                }
            }
            else {
                this.$header.css('top', '-1000px').width(); //move off and force redraw
                this.$header.hide();
            }
            this.options.dockScrollBar && this.updateDockedScrollBar();
        },

        handleContainerScroll: function() {
            // no-op if update has not been called yet or the dock is hidden
            if( !this.$table) {
                return;
            }
            
            var scrollLeft = this.$el.scrollLeft();
            
            
            if(this.scrollLeft === scrollLeft) {
                return;
            }

            this.$headerTable.css('marginLeft', -scrollLeft + 'px');
            this.$dockedScroll.scrollLeft(scrollLeft);
            this.scrollLeft = scrollLeft;
        },

        handleDockedScrollChange: function() {
            var scrollLeft = this.$el.scrollLeft(),
                dockScrollLeft = this.$dockedScroll.scrollLeft();

            if(scrollLeft !== dockScrollLeft) {
                this.$headerTable.css('marginLeft', -dockScrollLeft + 'px');
                this.$el.scrollLeft(dockScrollLeft);
            }
        },
        syncColumnWidths: function() {
            DetachedTableHeader.prototype.syncColumnWidths.apply(this, arguments);
            this.left = this.$el.position().left;
            this.$header.css({left: this.left});
        },
        syncScrollWidths: function() {
            if(!this.$table || !this.$table[0]) {
                return;
            }

            var tableWidth = parseFloat(this.$table[0].scrollWidth),
                $fullWidthDiv = $('<div />').width(tableWidth).height(1);

            this.$dockedScroll.html($fullWidthDiv);
        },

        updateDockedScrollBar: function() {
            var scrollTop = $(window).scrollTop(),
                tableOffsetTop = this.$table.offset().top,
                windowHeight = $(window).height(),
                tableHeight = this.$el.outerHeight();

            // If the table's scroll bar is below the fold, show the docked scroll bar.
            // But not if the body itself has a scroll bar (SPL-89945, no double rainbows!).
            if((tableOffsetTop + tableHeight > scrollTop + windowHeight) && $(window).width() >= $('body').width()) {
                if (!this.dockedScrollHidden) {
                    return;
                }
                this.$dockedScroll.show();
                this.dockedScrollHidden = false;
                this.handleContainerScroll();
            }
            else if (!this.dockedScrollHidden){
                this.$dockedScroll.hide();
                this.dockedScrollHidden = true;
            }
        },
        remove: function() {
            DetachedTableHeader.prototype.remove.apply(this);
            $(window).off('resize.' + this.eventNS);
            $(window).off('scroll.' + this.eventNS);
            this.$el.off('.' + this.eventNS);
            this.$headerTable && this.$headerTable.off('.' + this.eventNS);
            this.$dockedScroll && this.$dockedScroll.off('.' + this.eventNS);
            return this;
        }
    });

});
