define([
    'underscore',
    'models/Base',
    'module',
    'views/shared/Modal',
    'views/shared/controls/ControlGroup',
    'views/shared/FlashMessages',
    'uri/route',
    'util/pdf_utils',
    'util/math_utils'
    ],
    function(
        _,
        Base,
        module,
        Modal,
        ControlGroup,
        FlashMessagesV2,
        route,
        pdfUtils,
        mathUtils
    ) {
    return Modal.extend({
        moduleId: module.id,
         /**
         * @param {Object} options {
         *      model: {
         *          searchJob: <helpers.ModelProxy>,
         *          application: <models.Application>,
         *          report: <models.Report> (Only required for export to pdf. If passed in pdf will be a format option.)
         *      }
         * }
         */
        initialize: function(options) {
            Modal.prototype.initialize.apply(this, arguments);
            var Inmem = Base.extend({
                defaults: {
                    fileName: '',
                    format: 'csv',
                    limitResults: 'unlimited',
                    maxResults: 1000
                },
                validation: {
                    maxResults: {
                        fn: 'validateMaxResults'
                    },
                    format: {
                        fn: 'validateFormat'
                    }
                },
                validateMaxResults: function(value, attr, computedState){
                    if ((computedState.format === 'pdf' || computedState.limitResults === 'limit') &&
                        (!mathUtils.isInteger(value) || parseFloat(value) <= 0 )) {
                            return _('Max results must be an integer greater than 0').t();
                    }
                },
                validateFormat: function(value, attr, computedState){
                    if (computedState.format === 'pdf' && !(options.model.report && !options.model.report.isNew())) {
                        return _('Invalid format').t();
                    }
                }
            });

            this.model = {
                searchJob: this.model.searchJob,
                application: this.model.application,
                report: this.model.report,
                reportPristine: this.model.reportPristine,
                inmem: new Inmem()
            };

            this.children.flashMessage = new FlashMessagesV2({ model: this.model.inmem });

            this.deferredPdfAvailable = $.Deferred();
            this.deferredInitializeFormat = $.Deferred();

            this.usePanelType = options.usePanelType;
            this.resultTypeIsReport = this.usePanelType ? (this.model.report.entry.content.get('display.general.type') !== 'events') : this.model.searchJob.isReportSearch();
            
            if (this.model.report && !this.model.report.isNew()) {
                this.deferredPdfAvailable = pdfUtils.isPdfServiceAvailable();
            } else {
                this.deferredPdfAvailable.resolve(false);
            }

            $.when(this.deferredPdfAvailable).then(function(available) {
                var items = [
                    {
                        label: _('CSV').t(),
                        value: 'csv'
                    },
                    {
                        label: _('XML').t(),
                        value: 'xml'
                    },
                    {
                        label: _('JSON').t(),
                        value: 'json'
                    }
                ];

                if (available && (!this.model.reportPristine || !this.model.report.isDirty(this.model.reportPristine))) {
                    items.unshift({
                        label: _('PDF').t(),
                        value: 'pdf'
                    });
                }

                if (!this.resultTypeIsReport) {
                    items.unshift({
                        label: _('Raw Events').t(),
                        value: 'raw'
                    });
                }

                this.children.formatControl = new ControlGroup({
                    controlType: 'SyntheticSelect',
                    controlOptions: {
                        modelAttribute: 'format',
                        model: this.model.inmem,
                        items: items,
                        save: false,
                        toggleClassName: 'btn',
                        labelPosition: 'outside',
                        elastic: true,
                        popdownOptions: {
                            attachDialogTo: '.modal:visible',
                            scrollContainer: '.modal:visible .modal-body:visible'
                        }
                    },
                    label: _('Format').t()
                });

                this.deferredInitializeFormat.resolve();
            }.bind(this));

            this.children.filenameControl = new ControlGroup({
                controlType: 'Text',
                controlOptions: {
                    modelAttribute: 'fileName',
                    model: this.model.inmem,
                    placeholder: 'optional'
                },
                label: _('File Name').t()
            });

            this.children.limitResultsControl = new ControlGroup({
                controlType: 'SyntheticRadio',
                controlOptions: {
                    modelAttribute: 'limitResults',
                    model: this.model.inmem,
                    items: [
                        {
                            label: _('Unlimited').t(),
                            value: 'unlimited'
                        },
                        {
                            label: _('Limited').t(),
                            value: 'limit'
                        }
                    ],
                    save: false,
                    toggleClassName: 'btn',
                    labelPosition: 'outside',
                    elastic: true
                },
                label: _('Number of Results').t()
            });

            this.children.maxResultsControl = new ControlGroup({
                controlType: 'Text',
                controlOptions: {
                    modelAttribute: 'maxResults',
                    model: this.model.inmem
                },
                label: _('Max Results').t()
            });

            this.model.inmem.on('change:limitResults', this.toggleMaxResults, this);

            this.model.inmem.on('change:format', this.toggleByFormat, this);

            this.model.inmem.on('validated', function(isValid, model, invalidResults){
                if(isValid) {
                    var format = this.model.inmem.get('format'),
                        maxResults = this.model.inmem.get('maxResults'),
                        count = (this.model.inmem.get('limitResults') === 'limit') ? maxResults : 0;
                    if (format === 'pdf') {
                        var orientationSuffix = '',
                            orientation = this.model.report.entry.content.get('action.email.reportPaperOrientation'),
                            pageSize = this.model.report.entry.content.get('action.email.reportPaperSize') || 'a2';
                        if(orientation === 'landscape') {
                            orientationSuffix = '-landscape';
                        }
                        pdfUtils.getRenderURL(
                            this.model.report.entry.get('name'),
                            this.model.report.entry.acl.get('app'),
                            {
                                'sid': this.model.searchJob.id,
                                'paper-size': pageSize + orientationSuffix,
                                'max-rows-per-table': maxResults
                            },
                            'report'
                        ).done(function(url){
                            this.hide();
                            window.open(url);
                        }.bind(this));
                    } else {
                        window.location.href = route.exportUrl(this.model.application.get("root"), this.model.application.get("locale"), this.model.searchJob.get('id'),
                            this.model.inmem.get('fileName'), this.model.inmem.get('format'), this.model.inmem.get('limitResults'),
                            maxResults, count, this.resultTypeIsReport);
                        this.hide();
                    }
                }
            },this);
        },
        events: $.extend({}, Modal.prototype.events, {
            'click .btn-primary': function(e) {
                this.model.inmem.validate();
                e.preventDefault();
            }
        }),
        toggleByFormat: function() {
            if(this.model.inmem.get('format') === 'pdf') {
                this.children.filenameControl.$el.hide();
                this.children.limitResultsControl.$el.hide();
                this.children.maxResultsControl.$el.show();
            } else {
                this.children.filenameControl.$el.show();
                this.children.limitResultsControl.$el.show();
                this.toggleMaxResults();
            }
        },
        toggleMaxResults: function() {
            if(this.model.inmem.get('limitResults') === 'unlimited') {
                this.children.maxResultsControl.$el.hide();
            } else {
                this.children.maxResultsControl.$el.show();
            }
        },
        render : function() {
            $.when(this.deferredInitializeFormat).then(function() {
                this.$el.html(Modal.TEMPLATE);

                this.$(Modal.HEADER_TITLE_SELECTOR).html(_("Export Results").t());

                this.children.flashMessage.render().prependTo(this.$(Modal.BODY_SELECTOR));

                this.$(Modal.BODY_SELECTOR).append(Modal.FORM_HORIZONTAL_JUSTIFIED);

                this.children.formatControl.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.filenameControl.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.limitResultsControl.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));
                this.children.maxResultsControl.render().appendTo(this.$(Modal.BODY_FORM_SELECTOR));

                this.toggleByFormat();
                this.toggleMaxResults();

                this.$(Modal.FOOTER_SELECTOR).append(Modal.BUTTON_CANCEL);
                this.$(Modal.FOOTER_SELECTOR).append('<a href="#" class="btn btn-primary modal-btn-primary">' + _("Export").t() + '</a>');

                return this;
            }.bind(this));
        }
    });
});
