/**
 * @author sfishel
 *
 * Table renderer for search results
 *
 * Child Views:
 *
 * header <views/shared/results_table/ResultsTableHeader> the table header
 * rows <views/shared/results_table/ResultsTableRow> one for each row in the table
 *
 * Custom Events:
 *
 * cellClick - triggered when a cell is clicked for a drilldown action
 *     @param drilldownInfo {Object} {
 *         name {String} the name of the row-split field
 *         value {String} the value of the row-split field for the clicked row
 *         name2 {String} (only applicable if the user did not click on the first cell of the row)
 *                        the name of the column header for the clicked cell
 *         value2 {String} (only applicable if the user did not click on the first cell of the row)
 *                         the value contained in the clicked cell
 *         _span {Integer} (only applicable if the row-split field is time-based)
 *                         the number of seconds that the clicked row represents in time
 *         modifierKey {Boolean} whether the Ctrl/Cmd key was pressed when the cell was clicked
 *         rowContext {Object} a dictionary with key-value pairs for all of the cells in the clicked row
 *         rowIndex {Integer} index of the row in the table
 *     }
 *
 *     In the case of a multivalue field, the value that was clicked will appear as 'value2',
 *     but the rowContext will include the entire array of values.
 */

define([
            'jquery',
            'underscore',
            'module',
            'views/Base',
            './ResultsTableHeader',
            './ResultsTableRow',
            'views/shared/delegates/ColumnSort',
            'views/shared/delegates/TableDock',
            'views/shared/delegates/TableHeadStatic',
            'helpers/grid/RowIterator',
            'helpers/Printer',
            'splunk.util',
            'util/general_utils',
            'util/time',
            'util/math_utils',
            'jquery.sparkline',
            './renderers/BaseCellRenderer',
            './renderers/BaseRowExpansionRenderer',
            './renderers/NullCellRenderer',
            './renderers/NumberCellRenderer',
            './renderers/SparklineCellRenderer',
            './renderers/StringCellRenderer',
            './renderers/TimeCellRenderer',
            'util/console'
        ],
        function(
            $,
            _,
            module,
            Base,
            ResultsTableHeader,
            ResultsTableRow,
            ColumnSort,
            TableDock,
            TableHeadStatic,
            RowIterator,
            Printer,
            splunkUtils,
            generalUtils,
            timeUtils,
            mathUtils,
            _sparkline,
            BaseCellRenderer,
            BaseRowExpansionRenderer,
            NullCellRenderer,
            NumberCellRenderer,
            SparklineCellRenderer,
            StringCellRenderer,
            TimeCellRenderer,
            console
        ) {

    var DATA_FIELD_REGEX = /^[^_]|^_time$|^_raw$/;

    var TIMESTAMP_FORMATS = timeUtils.RESULTS_TIMESTAMP_FORMATS;

    var REPORT_FORMAT_ATTRIBUTES = [
        'display.statistics.wrap',
        'display.statistics.rowNumbers',
        'display.statistics.overlay',
        'display.statistics.sparkline.format'
    ];

    var MV_SUBCELL_SELECTOR = '.' + ResultsTableRow.MV_SUBCELL_CLASSNAME;

    var ENTER_KEY = 13;

    return Base.extend({

        moduleId: module.id,

        events: {
            'mouseover tbody td': function(e) {
                this.handleCellMouseover($(e.target));
            },
            'mouseout tbody td': function(e) {
                this.handleCellMouseout($(e.target));
            },
            'mousedown tbody td': function(e) {
                this.handleCellMousedown($(e.target), e);
            },
            'click tbody td': function(e) {
                this.handleCellClick($(e.target), e);
            },
            'keypress tbody td': function(e) {
                if(e.which === ENTER_KEY) {
                    this.handleCellClick($(e.target), e);
                }
            },
            'keypress tbody tr': function(e) {
                if(e.which === ENTER_KEY) {
                    // treat the target as the first cell in the row
                    this.handleCellClick($(e.target).find('td[data-cell-index]').first(), e);
                }
            }
        },

        /**
         * @constructor
         * @param options {Object} {
         *    model: {
         *        results <models/services/search/jobs/ResultsRow> the search results in row-major format
         *        metadata {Model} a model containing the fetch data for the results
         *        format <model> the table format settings
         *    },
         *    {String} enableTableDock: (Optional) whether or not to dock table header. defalut true,
         *    {Number} tableDockOffset: (Optional) the amount to offset the table header from the top when docked
         *    {String} sortKeyAttr: (Optional) attribute to set sort key on.  default is 'sortKey',
         *    {String} sortDirAttr: (Optional) attribute to set sort direction on. default is 'sortDirection',
         *    {Boolean} sortableFields (Optional) disable/enable the sortable fields. default is true,
         *    {String} countAttr: (Optional) attribute to set count on. default is 'count',
         *    {String} offsetAttr: (Optional) attribute to set offset on. default is 'offset'
         * }
         */

        initialize: function(options) {
            Base.prototype.initialize.call(this, options);

            var defaults = {
                sortKeyAttr: 'sortKey',
                sortDirAttr: 'sortDirection',
                sortableFields: true,
                countAttr: 'count',
                offsetAttr: 'offset',
                numRowSplits: 1,
                headerClass: ResultsTableHeader,
                enableTableDock: true
            };
            this.options = $.extend(true, defaults, this.options);


            if(this.options.enableStaticHeader && this.options.enableTableDock) {
                throw new Error('Static header and docking header cannot be enabled at the same time');
            }

            // Since this view is often subclassed, add a class here as a hook for CSS styles, as opposed to the one
            // derived from its module id.
            this.$el.addClass('results-table');

            this._cellRenderers = [];
            this._rowExpansionRenderers = [];

            // delegates
            this.children.columnSort = new ColumnSort({
                el: this.el,
                model: this.model.metadata,
                autoUpdate: true,
                sortKeyAttr: this.options.sortKeyAttr,
                sortDirAttr: this.options.sortDirAttr
            });
            if (this.options.enableTableDock) {
                this.children.tableDock = new TableDock({ el: this.el, offset: this.options.tableDockOffset });
            }

            this.addCellRenderer(new StringCellRenderer());
            this.addCellRenderer(new TimeCellRenderer());
            this.addCellRenderer(new NumberCellRenderer());
            this.addCellRenderer(new SparklineCellRenderer());
            this.addCellRenderer(new NullCellRenderer());
            
            this.rawFields = [];
            this.rawRows = [];
            this.dataFields = [];
            this.dataRows = [];

            this.columnTypes = {};
            this.fieldHeatRanges = {};
            this.fieldHeatOffsets = {};            
            this.fieldExtremes = {};
            this.fieldTimestampFormats = {};
            this.sparklineFormatSettings = {};
            
            this.updateFormatSettings();
            this.readResults();

            this.activate({skipUpdate: true});
        },
        startListening: function() {
            this.listenTo(this.model.results, 'change', _.debounce(function(){
                if (this.active) {
                    this.handleResultsChange();
                }
            }, 0));
            this.listenTo(this.model.format, 'change', this.handleFormatChange);
            if (this.options.enableStaticHeader) {
                this.listenTo(Printer, Printer.PRINT_END, function() {
                    this.children.staticHead.update();
                });
            }
        },
        activate: function(options) {
            options = options || {};
            if (this.active) {
                return Base.prototype.activate.apply(this, arguments);
            }
            if (this.options.disable) {
                this.$el.addClass("table-disabled");
            }
            if (!options.skipUpdate) {
                this.updateFormatSettings();
                this.handleResultsChange();
            }
            
            return Base.prototype.activate.apply(this, arguments);
        },

        deactivate: function(options) {
            if (!this.active) {
                return Base.prototype.deactivate.apply(this, arguments);
            }
            Base.prototype.deactivate.apply(this, arguments);

            this.collapseRow();
            return this;
        },

        addCellRenderer: function(cellRenderer) {
            if ((typeof cellRenderer !== "function") && !(cellRenderer instanceof BaseCellRenderer)) {
                throw new Error("Parameter cellRenderer must be a function or an instance of views/shared/results_table/renderers/BaseCellRenderer.");
            }

            this._cellRenderers.unshift(cellRenderer);
        },

        removeCellRenderer: function(cellRenderer) {
            if ((typeof cellRenderer !== "function") && !(cellRenderer instanceof BaseCellRenderer)) {
                throw new Error("Parameter cellRenderer must be a function or an instance of views/shared/results_table/renderers/BaseCellRenderer.");
            }

            var cellRenderers = this._cellRenderers;
            for (var i = 0, l = cellRenderers.length; i < l; i++) {
                if (cellRenderers[i] === cellRenderer) {
                    cellRenderers.splice(i, 1);
                    break;
                }
            }
        },

        getCellRenderers: function() {
            return this._cellRenderers.concat();
        },

        addRowExpansionRenderer: function(rowExpansionRenderer) {
            if (!(rowExpansionRenderer instanceof BaseRowExpansionRenderer)) {
                throw new Error("Parameter rowExpansionRenderer must be an instance of views/shared/results_table/renderers/BaseRowExpansionRenderer.");
            }

            this._rowExpansionRenderers.unshift(rowExpansionRenderer);
        },

        removeRowExpansionRenderer: function(rowExpansionRenderer) {
            if (!(rowExpansionRenderer instanceof BaseRowExpansionRenderer)) {
                throw new Error("Parameter rowExpansionRenderer must be an instance of views/shared/results_table/renderers/BaseRowExpansionRenderer.");
            }

            var rowExpansionRenderers = this._rowExpansionRenderers;
            for (var i = 0, l = rowExpansionRenderers.length; i < l; i++) {
                if (rowExpansionRenderers[i] === rowExpansionRenderer) {
                    rowExpansionRenderers.splice(i, 1);
                    break;
                }
            }
        },

        getRowExpansionRenderers: function() {
            return this._rowExpansionRenderers.concat();
        },

        expandRow: function(rowIndex) {
            if (typeof rowIndex !== "number") {
                throw new Error("Parameter rowIndex must be a Number.");
            }

            this.collapseRow();

            var rowView = this.children["row_" + rowIndex];
            if (!rowView) {
                return;
            }

            var rowData = rowView.getRowData();
            var rowExpansionRenderers = this._rowExpansionRenderers;
            var rowExpansionRenderer;
            for (var i = 0, l = rowExpansionRenderers.length; i < l; i++) {
                rowExpansionRenderer = rowExpansionRenderers[i];
                if (rowExpansionRenderer.canRender(rowData)) {
                    this._expandedRow = rowView;
                    this._expandedRowIndex = rowIndex;
                    rowView.expandRow(rowExpansionRenderer, rowData);
                    break;
                }
            }
        },

        collapseRow: function() {
            var rowView = this._expandedRow;
            if (!rowView) {
                return;
            }

            this._expandedRowIndex = -1;
            this._expandedRow = null;
            rowView.collapseRow();
        },

        getExpandedRowIndex: function() {
            return (this._expandedRowIndex >= 0) ? this._expandedRowIndex : -1;
        },

        remove: function() {
            this.collapseRow();
            return Base.prototype.remove.apply(this, arguments);
        },

        handleResultsChange: function() {
            this.readResults();
            // no need to debounce the render, this method is already debounced
            this.render();
        },

        readResults: function() {
            this.rawFields = this.getFieldList();
            this.rawRows = this.model.results.get('rows') || [];
            this.dataFields = _(this.rawFields).filter(this.isDataField, this);
            this.dataRows = _(this.rawRows).map(function(row) {
                return _(row).filter(function(data, i) { return this.isDataField(this.rawFields[i]); }, this);
            }, this);
        },

        handleFormatChange: function() {
            var changedFormatAttributes = this.model.format.changedAttributes();
            // only do work if an attribute changed that actually requires a re-render
            if(_(changedFormatAttributes).chain().pick(REPORT_FORMAT_ATTRIBUTES).isEmpty().value()) {
                // changes to drilldown do not require a re-render, but do have to be handled if a re-render is not happening
                if(changedFormatAttributes.hasOwnProperty('display.statistics.drilldown')) {
                    this.updateDrilldownClasses();
                }
                return;
            }
            this.updateFormatSettings();
            this.debouncedRender();
        },

        render: function() {
            if(this.options.enableStaticHeader && !this.el.innerHTML) {
                this.$el.html(this.staticHeaderTemplate);
                this.children.staticHead = new TableHeadStatic({
                    el: this.el,
                    flexWidthColumn: false
                });
            }
            var html = this.compiledTemplate({
                    wrapResults: splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.wrap')),
                    numDataColumns: this.dataFields.length
                }),
                $html = $(html);

            this.disposeOfChildren();
            if(this.dataFields.length > 0 && this.dataRows.length > 0) {
                console.debug('rendering results table with data:', { fields: this.dataFields, rows: this.dataRows });
                var columns = this.resultsAsColumns();
                this.updateColumnTypes(columns);
                this.insertHeader($html, this.dataFields);
                this.insertRows($html, this.dataFields, this.dataRows, columns);
                this.children.columnSort.update($html);
            }
            if(this.options.enableStaticHeader) {
                this.$('.scroll-table-wrapper').html($html);
                this.children.staticHead.update();
            }
            else {
                this.$el.html($html);
            }
            // render the sparklines that have been created with hidden elements or elements not attached to DOM
            $.sparkline_display_visible();
            if(this.children.tableDock) {
                this.children.tableDock.update();
            }
            
            if (this.options.disable) {
                this.$tableDisabled = $('<div class="table-disabled-screen"></div>');
                this.$el.append(this.$tableDisabled);
            }
            this.updateDrilldownClasses();

            // SPL-69820 trigger event when content has been updated
            this.trigger('rendered');


            // store a reference to the base table (not any created by delegates)
            this.$table = this.$('table').last();
            this.reflow();
            return this;
        },
        reflow: function() {
            if(this.children.tableDock) {
                this.children.tableDock.update(); 
            }
            Base.prototype.reflow.apply(this, arguments);
            if(this.$tableDisabled && this.$table.length > 0) {
                this.$tableDisabled.width(this.$table[0].scrollWidth || '100%');
            }
        },
        onAddedToDocument: function() {
            Base.prototype.onAddedToDocument.apply(this, arguments);
            $.sparkline_display_visible();
        },
        handleCellMouseover: function($target) {
            var drilldown = this.model.format.get('display.statistics.drilldown'),
                rowNumbers = splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers')),
                rowExpansion = (this._rowExpansionRenderers.length > 0);

            if(drilldown === 'none') {
                return;
            }
            var $cell = $target.closest('td'),
                cellInfo = this.getCellInfo($cell),
                rowView = this.children['row_' + cellInfo.rowIndex];

            // If we didn't get a row view, then we should just stop
            if (!rowView) {
                return;
            }

            if(drilldown === 'cell') {
                // if the cell is multivalue but the mouse is not over one of the subcells, ignore the event
                if($cell.find(MV_SUBCELL_SELECTOR).length > 0 && !$target.is(MV_SUBCELL_SELECTOR)) {
                    return;
                }
                this.children.header.highlightColumn(cellInfo.cellIndex);
                // FIXME: this is pretty lame
                var cellIndexOffset = (rowNumbers ? 1 : 0) + (rowExpansion ? 1 : 0);
                if(this.children.tableDock) {
                    this.children.tableDock.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + cellIndexOffset).addClass('highlighted');
                }
                if(this.children.staticHead) {
                    this.children.staticHead.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + cellIndexOffset).addClass('highlighted');
                }
                rowView.highlightCell($target, cellInfo.cellIndex);
            }
            else {
                // drilldown === 'row'
                rowView.highlightRow();
            }
        },

        handleCellMouseout: function($target) {
            var drilldown = this.model.format.get('display.statistics.drilldown'),
                rowNumbers = splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers')),
                rowExpansion = (this._rowExpansionRenderers.length > 0);

            if(drilldown === 'none') {
                return;
            }
            var $cell = $target.closest('td'),
                cellInfo = this.getCellInfo($cell),
                rowView = this.children['row_' + cellInfo.rowIndex];

            // If we didn't get a row view, then we should just stop
            if (!rowView) {
                return;
            }

            if(drilldown === 'cell') {
                // if the cell is multivalue but the mouse did not leave one of the subcells, ignore the event
                if($cell.find(MV_SUBCELL_SELECTOR).length > 0 && !$target.is(MV_SUBCELL_SELECTOR) && !$target.parent().is(MV_SUBCELL_SELECTOR)) {
                    return;
                }
                this.children.header.unHighlightColumn(cellInfo.cellIndex);
                // FIXME: this is pretty lame
                var cellIndexOffset = (rowNumbers ? 1 : 0) + (rowExpansion ? 1 : 0);
                if(this.children.tableDock) {
                    this.children.tableDock.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + cellIndexOffset).removeClass('highlighted');
                }
                if(this.children.staticHead) {
                    this.children.staticHead.$headerTable.find('th').eq(parseInt(cellInfo.cellIndex, 10) + cellIndexOffset).removeClass('highlighted');
                }
                rowView.unHighlightCell($target, cellInfo.cellIndex);
            }
            else {
                // drilldown === 'row'
                rowView.unHighlightRow();
            }
        },

        handleCellMousedown: function($target, event) {
            this.mouseDownX = event.pageX;
            this.mouseDownY = event.pageY;
        },

        handleCellClick: function($target, clickEvent) {
            clickEvent.preventDefault();

            // compute mouse deltas, carefully handling missing/invalid mouse coordinates
            var deltaMouseX = (clickEvent.pageX && this.mouseDownX) ? Math.abs(clickEvent.pageX - this.mouseDownX) : 0;
            var deltaMouseY = (clickEvent.pageY && this.mouseDownY) ? Math.abs(clickEvent.pageY - this.mouseDownY) : 0;

            // zero out mouse coordinates to handle cases where mousedown is not triggered (ENTER key press)
            this.mouseDownX = 0;
            this.mouseDownY = 0;

            // if mouse deltas are greater than threshold, abort click actions
            var dragThreshold = 2;
            if ((deltaMouseX > dragThreshold) || (deltaMouseY > dragThreshold)) {
                // trigger a custom event in case observers or subclasses want to add text highlighting behavior
                this.trigger('cellHighlight', $target, clickEvent);
                return;
            }

            var $cell = $target.closest('td');
            if ($cell.hasClass(ResultsTableRow.ROW_EXPANSION_CLASSNAME)) {
                var $row = $cell.closest('tr');
                var rowIndex = Number($row.attr(ResultsTableRow.ROW_INDEX_ATTR));
                if (this.getExpandedRowIndex() != rowIndex) {
                    this.expandRow(rowIndex);
                } else {
                    this.collapseRow();
                }
                return;
            }

            var drilldown = this.model.format.get('display.statistics.drilldown');
            if(drilldown === 'none') {
                return;
            }

            var isMultiValue = $cell.find(MV_SUBCELL_SELECTOR).length > 0;

            // if the cell is multivalue but the click was not on one of the subcells, ignore the event
            if(drilldown === 'cell' && isMultiValue && !$target.is(MV_SUBCELL_SELECTOR) && !$target.parent().is(MV_SUBCELL_SELECTOR)) {
                return;
            }

            var cellInfo = this.getCellInfo($cell),
                modifierKey = !!(clickEvent.ctrlKey || clickEvent.metaKey);

            if(isMultiValue) {
                cellInfo.mvIndex = this.getMultiValueIndex($target);
            }
            this.emitDrilldownEvent(cellInfo, modifierKey, clickEvent);
        },

        emitDrilldownEvent: function(cellInfo, modifierKey, clickEvent) {
            // If somebody clicked on an invalid row, we cannot create a drilldown
            // event for them.

            if (cellInfo.rowIndex === undefined || cellInfo.cellIndex === undefined) {
                return;
            }
            
            var rowIndex = cellInfo.rowIndex,
                cellIndex = cellInfo.cellIndex,
                rowName = this.dataFields[0],
                rowSplitIsTime = this.isTimeField(rowName),
                rowValue = this.dataRows[rowIndex][0],
                drilldownInfo = {
                    modifierKey: modifierKey,
                    name: rowName,
                    value: rowSplitIsTime ? splunkUtils.getEpochTimeFromISO(rowValue) : rowValue
                };

            // for a time-based row split, add the _span information for the row
            if(rowSplitIsTime) {
                var spanSeries = this.getSpanSeriesForTimeField(rowName);
                if(spanSeries) {
                    drilldownInfo._span = parseInt(spanSeries[rowIndex], 10);
                }
                else {
                    // If there is no span series (we are drilling into discrete time points), set the span to one
                    // millisecond since this is the finest granularity the UI has.
                    drilldownInfo._span = 0.001;
                }
            }

            // can only add a name2/value2 if the click was not on the first cell of the row, unless it's a multi-value field
            var isMultiValue = _.isArray(this.dataRows[rowIndex][cellIndex]);
            if((cellIndex > this.options.numRowSplits - 1) || isMultiValue) {
                drilldownInfo.name2 = this.dataFields[cellIndex];
                // if value2 is an array (a multivalue field), use the mvIndex to pick the right value (or just take the first one)
                drilldownInfo.value2 = isMultiValue ? this.dataRows[rowIndex][cellIndex][cellInfo.mvIndex || 0] :
                                            this.dataRows[rowIndex][cellIndex];
            }

            // add in the row context information
            var rowContext = {};
            _(this.dataRows[rowIndex]).each(function(value, i) {
                rowContext['row.' + this.dataFields[i]] = value;
            }, this);
            drilldownInfo.rowContext = rowContext;
            drilldownInfo.rowIndex = parseInt(rowIndex, 10);
            drilldownInfo.cellIndex = parseInt(cellIndex, 10);
            drilldownInfo.originalEvent = clickEvent;

            this.trigger('cellClick', drilldownInfo);
        },

        insertHeader: function($rootEl, dataFields) {
            this.children.header = new this.options.headerClass(this.generateHeaderViewOptions(dataFields));
            this.children.header.render().replaceAll($rootEl.find('thead'));
        },

        generateHeaderViewOptions: function(dataFields) {
            return ({
                rowNumbers: splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers')),
                rowExpansion: (this._rowExpansionRenderers.length > 0),
                fields: dataFields,
                columnTypes: this.columnTypes,
                sortKeyAttribute: ColumnSort.SORT_KEY_ATTR,
                sortableFields: this.options.sortableFields
            });
        },

        insertRows: function($rootEl, dataFields, dataRows, dataColumns) {
            var dataOverlay = this.model.format.get('display.statistics.overlay'),
                showRowNumbers = splunkUtils.normalizeBoolean(this.model.format.get('display.statistics.rowNumbers')),
                rowExpansion = (this._rowExpansionRenderers.length > 0);

            if(dataOverlay === 'heatmap') {
                this.updateHeatRanges(dataColumns);
            }
            else if(dataOverlay === 'highlow') {
                this.updateExtremes(dataColumns);
            }

            if(_(this.columnTypes).contains('timestamp')) {
                this.updateFieldTimestampFormats(dataFields, dataRows, dataColumns);
            }

            var rowIterator = new RowIterator({
                    offset: this.model.metadata.get(this.options.offsetAttr)
                }),
                $tbody = $rootEl.find('tbody');

            rowIterator.eachRow(dataRows, function(row, i, rowNumber) {
                var viewOptions = {
                    table: this,
                    fields: dataFields,
                    values: row,
                    cells: this.generateCellObjects(row, dataFields, dataOverlay, i),
                    dataOverlay: dataOverlay,
                    rowIndex: i,
                    numRowSplits: this.options.numRowSplits,
                    rowExpansion: rowExpansion,
                    splitHighlight: this.options.splitHighlight
                };

                if(showRowNumbers) {
                    viewOptions.rowNumber = rowNumber;
                }
                var rowView = this.children['row_' + i] = new ResultsTableRow(viewOptions);
                rowView.render().appendTo($tbody);

            }, this);
        },

        generateCellObjects: function(row, fields, dataOverlay, rowNumber) {
            return _(row).map(function(value, i) {
                var field = fields[i];
                // TODO add configured format
                return ({
                    value: value,
                    field: field,
                    dataOverlay: dataOverlay,
                    columnType: this.columnTypes[field] || 'string',
                    heatRange: this.fieldHeatRanges[field],
                    heatOffset: this.fieldHeatOffsets[field],        
                    extremes: this.fieldExtremes[field],
                    timestampFormat: this.fieldTimestampFormats[field],
                    sparklineFormat: this.sparklineFormatSettings[field],
                    index: i
                });
            }, this);
        },

        disposeOfChildren: function() {
            this.collapseRow();
            _(this.children).each(function(child, name) {
                if(name === 'header' || name.substr(0, 4) === 'row_') {
                    child.stopListening();
                    delete this.children[name];
                }
            }, this);
        },

        resultsAsColumns: function() {
            var columns = {},
                rows = this.model.results.get('rows') || [];

            // make 'columns' into a dictionary of field name to array of values in the column
            _(this.getFieldList()).each(function(field, i) {
                columns[field] = _(rows).pluck(i);
            });
            return columns;
        },

        updateDrilldownClasses: function() {
            var drilldown = this.model.format.get('display.statistics.drilldown'),
                $table = this.$('.table'),
                $rows = $table.find('tbody > tr'),
                $cells = $table.find('tbody > tr > td'),
                $mvSubCells = $cells.find('.multivalue-subcell');

            if(drilldown === 'none') {
                $table.removeClass('table-drilldown table-drilldown-row table-drilldown-cell');
                $rows.removeAttr('tabindex');
                $cells.removeAttr('tabindex');
                $mvSubCells.removeAttr('tabindex');
            }
            else if(drilldown === 'cell') {
                $table.addClass('table-drilldown table-drilldown-cell').removeClass('table-drilldown-row');
                $rows.removeAttr('tabindex');
                $cells.attr('tabindex', '0');
                $mvSubCells.attr('tabindex', '0');
            }
            else {
                // drilldown === 'row'
                $table.addClass('table-drilldown table-drilldown-row').removeClass('table-drilldown-cell');
                $rows.attr('tabindex', '0');
                $cells.removeAttr('tabindex');
                $mvSubCells.removeAttr('tabindex');
            }
            $table.find('.row-expansion-toggle').removeAttr('tabindex');
        },

        updateFormatSettings: function() {
            var settings = this.sparklineFormatSettings = {};
            if(this.model.format.has('display.statistics.sparkline.format')) {
                _.extend(settings, this.model.format.get('display.statistics.sparkline.format'));
            }
        },

        updateColumnTypes: function(columns) {
            this.columnTypes = {};
            _(columns).each(function(column, field) {
                // non-data fields (i.e. _span) do not get assigned a column type
                if(!this.isDataField(field)) {
                    return;
                }
                if(this.isTimeField(field)) {
                    this.columnTypes[field] = 'timestamp';
                }
                else {
                    this.columnTypes[field] = generalUtils.valuesAreNumeric(column) ? 'number' : 'string';
                }
            }, this);
        },

        updateHeatRanges: function(columns) {
            // create a data structure that holds the adjusted range (5th percentile to 95th) for each numeric field
            // the current implementation applies the same global heat delta to all numeric series
            // but the interface is designed so that different deltas could be applied to different columns in future
            var allNumericValues = this.getAllNumericValues(columns),
                bounds = generalUtils.getPercentiles(allNumericValues.sort(function(a, b) { return a - b; }), 0.05, 0.95),
                heatRange = bounds.upper - bounds.lower,
                heatOffset = bounds.lower;

            this.fieldHeatRanges = {};
            this.fieldHeatOffsets = {};

            _(columns).each(function(column, field) {
                if(this.columnTypes[field] === 'number') {
                    this.fieldHeatRanges[field] = heatRange;
                    this.fieldHeatOffsets[field] = heatOffset;
                }
            }, this);
        },

        updateExtremes: function(columns) {
            // create a data structure that holds the extremes (min and max) for each numeric field
            // the current implementation applies the same global extremes to all numeric series
            // but the interface is designed so that different extremes could be applied to different columns in future
            var allNumericValues = this.getAllNumericValues(columns),
                extremes = {
                    min: _(allNumericValues).chain().without(null).min().value(),
                    max: _(allNumericValues).max()
                };

            this.fieldExtremes = {};
            _(columns).each(function(column, field) {
                if(this.columnTypes[field] === 'number') {
                    this.fieldExtremes[field] = extremes;
                }
            }, this);
        },

        // used for calculating heat deltas and extremes
        // finds all columns that contain numeric data, then returns all of their contents as a single flattened array
        getAllNumericValues: function(columns) {
            var normalizeToNumber = function(value) {
                // convert to null if the value can't be parsed to a number
                return mathUtils.strictParseFloat(value) || null;
            };

            // find all of the columns that have been deemed numeric, flatten, then normalize all values to numbers
            return _(columns).chain()
                .filter(function(column, field) { return this.columnTypes[field] === 'number'; }, this)
                // this is a deep flattening operation, so multi-value fields will also be flattened out
                .flatten()
                .map(normalizeToNumber)
                .value();
        },

        updateFieldTimestampFormats: function(fields, rows, columns) {
            this.fieldTimestampFormats = {};
            _(fields).each(function(field, i) {
                if(this.isTimeField(field)) {
                    if(this.isBucketedTimeField(field)) {
                        var labelGranularity = timeUtils.determineLabelGranularity(columns[field]);
                        this.fieldTimestampFormats[field] = TIMESTAMP_FORMATS[labelGranularity];
                    }
                    else {
                        this.fieldTimestampFormats[field] = TIMESTAMP_FORMATS['second'];
                    }
                }
            }, this);
        },

        // TODO [sff] a lot of these next methods should probably be extracted into a share-able helper

        // helper to normalize getting the list of field names
        // depending on whether 'show_metadata=true' was used to fetch the results
        // the "fields" could be a list of string names or a list of objects with a "name" property
        getFieldList: function() {
            return _(this.model.results.get('fields') || []).map(function(field) {
                return _(field).isString() ? field : field.name;
            });
        },

        isDataField: function(fieldName) {
            return DATA_FIELD_REGEX.test(fieldName);
        },

        isTimeField: function(fieldName) {
            return (fieldName in { _time: true, earliest_time: true, latest_time: true });
        },

        isBucketedTimeField: function(fieldName) {
            return (fieldName === '_time' && !!this.getSpanSeriesForTimeField(fieldName));
        },

        // hard-coded to use '_span' at the moment,
        // but we want an interface that could potentially support multiple time series with different spans
        getSpanSeriesForTimeField: function(fieldName) {
            var spanSeriesName = '_span';
            // find the column whose corresponding field matches the span series name
            return this.resultsAsColumns()[spanSeriesName];
        },

        getCellInfo: function($cell) {
            return ({
                cellIndex: $cell.attr(ResultsTableRow.CELL_INDEX_ATTR),
                rowIndex: $cell.closest('tr').attr(ResultsTableRow.ROW_INDEX_ATTR)
            });
        },

        getMultiValueIndex: function($el) {
            return $el.attr(ResultsTableRow.MV_INDEX_ATTR);
        },

        template: '\
            <table class="table table-chrome table-striped <%= wrapResults ? "wrapped-results" : "not-wrapped-results" %> <%= numDataColumns === 1 ? \'single-column-table\' : \'\' %>">\
                <thead></thead>\
                <tbody></tbody>\
            </table>\
        ',

        staticHeaderTemplate: '\
            <div class="header-table-static"></div>\
            <div class="scroll-table-wrapper"></div>\
        '

    });

});
