define(
    [
        'jquery',
        'underscore',
        'module',
        'views/Base',
        'views/shared/delegates/TextareaResize',
        'util/keyboard',
        'util/dom_utils'
    ],
    function($, _, module, Base, TextareaResize, keyboard_utils, dom_utils) {
        return Base.extend({
            moduleId: module.id,
            className: 'search-field-wrapper',
            initialize: function() {
                Base.prototype.initialize.apply(this, arguments);
                this.multiline = false;
                this.useSyntheticPlaceholder = !dom_utils.supportsNativePlaceholder();
                
                this.activate();
            },
            startListening: function() {
                this.listenTo(this.model.searchBar, 'searchFieldfocus', this.searchFieldfocus);

                this.listenTo(this.model.searchBar, 'change:search', function() {
                    this.setSearchString(this.model.searchBar.get('search') || "");
                });
                
                this.listenTo(this.model.content, 'applied', function(options) {
                    this.model.searchBar.trigger('closeAssistant');
                    this._onFormSubmit(options);
                });
                
                this.listenTo(this.model.content, 'change:search', function() {
                    this.setSearchString(this.model.content.get('search') || "");
                });
                
                this.listenTo(this.model.content, 'change:display.page.search.mode', function() {
                    this._onFormSubmit({silent:true});
                });
                
                if (this.options.disableOnSubmit) {
                    this.listenTo(this.model.content, 'enableSearchInput', function() {
                        this.$(".search-field").attr('disabled', false);
                    });
                }

                this.listenTo(this.model.searchBar, 'setCaretPositionToEnd', this.setCaretPositionToEnd);
            },
            activate: function(options) {
                if (this.active) {
                    return Base.prototype.activate.apply(this, arguments);
                }
                
                if (this.$el.html()) {
                    if (this.options.disableOnSubmit) {
                        this.$(".search-field").attr('disabled', false);
                        this.searchFieldfocus();
                    }
                    this.render();
                }

                return Base.prototype.activate.apply(this, arguments);
            },
            deactivate: function(options) {
                if (!this.active) {
                    return Base.prototype.deactivate.apply(this, arguments);
                }
                Base.prototype.deactivate.apply(this, arguments);
                if (this.options.disableOnSubmit) {
                    if (this.$el.html()) {
                        this.$(".search-field").attr('disabled', true);
                    }
                }
                return this;
            },
            events: {
                'change textarea': function(e) {
                    this.updatePlaceholder();
                },
                'propertychange .search-field': function(e) {
                    this.updatePlaceholder();
                },
                'click .placeholder': function(e) {
                    //can only happen if you are using the synthetic placeholder when no HTML5 support
                    this.$(".search-field").focus(); 
                },
                'mouseup textarea': function(e) { //could result in pasted text
                    this.updatePlaceholder();
                },
                'focus .search-field' : function(e) {
                    this.model.searchBar.set('assistantCursor', - 1);
                },
                'keyup .search-field' : 'onSearchFieldKeyUp',
                'keydown .search-field' : 'onSearchFieldKeyDown'
            },
            updatePlaceholder: function() {
                if (this.useSyntheticPlaceholder) {
                    if (!this.$placeholder) {
                        this.$placeholder = this.$(".placeholder");
                    }
                    this.$placeholder[this.$(".search-field").val() === '' ? 'show' : 'hide']();
                }
            },
            onSearchFieldKeyUp: function(e) {
                if (!e.metaKey && !e.ctrlKey) {    
                    this.updateMultiline();
                    switch (e.keyCode) {
                        case keyboard_utils.KEYS['ESCAPE']:
                        case keyboard_utils.KEYS['LEFT_ARROW']:
                        case keyboard_utils.KEYS['RIGHT_ARROW']:
                        case keyboard_utils.KEYS['UP_ARROW']:
                        case keyboard_utils.KEYS['TAB']:
                            return; //don't open the search assistant.
                        case keyboard_utils.KEYS['ENTER']:
                            e.preventDefault();
                            return;
                        default:
                            break;
                    }
                    this.model.searchBar.set({search: this.$(".search-field").val()});
                    this.model.searchBar.trigger('openAssistant');
                    this.updatePlaceholder();
                }
            },
            onSearchFieldKeyDown: function(e) {
                if (!e.metaKey && !e.ctrlKey) {
                    switch (e.keyCode) {
                        case keyboard_utils.KEYS['TAB']:
                            this.model.searchBar.trigger('closeAssistant');
                            break;
                        case keyboard_utils.KEYS['DOWN_ARROW']:
                            // Left bracket and down arrow register as 40. If the shift key is down, then it must be a bracket.
                            if (!e.shiftKey) {
                                if (!this.children.resize.isMultiline() || this.children.resize.caretIsLast()) {
                                    this.model.searchBar.trigger('downArrow');
                                    e.preventDefault();
                                }
                            }
                            break;
                        case keyboard_utils.KEYS['ENTER']:
                            if (e.shiftKey) {
                                return;
                            }
                            this.model.searchBar.trigger('closeAssistant');
                            this._onFormSubmit();
                            e.preventDefault();
                            break;
                        default:
                            break;
                    }
                }
            },
            updateMultiline: function() {
                if (this.children.resize) {
                    // add mulitline class for state style
                    var multiline = this.children.resize.isMultiline();

                    if (multiline && (multiline != this.multiline)) {
                        this.$el.addClass('multiline');
                        this.multiline = true;
                    } else if (!multiline && (multiline != this.multiline)) {
                        this.$el.removeClass('multiline');
                        this.multiline = false;
                    }
                }
            },
            getSearchFieldValue: function(){
                return $.trim(this.$(".search-field").val());
            },
            _onFormSubmit: function(options) {
                options = options || {};
                // don't do anything if there's nothing in the search box
                var search = this.model.searchBar.get('search'),
                    currentSearch = this.model.content.get('search'),
                    searchFromTextarea = this.getSearchFieldValue();
                
                if (search !== searchFromTextarea) {
                    this.model.searchBar.set('search', searchFromTextarea);
                    search = searchFromTextarea;
                }
                
                if (search) {
                    if (this.options.disableOnSubmit) {
                        this.$(".search-field").attr('disabled', true);
                    }
                    if (currentSearch !== search){
                        this.model.content.set({search: search}, options);
                    } else {
                        if (!options.silent) {
                            this.model.content.unset("search", {silent: true});
                            this.model.content.set({search: search});
                        }
                    }
                }
            },
            /**
             * Sometimes, like when we're resurrecting a search, we will
             * write our own input value.
             */
            setSearchString: function(search) {
                var currentVal = this.$(".search-field").val();
                this.model.searchBar.set('search', search);
                if (search !== currentVal) {
                    this.$(".search-field").val(search);
                }
                this.updatePlaceholder();
                this.searchFieldfocus();
                if (this.children.resize) {
                    this.children.resize.resizeTextarea();
                    this.updateMultiline();
                }
            },
            searchFieldfocus: function() {
                var $searchField = this.$(".search-field");
                if (!$searchField.attr('disabled')) {
                    $searchField.focus();
                }
            },
            setCaretPositionToEnd: function() {
                var $searchField = this.$(".search-field");
                dom_utils.setCaretPosition($searchField.get(0), $searchField.val().length);                
            },
            render: function() {
                var $searchField;
                if (this.$el.html()) {
                    $searchField = this.$(".search-field");
                    this.$(".search-field").val(this.model.searchBar.get('search'));
                } else {
                    var template = _.template(this.template, {
                        placeholder: _("enter search here...").t(),
                        useSyntheticPlaceholder: this.useSyntheticPlaceholder,
                        inputValue: this.model.searchBar.get('search')
                    });
                    this.$el.html(template);
                    $searchField = this.$(".search-field");
                    
                    _.defer(function(){
                        if (this.options.useAutoFocus) {
                            this.searchFieldfocus();
                        }
                        var maxLines = Math.floor(($(window).height() - 100) / parseInt($searchField.css('lineHeight'), 10));
                        this.children.resize = new TextareaResize({el: $searchField.get(0), maxLines: maxLines, minHeight: 20});
                        this.updateMultiline();
                        this.setCaretPositionToEnd();
                    }.bind(this));
                }
                
                this.updatePlaceholder();
                
                if (this.options.disableOnSubmit) {
                    $searchField.attr('disabled', false);
                }
                if (this.children.resize) {
                    this.children.resize.resizeTextarea();
                    this.updateMultiline();
                }
                
                return this;
            },
            reflow: function() {
                var el = this.$(".search-field").get(0),
                    inputValue = this.model.searchBar.get('search') || "",
                    currentCaretPos = dom_utils.getCaretPosition(el);
                dom_utils.setCaretPosition(el, (currentCaretPos || inputValue.length));
                Base.prototype.reflow.apply(this, arguments);
            },
            template: '\
                <textarea rows="1" name="q" spellcheck="false" class="search-field" autocorrect="off" autocapitalize="off"\
                <% if (!useSyntheticPlaceholder) { %> placeholder="<%- placeholder %>"\ <% } %>><%- inputValue %></textarea>\
                <% if (useSyntheticPlaceholder) { %> \
                <span class="placeholder"><%- placeholder %></span>\
                <% } %>\
                '
        });
    }
);
