define(
    [
        'jquery',
        'underscore',
        'module',
        'views/Base',
        'views/shared/timerangepicker/dialog/Master',
        'views/shared/delegates/Popdown',
        'collections/SplunkDsBase',
        'splunk.util',
        'util/console'
    ],
    function($, _, module, Base, Dialog, Popdown, SplunkDsBaseCollection, splunk_util, console) {
        /**
         * @param {Object} options {
         *      model: {
         *          state: <models.services.SavedSearch.entry.content>,
         *          timeRange: <models.TimeRange>,
         *          user: <models.services.admin.User>,
         *          appLocal: <models.services.AppLocal>,
         *          application: <models.Application>
         *      },
         *      collection: <collections.services.data.ui.TimesV2>
         *      {String} timerangeClassName (Optional) Class attribute to the button element. Default is btn.
         *      {Object} dialogOptions: (Optional) Keys and values passed to the dialog for customization. See views/shared/timerangepicker/dialog/Master.
         *      {Object} popdownOptions: (Optional) Keys and values passed to the popdown for customization.
         *      {Boolean} forceTimerangeChange: (Optional) Force timerange change event when applied
         *                    is triggered on the timeRange model even if timerange has not changed. default false.
         * }
         */
        return Base.extend({
            moduleId: module.id,
            initialize: function(options) {
                var defaults = {
                    timerangeClassName: 'btn'
                };

                _.defaults(this.options, defaults);

                Base.prototype.initialize.call(this, options);

                this.children.dialog = new Dialog(
                    $.extend(
                        {
                            model: {
                                timeRange: this.model.timeRange,
                                user: this.model.user,
                                appLocal: this.model.appLocal,
                                application: this.model.application
                            },
                            collection: this.collection
                        },
                        this.options.dialogOptions || {}
                     )
                );

                this.activate({skipSetLabel: true});
            },
            startListening: function() {
                if (this.collection) {
                    this.listenTo(this.collection, 'reset', function(){
                        console.warn("timerangepicker setting label because of collection reset");
                        this.setLabel();
                    });
                }

                this.listenToModels();                
            },
            listenToModels: function() {
                this.listenTo(this.model.timeRange, 'change:earliest change:latest', _.debounce(function() {
                    if (this.active) {
                        this.timeRangeChange();
                    }
                }, 0));

                this.listenTo(this.model.timeRange, 'applied', function() {
                    this.children.popdown.hide();
                    
                    this.model.state.trigger("applied", {
                        silent: splunk_util.normalizeBoolean(this.model.state.get('disable_timerange_trigger'))
                    });

                    if (this.options.forceTimerangeChange && !this.model.timeRange.hasChanged('earliest') && !this.model.timeRange.hasChanged('latest')) {
                        this.model.state.unset("dispatch.earliest_time", {silent: true});
                        this.model.state.unset("dispatch.latest_time", {silent: true});
                        this.model.state.set({
                            'dispatch.earliest_time': this.model.timeRange.get('earliest'),
                            'dispatch.latest_time': this.model.timeRange.get('latest')
                        });
                    }
                });

                this.listenTo(this.model.timeRange, 'change:earliest_epoch change:latest_epoch change:earliest change:latest', _.debounce(function() {
                    if (this.active) {
                        this.setLabel();
                    }
                }, 0));

                this.listenTo(this.model.state, 'change:dispatch.earliest_time change:dispatch.latest_time', _.debounce(function(){
                    if (this.active) {
                        this.stateChange();
                    }
                }, 0));
            },
            stopListeningToModels: function() {
                this.stopListening(this.model.timeRange);
                this.stopListening(this.model.state);
            },
            activate: function(options) {
                options = options || {};
                
                if (this.active) {
                    return Base.prototype.activate.apply(this, arguments);
                }

                if (!options.skipSetLabel) {
                    this.setLabel();
                }

                Base.prototype.activate.apply(this, arguments);

                this.model.timeRange.trigger("prepopulate");

                return this;
            },
            stateChange: function() {
                this.stopListeningToModels();
                this.model.timeRange.clear({silent:true, setDefaults: true});
                this.model.timeRange.save(
                    {
                        'earliest': this.model.state.get('dispatch.earliest_time'),
                        'latest':this.model.state.get('dispatch.latest_time')
                    },
                    {
                        wait: true,
                        success: function(model, response) {
                            this.listenToModels();
                            this.setLabel();
                        }.bind(this),
                        error: function(model, response) {
                            this.listenToModels();
                            this.setLabel();
                        }.bind(this)
                    }
                );
            },
            timeRangeChange: function() {
                this.stopListeningToModels();
                this.model.state.set({
                    'dispatch.earliest_time': this.model.timeRange.get('earliest'),
                    'dispatch.latest_time':this.model.timeRange.get('latest')
                });
                this.listenToModels();
            },
            setLabel: function() {
                if (this.$el.html()) {
                    var timeLabel = this.model.timeRange.generateLabel(this.collection || new SplunkDsBaseCollection());
                    this.$el.children('a').find(".time-label").text(_(timeLabel).t()).attr('title', _(timeLabel).t());
                }
            },
            render: function() {
                if (this.$el.html()) {
                    return this;
                }

                this.$el.html(this.compiledTemplate({
                    options: this.options
                }));

                this.children.dialog.render().appendTo(this.$('.popdown-dialog'));

                this.children.popdown = new Popdown(_.extend({
                    el: this.el,
                    toggle:'> a',
                    mode: "dialog",
                    attachDialogTo: 'body',
                    ignoreClasses: [
                        "ui-datepicker",
                        "ui-datepicker-header",
                        "dropdown-menu"
                    ]
                }, this.options.popdownOptions || {}));

                this.children.popdown.on('shown', function() {
                    if (this.children.dialog.$(".accordion-group.active").length){
                        this.children.dialog.onShown();
                        return;
                    }
                    var timePanel = "presets";
                    this.children.dialog.children[timePanel].$el.closest(".accordion-group").find(".accordion-toggle").first().click();
                    this.children.dialog.onShown();
                }, this);

                this.setLabel();

                return this;
            },
            template: '\
                <a class=" splBorder splBorder-nsew splBackground-primary <%- options.timerangeClassName %>" href="#"><span class="time-label"></span><span class="caret"></span></a>\
                <div class="popdown-dialog">\
                    <div class="arrow"></div>\
                </div>\
                '
        });
    }
);
