define([
        "jquery",
        "underscore",
        "backbone",
        "module",
        "models/shared/DateInput",
        "models/shared/TimeRange",
        'views/Base',
        "views/shared/controls/DateControl",
        "views/shared/FlashMessages"
      ],
     function($, _, Backbone, module, DateInputModel, TimeRangeModel, Base, DateControl, FlashMessages) {
        return Base.extend({
            // tagName: 'tbody',
            moduleId: module.id,
            initialize: function() {
                Base.prototype.initialize.apply(this, arguments);

                this.model = {
                    timeRange: this.model,
                    workingRange: new TimeRangeModel(),
                    earliestDateInput: new DateInputModel(),
                    latestDateInput: new DateInputModel()
                };
                this.syncInternals();

                this.children.earliestTimeInput = new DateControl({
                    model: this.model.earliestDateInput,
                    inputClassName: "timerangepicker-earliest-date",
                    validate: true
                });

                this.children.latestTimeInput = new DateControl({
                    model: this.model.latestDateInput,
                    inputClassName: "timerangepicker-latest-date",
                    validate: true
                });

                this.children.flashMessages = new FlashMessages({
                    model: {
                        workingRange: this.model.workingRange,
                        earliestDateInput: this.model.earliestDateInput,
                        latestDateInput: this.model.latestDateInput
                    }
                });

                this.activate();
            },
            startListening: function() {
                this.listenTo(this.model.timeRange, 'change:earliest_date change:latest_date prepopulate', _.debounce(this.prepopulate, 0));

                this.listenTo(this.model.earliestDateInput, "validated:invalid", this.disableApply);
                this.listenTo(this.model.earliestDateInput, "validated:valid", this.enableApply);

                this.listenTo(this.model.earliestDateInput, "change", function(){
                    this.model.workingRange.trigger("validated", true, this.model.workingRange, {});
                    this.model.workingRange.trigger("serverValidated", true, this.model.workingRange, []);
                });

                this.listenTo(this.model.latestDateInput, "validated:invalid", this.disableApply);
                this.listenTo(this.model.latestDateInput, "validated:valid", this.enableApply);

                this.listenTo(this.model.latestDateInput, "change", function(){
                    this.model.workingRange.trigger("validated", true, this.model.workingRange, {});
                    this.model.workingRange.trigger("serverValidated", true, this.model.workingRange, []);
                });
            },
            syncInternals: function() {
                this.yesterday = new Date();
                this.yesterday.setDate(this.yesterday.getDate() - 1);
                this.yesterday.setHours(0, 0, 0, 0);
                this.model.earliestDateInput.setFromJSDate(this.yesterday, {silent:true});

                this.today = new Date();
                this.today.setHours(0, 0, 0, 0);
                this.model.latestDateInput.setFromJSDate(this.today, {silent:true});
            },
            activate: function(options) {
                if (this.active) {
                    return Base.prototype.activate.apply(this, arguments);
                }
                this.syncInternals();
                return Base.prototype.activate.apply(this, arguments);
            },
            deactivate: function(options) {
                if (!this.active) {
                    return Base.prototype.deactivate.apply(this, arguments);
                }
                Base.prototype.deactivate.apply(this, arguments);
                this.model.workingRange.fetchAbort();
                this.model.workingRange.clear({setDefaults: true});
                this.model.earliestDateInput.clear({setDefaults: true});
                this.model.latestDateInput.clear({setDefaults: true});

                return this;
            },
            prepopulate: function() {
                var earliest = this.model.timeRange.get('earliest_date'),
                    latest =  this.model.timeRange.get('latest_date');

                if (!this.model.timeRange.hasNoEarliest() && earliest) {
                    this.model.earliestDateInput.setFromJSDate(earliest, {validate: true});
                }
                if (latest) {
                    var latestDate= new Date(latest.getTime());

                    if (latestDate.getHours() == 0) {
                        latestDate.setDate((latestDate.getDate() - 1), {validate: true});
                    }
                    this.model.latestDateInput.setFromJSDate(latestDate, {validate: true});
                }
            },
            events: {
                "click .apply": function(event) {
                    if ($(event.currentTarget).hasClass("disabled")) {
                        event.preventDefault();
                        return;
                    }

                    this.model.earliestDateInput.setHoursMinSecFromStr("00:00:00.000", {silent:true, validate: true});
                    this.model.latestDateInput.setHoursMinSecFromStr("24:00:00.000", {silent:true, validate: true});

                    var that = this,
                    earliestIsoWithoutTZ = this.model.earliestDateInput.isoWithoutTZ(),
                    latestIsoWithoutTZ = this.model.latestDateInput.isoWithoutTZ();

                    this.disableApply();

                    this.model.workingRange.save(
                            {
                                earliest: earliestIsoWithoutTZ,
                                latest: latestIsoWithoutTZ

                            },
                            {
                                success: function(model) {
                                    that.enableApply();
                                    that.model.timeRange.set({
                                        'earliest': model.get('earliest_epoch'),
                                        'latest': model.get('latest_epoch'),
                                        'earliest_epoch': model.get('earliest_epoch'),
                                        'latest_epoch': model.get('latest_epoch'),
                                        'earliest_iso': model.get('earliest_iso'),
                                        'latest_iso': model.get('latest_iso'),
                                        'earliest_date': new Date(model.get('earliest_date').getTime()),
                                        'latest_date': new Date(model.get('latest_date').getTime())
                                    });
                                    that.model.timeRange.trigger("applied");
                                },
                                error: function(model, error) {
                                    that.enableApply();
                                }
                            }
                    );
                    event.preventDefault();
                }
            },
            generateLabel: function() {
                //This is a temporary generator. Should move this into the model?
                var earliest = this.model.earliestDateInput.jsDate();
                var latest = this.model.latestDateInput.jsDate();
                latest.setDate(latest.getDate() -1);

                return earliest.toDateString().slice(4, -5) + " - " + latest.toDateString().slice(4, -5);

            },
            enableApply: function() {
                this.$("#apply_" + this.cid).removeClass('disabled');
            },
            disableApply: function() {
                this.$("#apply_" + this.cid).addClass('disabled');
            },
            render: function() {
                var template = _.template(this.template, {
                    _: _,
                    cid: this.cid
                });
                this.$el.html(template);
                this.children.flashMessages.render().appendTo(this.$(".flash_messages"));
                this.children.earliestTimeInput.render().prependTo(this.$('.earliest-date'));
                this.children.latestTimeInput.render().prependTo(this.$('.latest-date'));

                return this;
            },
            template: '\
            <div class="time-between-dates">\
                <div class="flash_messages"></div>\
                <div class="earliest-date col-1">\
                    <span class="help-block"><%- _("00:00:00").t() %></span>\
                </div>\
                <label class="control-label"><%- _("and").t() %></label>\
                <div class="latest-date col-2">\
                    <span class="help-block"><%- _("24:00:00").t() %></span>\
                </div>\
                <div class="col-3">\
                    <button class="apply btn" id="apply_<%- cid %>"><%- _("Apply").t() %></button>\
                </div>\
            </div>\
            '
        });
    }
);
