define(
    [
        'jquery',
        'underscore',
        'backbone',
        'module',
        'views/Base',
        'views/shared/controls/SyntheticSelectControl',
        'views/shared/vizcontrols/Format',
        'models/services/search/IntentionsParser',
        'models/shared/Visualization',
        'util/vizTypeUtil'
    ],
    function($, _, Backbone, module, BaseView, SyntheticSelectControl, Format, IntentionsParser, VisualizationModel, vizTypeUtil) {
        return BaseView.extend({

            moduleId: module.id,

            reportTree: {
                'match': {
                    'display.general.type': {
                        'visualizations' : {
                            'match': {
                                'display.visualizations.type': {
                                    'singlevalue': {
                                        'view': 'single'
                                    },
                                    'charting': {
                                        'match': {'display.visualizations.charting.chart': {
                                                'line': {  'view': 'line' },
                                                'area': { 'view': 'area' },
                                                'column': { 'view': 'column' },
                                                'bar': { 'view': 'bar' },
                                                'pie': { 'view': 'pie' },
                                                'scatter': { 'view': 'scatter' },
                                                'bubble': { 'view': 'bubble' },
                                                'radialGauge': { 'view': 'radialGauge' },
                                                'fillerGauge': { 'view': 'fillerGauge' },
                                                'markerGauge': { 'view': 'markerGauge' }
                                            }
                                        }
                                    },
                                    'mapping': {
                                        view: 'mapping'
                                    }
                                }
                            }
                        },
                        'statistics': {
                            'view': 'statistics'
                        },
                        'events': {
                            'view': 'events'
                        }
                    }
                }
            },

            commandToChartType: {
                "timechart" : ["line", "area", "column", "statistics"],
                //"chart" : ["column", "line", "area", "bar", "pie", "scatter", "bubble", "radialGauge", "fillerGauge", "markerGauge", "statistics"],
                "top" : ["column", "bar", "pie", "statistics"],
                "rare" : ["column", "bar", "pie", "statistics"],
                "predict": ["line", "statistics"],
                "geostats": ["mapping"]
            },

            allowedVizTypes: ['events', 'statistics', 'visualizations'],

            events: {
                'click .format': function(e) {
                    e.preventDefault();
                    
                    var $target = $(e.currentTarget);
                    this.setVisualizationFromReport();
                    this.children.format = new Format({
                        model: {
                            report: this.model.report,
                            visualization: this.model.visualization
                        },
                        onHiddenRemove: true,
                        saveOnApply: this.options.saveOnApply,
                        ignoreClasses: ['color-picker-container', 'select2-drop', 'select2-drop-mask', 'dropdown-menu']
                    });
                    this.children.format.render().activate().appendTo($('body'));
                    this.children.format.show($target);
                    $target.addClass('active');

                    this.listenTo(this.children.format, 'hidden', function() {
                        $target.removeClass('active');
                    });
                }
            },

            /**
             * @param {Object} options {
             *     model: {
             *         report: <models.search.Report>,
             *         application: <models.shared.Application>
             *     },
             *     visualizationTypes (required): [events &| statistics &| visualizations],
             *     saveOnApply: <Boolean> whether to save the report when any changes are submitted
             *     bindToChangeOfSearch: <Boolean> whether to bind to changes of the search string and update recommendations
             *     dashboard: TODO [sff] what does this do?
             * }
             */
            initialize: function() {
                BaseView.prototype.initialize.apply(this, arguments);

                var defaults = {
                    bindToChangeOfSearch: true,
                    saveOnApply: false
                };
                this.options = $.extend(true, defaults, this.options);
                this.recommendations = [];

                if(!_.isArray(this.options.vizTypes) || this.options.vizTypes.length === 0) {
                    throw new Error('Vizcontrols Master must be instantiated with at least one viz type');
                }
                _(this.options.vizTypes).each(function(type) {
                    if(!_(this.allowedVizTypes).contains(type)) {
                        throw new Error(type + ' is not an allowed viz type');
                    }
                }, this);

                this.model.intentionsParser = new IntentionsParser();
                this.model.visualization = new VisualizationModel();
                this.activate();
            },
            startListening: function() {
                this.listenTo(this.model.intentionsParser, 'change', this.handleIntentionsParserUpdate);
                this.listenTo(this.model.report.entry.content, 'change:display.general.type', function(model, value, options) {
                    if (_.indexOf(this.options.vizTypes, value)>-1) {
                        this.updateVizTypeChildView({ activate: true });
                    }
                });
                
                this.listenTo(this.model.visualization, 'change:viz_type', function() {
                    var viz = this.model.visualization.get('viz_type');
                    this.model.report.entry.content.set('viz_type', viz);
                    this.model.report.entry.content.set(vizTypeUtil.getReportSettingsForType(viz));
                    this.model.visualization.set(vizTypeUtil.getReportSettingsForType(viz));
                    if(this.options.saveOnApply) {
                        this.model.report.save();
                    }
                });
                
                this.listenTo(this.model.report.entry.content, 'change:search change:id', this.updateVizTypeVisibility);

                if (this.options.bindToChangeOfSearch) {
                    this.listenTo(this.model.report.entry.content, 'change:search', this.intentionsFetch); //simplexml
                }
            },
            activate: function(options) {
                if (this.active) {
                    return BaseView.prototype.activate.apply(this, arguments);
                }

                this.model.visualization.set({dashboard: this.options.dashboard});
                // the child view will be activated when we call super below
                this.updateVizTypeChildView({ activate: false });

                this.intentionsFetch();
                return BaseView.prototype.activate.apply(this, arguments);
            },

            deactivate: function(options) {
                if (!this.active) {
                    return BaseView.prototype.deactivate.apply(this, arguments);
                }
                BaseView.prototype.deactivate.apply(this, arguments);
                this.model.intentionsParser.fetchAbort();
                this.model.intentionsParser.clear({setDefaults: true});
                this.model.visualization.clear({setDefaults: true});
                return this;
            },

            intentionsFetch: function() {
                this.model.intentionsParser.fetch({
                    data: {
                        q: this.model.report.entry.content.get('search'),
                        app: this.model.application.get('app'),
                        owner: this.model.application.get('owner'),
                        parse_only: true
                    }
                });
            },

            handleIntentionsParserUpdate: function() {
                var newRecommendations = this.getRecommendations();
                if(!_.isEqual(newRecommendations, this.recommendations)) {
                    this.recommendations = newRecommendations;
                    this.updateVizTypeChildView({ activate: true });
                }
            },

            updateVizTypeVisibility: function() {
                if (!this.children.visualizationType || !this.model.visualization.get('dashboard')){
                    return;
                }

                if (this.model.report.isPivotReport() && !this.model.report.isNew()){
                    this.children.visualizationType.disable();
                    this.children.visualizationType.tooltip({animation:false, title:_('Edit visualization with the pivot tool.').t(), container: 'body'});
                } else {
                    this.children.visualizationType.enable();
                    this.children.visualizationType.tooltip('disable');
                }
            },

            getRecommendations: function() {
                var reportsSearch = this.model.intentionsParser.get('reportsSearch');
                if(!reportsSearch) {
                    return ['events'];
                }
                var reportingCommand = reportsSearch.replace(/\s{2,}/g, ':::').split(':::')[0];
                if(this.model.intentionsParser.has('commands')) {
                    var commands = _(this.model.intentionsParser.get('commands')).pluck('command');
                    if(_(commands).contains('predict')) {
                        reportingCommand = 'predict';
                    }
                    else if(_(commands).contains('geostats')) {
                        reportingCommand = 'geostats';
                    }
                }
                return this.commandToChartType[reportingCommand] || [];
            },

            setVisualizationFromReport: function() {
                this.model.report.entry.content.unset('viz_type');
                var toSetOnViz = $.extend({}, this.model.report.entry.content.toJSON());
                // SPL-76456, pre-filter out any invalid gauge range values from the incoming attributes
                // this will be consistent with the renderer which uses the valid values even if some are invalid
                var validRanges = VisualizationModel.filterToValidRangeValues(toSetOnViz['display.visualizations.charting.chart.rangeValues']);
                toSetOnViz['display.visualizations.charting.chart.rangeValues'] = validRanges.length > 0 ? JSON.stringify(validRanges) : '';
                // SPL-80495, also pre-validate the gauge color values
                var validColors = VisualizationModel.parseStringifiedColorArray(toSetOnViz['display.visualizations.charting.gaugeColors']);
                toSetOnViz['display.visualizations.charting.gaugeColors'] = (validColors && validColors.length > 0) ? validColors : '';
                // make sure the visualization model will be in a valid state
                // any attributes that are not valid are not set
                _(toSetOnViz).each(function(value, key) {
                    if(this.model.visualization.preValidate(key, value)) {
                        delete toSetOnViz[key];
                    }
                }, this);
                this.model.visualization.set(toSetOnViz);
                var viewObj = this.setVizType(this.reportTree), view;
                view = (viewObj && viewObj.view) ? viewObj.view : 'column';
                this.model.visualization.set('viz_type', view);
            },

            updateVizTypeChildView: function(options) {
                if(this.children.visualizationType) {
                    this.children.visualizationType.remove();
                }
                this.setVisualizationFromReport();
                var vizDropdown = [];
                _(this.options.vizTypes).each(function(value) {
                    var groupConfig = $.extend(true, [], vizTypeUtil.getVisualizationGroupings()[value]);
                    _(groupConfig).each(function(item) {
                        if(_(this.recommendations).contains(item.value)) {
                            item.description = _("Recommended").t();
                        }
                    }, this);
                    vizDropdown.push(groupConfig);
                },this);

                if(vizDropdown.length === 0) {
                    throw new Error('Vizcontrols Master updateVizTypeChildView(), no compatible groups');
                }
                this.children.visualizationType = new SyntheticSelectControl({
                    model: this.model.visualization,
                    items: vizDropdown,
                    className: "dropdown pull-left",
                    toggleClassName: 'btn-pill',
                    modelAttribute: 'viz_type',
                    menuClassName: 'viz-dropdown',
                    popdownOptions: {
                        attachDialogTo: 'body'
                    }
                });
                this.children.visualizationType.render().prependTo(this.el);
                if(options && options.activate) {
                    this.children.visualizationType.activate();
                }
                this.updateVizTypeVisibility();
            },

            setVizType: function(reportTree) {
                if (reportTree && reportTree.view){
                    return reportTree;
                } else if (reportTree && reportTree.match){
                    var match;
                    _(reportTree.match).each(function(v, k){
                       match = v[this.model.report.entry.content.get(k)];
                    }, this);
                    if (match) return this.setVizType(match);
                }
            },

            render: function() {
                if(this.children.visualizationType) {
                    this.children.visualizationType.detach();
                }
                this.$el.empty();
                if(this.children.visualizationType) {
                    this.children.visualizationType.prependTo(this.el);
                }
                this.$el.append(this.compiledTemplate());
                return this;
            },

            template: '\
                <div class="btn-group pull-left">\
                    <a class="btn-pill popdown-toggle format" href="#">\
                        <i class="icon-paintbrush"/><span class="link-label"><%- _("Format").t() %></span><span class="caret"></span>\
                    </a>\
                </div>\
            '
        });
    }
);
